/*************************************************/
/* Program transformation for PC++               */
/* Created by Jenq-Kuen Lee Feb 1, 1991          */
/*************************************************/

#include "vparse.h"
#include "tag"
#include "vpc.h"
#include "db.h"
#include "vextern.h"
#include "distribution.h"


#define DOALLStringName         "DOALL"
extern PTR_FILE fi;
int    full_temp_counter =0;
int *pllnd_pool;
PTR_LLND find_suit_declarator();
PTR_LLND convert_ll_node();
PTR_LLND convert_declarator_expression();
PTR_LLND follow_to_no_single_expr_list();
PTR_LLND find_suit_declarator();
PTR_TYPE follow_type_2();
PTR_SYMB gen_next_symbol();
PTR_BFND the_closest_root();
PTR_SYMB find_to_functor();

convert_reference_to_value()
{
  
  alloc_temp();
  convert_bif_node(global_bfnd,1);
  convert_bif_node(global_bfnd,2);
  update_reference_type();
  free_temp();
}

/*********************************************************/
/* Convert Call by reference to Call by value and Pointer*/
/*********************************************************/

/*
   Rule to convert call-by-reference to call-by-value and pointer
   (1) &* this  = this
   (2) If Y is formal parameter of function Z and Y is called by Reference
        (2.1) Change Y to be call by value and Pointer
        (2.2) Substitute every Y to  *Y inside the body of the function Z
        (2.3) Apply &*Y = Y reduction 
   (3) In the invocation of function Z, Suppose S is the object to be passed
       to Y. We will use &S as the parameter. Still remember to apply
       &*S =S rule if that exists.

*/


/***************************************************************************/
/* update_reference_type():                                                */
/* (1) Change all the types of reference to pointer type.                  */
/* (2) This should be done in the final stage, since if you change it too  */
/*   early there is no way to know where the reference node is             */
/***************************************************************************/
update_reference_type()
{
  PTR_TYPE type;

 for ( type = fi->head_type; type; type =type->thread)
   {
       if (type->variant==T_REFERENCE)  type->variant = T_POINTER;
   }
}


/*************************************************************************/
/* convert_bif_node()                                                    */
/*                                                                       */
/* flag ==1 : we only mark the expression that a &expression is needed   */
/* flag ==2 : we substitute the variable which is of reference type      */
/*************************************************************************/
convert_bif_node(pbf,flag)
PTR_BFND pbf;
{
  PTR_BLOB blob;

  if (!pbf) return;
  if (pbf->variant == CONTROL_END) return;
  for(blob = pbf->entry.template.bl_ptr1;blob; blob=blob->next)
    convert_bif_node(blob->ref,flag);
  for(blob = pbf->entry.template.bl_ptr2;blob; blob=blob->next)
    convert_bif_node(blob->ref,flag);
  switch (pbf->variant) {
  case IF_NODE    :
  case FOR_NODE   :
  case WHILE_NODE :
  case DO_WHILE_NODE:             /* New added for VPC */
  case SWITCH_NODE :              /* New added for VPC */
  case COLOOP_NODE  :             /* New added for VPC */
  case COEXEC_NODE  :             /* New added for VPC */
  case EXPR_STMT_NODE:
   pbf->entry.template.ll_ptr1=
     convert_ll_node(pbf->entry.template.ll_ptr1,flag,1);
   pbf->entry.template.ll_ptr2=
     convert_ll_node(pbf->entry.template.ll_ptr2,flag,1);
   pbf->entry.template.ll_ptr3=
     convert_ll_node(pbf->entry.template.ll_ptr3,flag,1);
    break;
  case COLLECTION_DECL:
  case CLASS_DECL:
  case VAR_DECL:
  case UNION_DECL:
  case STRUCT_DECL:
  case ENUM_DECL:
   if (flag ==1) mark_declarator_expression(pbf->entry.template.ll_ptr1);
   else {
     pbf->entry.template.ll_ptr1=
       convert_declarator_expression(pbf->entry.template.ll_ptr1);
   }
   break;
  case RETURN_STAT:
  case RETURN_NODE  :             /* New added for VPC */
   if (flag ==1)     mark_return_expression(pbf,pbf->entry.template.ll_ptr1);
   else {
   pbf->entry.template.ll_ptr1=
     convert_ll_node(pbf->entry.template.ll_ptr1,flag,2);
 }
    break;
  default:
   pbf->entry.template.ll_ptr1=
     convert_ll_node(pbf->entry.template.ll_ptr1,flag,1);
   pbf->entry.template.ll_ptr2=
     convert_ll_node(pbf->entry.template.ll_ptr2,flag,1);
   pbf->entry.template.ll_ptr3=
     convert_ll_node(pbf->entry.template.ll_ptr3,flag,1);
    break;
 }
}


/*************************************************************************/
/* convert_ll_node(pllnd,flag,sw)                                        */
/*                                                                       */
/* sw==1  if ordinary expression                                         */
/* sw==2  if the beginning of declarator and initlist                    */
/*************************************************************************/
PTR_LLND
convert_ll_node(pllnd,flag,sw)
PTR_LLND        pllnd;
int sw;
{
  PTR_LLND p,p1;
  PTR_TYPE n_type;
  int k;

  if (pllnd == NULL) return(NULL);
  if (flag ==1) {
    switch (pllnd->variant) {
    case PROC_CALL  :
    case FUNC_CALL  :
      mark_expression(pllnd,sw);
      break;
    case VAR_REF    :
    case FUNCTION_OP :               /* New added for VPC */
    default                 :
      convert_ll_node(pllnd->entry.template.ll_ptr1,flag,sw);
      convert_ll_node(pllnd->entry.template.ll_ptr2,flag,sw);
      break;
    }
   return(pllnd);
  }
  
 /* for flag ==2  (second pass)   */
  switch (pllnd->variant) {
     case INT_VAL    :
     case FLOAT_VAL  :
     case DOUBLE_VAL :
     case STRING_VAL :
     case CHAR_VAL   :
       break;
     case VAR_REF:
       k = is_call_by_reference(pllnd->type);
       if ((pllnd_pool[pllnd->id])&&k) return(pllnd);
       if (k){
	 p =make_llnd(fi,DEREF_OP,pllnd,NULL,NULL);	 
	 n_type = get_type(fi,T_POINTER) ;
	 n_type->entry.template.dummy1 = -1 ;
	 n_type->entry.template.base_type = pllnd->type ;
	 p->type = n_type;
	 return(p);
       }
       if (pllnd_pool[pllnd->id]){
	 p =make_llnd(fi,ADDRESS_OP,pllnd,NULL,NULL);	 
	 n_type = get_type(fi,T_POINTER) ;
	 n_type->entry.template.base_type = pllnd->type;
	 p->type =n_type;
	 return(p);
       }

       break;
     case EXPR_LIST:
       if ((pllnd_pool[pllnd->id])&&(!pllnd->entry.template.ll_ptr2)) {
	 p=follow_to_no_single_expr_list(pllnd);
	 k = is_call_by_reference(p->type);
	 if (k) {
	   p->entry.template.ll_ptr1=convert_ll_node(p->entry.template.ll_ptr1,flag,sw);
	   p->entry.template.ll_ptr2=convert_ll_node(p->entry.template.ll_ptr2,flag,sw);
	   return(pllnd);
	 }
	 else {
	   p1 =make_llnd(fi,ADDRESS_OP,p,NULL,NULL);	 
	   pllnd->entry.template.ll_ptr1 = p1;
	   n_type = get_type(fi,T_POINTER) ;
	   n_type->entry.template.base_type = p->type;
	   p1->type =n_type;
	   return(pllnd);
	 }
	}
      else {	 
	/*default */
       if (pllnd_pool[pllnd->id]) {
	 p =make_llnd(fi,ADDRESS_OP,pllnd,NULL,NULL);	 
	 n_type = get_type(fi,T_POINTER) ;
	 n_type->entry.template.base_type = pllnd->type;
	 p->type =n_type;
	 pllnd->entry.template.ll_ptr1=convert_ll_node(pllnd->entry.template.ll_ptr1,flag,sw);
	 pllnd->entry.template.ll_ptr2=convert_ll_node(pllnd->entry.template.ll_ptr2,flag,sw);
	 return(p);
       }
       else {
	 pllnd->entry.template.ll_ptr1=convert_ll_node(pllnd->entry.template.ll_ptr1,flag,sw);
	 pllnd->entry.template.ll_ptr2=convert_ll_node(pllnd->entry.template.ll_ptr2,flag,sw);
	 return(pllnd);
       }
     }
	 break;

     case FUNC_CALL:
     case ARRAY_REF  :
       k = is_call_by_reference(pllnd->type);
       if ((pllnd_pool[pllnd->id])&&k) {
	 pllnd->entry.template.ll_ptr1=convert_ll_node(pllnd->entry.template.ll_ptr1,flag,sw);	 
	 return(pllnd);
       }
       if (k){
	 p =make_llnd(fi,DEREF_OP,pllnd,NULL,NULL);	 
	 n_type = get_type(fi,T_POINTER) ;
	 n_type->entry.template.dummy1 = -1 ;
	 n_type->entry.template.base_type = pllnd->type ;
	 p->type = n_type;
	 pllnd->entry.template.ll_ptr1=convert_ll_node(pllnd->entry.template.ll_ptr1,flag,sw);	 
	 return(p);
       }       
       if (pllnd_pool[pllnd->id]){
	 p =make_llnd(fi,ADDRESS_OP,pllnd,NULL,NULL);	 
	 n_type = get_type(fi,T_POINTER) ;
	 n_type->entry.template.base_type = pllnd->type;
	 p->type =n_type;
	 pllnd->entry.template.ll_ptr1=convert_ll_node(pllnd->entry.template.ll_ptr1,flag,sw);
	 return(p);

       }
       pllnd->entry.template.ll_ptr1=convert_ll_node(pllnd->entry.template.ll_ptr1,flag,sw);
       return(pllnd);

     case RECORD_REF:
     case POINTST_OP:
       k = is_call_by_reference(pllnd->type);
       if ((pllnd_pool[pllnd->id])&&k) {
	 pllnd->entry.template.ll_ptr1=convert_ll_node(pllnd->entry.template.ll_ptr1,flag,sw);	 
	 convert_inner_node(pllnd->entry.template.ll_ptr2,flag,sw);	 
	 return(pllnd);
       }
       if (k){
	 p =make_llnd(fi,DEREF_OP,pllnd,NULL,NULL);	 
	 n_type = get_type(fi,T_POINTER) ;
	 n_type->entry.template.dummy1 = -1 ;
	 n_type->entry.template.base_type = pllnd->type ;
	 p->type = n_type;
	 pllnd->entry.template.ll_ptr1=convert_ll_node(pllnd->entry.template.ll_ptr1,flag,sw);	 
	 convert_inner_node(pllnd->entry.template.ll_ptr2,flag,sw);	 
	 return(p);
       }       
       if (pllnd_pool[pllnd->id]){
	 p =make_llnd(fi,ADDRESS_OP,pllnd,NULL,NULL);	 
	 n_type = get_type(fi,T_POINTER) ;
	 n_type->entry.template.base_type = pllnd->type;
	 p->type =n_type;
	 pllnd->entry.template.ll_ptr1=convert_ll_node(pllnd->entry.template.ll_ptr1,flag,sw);
	 convert_inner_node(pllnd->entry.template.ll_ptr2,flag,sw);
	 return(p);
       }
       pllnd->entry.template.ll_ptr1=convert_ll_node(pllnd->entry.template.ll_ptr1,flag,sw);
       convert_inner_node(pllnd->entry.template.ll_ptr2,flag,sw);
       return(pllnd);

     case FUNCTION_OP :               /* New added for VPC */
     case ARRAY_OP:
     default:
       if (pllnd_pool[pllnd->id]) {
	 p =make_llnd(fi,ADDRESS_OP,pllnd,NULL,NULL);	 
	 n_type = get_type(fi,T_POINTER) ;
	 n_type->entry.template.base_type = pllnd->type;
	 p->type =n_type;
	 pllnd->entry.template.ll_ptr1=convert_ll_node(pllnd->entry.template.ll_ptr1,flag,sw);
	 pllnd->entry.template.ll_ptr2=convert_ll_node(pllnd->entry.template.ll_ptr2,flag,sw);
	 return(p);
       }
       else {
	 pllnd->entry.template.ll_ptr1=convert_ll_node(pllnd->entry.template.ll_ptr1,flag,sw);
	 pllnd->entry.template.ll_ptr2=convert_ll_node(pllnd->entry.template.ll_ptr2,flag,sw);
	 return(pllnd);
       }
     }
  return(pllnd);
}


convert_inner_node(pllnd,flag,sw)
PTR_LLND pllnd;
{
  PTR_LLND p;

  
  /* for flag ==2  (second pass)   */
  switch (pllnd->variant) {
  case VAR_REF:
    break;
  case EXPR_LIST:
    p=follow_to_no_single_expr_list(pllnd);
    convert_inner_node(p,flag,sw);
    break;
  case FUNC_CALL:
  case ARRAY_REF  :
    pllnd->entry.template.ll_ptr1 = convert_ll_node(pllnd->entry.template.ll_ptr1,flag,sw);
    break;
  default:
    pllnd->entry.template.ll_ptr1 = convert_ll_node(pllnd->entry.template.ll_ptr1,flag,sw);
    pllnd->entry.template.ll_ptr2 = convert_ll_node(pllnd->entry.template.ll_ptr2,flag,sw);
  }
}





int
is_call_by_reference(passing_type)
PTR_TYPE passing_type;
{

  PTR_TYPE type;

  for ( type=passing_type ;type ; )
    {
	switch(type->variant) {
	case T_ARRAY :
	  type = type->entry.template.base_type ;
	  break;
	case T_FUNCTION :
	  type = type->entry.template.base_type ;
	  break;
	case T_POINTER  :
	  return(0);
	case T_REFERENCE:
	  return(1);
	case T_DESCRIPT :
	  type = type->entry.descriptive.base_type ;
	  break;
	case T_DERIVED_TYPE :
	  return(0);
       default :
	 type = (PTR_TYPE) NULL ;
	}
    }
   return(0);
}


/*************************************************************************/
/* Given a foo (a1,b1,c1) low level node and we mark a1, b1, c1 if       */
/* any of them is called by reference in the function foo                */
/*************************************************************************/
mark_expression(pllnd,sw)
PTR_LLND pllnd;
int sw;
{
  int i;
  PTR_LLND expr_list,p;
  PTR_SYMB functor,symb;
  PTR_LLND arg_list1,arg_list,p1;


  if (!pllnd) return;
  if (pllnd->variant != FUNC_CALL) return;

  /* the following is default for standard functions */
  expr_list = pllnd->entry.template.ll_ptr1;
  functor = pllnd->entry.template.symbol;
  symb = functor->entry.member_func.in_list;
  p =expr_list;

  if  (!strcmp(functor->ident,DOALLStringName))
    {
      arg_list = pllnd->entry.template.ll_ptr1;
      arg_list1= arg_list->entry.template.ll_ptr1;
      if (arg_list1->variant==VECTOR_CONST)
	{
	  p1= arg_list->entry.template.ll_ptr2;
	  functor = find_to_functor(p1);
	  arg_list1 = p1->entry.template.ll_ptr2;
	}
      else {
	functor = find_to_functor(arg_list1);
	arg_list1 = arg_list->entry.template.ll_ptr2;
      }
      symb = functor->entry.member_func.in_list;
      if (symb) symb = symb->entry.var_decl.next_in;
      p =arg_list1;
    }
  else if ((!strcmp(functor->ident,"DoSubSet"))||
	   (!strcmp(functor->ident,"DOSUBSET")))
     {
       arg_list = pllnd->entry.template.ll_ptr1;
       arg_list1= arg_list->entry.template.ll_ptr2;
       functor = find_to_functor(arg_list1);
       arg_list1 = arg_list1->entry.template.ll_ptr2;
       symb = functor->entry.member_func.in_list;
       p =arg_list1;      
      }

  for (    ;symb ;symb = symb->entry.var_decl.next_in)
    {
      if (is_call_by_reference(symb->type)){
	i =p->entry.template.ll_ptr1->id;
	pllnd_pool[i]=1;
      }
      if (p&&p->entry.template.ll_ptr1) 
	convert_ll_node(p->entry.template.ll_ptr1,1,sw);
      p=p->entry.template.ll_ptr2;
    }

}

/*************************************************************************/
/* Expecting format like                                                 */
/*    int  &x = c, &y=d, f =2;                                           */
/* Mark c, d low level node in this case                                 */
/*************************************************************************/  
mark_declarator_expression(pllnd)
PTR_LLND pllnd;
{
 int i;
 PTR_LLND expr_list,p1,p2,p;
 PTR_SYMB symbol;

 for (expr_list=pllnd;expr_list;expr_list = expr_list->entry.template.ll_ptr2)
    {
      p = expr_list->entry.template.ll_ptr1;
      if (p->variant != ASSGN_OP) continue;
      else {
	p1 = p->entry.template.ll_ptr1;
	p2 = p->entry.template.ll_ptr2;
	p = find_suit_declarator(p1);
	symbol = p->entry.template.symbol;
	if (is_call_by_reference(symbol->type)){
	  i =p2->id;
	  pllnd_pool[i]=1;
	}	
      }
    }
}

/* 
  second pass:
  change 
       int &x=c, &y=d, f=2; 
   ==>
       int *x=c, *y=d, f=2;                  
*/
PTR_LLND
convert_declarator_expression(pllnd)
PTR_LLND pllnd;
{
 int i;
 PTR_LLND expr_list,p1,p2,p;
 PTR_SYMB symbol;

 for (expr_list=pllnd;expr_list;expr_list = expr_list->entry.template.ll_ptr2)
    {
      p = expr_list->entry.template.ll_ptr1;
      for ( ; p ;  )
	{
	  switch (p->variant) {
	  case BIT_NUMBER:
	  case ASSGN_OP :
	    p->entry.template.ll_ptr2= convert_ll_node(p->entry.template.ll_ptr2,2,1);
	    p = p->entry.template.ll_ptr1;
	    break;
	  case ARRAY_OP:
	  case FUNCTION_OP :
	  case CLASSINIT_OP:
	  case DEREF_OP : 
	    p = p->entry.template.ll_ptr1;
	    break;
	  case FUNCTION_REF:
	  case ARRAY_REF:
	  case VAR_REF:
	    p = NULL;
	    break ;
	  case ADDRESS_OP:
	    p->variant = DEREF_OP;
	    p =NULL;
	    break;
	  }
	  
	}

    }
 return(pllnd);
}





/*************************************************************************/
/* Expecting format like                                                 */
/*    return(p) or return (*q)                                           */
/* Mark p or *q if the return value of the functor is of reference type  */
/*************************************************************************/  
mark_return_expression(pbf,pllnd)
PTR_LLND pllnd;
PTR_BFND pbf;
{
 int i;
 PTR_LLND expr_list,p1,p2,p;
 PTR_SYMB functor;
 PTR_BFND bf;


 for (bf= pbf;(bf->variant!=FUNC_HEDR) ; bf= bf->control_parent) ;
 functor = bf->entry.template.symbol;
 if (is_call_by_reference(functor->type))
  {
    i =pllnd->id;
    pllnd_pool[i]=1;
  }	 
}




alloc_temp()
{
  /* To avoid boundary +1 is needed at least */
   pllnd_pool=(int *) calloc(1,sizeof(char *)*(fi->num_llnds+4)); 
  

}

free_temp()
{
  free(pllnd_pool);
}




PTR_LLND
follow_to_no_single_expr_list(pllnd)
PTR_LLND pllnd;
{
  PTR_LLND expr_list;

  for (expr_list=pllnd;expr_list;expr_list=expr_list->entry.template.ll_ptr1)
    if ((expr_list->variant!=EXPR_LIST)||
	(expr_list->entry.template.ll_ptr2)) return(expr_list);

  return(expr_list);

}



/*************************************************************************/
/* expand_expr_for_structure_fun_ref(global_bfnd)                        */
/* (1) For the expression                                                */
/*          foo( & goo(..), int  arg2,int  arg3);                        */
/*    where                                                              */
/*       goo is declared as struct object goo(....);                     */
/*       foo is declared as foo(object &, int arg2, int arg3);           */
/*                                                                       */
/*    transformed into                                                   */
/*     {                                                                 */
/*       struct object H1;                                               */
/*        foo((h1=goo(..), &h1), arg2,arg3);                             */
/*     }                                                                 */
/*    Or                                                                 */
/*             foo( & f.goo(..), int  arg2,int  arg3);                   */
/*    transformed into                                                   */
/*     {                                                                 */
/*       struct object H1;                                               */
/*        foo((h1=f.goo(..), &h1), arg2,arg3);                           */  
/*     }                                                                 */
/*                                                                       */
/* (2) This routine have to be used very late, since we need to have H1  */
/*     to be realy local variables in the parallel thread                */
/*************************************************************************/

expand_expr_for_struct_fun_ref(pbf)
PTR_BFND pbf;
{
   
  PTR_BLOB blob;
  PTR_BFND fbf;

  if (!pbf) return;
  if (pbf->variant == CONTROL_END) return;
  for(blob = pbf->entry.template.bl_ptr1;blob; blob=blob->next)
    expand_expr_for_struct_fun_ref(blob->ref);
  for(blob = pbf->entry.template.bl_ptr2;blob; blob=blob->next)
    expand_expr_for_struct_fun_ref(blob->ref);
  expand_struct_fun_ref_llnd(pbf,pbf->entry.template.ll_ptr1);
  expand_struct_fun_ref_llnd(pbf,pbf->entry.template.ll_ptr2);
  expand_struct_fun_ref_llnd(pbf,pbf->entry.template.ll_ptr3);
}

expand_struct_fun_ref_llnd(pbf,pllnd)
PTR_BFND pbf;
PTR_LLND  pllnd;
{
  PTR_LLND p1,p2,p,n1,expr_list;
  PTR_SYMB symb;


  if (!pllnd) return;
  expand_struct_fun_ref_llnd(pbf,pllnd->entry.template.ll_ptr1);
  expand_struct_fun_ref_llnd(pbf,pllnd->entry.template.ll_ptr2);
  if (pllnd->variant== FUNC_CALL) {
    expr_list = pllnd->entry.template.ll_ptr1;
    for (p =expr_list;p;p=p->entry.template.ll_ptr2)
    {
      p1= p->entry.template.ll_ptr1;
      if ((p1->variant==ADDRESS_OP)&& (n1=p1->entry.template.ll_ptr1)&&
	  (n1->variant==FUNC_CALL)) {
	symb= n1->entry.template.symbol;
	if (is_complicated_type(symb->type))
	  intro_temp_vars_for_func_ref(pbf,p1,symb);
      }
      else if ((p1->variant==ADDRESS_OP)&&
	       (n1=p1->entry.template.ll_ptr1)&&
	       ((n1->variant==RECORD_REF)||(n1->variant==POINTST_OP))&&
	       (n1->entry.template.ll_ptr2) &&
	       (n1->entry.template.ll_ptr2->variant==FUNC_CALL)){
	symb= n1->entry.template.ll_ptr2->entry.template.symbol;
	if (is_complicated_type(symb->type))
	  intro_temp_vars_for_func_ref(pbf,p1,symb);
      }
    }
  }
}


is_complicated_type(type1)
PTR_TYPE type1;
{
  PTR_TYPE t;
  int sw,level;

  t= follow_type_2(type1,&sw,&level);
  if (level)  return(0);
  switch(t->variant) {
  case T_INT:
  case T_FLOAT:
  case T_DOUBLE:
  case T_CHAR:
  case T_VOID:
  case DEFAULT:
    return(0);
  default:
    return(1);
  }
}

/*
  n1  is equivalent to ADDRESS_OP and &goo(..) in
             foo( & goo(..), int  arg2,int  arg3);    
    transformed into                                                   
     {                                                                 
       struct object H1;                                               
        foo((h1=goo(..), &h1), arg2,arg3);                             
     }         
   or
             foo( & f.goo(..), int  arg2,int  arg3);    
    transformed into                                                   
     {                                                                 
       struct object H1;                                               
        foo((h1=f.goo(..), &h1), arg2,arg3);                             
     }         
                                                       
  pbf is the current bif node

*/
intro_temp_vars_for_func_ref(pbf,n1,functor)
PTR_BFND pbf;
PTR_LLND n1;
PTR_SYMB functor;
{
  
  PTR_SYMB symb;
  PTR_LLND p1,pllnd,expr_list,func_llnd;
  PTR_BFND root,bf1;
  PTR_TYPE pointer_to_type;

  /* printf("we enter intro_vars_for_func_ref \n"); */
  /* generate proper info */
  root=the_closest_root(pbf);
  symb=gen_next_symbol(root);
  
  /* expand expression */
  func_llnd = n1->entry.template.ll_ptr1;

  symb->type = functor->type;
  pointer_to_type = get_type(fi,T_POINTER);
  pointer_to_type->entry.template.base_type = functor->type;
 
  p1=make_llnd(fi,VAR_REF,NULL,NULL,symb);  
  p1->type= functor->type;
  p1=make_llnd(fi,ADDRESS_OP,p1,NULL,NULL);
  p1->type= pointer_to_type;
  expr_list =make_llnd(fi,EXPR_LIST,p1,NULL,NULL);  
  expr_list->type= pointer_to_type;
  p1=make_llnd(fi,VAR_REF,NULL,NULL,symb);  
  p1->type = symb->type;
  pllnd =make_llnd(fi,ASSGN_OP,p1,n1->entry.template.ll_ptr1,NULL);  
  pllnd->type = symb->type;

  n1->variant=EXPR_LIST;
  n1->entry.template.ll_ptr1=pllnd;
  n1->entry.template.ll_ptr2=expr_list;
  n1->type= pointer_to_type;
  /* intro vars decl */
  p1=make_llnd(fi,VAR_REF,NULL,NULL,symb);  
  symb->type = functor->type;
  expr_list =make_llnd(fi,EXPR_LIST,p1,NULL,NULL);
  bf1=make_bfnd(fi,VAR_DECL,SMNULL,expr_list,NULL,NULL);
  chain_decl_up(root,bf1);

}


PTR_BFND
the_closest_root(pbf)
PTR_BFND pbf;
{
  PTR_BFND p;

  for (p = pbf;((p->variant != FUNC_HEDR)&&
		(p->variant != COEXEC_NODE)&&
		(p->variant != COLOOP_NODE)&&
		(p->variant != GLOBAL)) ; p = p->control_parent) ;
  if (p->variant== GLOBAL) warning("too complicated initial expressions");
  return (p);
}



PTR_SYMB
gen_next_symbol(pbf)
PTR_BFND pbf;
{
  int   index;
  char  buff1[1000];
  char *s;
  PTR_SYMB symb;

  index =full_temp_counter++;
  sprintf(buff1,"__H_i_%d",index);
  s= copys(buff1);
  symb= make_symb(fi,VARIABLE_NAME,s);
  symb->scope = pbf;
  return(symb);
}


chain_decl_up(root,bf1)
PTR_BFND root,bf1;
{
  PTR_SYMB functor;
  PTR_BLOB old_blob,blob,next_blob;

  switch(root->variant) {
  case FUNC_HEDR:
    functor=root->entry.template.symbol;
    for (old_blob=NULL,blob=root->entry.template.bl_ptr1;blob;blob=blob->next)
      {
           if (!is_param_decl(blob->ref,functor)) break;
	   old_blob=blob;
      }
    if (!old_blob) {
      next_blob = root->entry.template.bl_ptr1;
      root->entry.template.bl_ptr1 = make_blob(fi,bf1,next_blob);
      bf1->control_parent=root;
    }
    else {
      next_blob = root->entry.template.bl_ptr1;
      old_blob->next = make_blob(fi,bf1,next_blob);
      bf1->control_parent=root;
    }
    break;
  case COLOOP_NODE:
  case COEXEC_NODE:
    next_blob = root->entry.template.bl_ptr1;
    root->entry.template.bl_ptr1 = make_blob(fi,bf1,next_blob);
    bf1->control_parent=root;
  }
}






int
is_param_decl(var_bf, functor)
PTR_BFND var_bf ;
PTR_SYMB functor ;
{
  PTR_LLND flow_ptr,lpr ;
  int i ;
  PTR_SYMB s ;

   switch (var_bf->variant) {
   case VAR_DECL :
   case ENUM_DECL:
   case CLASS_DECL:
   case UNION_DECL:
   case STRUCT_DECL:
   case DERIVED_CLASS_DECL :
      lpr = var_bf->entry.template.ll_ptr1 ;
      for ( flow_ptr = lpr;flow_ptr ;flow_ptr = flow_ptr->entry.template.ll_ptr1)
       {  if ((flow_ptr->variant == VAR_REF) ||
              (flow_ptr->variant == ARRAY_REF) ||
              (flow_ptr->variant == FUNCTION_REF) ) break ;
        }
      if (!flow_ptr) { /* err("system crash");  */
                        return(NULL) ;
                     }

      for ( i=0, s = functor->entry.member_func.in_list ; s ; i++ )
      { 
            if (flow_ptr->entry.template.symbol == s) return(1);
            s = s->entry.var_decl.next_in ;
       }
      return(0);

      default :   return(0) ;
    }

}


/*************************************************************************/
/*                                                                       */
/* Normalize the overloading function call expressiion                   */
/* A op B  ==>  A.op(B)                                                  */
/*                                                                       */
/*************************************************************************/
normalize_overloading_expr(pbf)
PTR_BFND pbf;
{
  PTR_BLOB blob;
  PTR_BFND fbf;

  if (!pbf) return;
  if (pbf->variant == CONTROL_END) return;
  for(blob = pbf->entry.template.bl_ptr1;blob; blob=blob->next)
    normalize_overloading_expr(blob->ref);
  for(blob = pbf->entry.template.bl_ptr2;blob; blob=blob->next)
    normalize_overloading_expr(blob->ref);
  switch (pbf->variant) {
  case IF_NODE    :
    normalize_llnd(pbf->entry.template.ll_ptr1);
    break;
  case FOR_NODE   :
    normalize_llnd(pbf->entry.template.ll_ptr1);
    normalize_llnd(pbf->entry.template.ll_ptr2);
    normalize_llnd(pbf->entry.template.ll_ptr3);
    break;
  case WHILE_NODE :
    normalize_llnd(pbf->entry.template.ll_ptr1);
    break;
  case RETURN_STAT:
  case RETURN_NODE  :             /* New added for VPC */
    normalize_llnd(pbf->entry.template.ll_ptr1);
    break;
  case DO_WHILE_NODE:             /* New added for VPC */
    normalize_llnd(pbf->entry.template.ll_ptr1);
    break;
  case SWITCH_NODE :              /* New added for VPC */
    normalize_llnd(pbf->entry.template.ll_ptr1);
    break ;
  case COLOOP_NODE  :             /* New added for VPC */
    normalize_llnd(pbf->entry.template.ll_ptr2);
    break;
  case COEXEC_NODE  :             /* New added for VPC */
    normalize_llnd(pbf->entry.template.ll_ptr1);
    break;
  case EXPR_STMT_NODE:
    normalize_llnd(pbf->entry.template.ll_ptr1);
    break;
  default:
    break;
  }
}




normalize_llnd(pllnd)
PTR_LLND pllnd;
{
  PTR_LLND p1,p2,n1,n2;

  if (!pllnd) return;
  p1=pllnd->entry.template.ll_ptr1;
  p2=pllnd->entry.template.ll_ptr2;
  n1=NULL;
  n2=NULL;
  switch(pllnd->variant) {
   case INT_VAL    :
   case FLOAT_VAL  :
   case DOUBLE_VAL :
   case STRING_VAL :
   case CHAR_VAL   :
     return;
   case THIS_NODE:
   case VAR_REF    :
   case RECORD_REF:		    
   case DDOT       :
   case COPY_NODE :
   case VECTOR_CONST : 
   case INIT_LIST:
   case BIT_NUMBER:
     break;
   case ARRAY_REF  :
   case FUNC_CALL  :
     if (p2)  {
       n2=make_llnd(fi,pllnd->variant,p1,NULL,p2->entry.template.symbol);
       pllnd->variant = RECORD_REF;
       p2->entry.template.symbol = pllnd->entry.template.symbol;
       p2->variant= VAR_REF;
       p2->entry.template.ll_ptr1=NULL;
       p2->entry.template.ll_ptr2=NULL;
       pllnd->entry.template.ll_ptr1=p2;
       pllnd->entry.template.ll_ptr2=n2;
       pllnd->variant=RECORD_REF;
       normalize_llnd(pllnd->entry.template.ll_ptr1);
       normalize_llnd(pllnd->entry.template.ll_ptr2);
       return;
     }
     break;
   case MINUSMINUS_OP:        
   case PLUSPLUS_OP  :        
     if (pllnd->entry.template.symbol) {
       if (p1) {
	 n2=make_llnd(fi,FUNC_CALL,NULL,NULL,pllnd->entry.template.symbol);
	 pllnd->variant = RECORD_REF;
	 pllnd->entry.template.symbol=NULL;
	 pllnd->entry.template.ll_ptr1 = pllnd->entry.template.ll_ptr1;
	 pllnd->entry.template.ll_ptr2 = n2;
	 normalize_llnd(pllnd->entry.template.ll_ptr1);
       }
       else {
	 n2=make_llnd(fi,FUNC_CALL,NULL,NULL,pllnd->entry.template.symbol);
	 pllnd->variant = RECORD_REF;
	 pllnd->entry.template.symbol=NULL;
	 pllnd->entry.template.ll_ptr1=pllnd->entry.template.ll_ptr2;
	 pllnd->entry.template.ll_ptr2 = n2;
	 normalize_llnd(pllnd->entry.template.ll_ptr1);
       }
       return;
     }
     break;
   case EXPR_LIST  :
   case CLASSINIT_OP :  
   case DELETE_OP:
   case SCOPE_OP:
   case NEW_OP:
   case CAST_OP :             
   case EXPR_IF      :        
   case EXPR_IF_BODY :        
   case FUNCTION_REF :        
     break;
   case FUNCTION_OP :         
   case ARRAY_OP :
     if (pllnd->entry.template.symbol) {
       if (p2) n1=p2;
       n2=make_llnd(fi,FUNC_CALL,n1,NULL,pllnd->entry.template.symbol);
       pllnd->variant = RECORD_REF;
       pllnd->entry.template.symbol=NULL;
       pllnd->entry.template.ll_ptr1 = pllnd->entry.template.ll_ptr1;
       pllnd->entry.template.ll_ptr2 = n2;
       normalize_llnd(pllnd->entry.template.ll_ptr1);
       normalize_llnd(pllnd->entry.template.ll_ptr2);
       return;
     }
     break;
   default:
     if (pllnd->entry.template.symbol) {
       if (p2) n1=make_llnd(fi,EXPR_LIST,p2,NULL,SMNULL);
       n2=make_llnd(fi,FUNC_CALL,n1,NULL,pllnd->entry.template.symbol);
       pllnd->variant = RECORD_REF;
       pllnd->entry.template.symbol=NULL;
       pllnd->entry.template.ll_ptr1 = pllnd->entry.template.ll_ptr1;
       pllnd->entry.template.ll_ptr2 = n2;
       normalize_llnd(pllnd->entry.template.ll_ptr1);
       normalize_llnd(pllnd->entry.template.ll_ptr2);
       return;
     }
   }
  
  normalize_llnd(pllnd->entry.template.ll_ptr1);
  normalize_llnd(pllnd->entry.template.ll_ptr2);
  
}






/*************************************************************************/
/* check_usage_of_elementtype(pbf)                                       */
/*                                                                       */
/* pbf:  ElementType can not be used outside the collection              */
/*************************************************************************/
check_usage_of_elementtype(pbf,controller)
PTR_BFND pbf;
PTR_SYMB controller;
{
  PTR_BLOB blob;
  PTR_SYMB new_control,functor,base_name;
  
  if (!pbf) return;
  new_control = controller;
  switch (pbf->variant) {
    case FUNC_HEDR:
      functor = pbf->entry.template.symbol;
      if (functor->variant==MEMBER_FUNC) {
	base_name = functor->entry.member_func.base_name;
	if (base_name->variant== COLLECTION_NAME) return;
      }
      new_control=functor;
      check_param_usage_of_elementtype(functor->entry.member_func.in_list,new_control);            
      break;
    case VAR_DECL:
      check_declarator_usage_of_elementtype(pbf->entry.template.ll_ptr1,new_control);      
      break;
    case COLLECTION_DECL:
      return;
    }
  for(blob = pbf->entry.template.bl_ptr1;blob; blob=blob->next)
    check_usage_of_elementtype(blob->ref,new_control);
  for(blob = pbf->entry.template.bl_ptr2;blob; blob=blob->next)
    check_usage_of_elementtype(blob->ref,new_control);
  check_castop_usage_of_elementtype(pbf->entry.template.ll_ptr1,new_control);
  check_castop_usage_of_elementtype(pbf->entry.template.ll_ptr2,new_control);
  check_castop_usage_of_elementtype(pbf->entry.template.ll_ptr3,new_control);

}




/*************************************************************************/
/*   check_castop_usage_of_elementtype(pllnd)                            */
/*************************************************************************/
check_castop_usage_of_elementtype(pllnd,functor)
PTR_LLND        pllnd;
PTR_SYMB functor;
{

  if (!pllnd) return;
  if (pllnd->variant==CAST_OP)
    examine_elementtype(pllnd->type,functor);
  check_castop_usage_of_elementtype(pllnd->entry.template.ll_ptr1,functor);
  check_castop_usage_of_elementtype(pllnd->entry.template.ll_ptr2,functor);
}


check_declarator_usage_of_elementtype(pllnd,functor)
PTR_LLND pllnd;
PTR_SYMB functor;
{
  PTR_TYPE type;
  PTR_LLND p;

  p= find_suit_declarator(pllnd);
  if (p->entry.template.symbol) type =p->entry.template.symbol->type;
  else  type = p->type;
  examine_elementtype(type,functor);
}

check_param_usage_of_elementtype(s,functor)
PTR_SYMB s,functor;
{

  PTR_SYMB   next;

  for (next=s; next; next=next->entry.var_decl.next_in)
    examine_elementtype(next->type,functor);

}

examine_elementtype(q,functor)
PTR_TYPE q;
PTR_SYMB functor;
{
  int sw,level;
  PTR_TYPE type_a,base;
  
  type_a= follow_type_2(q,&sw,&level);
  if ((type_a->variant==T_DERIVED_TYPE)&&
      (!strcmp(type_a->entry.derived_type->ident,"ElementType"))){
    if (functor)
      warning_functor(functor->ident,"illegal usage of ElementType");
    else 
      warning_functor("global scoping","illegal usage of ElemenType");
  }
  else if (type_a->variant==T_DERIVED_COLLECTION) {
    base= type_a->entry.col_decl.base_type;
    if ((base->variant==T_DERIVED_TYPE)&&
      (!strcmp(base->entry.derived_type->ident,"ElementType"))){    
      if (functor)
	warning_functor(functor->ident,"illegal usage of ElementType");
      else 
	warning_functor("global scoping","illegal usage of ElemenType");
    }
  }
}


/*
   clean_void_argument(pbf)

    C++ class allows users to specify arguments as void to
    mean there is no argument, this routine eliminate
    the  argument
*/
clean_void_argument(fi)
PTR_FILE fi;
{
  PTR_SYMB symb1, s1,s2;
  PTR_TYPE type;
  int level,sw;

  for ( symb1=fi->head_symb; symb1; symb1=symb1->thread)
    {
      if ((symb1->variant==FUNCTION_NAME)|| 
	  (symb1->variant==MEMBER_FUNC))
	{
	    s1 = symb1->entry.member_func.in_list;	  
	    if (s1)  s2=s1->entry.var_decl.next_in;
	    if (s1 &&(!s2) ){
	      type =follow_type_2(s1->type,&level,&sw);
	      if ((type->variant==T_VOID)&&(level==0))
		symb1->entry.member_func.in_list = (PTR_SYMB) NULL;
	    }
	  }
    }
}



