/* $Header: /Users/nsundare/faust/VPC/RCS/verrors.c,v 2.0 92/04/02 21:24:44 nsundare Exp $ */

/*
 * errors.c
 *
 * Miscellaneous error routines
 */

#include <stdio.h>
#include "db.h"
#include "vextern.h"

static char buff[100];


/*
 * fatal -- print the fatal error message then exit
 *
 * input:
 *	  s - the message to be printed out
 */
void 
fatal(s)
	char   *s;
{
	fprintf(stderr, "Compiler error line %d of %s: %s\n", yylineno, line_pos_fname, s);
	exit(3);
}


/*
 * fatalstr -- sets up the error message according to the given format
 *		then call "fatal" to print it out
 *
 * input:
 *	  t - a string that specifies the output format
 *	  s - a string that contents the error messaged to be formatted
 */
void 
fatalstr(t, s)
	char   *t, *s;
{
/* VARARGS */ 
	sprintf(buff, t, s);
	fatal(buff);
}


/*
 * fatali -- formats a fatal error message which contains a number
 *	     then call "fatal" to print it
 *
 * input:
 *	  t - a string that specifies the output format
 *	  d - an integer to be converted according to t
 */
void 
fatali(t, d)
	char   *t;
	int	d;
{
/* VARARGS */ 
	sprintf(buff, t, d);
	fatal(buff);
}


/*
 * warn1 -- formats a warning message then call "warn" to print it out
 *
 * input:
 *	  s - string that specifies the conversion format
 *	  t - string that to be converted according to s
 */
void 
warn1(s, t)
	char   *s, *t;
{
/* VARARGS */ 
	sprintf(buff, s, t);
	warn(buff);
}


/*
 * warn -- print the warning message if specified
 *
 * input:
 *	  s - string to be printed
 */
warn(s)
	char   *s;
{
	if (!nowarnflag) {
		++nwarn;
		fprintf(stderr, "Warning on line %d of %s: %s\n", yylineno, line_pos_fname, s);
	}
}


/*
 * errstr -- formats the non-fatal error message then call "err" to print it
 *
 * input:
 *	  s - string that specifies the conversion format
 *	  t - string that to be formated according to s
 */
void 
errstr(s, t)
	char   *s, *t;
{
/* VARARGS */ 
	sprintf(buff, s, t);
	err(buff);
}


/*
 * errnm -- format an error number and print it
 *
 * input:
 *	  fmt - string that specified the format
 *	  l   - length of the string "s"
 *	  s   - string to be converted according to "fmt"
 */
void 
errnm(fmt, l, s)
	char   *fmt;
	int	l;
	register char *s;
{
	char	buf[50];
	register int i;

	for (i = 0; i < l; i++)
		buf[i] = s[i];
	buf[i] = '\0';

	errstr(fmt, buf);
}


/*
 * warnnm -- formats a warning message then prints it
 *
 * input:
 *	  fmt - string that specifies the output format
 *	  l   - length of the string "s"
 *	  s   - string to be formatted according to "fmt"
 */
void 
warnnm(fmt, l, s)
	char   *fmt;
	int	l;
	register char *s;
{
	register int i;
	char	buf[50];

	for (i = 0; i < l; i++)
		buf[i] = s[i];
	buf[i] = '\0';

	warn1(fmt, buf);
}


/*
 * erri -- formats an error number then prints it out
 *
 * input:
 *	  s - string that specifies the output format
 *	  t - error number to be formatted
 */
erri(s, t)
	char   *s;
	int	t;
{
/* VARARGS */ 
	sprintf(buff, s, t);
	err(buff);
}

extern int errors_cmnt;

/*
 * err -- prints the error message and does the bookkeeping
 *
 * input:
 *	  s - string to be printed out
 */
err(s)
	char   *s;
{
	FILE *vpc_errors, *fopen();
	if (! errcnt++)
		errline = yylineno;
	if(errors_cmnt)
		fprintf(stderr,"Error on line %d of %s:  %s\n", yylineno, line_pos_fname,s);
	
	exit(1);
}

/*
 * err_symb -- prints the error message and does the bookkeeping
 *
 * input:
 *	  s - string to be printed out
 *        s2- symb->ident to be printed
 */
err_symb(s,s2)
	char   *s;
        char *s2;
{
	FILE *vpc_errors, *fopen();
	if (! errcnt++)
		errline = yylineno;
	if(errors_cmnt)
		fprintf(stderr,"Error on line %d of %s:symbol \'%s\'  %s\n", yylineno, line_pos_fname,s2,s);
	
	exit(1);
}

/*
 *	  s - the error message to be printed
 */
void 
yyerror(s)
	char   *s;
{
	err(s);
}


/*
 * dclerr -- prints the error message when find error in declaration part
 *
 * input:
 *	  s - error message string
 *	  v - pointer to the symble table entry
 */
void 
dclerr(s, v)
	char   *s;
	PTR_SYMB v;
{
	char	buff[100];

	if (v) {
/* VARARGS */ 
		sprintf(buff, "Declaration error for %s: %s", v->ident, s);
		err(buff);
	} else
		errstr("Declaration error: %s", s);
}


/*
 * execerr -- prints error message for executable part
 *
 * input:
 *	  s - the error message string1
 *	  n - the error message string2
 */
void 
execerr(s, n)
	char   *s, *n;
{
	char	buf1[100], buf2[100];

/* VARARGS */ 
	sprintf(buf1, "Execution error: %s", s);
/* VARARGS */ 
	sprintf(buf2, buf1, n);
	err(buf2);
}


/*  moved from main.c */
warning(s)
char *s;
{
fprintf(stderr,"Warning on line %d of %s:  %s\n", yylineno, line_pos_fname,s); 

}


warning_symb(s,s2)
	char   *s;
        char *s2;
{
  fprintf(stderr,"Warning on line %d of %s:symbol \'%s\'  %s\n", yylineno, line_pos_fname,s2,s);

}


warning_functor(s,s2)
	char   *s;
        char *s2;
{
  fprintf(stderr,"Error inside function %s: %s\n",s,s2);

}



/*  moved from main.c */
info(s)
char *s;
{
fprintf(stderr,"Info: on line %d of %s:  %s\n", yylineno, line_pos_fname,s); 

}
