/* Employed by Jenq-Kuen Lee for VPC++ Nov 19 ,1987 */
/* Original Filename : hash.c                       */
/* New filename      : vhash.c                      */
/* $Header: /Users/nsundare/faust/VPC/RCS/vhash.c,v 2.0 92/04/02 21:24:46 nsundare Exp $ */

/*
 * hash.c -- hash table routines
 */

#include <stdio.h>
#include <strings.h>

#include "tag"
#include "vpc.h"
#include "db.h"
#include "vextern.h"

PTR_LLND find_suit_declarator();
PTR_SYMB find_object_symbol();
PTR_SYMB check_field_decl_3();
PTR_BFND find_current_class_bfnd();
PTR_BFND find_class_bfnd();
PTR_SYMB make_symb();
PTR_BLOB follow_blob();
PTR_LLND follow_llnd();
PTR_SYMB re_install_entry();
PTR_HASH hash_table[hashMax];
PTR_BFND cur_scope(),pre_scope(),real_scope() ;

/*
 * init_hash -- initialize the hash table
 */
void
init_hash()
{
	register int i;

	for (i = 0; i < hashMax; i++)
		hash_table[i] = HSNULL;
}


/*
 * Hash(string) -- compute hash value of string.
 */
int 
hash(string)
	register char *string;
{
	register int i;

	for (i = 0; *string;)
		i += *string++;
	return (i % hashMax);
}


/*
 * look_up(string) -- lookup string in the hash table and
 *		      install it if not there
 */
PTR_HASH 
look_up(string)     
	register char *string;
{
	int     index;
	register PTR_HASH entry;

	index = hash(string);
	for (entry = hash_table[index]; entry; entry = entry->next_entry)
		if (!strcmp(string, entry->ident))
			break;

	if (!entry) {	 /* since hash entry was not found make one */
		entry = ALLOC(hash_entry);
		entry->ident = copys(string);
		entry->next_entry = hash_table[index];
		hash_table[index] = entry;
	}
	return (entry);
}


/*
 * look_up_type(string,type_status) 
 *     <1> lookup string in the hash table and install it if not there
 *         exactly same behavior like look_up
 *     <2> type_status called by out
 *         ON  : If the string has been a typename already.
 *         OFF : If not foumd or Has been found but not typename
 *
 */
PTR_HASH 
look_up_type(string,type_status)     
	register char *string;
        int *type_status ;
{
	int     index;
	register PTR_HASH entry;
        PTR_SYMB symb_ptr;



        *type_status = OFF;
	index = hash(string);
	for (entry = hash_table[index]; entry; entry = entry->next_entry)
		if (!strcmp(string, entry->ident))
			break;

	if (!entry) {	 /* since hash entry was not found make one */
		entry = ALLOC(hash_entry);
		entry->ident = copys(string);
                entry->id_attr = SMNULL ;
		entry->next_entry = hash_table[index];
		hash_table[index] = entry;
	}
        else {
	  for (symb_ptr =entry->id_attr ;(symb_ptr && (! (*type_status))); symb_ptr = symb_ptr->outer)
	   { 	  if ((symb_ptr->variant==TYPE_NAME)||
		     (symb_ptr->variant == VARIABLE_NAME)||
		     (symb_ptr->variant == UNION_NAME)||
		     (symb_ptr->variant == CLASS_NAME)||
		     (symb_ptr->variant == COLLECTION_NAME)||
		     (symb_ptr->variant == STRUCT_NAME)||
                     (symb_ptr->variant == ENUM_NAME))
		       {  PTR_BFND flow ;
                          PTR_BFND type_scope ;                    
                          
                          type_scope = real_scope(symb_ptr->scope) ;
                          for ( flow = cur_scope(); flow ; flow = pre_scope(flow) )
                           {   if (flow == type_scope) 
                                {   
				  if ((symb_ptr->variant == TYPE_NAME)||(is_symbol_an_object(symb_ptr->variant)))
				    *type_status = ON ;
				  return(entry);
                                }
                           }
                       }
            }     
	}

	return (entry);
      }






/* Given a TYPENAME hash_entry, to find the exact symbol for this type
 * flag : OFF  stands for find a id don't care any kind of id
 *
 */
PTR_SYMB
find_type_symbol(entry,flag,regions)
PTR_HASH entry ;
int  flag ;
int regions;
{  PTR_SYMB symb_ptr ,old_symb ;
   int type_status ;
   PTR_BFND real_scope();

  type_status= OFF ;
  for (old_symb=symb_ptr =entry->id_attr ;(symb_ptr && (! type_status)); symb_ptr = symb_ptr->outer)
   { 
    if (is_same_variant_group(symb_ptr->variant,flag) || (flag==OFF))
      {  PTR_BFND flow ;
         PTR_BFND type_scope ;                    
       
         type_scope = real_scope(symb_ptr->scope) ;
         for ( flow = cur_scope(); flow ; flow = pre_scope(flow) )
            if (flow == type_scope) 
             { 
               if ((symb_ptr->variant== FIELD_NAME) &&
                   ((regions==ELEMENT_ONLY)&&
                     (!(symb_ptr->attr & ELEMENT_FIELD))))
		 {
                     /*no match yet */
                 }
               else  {
                    type_status = ON ;
                    break ;
                }
	     }
       }
            old_symb = symb_ptr ;
  }
     if (type_status == ON)  return(old_symb);
     else {
         /* default integer function */
         return(SMNULL);
       }

 }

/* find_type_symbol_by_inherited_class
 * 
 * (1) Find the possible elements through inherited class.
 * (2) Find elements through super collection.
 *  range: (Same statuis flag as check_field_decl_3())
 *    ALL_FIELDS: check all the legal fields for the object.
 *    CLASS_ONLY: check only for the ordinary C++ object field for the object
 *    ELEMENTS_ONLY: check methods of elements described in a collection only.
 *
 */
PTR_SYMB
fts_by_inherited_class(entry,flag,range)
PTR_HASH entry ;
int  flag ;
int range;
{  
  PTR_LLND ll;
  PTR_SYMB symb_ptr ,old_symb,found_symb,object,first,base_name;
  int type_status,status,sw ;
  PTR_LLND expr_list,list,list1,pllnd;
  PTR_BFND pbf;
  PTR_BLOB blob;

  pbf= find_current_class_bfnd(pred_bfnd,&status);
  if (!pbf) return(SMNULL);
  if (status == ON) {
    blob = pbf->entry.template.bl_ptr1;
    ll= find_suit_declarator(blob->ref->entry.template.ll_ptr1);   
    first = ll->entry.template.symbol;
    for (; first  ;  )
      { if ((!strcmp(first->ident,entry->ident))&&
	    ((range != ELEMENT_ONLY)||
	     ((range==ELEMENT_ONLY)&&(first->attr & ELEMENT_FIELD))))  break;
        else
	  {
	    if (first->variant == FIELD_NAME)
	      first = first->entry.field.next;
	    else
	      first = first->entry.member_func.next;
	  }
      }
    if (first) {
      if (first->variant == MEMBER_FUNC)
	 return(first);
      else            return(first);
    }
  }    
 if (expr_list =(pbf->entry.template.ll_ptr2)) {
   for ( list = expr_list; list; list =list->entry.template.ll_ptr2)
     { list1  = follow_llnd(list,1);
       object = list1->entry.template.symbol;
       if (found_symb=check_field_decl_3(object,entry,0,range)) return(found_symb);
     }
  }
  if ((range==ALL_FIELDS)&&(pllnd=pbf->entry.template.ll_ptr3))   {
    base_name = pllnd->entry.template.symbol;
    return(check_field_decl_3(base_name,entry,sw,ELEMENT_ONLY));
    
  }
  
 return(SMNULL);

}




/* Given a TYPENAME hash_entry, to find the object base for this type
 * flag : OFF  stands for find a id don't care any kind of id
 *
 */
PTR_SYMB
find_object_symbol(entry)
PTR_HASH entry ;
{  PTR_SYMB symb_ptr ,old_symb ;
   int type_status ;
   PTR_BFND real_scope();

  type_status= OFF ;
  for (old_symb=symb_ptr =entry->id_attr ;(symb_ptr && (! type_status)); symb_ptr = symb_ptr->outer)
	   { 	  if (is_symbol_an_object(symb_ptr->variant))
		       {  PTR_BFND flow ;
                          PTR_BFND type_scope ;                    
                          
                          type_scope = real_scope(symb_ptr->scope) ;
                          for ( flow = cur_scope(); flow ; flow = pre_scope(flow) )
                              if (flow == type_scope) 
                                {   type_status = ON ;
                                    break ;
                                 }
                         }   
                  old_symb = symb_ptr ;
            }
     if (type_status == ON)  return(old_symb);
     else {
         /* default integer function */
         return(SMNULL);
       }

 }



/* Given a hash_entry, to find the symbol whose scope is global bif node
 * 
 *
 */
PTR_SYMB
find_global_symbol(entry,flag)
PTR_HASH entry ;
int  flag ;
{  PTR_SYMB symb_ptr ,old_symb ;
   int type_status ;
   PTR_BFND real_scope();

  type_status= OFF ;
  for (old_symb=symb_ptr =entry->id_attr ;(symb_ptr && (! type_status)); symb_ptr = symb_ptr->outer)
	   { 	  if (is_same_variant_group(symb_ptr->variant,flag) || (flag==OFF))
		       {  PTR_BFND flow ;
                          PTR_BFND type_scope ;                    
			  
			  type_scope =real_scope(symb_ptr->scope) ;
                          if (type_scope->variant == GLOBAL)
			    type_status = ON;
			  
			}
                  old_symb = symb_ptr ;
            }
     if (type_status == ON)  return(old_symb);
     else {
         /* default integer function */
         return(SMNULL);
       }
 }


/* Very Imortant basic routine
 * install_entry takes
 *   <1>  a pointer to a hash entry and makes another symbol table entry for 
 *        the same id
 *   <2>  var_decl : variant for new symbol
 *                   if (variant == FUNCTION_NAME) and (sw== ONE)
 *                      1. if it is a new symbol, then 
 *                           FUNCTION_NAME installed, scope = Global_Bfnd,
 *                           seen = NO ;
 *                      2. if has same name symbol,
 *                         a. if not visible don't care same as 1.
 *                         b. report error except for the same name id
 *                            is a function and its scope is global_bfnd
 *                         c. if so, check seen or not seen, anyway nothing
 *                            can do at this time.
 *                  if (variant == FUNCTION_NAME) and (sw >= TWO)
 *                      1. check as normal field_name is ok.
 *                      2. MEMBER_FUNC installed, scope is same as normal field
 *                      3. Seen = No
 *                  if normal data then
 *                      1. only check same name id if appear in the same scope
 *                      2. install entry
 *   <3>  first_name : if the new entry is only a field of a T_ENUM, then
 *        first_name represent the first member in the same structure
 *        it will be NULL if not a member.
 *   <4>  sw         : ONE if  the entry is not a member
 *                     TWO if  the entry is the first member of the structure
 *                    THREE if the entry is a member but not first one
 */
PTR_SYMB 
install_entry(entry,var_decl,err_sw)
	PTR_HASH entry;
        int      var_decl ;
        int      err_sw;

{
	register PTR_SYMB symb_ptr;
	PTR_BFND decl_block();
	PTR_SYMB temp_ptr,old_symb;
        PTR_BFND flow,type_scope ;
        int type_status;
      

      type_status= OFF ;
      for (old_symb=symb_ptr =entry->id_attr ;(symb_ptr && (! type_status)); )
	   { 
                type_scope = symb_ptr->scope ;
                if (type_scope == decl_block() )
		  {
                     if (err_sw == YES) {
		       err("Redeclaration of Identifier");
		       type_status = ON ;
		       return(SMNULL);
		     }
		     else
		       { type_status = ON ;
			 return(symb_ptr);
		       }
                  }
                else      
                  {
                    symb_ptr =symb_ptr->outer ;
                    old_symb = symb_ptr ;
                  }
           }
       
	symb_ptr = make_symb(fi,var_decl, entry->ident);
        symb_ptr->scope = decl_block();
	symb_ptr->outer = entry->id_attr;
	symb_ptr->parent = entry;
	symb_ptr->attr = NULL;
	entry->id_attr = symb_ptr;
        symb_ptr->dovar = (int ) NULL ;
	symb_ptr->id_list = SMNULL;
	return (symb_ptr);
      }



/* install_entry_2 is a division from install_entry
 * <1> No check for scope is done here. Absolutely install one entry
 * <2> For parameter,  empty string name, and label name
 * <3> If scope_sw == YES we use normal way <pred_bfnd> for variable scope
 * <4> If scope_sw == NO  we use current_function bif_node as scope
 * <5> if scope_sw == OTHER the global_bfnd as scope ;
 * <6> if scope_sw == UP_TO_FUNC_HEDR, we use the near predecessor func_hedr
 * <5> if scope_sw == UP_TO_NODECL, we use the first non-decl predecessor
 */
PTR_SYMB 
install_entry_2(entry,var_decl,scope_sw)
PTR_HASH entry;
int      var_decl ;
int      scope_sw ;
{
	register PTR_SYMB symb_ptr;
	PTR_BFND decl_block(),non_decl_scope(),function_block();
	PTR_SYMB temp_ptr,old_symb;
        PTR_BLOB blob ;

      for (symb_ptr =entry->id_attr ;(symb_ptr) ; )
	   { 
                symb_ptr =symb_ptr->outer ;
           }
       
	symb_ptr = make_symb(fi,var_decl, entry->ident);
	switch (scope_sw) {
	case YES:
              symb_ptr->scope = decl_block() ;
	      break;
	case NO:
              blob = follow_blob(pred_bfnd);
              if (!blob) err("system crash at install_entry_2 ");
              symb_ptr->scope = blob->ref ;
	      break;
	case OTHER:
              symb_ptr->scope =  global_bfnd;
	      break;
	case UP_TO_FUNC_HEDR:
              symb_ptr->scope = function_block(pred_bfnd);
	      break;
	case UP_TO_NODECL:
	      symb_ptr->scope = non_decl_scope(pred_bfnd);
	      break;
        }
	symb_ptr->outer = entry->id_attr;
	symb_ptr->parent = entry;
	symb_ptr->attr = NULL;
	entry->id_attr = symb_ptr;
        symb_ptr->dovar = (int ) NULL ;
	symb_ptr->id_list = SMNULL;
	return (symb_ptr);
      }

/* Needs change */
PTR_SYMB 
install_procedure_entry(entry)
	PTR_HASH entry;
{
	register PTR_SYMB symb_ptr;
	PTR_BFND cur_scope();

	for (symb_ptr = entry->id_attr; symb_ptr; symb_ptr = symb_ptr->outer)
	  ;
	if (!symb_ptr || symb_ptr->variant != PROCEDURE_NAME)
	  symb_ptr = make_symb(fi,PROCEDURE_NAME, entry->ident);

	symb_ptr->scope = cur_scope();
	symb_ptr->outer = entry->id_attr;
	symb_ptr->parent = entry;
	entry->id_attr = symb_ptr;

	return (symb_ptr);
}

PTR_BFND
decl_block()
{
   return(pred_bfnd);
 }


/* Mofified */
PTR_BFND
cur_scope()
{
	register PTR_BFND p;
        p = real_scope(pred_bfnd);
        return(p);
}


/* Mofified */
PTR_BFND
real_scope(in_bfnd)
PTR_BFND in_bfnd;

{
	register PTR_BFND p;

/*	for (p = in_bfnd;(is_var_decl(p->variant) == YES ) ; p = p->control_parent) ;
	return (p);
*/
/* Modified Apr 16,1988 */
        return(in_bfnd);
}


/* Mofified */
PTR_BFND
pre_scope(fix_bfnd)
        PTR_BFND fix_bfnd;     
{
	register PTR_BFND p;
        if (fix_bfnd->variant == GLOBAL) return(BFNULL);
/*	for (p = fix_bfnd->control_parent ; (is_var_decl(p->variant) == YES) ; p = p->control_parent);
	return (p);
 */
        p = fix_bfnd->control_parent ;
        return(p);
}


int
is_var_decl(variant_value)
int variant_value;
{
   switch (variant_value) {
   case UNION_DECL :
   case VAR_DECL :
   case CLASS_DECL :
   case COLLECTION_DECL :
   case STRUCT_DECL :
   case ENUM_DECL :
                     return(YES);
   default :         return(NO);

   }

 }


/*****************************************************************************/
/* find_file_entry: find the file_entry at include file_table                */
/*                  A new one will be allocated if it doesnot exist          */
/*****************************************************************************/
PTR_FNAME
find_file_entry( name)
	char *name ;
{
  register PTR_FNAME file_entry ;
  PTR_FNAME make_file();

  for( file_entry = head_file; file_entry ; file_entry = file_entry->next)
      if (!strcmp(name,file_entry->name)) break;
  if (!file_entry) file_entry = make_file(name);
  return(file_entry);
}



/* Mofified */
PTR_BFND
non_decl_scope(in_bfnd)
PTR_BFND in_bfnd;

{
	register PTR_BFND p;

	for (p = in_bfnd;(is_var_decl(p->variant) == YES ) ; p = p->control_parent) ;
	return (p);
}



/* Mofified */
PTR_BFND
function_block(in_bfnd)
PTR_BFND in_bfnd;

{
	register PTR_BFND p;

	for (p = in_bfnd;((p->variant != FUNC_HEDR)&&(p->variant != GLOBAL)) ; p = p->control_parent) ;
	return (p);

}

/**************************************************************************/
/* check scope for struct tag .....                                       */
/**************************************************************************/
same_scope(old_scope,parent_node)
PTR_BFND old_scope,parent_node;
{
  register PTR_BFND p ;
  int sw;
  p = non_decl_scope(parent_node);
  sw = (p == old_scope) ? 1 : 0 ;
  return(sw);
}


/*****************************************************************************/
/* Remove_entry: Remove a installed symbol entry based on that symbol        */
/*               after the remove will be re-installed again.                */
/*               It is limited used for change the scope of a symbol.        */
/*               Remove it and then re-insert it                             */
/*****************************************************************************/
PTR_SYMB 
remove_entry(symb)
PTR_SYMB symb;
{
  PTR_SYMB old_symb, first;

  old_symb = (PTR_SYMB) NULL;
  for (first =symb->parent->id_attr ;first ;first = first->outer )
    { 
      if (first == symb) break;
      old_symb = first;
    }
  if (!old_symb) {
   /* symb is supposed to be the first one */
    symb->parent->id_attr = first->outer;
  }
  else {
    old_symb->outer =first->outer;
  }
}

change_scope(symb, new_scope)
PTR_SYMB symb;
PTR_BFND new_scope;
{

  remove_entry(symb);
  re_install_entry(symb,new_scope);

}

/****************************************************************************/
/* re_install_entry with symbol instead of hash_entry                       */
/* It is just effective for installing the scope of GLOBAL                  */
/* The symbol is installed as to be found as last one, since it is GLOBAL   */
/****************************************************************************/
PTR_SYMB 
re_install_entry(symb,new_scope)
PTR_SYMB symb;
PTR_BFND new_scope;
{
	register PTR_SYMB symb_ptr;
	PTR_BFND decl_block();
	PTR_SYMB temp_ptr,old_symb;
        PTR_BFND flow,type_scope ;
        int type_status;

      type_status= OFF ;
      for (old_symb=symb_ptr =symb->parent->id_attr ;(symb_ptr && (! type_status)); )
	   { 
                type_scope = symb_ptr->scope ;
                if (type_scope == new_scope)
		  {
                     err("Redeclaration of Identifier");
                     type_status = ON ;
                     return(SMNULL);
                  }
                else      
                  {
                    old_symb = symb_ptr ;
                    symb_ptr =symb_ptr->outer ;
                  }
           }
       
        symb->scope = new_scope;
	symb->outer = (PTR_SYMB) NULL;
	if (old_symb) old_symb->outer = symb;
	else          symb->parent->id_attr = symb;
   }


/***********************************************************************/
/* install_duplicate_entry(old_symb):                                  */
/*     This routine installs overloaded function entry following the   */
/*     last entry. Remember the new entry will have the same scope     */
/*      and the same name as the old one                               */
/*                                                                     */
/***********************************************************************/
PTR_SYMB
install_duplicate_entry(old_symb,new_variant)
PTR_SYMB old_symb;
int new_variant;
{
 PTR_SYMB next_symb,symb_ptr;

  next_symb = old_symb->outer;
  symb_ptr = make_symb(fi,new_variant, old_symb->ident);
  symb_ptr->scope = old_symb->scope;
  symb_ptr->outer = next_symb;
  old_symb->outer = symb_ptr;
  symb_ptr->parent = old_symb->parent;
  symb_ptr->attr = NULL;
  symb_ptr->dovar = (int ) NULL ;
  symb_ptr->id_list = SMNULL;
  return (symb_ptr);
}









is_symbol_an_object(variant)
int variant;
{
  if ((variant==CLASS_NAME)||(variant == UNION_NAME) || 
      (variant==COLLECTION_NAME)|| (variant==STRUCT_NAME)||
      (variant==ENUM_NAME)) return(1);
  else return(0);
}


/* from the pred_bfnd to find the current function's class_bfnd  
   status will be returned as OFF if the current function is declared in
   the body of the class, ON otherwise (i.e. class::foo(...)
*/
PTR_BFND
find_current_class_bfnd(parent,status)
PTR_BFND parent;
int *status;
{
   PTR_BFND old,p;
   PTR_LLND ll;
   PTR_SYMB functor,classtype_name,classname;
        *status = OFF;
	for (old = p = parent;((p->variant!=CLASS_DECL)&&
			       (p->variant!=STRUCT_DECL)&&
			       (p->variant!=COLLECTION_DECL)&&
			       (p->variant!=ENUM_DECL)&&
			       (p->variant!=DERIVED_CLASS_DECL)&&
			       (p->variant!=UNION_DECL)&&
	                       (p->variant != GLOBAL)) ;
	 p = p->control_parent) old = p ;
        if (p->variant != GLOBAL)  return(p);
        if (old->variant == FUNC_HEDR) {
	  functor = old->entry.template.symbol;
	  *status = ON;
          if (functor->variant==MEMBER_FUNC) return(functor->scope);
	}
   else err("unexpected bif form or undefine variables");
   return((PTR_BFND)NULL);
 }





/* variant2 is the one needed, so this routine check if variant1 is in the same
   group of variant2, not the other way around.
*/
is_same_variant_group(variant1,variant2)
int variant1,variant2;
{
  if (((variant1== VARIABLE_NAME)||(variant1==FIELD_NAME)||
       (variant1==FUNCTION_NAME)||(variant1== MEMBER_FUNC))&&
      (((variant2== VARIABLE_NAME)||(variant2==FIELD_NAME)||
       (variant2==FUNCTION_NAME)||(variant2== MEMBER_FUNC)))) return(1);
  if (variant1==variant2) return(1);
  if ((variant2== TYPE_NAME)&&is_symbol_an_object(variant1)) return(1);
  if (((variant1== STRUCT_NAME)||(variant1==CLASS_NAME)||(variant1==COLLECTION_NAME))&&
      ((variant2== STRUCT_NAME)|| (variant2==CLASS_NAME)||(variant1==COLLECTION_NAME))) return(1);
  return(0);
}
