/*
 * sym.c -- hash table routines
 */

#include <stdio.h>
#include <strings.h>

#include "defs.h"
#include "symb.h"
#include "defines.h"
#include "bif.h"
#include "extern.h"
#include "f90.h"
#include "tokdefs.h"

extern int hash(), ndim;
extern PTR_BFND cur_bfnd, pred_bfnd, global_bfnd;
extern PTR_TYPE vartype, global_default, impltype[], make_type();
extern PTR_LLND make_llnd();
extern PTR_SYMB make_symb();
/* added for FORTRAN 90 */
extern PTR_LLND first_unresolved_call;
extern PTR_LLND last_unresolved_call;
extern void err(), errstr();
char *chkalloc();

extern PTR_HASH hash_table[];

/* Scope table variables */
PTR_BFND scope_table[1000];
int top_scope_level;
int scope_starts[10];
int top_scope_starts;

struct operand
{
     int opval;
     char *opname;
} oplist [] = {
        {PLUS, "+"},
	{MINUS, "-"},
	{ASTER, "*"},
	{DASTER, "**"},
	{SLASH, "/"},
	{DSLASH, "//"},
	{AND, ".and."},
	{OR, ".or."},
	{XOR, ".xor."},
	{NOT, ".not."},
	{EQ, ".eq."},
	{NE, ".ne."},
	{GT, ".gt."},
	{GE, ".ge."},
	{NEQV, ".neqv."},
	{EQV, ".eqv."},
	{0, 0}
};

void
init_scope_table()
{
     scope_table[0] = global_bfnd;
     top_scope_level = 0;
     scope_starts[0]= 0;
     top_scope_starts = 0;
}

PTR_BFND
cur_scope()
{
     return (scope_table[top_scope_level]);
}

PTR_BFND
parent_scope(present_scope_level)
int present_scope_level;
{
     if (present_scope_level >= 0)
	  return (scope_table[--present_scope_level]);
     else errstr("sym.c:parent_scope: requested scope level non-existent");
}

void
add_scope_level(new_scope, due_to_use_stat)
PTR_BFND new_scope;
int due_to_use_stat;
{
     PTR_BFND tmp;
     
     if (due_to_use_stat)
     {
	  tmp = scope_table[top_scope_level];
	  scope_table[top_scope_level] = new_scope;
	  scope_table[++top_scope_level] = tmp;
     }
     else 
     {
	  scope_table[++top_scope_level] = new_scope;
	  scope_starts[++top_scope_starts] = top_scope_level;
     }
}

void
delete_beyond_scope_level(level)
PTR_BFND level;
{
     top_scope_level = scope_starts[top_scope_starts] -  1;
     top_scope_starts --;
     if (top_scope_level < 0) 
	  errstr("sym.c:delete_scope: requested scope level non-existent");
}	  

int
cur_scope_level()
{
     return (top_scope_level);
}

PTR_BFND
scope_at_level(level)
int level;
{
     if (level >= 0)
	  return (scope_table[level]);
     else  errstr("sym.c:scope_at_level: requested scope non-existent");
}
     
/*
 look_up_sym(string) : 
   lookup string in the hash table. If a hash table entry having string 
   as it's name, in curent scope is found then return the hash table entry, 
   else make a hash  table entry with string as it's name and return it.
 */
PTR_HASH 
look_up_sym(string)
register char *string;
{
	int i, index, cur_scope_level();
	register PTR_HASH entry;
	PTR_HASH  make_hash_entry();
	PTR_BFND parent_scope(), scope_at_level(), p;

	i = hash(string);
        for (index = cur_scope_level(); index >= 0; index--) 
	{
	     p = scope_at_level(index);
	     for (entry = hash_table[i]; entry; entry = entry->next_entry) {
		  if (!strcmp(string, entry->ident) && (entry->id_attr) && 
                      (entry->id_attr->scope == p)) {
		       return (entry);
		  }
	     }
	}

	return (make_hash_entry(string));
}

PTR_HASH 
just_look_up_sym_in_scope(scope, string)
PTR_BFND scope;
register char *string;
{
	int i, index, cur_scope_level();
	register PTR_HASH entry;
	PTR_HASH  make_hash_entry();
	PTR_BFND parent_scope(), scope_at_level(), p;

	i = hash(string);
	p = scope;
	for (entry = hash_table[i]; entry; entry = entry->next_entry) {
	     if (!strcmp(string, entry->ident) && (entry->id_attr) && 
		 (entry->id_attr->scope == p)) {
		  return (entry);
	     }
	}

	return (HSNULL);
}


PTR_HASH
look_up_op(operator)
int operator;
{
     struct operand *p;
     
     for (p = oplist; p->opname; p++)
	  if (p->opval == operator)
	       return (look_up_sym(p->opname));
     errstr("sym.c:look_up_op: unknown operator %d", operator);
}

PTR_HASH 
just_look_up_sym(string)
register char *string;
{
	int i, index, cur_scope_level();
	register PTR_HASH entry;
	PTR_HASH  make_hash_entry();
	PTR_BFND parent_scope(), scope_at_level(), p;

	i = hash(string);
	p = cur_scope();
        for (index = cur_scope_level(); index >= 0; index--) 
	{
	     p = scope_at_level(index);
	     for (entry = hash_table[i]; entry; entry = entry->next_entry) {
		  if (!strcmp(string, entry->ident) && (entry->id_attr) && 
                      (entry->id_attr->scope == p)) {
		       return (entry);
		  }
	     }
	}

	return (HSNULL);
}

PTR_HASH 
make_hash_entry(string)
register char *string;
{
     int i;
     register PTR_HASH entry;

     i = hash(string);

     entry = (struct hash_entry *) chkalloc(sizeof(struct hash_entry));
     entry->ident = copys(string);
     entry->next_entry = hash_table[i];
     hash_table[i] = entry;
     return (entry);
}    
 
PTR_SYMB
make_sym_entry(var_hash_entry, variant, type, scope, kind)
PTR_HASH var_hash_entry;
int variant, kind;
PTR_TYPE type;
PTR_BFND scope;
{
     PTR_SYMB var_sym_entry;
     
     /* If type is undefined, then obtain type from implicit type table. */
     if ((variant != PROGRAM_NAME) && (variant != DEFAULT) &&
	 (variant != PROCEDURE_NAME)) 
     {
	  
	  if (type == TYNULL)
	       type =  impltype[*var_hash_entry->ident - 'a'];
	  else if ((type->variant == T_ARRAY) &&
		   (type->entry.ar_decl.base_type == TYNULL))
	       type->entry.ar_decl.base_type = 
		    impltype[*var_hash_entry->ident - 'a'];
	
	  if ((type == TYNULL) ||
              ((type->variant == T_ARRAY) &&
               (type->entry.ar_decl.base_type == TYNULL)))
	  {
	       errstr("sym.c:make_sym_entry: type unknown of %s", var_hash_entry->ident);
	       type = global_default;
	  }
     }
     

     var_sym_entry = make_symb(fi, variant, var_hash_entry->ident);

     /* Point the hash entry to the symbol table entry */
     var_hash_entry->id_attr = var_sym_entry;

     var_sym_entry->variant = variant;
     var_sym_entry->type = type;
     var_sym_entry->scope = scope;
     var_sym_entry->id_list = SMNULL;
     switch (variant) {
     case VARIABLE_NAME: 
	  /* if not a formal parameter, 
             then mark it as local. */
	  if (var_sym_entry->entry.var_decl.local != IO) 
	       var_sym_entry->entry.var_decl.local = kind;
	  break;
     case FUNCTION_NAME:
     case PROCEDURE_NAME:
	  var_sym_entry->entry.proc_decl.seen = kind;
	  break;
     default:
	  break;
     }

     /* initialize administrative stuff */
     var_sym_entry->outer = SMNULL;
     var_sym_entry->parent = var_hash_entry;
/*     var_sym_entry->decl = HARD; */
     
     return(var_sym_entry);
}

PTR_SYMB
make_constant(var_hash_entry, type)
PTR_HASH var_hash_entry;
PTR_TYPE type;
{
     PTR_SYMB var_sym_entry;
     int var_variant;
   

     var_sym_entry = var_hash_entry->id_attr;
     if (var_sym_entry == SMNULL)
	  return (make_sym_entry(var_hash_entry, CONST_NAME, type, cur_scope(), 
                                 0));

     var_variant = var_sym_entry->variant;
     if (var_sym_entry->scope == global_bfnd)
	  return (make_sym_entry(make_hash_entry(var_hash_entry->ident),
                                 CONST_NAME, type, cur_scope(), 0));
     	       
     if (var_sym_entry->scope == cur_scope())
	  switch(var_variant) 
	  {
	  case VARIABLE_NAME:
	       var_sym_entry->variant = CONST_NAME;
	       /*  var_sym_entry's type */
	       switch (var_sym_entry->type->variant) 
	       {
	       case T_STRING:
	       case T_ARRAY:
                    if (type != TYNULL)
		       var_sym_entry->type->entry.ar_decl.base_type = type;
                    return (var_sym_entry);
	       case (T_POINTER):
                    if (type != TYNULL)
		       var_sym_entry->type->entry.template.base_type = type;
                    return (var_sym_entry);
	       case T_INT:
	       case T_FLOAT:
	       case T_DOUBLE:
	       case T_COMPLEX:
	       case T_BOOL:
	       case T_STRUCT:
		    if (type != TYNULL) 
			 var_sym_entry->type = type;
		    return (var_sym_entry);
	       }
	       break;
	  case CONST_NAME:
	       if (type != TYNULL)
		    var_sym_entry->type = type;
	       return (var_sym_entry);
	  default:
	       errstr("sym.c:make_constant:1: inconsistent constant declaration %s.", var_hash_entry->ident);
		    return (SMNULL);
	  }
     
     return (make_sym_entry(make_hash_entry(var_hash_entry->ident), CONST_NAME,
                            type, cur_scope(), 0));
}     

/* Occurence of Function name in it's body not properly taken care of. */
PTR_SYMB
make_scalar(var_hash_entry, type, kind)
PTR_HASH var_hash_entry;
PTR_TYPE type;
int kind;
{
     PTR_SYMB var_sym_entry, cur_scope_sym_ptr, result_sym_ptr;
     int var_variant;

     var_sym_entry = var_hash_entry->id_attr;
     if (var_sym_entry == SMNULL)
	  return (make_sym_entry(var_hash_entry, VARIABLE_NAME, type, 
                                 cur_scope(), kind));

     var_variant = var_sym_entry->variant;
     cur_scope_sym_ptr = cur_scope()->entry.template.symbol;
     if (var_sym_entry->scope == global_bfnd)
	  if ((cur_scope()->variant == FUNC_HEDR) &&
	      (!(strcmp(var_hash_entry->ident, cur_scope_sym_ptr->ident))))
	       if (type == TYNULL)
		    return (var_sym_entry);
               else 
	       {
		    cur_scope_sym_ptr->type = type;
		    if (result_sym_ptr =  cur_scope_sym_ptr->entry.template.declared_name)
			 result_sym_ptr->type = type;
		    return (var_sym_entry);
	       }
     	  else return (make_sym_entry(make_hash_entry(var_hash_entry->ident),
                                 VARIABLE_NAME, type, cur_scope(), kind));
     	       
     if (var_sym_entry->scope == cur_scope())
	  switch(var_variant) 
	  {
	  case VARIABLE_NAME:
	       /*  var_sym_entry's type */
	       switch (var_sym_entry->type->variant) 
	       {
	       case T_STRING:
	       case T_ARRAY:
                    if (type != TYNULL)
		       var_sym_entry->type->entry.ar_decl.base_type = type;
                    return (var_sym_entry);
	       case T_POINTER:
                    if (type != TYNULL)
		       var_sym_entry->type->entry.template.base_type = type;
                    return (var_sym_entry);
	       case T_INT:
	       case T_FLOAT:
	       case T_DOUBLE:
	       case T_COMPLEX:
	       case T_BOOL:
	       case T_STRUCT:
		    if (type != TYNULL)
			 var_sym_entry->type = type;
		    return (var_sym_entry);
	       }
	       break;
	  case CONST_NAME:
	       if (type != TYNULL)
		    var_sym_entry->type = type;
	       return (var_sym_entry);
	  case ROUTINE_NAME:
	       if (type != TYNULL)
		    var_sym_entry->type = type;
	       return (var_sym_entry);
	  case LABEL_VAR:
	       return (var_sym_entry);
	  default:
	       errstr("sym.c:make_scalar:1: inconsistent declaration of identifier %s.", var_hash_entry->ident);
	       return (SMNULL);
	  }
     
     return (make_sym_entry(make_hash_entry(var_hash_entry->ident), 
                            VARIABLE_NAME, type, cur_scope(), kind));
}     

/*
     make_array(var_hash_entry:PTR_HASH, base_type: PTR_TYPE, dcl_type:int,
                ranges: PTR_LLND, ndim : int)
       makes a array symbol table entry.
*/

PTR_SYMB
make_array(var_hash_entry, base_type, ranges, ndim, kind)
PTR_HASH var_hash_entry;
PTR_TYPE base_type;
PTR_LLND ranges;
int ndim, kind;
{
     PTR_SYMB var_sym_entry;
     PTR_TYPE array_type;
     int var_variant;
     
     
     var_sym_entry = var_hash_entry->id_attr;
     if (var_sym_entry == SMNULL)
     {
	  array_type = make_type(fi, T_ARRAY);
	  array_type->entry.ar_decl.base_type = base_type;
	  array_type->entry.ar_decl.num_dimensions = ndim;
	  return (make_sym_entry(var_hash_entry, VARIABLE_NAME, array_type,
                                 cur_scope(), kind));
     }
     
     var_variant = var_sym_entry->variant;
     if (var_sym_entry->scope == global_bfnd)
     {
	  array_type = make_type(fi, T_ARRAY);
	  array_type->entry.ar_decl.base_type = base_type;
	  array_type->entry.ar_decl.num_dimensions = ndim;
	  return (make_sym_entry(var_hash_entry, VARIABLE_NAME, array_type, 
                                 cur_scope(), kind));
     }

     if (var_sym_entry->scope == cur_scope())
	  switch(var_variant) 
	  {
	  case VARIABLE_NAME:
	       /*  var_sym_entry's type */
	       switch (var_sym_entry->type->variant) 
	       {
	       case T_ARRAY:
                    if (base_type != TYNULL)
                         var_sym_entry->type->entry.ar_decl.base_type = 
			      base_type;
		    if (ndim) 
			 var_sym_entry->type->entry.ar_decl.num_dimensions = ndim;
                    return (var_sym_entry);
	       case T_STRING:
	       case T_POINTER:
	       case T_INT:
	       case T_FLOAT:
	       case T_DOUBLE:
	       case T_COMPLEX:
	       case T_BOOL:
	       case T_STRUCT:
	       {
		    array_type = make_type(fi, T_ARRAY);
		    if (base_type == TYNULL)
			 array_type->entry.ar_decl.base_type = 
			      var_sym_entry->type;
		    else array_type->entry.ar_decl.base_type = base_type;
		    array_type->entry.ar_decl.num_dimensions = ndim;
		    var_sym_entry->type = array_type;
                    return (var_sym_entry);
	       }
               default:
		    return (var_sym_entry);
	       }
	       break;
	  default:
	       errstr("sym.c:make_array:1: inconsistent array declaration of identifier %s.", var_hash_entry->ident);
		    return (SMNULL);
	  }

     array_type = make_type(fi, T_ARRAY);
     array_type->entry.ar_decl.base_type = base_type;
     array_type->entry.ar_decl.num_dimensions = ndim;
     return (make_sym_entry(var_hash_entry, VARIABLE_NAME, array_type,
                             cur_scope(), kind));
}

PTR_SYMB
make_pointer(var_hash_entry, base_type, kind)
PTR_HASH var_hash_entry;
PTR_TYPE base_type;
int kind;
{
     PTR_SYMB var_sym_entry;
     PTR_TYPE pointer_type;
     int var_variant;
     
     
     var_sym_entry = var_hash_entry->id_attr;
     if (var_sym_entry == SMNULL)
     {
	  pointer_type = make_type(fi, T_POINTER);
	  pointer_type->entry.template.base_type = base_type;
	  return (make_sym_entry(var_hash_entry, VARIABLE_NAME, pointer_type,
                                 cur_scope(), kind));
     }
     
     var_variant = var_sym_entry->variant;
     if (var_sym_entry->scope == global_bfnd)
     {
	  pointer_type = make_type(fi, T_POINTER);
	  pointer_type->entry.template.base_type = base_type;
	  return (make_sym_entry(var_hash_entry, VARIABLE_NAME, pointer_type, 
                                 cur_scope(), kind));
     }

     if (var_sym_entry->scope == cur_scope())
	  switch(var_variant) 
	  {
	  case VARIABLE_NAME:
	       /*  var_sym_entry's type */
	       switch (var_sym_entry->type->variant) 
	       {
	       case T_POINTER:
                    if (base_type != TYNULL)
                         var_sym_entry->type->entry.template.base_type = 
			      base_type;
                    return (var_sym_entry);
	       case T_STRING:
	       case T_ARRAY:
	       case T_INT:
	       case T_FLOAT:
	       case T_DOUBLE:
	       case T_COMPLEX:
	       case T_BOOL:
	       case T_STRUCT:
	       {
		    pointer_type = make_type(fi, T_POINTER);
		    if (base_type == TYNULL)
			 pointer_type->entry.template.base_type = 
			      var_sym_entry->type;
		    else pointer_type->entry.template.base_type = base_type;
		    var_sym_entry->type = pointer_type;
                    return (var_sym_entry);
	       }
               default:
		    return (var_sym_entry);
	       }
	       break;
	  default:
	       errstr("sym.c:make_pointer:1: inconsistent pointer declaration of identifier %s.", var_hash_entry->ident);
		    return (SMNULL);
	  }

     pointer_type = make_type(fi, T_POINTER);
     pointer_type->entry.template.base_type = base_type;
     return (make_sym_entry(var_hash_entry, VARIABLE_NAME, pointer_type,
                             cur_scope(), kind));
}
     
PTR_SYMB
make_function(var_hash_entry, type, kind)
PTR_HASH var_hash_entry;
PTR_TYPE type;
int kind;
{
     PTR_SYMB var_sym_entry;
     int var_variant;
     
     var_sym_entry = var_hash_entry->id_attr;
     if (var_sym_entry == SMNULL)
	  if (kind == LOCAL)
	       return (make_sym_entry(var_hash_entry, FUNCTION_NAME, type,
                                      cur_scope(), kind));
          else return (make_sym_entry(var_hash_entry, FUNCTION_NAME, type,
                                      global_bfnd, kind));
     
     var_variant = var_sym_entry->variant;
     if (var_sym_entry->scope == global_bfnd) 
	  if (var_variant == FUNCTION_NAME)
	       if (kind == LOCAL)
		    return (make_sym_entry(var_hash_entry, FUNCTION_NAME, type,
                                      cur_scope(), kind));
	       else return var_sym_entry;
          else if (var_variant == ROUTINE_NAME)
	  {
	       var_sym_entry->variant == FUNCTION_NAME;
	       return var_sym_entry;
	  }
          else if (var_variant == INTERFACE_NAME)
	       return var_sym_entry;
          else if (var_variant == VARIABLE_NAME)
	  {
	       var_sym_entry->variant == FUNCTION_NAME;
	       return var_sym_entry;
	  }
	  else
	  {
	       errstr("sym.c:make_function:1:inconsistent function %s.", var_hash_entry->ident);
	       return (SMNULL);
	  }     	       
     if (var_sym_entry->scope == cur_scope())
	  switch(var_variant) 
	  {
	  case VARIABLE_NAME:
	       /*  var_sym_entry's type */
	       switch (var_sym_entry->type->variant) 
	       {
	       case T_STRING:
	       case T_INT:
	       case T_FLOAT:
	       case T_DOUBLE:
	       case T_COMPLEX:
	       case T_BOOL:
	       case T_POINTER:
	       case T_STRUCT:
		    /* if not a formal parameter, convert it into 
		       a global function. */
		    if (var_sym_entry->entry.var_decl.local != IO)
		    {
			 var_sym_entry->variant = FUNCTION_NAME;
			 var_sym_entry->scope = global_bfnd;
		    }
		    if (type != TYNULL)
			 var_sym_entry->type = type;
		    return (var_sym_entry);
	       case T_ARRAY:
                    if (kind == LOCAL)
		    {
			 errstr("sym.c:make_function:2: inconsistent function %s.", var_hash_entry->ident);
			 return (SMNULL);
		    } 
                    else return (var_sym_entry);
	       }
	  case FUNCTION_NAME:
	       var_sym_entry->variant = FUNCTION_NAME;
	       return (var_sym_entry);
	  case ROUTINE_NAME:
	       var_sym_entry->variant = FUNCTION_NAME;
	       return (var_sym_entry);
	  case INTERFACE_NAME:
	       return (var_sym_entry);
	  default:
	       errstr("sym.c:make_function:3: inconsistent function %s.", var_hash_entry->ident);
	       return (SMNULL);
	  }
     
     return (make_sym_entry(make_hash_entry(var_hash_entry->ident), FUNCTION_NAME, type, cur_scope(), kind));
}     

PTR_SYMB
make_external(var_hash_entry, type)
PTR_HASH var_hash_entry;
PTR_TYPE type;
{
     PTR_SYMB var_sym_entry;
     int var_variant;
     
     var_sym_entry = var_hash_entry->id_attr;
     if (var_sym_entry == SMNULL)
	  return (make_sym_entry(var_hash_entry, ROUTINE_NAME, type,
				 global_bfnd, NO));
     
     var_variant = var_sym_entry->variant;
     if (var_sym_entry->scope == global_bfnd) 
	  if ((var_variant == FUNCTION_NAME) ||
              (var_variant == PROCEDURE_NAME))
	       return var_sym_entry;
          else if (var_variant == ROUTINE_NAME)
	  {
	       return var_sym_entry;
	  }
	  else
	  {
	       errstr("sym.c:make_external:1:inconsistent procedure %s.", var_hash_entry->ident);
	       return (SMNULL);
	  }     	       
     if (var_sym_entry->scope == cur_scope())
	  switch(var_variant) 
	  {
	  case VARIABLE_NAME:
	       /*  var_sym_entry's type */
	       switch (var_sym_entry->type->variant) 
	       {
	       case T_STRING:
	       case T_INT:
	       case T_FLOAT:
	       case T_DOUBLE:
	       case T_COMPLEX:
	       case T_BOOL:
	       case T_POINTER:
	       case T_STRUCT:
	       case T_ARRAY:
		    /* if not a formal parameter, convert it into 
		       a global function. */
		    if (var_sym_entry->entry.var_decl.local != IO)
		    {
			 var_sym_entry->variant = ROUTINE_NAME;
			 var_sym_entry->scope = global_bfnd;
		    }
		    if (type != TYNULL)
			 var_sym_entry->type = type;
		    return (var_sym_entry);
	       }
	  case ROUTINE_NAME:
	       return (var_sym_entry);
	  default:
	       errstr("sym.c:make_external:2: inconsistent procedure %s.", var_hash_entry->ident);
	       return (SMNULL);
	  }     
     return (make_sym_entry(make_hash_entry(var_hash_entry->ident), ROUTINE_NAME, type, global_bfnd, NO));
}     
     
PTR_SYMB
make_procedure(var_hash_entry, kind)
PTR_HASH var_hash_entry;
int kind;
{
     PTR_SYMB var_sym_entry;
     
     var_sym_entry = var_hash_entry->id_attr;
     if (var_sym_entry == SMNULL)
	  if (kind == LOCAL)
	       return (make_sym_entry(var_hash_entry, PROCEDURE_NAME, TYNULL,
                                      cur_scope(), kind));
          else return (make_sym_entry(var_hash_entry, PROCEDURE_NAME, TYNULL,
                                      global_bfnd, kind));
     if (var_sym_entry->scope == global_bfnd)
	  if (var_sym_entry->variant == PROCEDURE_NAME)
	       return (var_sym_entry);
          else if (var_sym_entry->variant == ROUTINE_NAME)
	  {
	       var_sym_entry->variant = PROCEDURE_NAME;
	       return var_sym_entry;
	  }
          else if (var_sym_entry->variant == INTERFACE_NAME)
	       return var_sym_entry;
	  else
	  {
	       errstr("sym.c:make_procedure:1:inconsistent subroutine %s.", var_hash_entry->ident);
	       return (SMNULL);
	  }
     	       
     if (var_sym_entry->scope == cur_scope())
	  switch(var_sym_entry->variant)
	  {
	  case PROCEDURE_NAME:
	       return (var_sym_entry);
	  case VARIABLE_NAME:
	       /* if not a formal parameter, convert it into 
                  a global procedure. */
	       if (var_sym_entry->entry.var_decl.local != IO)
	       {
		    var_sym_entry->variant = PROCEDURE_NAME;
		    var_sym_entry->scope = global_bfnd;
	       }
	       return (var_sym_entry);
	  case ROUTINE_NAME:
	       var_sym_entry->variant = PROCEDURE_NAME;
	       return (var_sym_entry);
	  case INTERFACE_NAME:
	       return (var_sym_entry);
	  default:
	       errstr("sym.c:make_procedure:2:inconsistent subroutine %s.", var_hash_entry->ident);
	       return (SMNULL);
	  }
     
     return (make_sym_entry(make_hash_entry(var_hash_entry->ident),
                            PROCEDURE_NAME, TYNULL, global_bfnd, kind));
}     
     

PTR_SYMB
make_program(var_hash_entry)
PTR_HASH var_hash_entry;
{
     PTR_SYMB var_sym_entry;
     
     var_sym_entry = var_hash_entry->id_attr;
     if (var_sym_entry == SMNULL)
	  return (make_sym_entry(var_hash_entry, PROGRAM_NAME, TYNULL, 
                                 global_bfnd, 0));

     if (var_sym_entry->scope == global_bfnd)
	  {
	       errstr("sym.c:make_program:1: inconsistent program declaration %s.", var_hash_entry->ident);

	       return (SMNULL);
	  }
     	       
     if (var_sym_entry->scope == cur_scope())
	  {
	       errstr("sym.c:make_program:2: inconsistent program declaration %s.", var_hash_entry->ident);

	       return (SMNULL);
	  }
     
     return (make_sym_entry(make_hash_entry(var_hash_entry->ident), PROGRAM_NAME,
                            TYNULL, global_bfnd, 0));
}     

PTR_SYMB
make_common(var_hash_entry)
PTR_HASH var_hash_entry;
{
     PTR_SYMB var_sym_entry;
     
     var_sym_entry = var_hash_entry->id_attr;
     if (var_sym_entry == SMNULL)
	  return (make_sym_entry(var_hash_entry, DEFAULT, TYNULL, global_bfnd,
                                 0));

     if (var_sym_entry->scope == global_bfnd)
	  if (var_sym_entry->variant == DEFAULT)
	       return (var_sym_entry);
	  else 
	  {
	       errstr("sym.c:make_common:1: inconsistent common declaration %s.", var_hash_entry->ident);

	       return (SMNULL);
	  }
     	       
     /* A local entity can have same name as a common block name. So, ignore
        the local entity. */
     
     return (make_sym_entry(make_hash_entry(var_hash_entry->ident), DEFAULT,
                            TYNULL, global_bfnd, 0));
}     

PTR_TYPE 
make_type_node(typespec, lengspec)
PTR_TYPE typespec;
PTR_LLND lengspec;
{
     PTR_TYPE t;
     
     t = typespec;
     if (lengspec != LLNULL) 
     {
	  t = make_type(fi, typespec->variant);
	  t->entry.template.ranges = lengspec;
     }
     return (t);
}

PTR_SYMB
make_derived_type(var_hash_entry, type, kind)
PTR_HASH var_hash_entry;
PTR_TYPE type;
int kind;
{
     PTR_SYMB var_sym_entry;
     PTR_TYPE struct_type;
     int var_variant;
     
     
     var_sym_entry = var_hash_entry->id_attr;
     if (var_sym_entry == SMNULL)
     {
	  struct_type = make_type(fi, T_STRUCT);
	  return (make_sym_entry(var_hash_entry, TYPE_NAME, struct_type,
                                 cur_scope(), kind));
     }
     
     var_variant = var_sym_entry->variant;
     if (var_sym_entry->scope == global_bfnd)
     {
	  struct_type = make_type(fi, T_STRUCT);
	  return (make_sym_entry(var_hash_entry, TYPE_NAME, struct_type, 
                                 cur_scope(), kind));
     }

     if (var_sym_entry->scope == cur_scope())
	  switch(var_variant) 
	  {
	  case VARIABLE_NAME:
	       /*  var_sym_entry's type */
	       switch (var_sym_entry->type->variant) 
	       {
	       case T_STRING:
	       case T_ARRAY:
                    if (type != TYNULL)
                         var_sym_entry->type->entry.ar_decl.base_type = 
			      type;
                    return (var_sym_entry);
	       case T_INT:
	       case T_FLOAT:
	       case T_DOUBLE:
	       case T_COMPLEX:
	       case T_BOOL:
	       case T_POINTER:
	       case T_STRUCT:
	       {
		    struct_type = make_type(fi, T_STRUCT);
		    var_sym_entry->type = struct_type;
                    return (var_sym_entry);
	       }
               default:
		    return (var_sym_entry);
	       }
	       break;
	  default:
	       errstr("sym.c:make_struct:1: inconsistent struct declaration of identifier %s.", var_hash_entry->ident);
		    return (SMNULL);
	  }

     struct_type = make_type(fi, T_STRUCT);
     return (make_sym_entry(var_hash_entry, TYPE_NAME, struct_type,
                             cur_scope(), kind));
}

PTR_SYMB
make_local_entity(var_hash_entry, variant, type, kind)
PTR_HASH var_hash_entry;
PTR_TYPE type;
int variant, kind;
{
     PTR_SYMB var_sym_entry;
     int var_variant;
     
     var_sym_entry = var_hash_entry->id_attr;
     if (var_sym_entry == SMNULL)
     {
	  return (make_sym_entry(var_hash_entry, variant, type,
                                 cur_scope(), kind));
     }
     
     var_variant = var_sym_entry->variant;
     if (var_sym_entry->scope == global_bfnd)
     {
	  return (make_sym_entry(var_hash_entry, variant, type, 
                                 cur_scope(), kind));
     }

     if (var_sym_entry->scope == cur_scope())
	  switch(var_variant) 
	  {
	  default:
	       if (variant == var_variant)
		    return(var_sym_entry);
	       else 
               {
                 errstr("sym.c:make_local_entity:1: inconsistent declaration of identifier %s.", var_hash_entry->ident);
		 return (SMNULL);
	       }
	  };

     return (make_sym_entry(var_hash_entry, variant, type, cur_scope(), kind));
}

PTR_SYMB
make_global_entity(var_hash_entry, variant, type, kind)
PTR_HASH var_hash_entry;
PTR_TYPE type;
int variant, kind;
{
     PTR_SYMB var_sym_entry;
     int var_variant;
     
     var_sym_entry = var_hash_entry->id_attr;
     if (var_sym_entry == SMNULL)
     {
	  return (make_sym_entry(var_hash_entry, variant, type,
                                 global_bfnd, kind));
     }
     
     var_variant = var_sym_entry->variant;
     if (var_sym_entry->scope == global_bfnd)
	  switch(var_variant) 
	  {
	  default:
	       if (variant == var_variant)
		    return(var_sym_entry);
	       else 
               {
                 errstr("sym.c:make_global_entity:1: inconsistent declaration of identifier %s.", var_hash_entry->ident);
		 return (SMNULL);
	       }
	  };
     errstr("sym.c:make_global_entity:1: inconsistent declaration of identifier %s.", var_hash_entry->ident);
     return (SMNULL);
}

void
process_type(type_var, type_fields_end)
PTR_SYMB type_var;
PTR_BFND type_fields_end;
{
     PTR_SYMB sym_temp, last_field;
     PTR_LLND ll_temp;
     PTR_BFND temp;
     PTR_BLOB blob_temp;
     int count = 0;
     
     temp = cur_scope();
     blob_temp = temp->entry.template.bl_ptr1;
     
     last_field = blob_temp->ref->entry.template.ll_ptr1->entry.list.item->entry.template.symbol;
     type_var->type->variant = T_STRUCT;
     /* The field derived class is a constraint put by the design. Maybe instead of T_STRUCT, T_RECORD 
        might be better. But then it would be inconsistent with C++. */
     type_var->type->entry.derived_class.first = last_field;
     while (blob_temp && (blob_temp->ref != type_fields_end))
     {
          ll_temp = blob_temp->ref->entry.template.ll_ptr1;
	  while (ll_temp != LLNULL) 
	  {
	       count++;
	       sym_temp = ll_temp->entry.list.item->entry.template.symbol;
	       sym_temp->entry.template.tag = FIELD_NAME;
	       last_field->entry.template.next = sym_temp;
	       sym_temp->entry.template.base_name = type_var;
	       ll_temp = ll_temp->entry.list.next;
	  }
	  blob_temp = blob_temp->next;
     }
     type_var->type->entry.derived_class.num_fields = count;
}     

void
process_interface(end_of_interface)
PTR_BFND end_of_interface;
{
     PTR_SYMB sym_temp, last_symbol, interface_symbol;
     PTR_BFND temp;
     PTR_BLOB blob_temp;
     
     temp = pred_bfnd;
     blob_temp = temp->entry.template.bl_ptr1;
     interface_symbol = pred_bfnd->entry.template.symbol;
     if (interface_symbol == SMNULL) return;

     last_symbol = blob_temp->ref->entry.template.symbol;
     interface_symbol->entry.template.declared_name = last_symbol;
     while (blob_temp && (blob_temp->ref != end_of_interface))
     {
	  sym_temp = blob_temp->ref->entry.template.symbol;
	  last_symbol->entry.template.declared_name = sym_temp;
	  blob_temp = blob_temp->next;
	  last_symbol = sym_temp;
     }
     last_symbol->entry.template.declared_name = SMNULL;
}     

PTR_TYPE
lookup_type(name)
PTR_HASH name;
{
     PTR_HASH hash_temp;
     PTR_SYMB sym_temp;
     PTR_TYPE ty_temp;
     
     hash_temp = just_look_up_sym(name->ident);
     if (hash_temp && (sym_temp = hash_temp->id_attr) && (sym_temp->variant == TYPE_NAME))
     {
	  ty_temp = make_type(fi, T_DERIVED_TYPE);
	  ty_temp->name = sym_temp;
	  return (ty_temp);
     }
     errstr("sym.c:lookup_type:1: undefined type %s", name->ident);
     return (TYNULL);
}

PTR_SYMB
component(type_sym_entry, fieldname)
PTR_SYMB type_sym_entry;
char *fieldname;
{
	int i;
	register PTR_HASH entry;
	register PTR_SYMB t;

	i = hash(fieldname);
	for (entry = hash_table[i]; entry; entry = entry->next_entry) {
	     t = entry->id_attr;
	     if (!strcmp(fieldname, entry->ident) && t && 
		 (t->entry.template.tag == FIELD_NAME) && (t->entry.template.base_name == type_sym_entry)) {
		  return (t);
	     }
	}

	return (SMNULL);
}

PTR_LLND
deal_with_options(name, type, attributes, dims, ndim, value)
PTR_HASH name;
PTR_TYPE type;
int attributes, ndim;
PTR_LLND dims, value;
{
     PTR_SYMB s;
     PTR_LLND l;

     l = LLNULL;
     if (attributes & PARAMETER_BIT) 
     {
	  s = make_constant(name, type);
	  s->entry.const_value = value;
	  s->attr = attributes;
	  l = make_llnd(fi, CONST_REF, LLNULL, LLNULL, s);
	  return (l);
     }
     if (attributes & EXTERNAL_BIT) 
     {
	  s = make_scalar(name, type, NO);
	  s->attr = attributes;
	  s->variant = ROUTINE_NAME;
	  l = make_llnd(fi, VAR_REF, LLNULL, LLNULL, s);
	  return (l);
     }
     if (attributes & INTRINSIC_BIT) 
     {
	  s = make_function(name, type, NO);
	  s->attr = attributes;
	  l = make_llnd(fi, VAR_REF, LLNULL, LLNULL, s);
	  return (l);
     }

     s = make_scalar(name, type, LOCAL);
     s->attr = attributes;
     if ((attributes & DIMENSION_BIT) || (dims != LLNULL))
     {
	  s = make_array(name, TYNULL, dims, ndim, LOCAL);
	  l = make_llnd(fi, ARRAY_REF, dims, LLNULL, s);
	  s->type->entry.ar_decl.ranges = dims;
     }
     if (attributes & POINTER_BIT) 
     {
	  s = make_pointer(name, TYNULL, LOCAL);
	  l = make_llnd(fi, VAR_REF, LLNULL, LLNULL, s);
     }

     if (l == LLNULL)
     {
	  l = make_llnd(fi, VAR_REF, LLNULL, LLNULL, s);
     }

     if (value != LLNULL)
	  l = make_llnd(fi, ASSGN_OP, l, value, SMNULL);       
     return (l);
}

/*    
PTR_BFND
parent_scope(p)
register PTR_BFND p;
{
     for (p = p->control_parent;
	     (p != BFNULL) &&
     	     (p->variant != PROG_HEDR) &&
	     (p->variant != PROC_HEDR) &&
	     (p->variant != FUNC_HEDR) &&
	     (p->variant != BLOCK_DATA) &&
	     (p->variant != FORALL_NODE) &&
	     (p->variant != GLOBAL) &&
	     (p->variant != CDOALL_NODE) &&
	     (p->variant != SDOALL_NODE) &&
	     (p->variant != DOACROSS_NODE) &&
	     (p->variant != CDOACROSS_NODE) &&
	     (p->variant != STRUCT_DECL);
	     p = p->control_parent)
	  ;
	return (p);
}
*/

int is_array_section_ref(list)
PTR_LLND list;
{
     int num_triplets = 0;
     PTR_LLND p;
     PTR_TYPE t;

     p = list;
     while (p != LLNULL) 
     {
	  /* subscript range */
	  if (p->entry.template.ll_ptr1->variant == DDOT) num_triplets++;
	  if (t = list->entry.template.ll_ptr1->type) 
	  {
	       if (t->variant == KEYWORD_ARG) return (0);
	       /* vector subscript */
	       if ((t->variant == T_ARRAY) && (t->entry.ar_decl.base_type->variant == T_INT))
	       num_triplets++;
	  }
	  p = p->entry.template.ll_ptr2;
     }
     
     return (num_triplets);
}

int is_substring_ref(list)
PTR_LLND list;
{    PTR_LLND l1;
     
     l1 = list->entry.template.ll_ptr1;
     if (l1->variant == KEYWORD_ARG) return (0);
     if (list && (list->entry.template.ll_ptr2 == LLNULL) && 
	 l1 && (l1->variant == DDOT) && 
	 (l1->entry.template.ll_ptr1->variant == DDOT))
	  return (1);
     else return (0);
}

void
bind()
{
     PTR_SYMB s, sym;
     PTR_LLND tmp;
     PTR_HASH hash_node;
     
     while (first_unresolved_call != LLNULL) 
     {
	  s = first_unresolved_call->entry.template.symbol;
	  if (s->decl == 0)
	  {
	       hash_node = just_look_up_sym_in_scope(s->scope->control_parent, s->ident);
	       if (hash_node == HSNULL) 
	       {
		    s->scope = global_bfnd;
	       }
	       else if (hash_node->id_attr->variant == s->variant) 
	       {
		    sym = hash_node->id_attr;
		    /*  remove s from symbol table. */
		    s->parent->id_attr = SMNULL;
		    first_unresolved_call->entry.template.symbol = sym;
	       }
	       else  errstr("sym.c:bind: inconsistent call %s", s->ident);
	  }
			 
	  tmp = first_unresolved_call;
	  first_unresolved_call = tmp->entry.template.ll_ptr2;
	  tmp->entry.template.ll_ptr2 = LLNULL;
     }
}     

void
late_bind_if_needed(ll_node)
PTR_LLND ll_node;
{
     PTR_SYMB s;
     
     s = ll_node->entry.template.symbol;
     if ((s->entry.var_decl.local == IO) || 
	 (s->decl == YES))
	  return;
     else 
     {
	  if (first_unresolved_call == LLNULL) 
	       last_unresolved_call = first_unresolved_call = ll_node;
	  else 
	  {
	       last_unresolved_call->entry.template.ll_ptr2 = ll_node;
	       last_unresolved_call = ll_node;
	  }
     }	  
}       
