
/* Brian Bliss Demo- 
   This program illustrates the features that Brian needs for the CSRD
   project at this point.  It does the following.
	1. creates logical declarations of arrays.
	2. creates labeled common statements
	3. get a list of all symbols that are implicitly or explicitly
	   declared in a subroutine or function.
	4. create a conditional.
	5. create a call statement with a string parameter.
	6. dump the dep file.

	run the probram with brian x.proj
	output goes to stdout and file test.f
	the dep file is in test.dep
*/


#include <stdio.h>
#include "SiDefines.h"
#include "SiEList.h"
#include "SiFuns.h"
#include "tag"

int SiMakeCommStat();
ELIST SiLocalVarList();

int debug = 0;
int show_deps = 0;/* these two definitions are needed for the library */

int event_id = 0;
FILE *outfile;
     
main(argc, argv)
int argc;
char **argv;
{

	ELIST e, enames, subname,procs;
	char *proj_name;
	int i, j;
	char *filename;
        debug = 0; show_deps = 0;
	
	if(argc == 1) {
		printf("usage: instr project_name.proj\n");
		exit(0);
		}
	proj_name = argv[1];	
	printf("opening project: %s\n",proj_name);
	if( SiOpenProject(proj_name,proj_name) < 0){
		printf("*** could not open project ***\n");
		exit(0);
		}
	
	if((outfile = fopen("test.f","w")) == NULL){
		printf("could not open file for output\n");
		exit(0);
	      }

	procs = SiGetProcIdList(proj_name);
	printf("\n proc list:\n");
	enames =  SiGetProcList(proj_name);
	SiUnparseEList(enames); 
	filename = SiECadar(enames);
	SiSelectFile(filename);
 	printf("start analysis\n");
	for(e = procs; e; e = SiECdr(e)){
		filename = SiECadar(enames);
		printf("\nIn source file %s  and doing analysis for-\n",filename);
		subname = SiECddar(enames);
		printf("subroutine =%s=\n",SiECar(subname));
		SiSelectFile(filename);
                transform_routine(e);
		enames = SiECdr(enames);
		}
	fclose(outfile);
	SiWriteDepFile("test.dep");
}

transform_routine(e)
ELIST e;
{
      ELIST x;
      insert_common_vars((int) SiECar(e));
      add_conditional((int) SiECar(e));
      x = SiUnparseBlock((int) SiECar(e)); 
      /* SiUnparseEList(x); */
     fprintf(outfile, (char *) SiECar(x));
}
         


insert_common_vars(fun_hdr)
int fun_hdr;
{
	int stmt, e, s, ref, t;
	ELIST el, text;
	ELIST EAppend();
        int arglist[2];
	int last_decl, stmt2, stmt3;

	/* make a declaration for bpcfs as an int and make a common statement */
	s = SiNewVarInt(TYPE_INT, fun_hdr, "bpcfs");
	ref = SiNewSymbolRef(VAR_REF, s, -1);
	last_decl = SiMakeCommonStat(fun_hdr, "bpcf", ref);

	/* make a declaration for LOGICAL bpcfsa(1) and a "bpcfs" common block */
	/* as per request from bb. */
        /* make a (1) expression */
	e = SiNewExpr(EXPR_LIST, SiNewConstant(EAppend(ENumber(TYPE_INT),ENumber(1))), -1);
	/* make an array type of the from LOGICAL(1) */
        t = SiMakeArrayType(TYPE_BOOL,1, e);
	s = SiNewVarInt(t, fun_hdr, "bpcfsa");
	ref = SiNewSymbolRef(ARRAY_REF, s, e);
	stmt = SiMakeCommonStat(fun_hdr, "bpcfs", ref);

	/* now for fun let's list all the local variables */
	el = SiLocalVarList(SiGetStmtSymbol(fun_hdr));
	while(el){
		printf("\n local var = %s", SiVarIdent((int) SiECar(el)) );
		el = SiECdr(el);	
		}
	printf("\n end local list \n");
}

int executable_stmt(stmt_id)
int stmt_id;
{

  if( SiStmtVariant(stmt_id) <= 168 && SiStmtVariant(stmt_id) >=152)
   return 0;
  else return 1;
}


add_conditional(fun_hdr)
int fun_hdr;
{
	int st, stmt, e, s, ref, ref2, t;
	ELIST el, text, q;
	ELIST EAppend();
	int last_decl, stmt2, stmt3;

	/* Now add an if statment with a subroutine call */
	/* we shall use the same subroutine as the give scope, i.e. an illegal */
	/* recursive call. the parameters to the call will be */
	/*  call <sub_name>(ready(23),"hello"))               */

	/* first make the call stmt */
	st= SiNewExpr(EXPR_LIST, SiNewConstant(EAppend(ENumber(TYPE_STRING), EString("hello"))),
		     -1);
	e = SiNewExpr(EXPR_LIST, SiNewConstant(EAppend(ENumber(TYPE_INT),ENumber(23))), -1);
        t = SiMakeArrayType(TYPE_BOOL,1, e);
	s = SiNewVarInt(t, fun_hdr, "ready");
	ref = SiNewSymbolRef(ARRAY_REF, s, e);
	ref = SiCopyExpr(ref);
	ref2 = SiNewExpr(EXPR_LIST, ref, st);
	stmt3 = SiNewStmtInt(PROC_STAT, SiGetStmtSymbol(fun_hdr), ref2, -1, -1);
	text = SiUnparseStmt(stmt3);
	SiUnparseEList(text);

	/* make the if statement */
	ref = SiNewExpr(EXPR_LIST, SiCopyExpr(ref), -1);
	stmt = SiNewStmtInt(IF_NODE, -1, ref, -1, -1);
	SiAddTrueChild(stmt, stmt3);
	stmt3 = SiNewStmtInt(CONTROL_END, -1, -1, -1, -1);
	SiAddTrueChild(stmt, stmt3);
	/*
	text = SiUnparseBlock(stmt);
	SiUnparseEList(text);
	*/

        /* find insertion point */
	q = SiControlChildren(fun_hdr);
	q = (ELIST) SiECar(q);
	stmt3 = -1;
	while(!executable_stmt(stmt2 = (int) SiECar(q))){
		stmt3 = stmt2;
		q = SiECdr(q);
		}
	if(stmt3 > 0){
	    last_decl = stmt3;
	    el = EAppend(NULL, ENumber(last_decl));
	    el = EAppend(el, ENumber(1));
	    el = EAppend(el, ENumber(stmt));
	    SiInsertStmt(el);
	   }
}
