      subroutine conduct(dtnph,dte,kenn,lenn,eloss)
      implicit double precision (a-h,o-z)
 
c         this routine performs the heat conduction sweep using
c      the lu decomposition algorithm with operator splitting.
c      the heat conduction delta_t (-dte-) is also computed here,
c      as well as the calculation for the heat flow across the
c      boundary (-hn-).

      parameter( MAXKM=54, MAXLM=54 )

      common /main/
     $     r(MAXKM,MAXLM),z(MAXKM,MAXLM),u(MAXKM,MAXLM),v(MAXKM,MAXLM),
     $     aj(MAXKM,MAXLM),energy(MAXKM,MAXLM),p(MAXKM,MAXLM),
     $     q(MAXKM,MAXLM),temp(MAXKM,MAXLM),rho(MAXKM,MAXLM),
     $     dtau(MAXKM,MAXLM),mass(MAXKM,MAXLM),nbc(MAXKM,MAXLM)
      double precision mass

      common /workspc/
     $    a(MAXKM,MAXLM),b(MAXKM,MAXLM),cbb(MAXKM,MAXLM),
     $    dbb(MAXKM,MAXLM),oldtemp(MAXKM,MAXLM),sig(MAXKM,MAXLM),
     $    cc(MAXKM,MAXLM)
      common /klspace/ kmn,lmn,kmx,lmx
 
 
      double precision kappa
      data tflr /0.0001/
 
c      electron conduction -lu decomposition-
 
      kmnp = kmn + 1
      lmnp = lmn + 1
 
c      set up material properties for all internal zones
 
CVD$ CNCALL
      do 10 l=lmnp,lmx
CVD$ CNCALL
CVD$ NODEPCHK
        do 10 k=kmnp,kmx
          cc(k,l) = (0.0001*sqrt(temp(k,l))*temp(k,l)**2)/aj(k,l)
          kappa = eos(dummy,dummy,3)
          sig(k,l) = mass(k,l)*kappa/dtnph
          oldtemp(k,l) = temp(k,l)
   10 continue
 
c      set up left-hand boundary to have same properties as neighbor
c      zone to the right.
 
      do 12 l=lmnp,lmx
        cc(kmn,l) = cc(kmn+1,l)
   12 continue
 
c      set up right-hand boundary to have same properties as neighbor
c      zone to the left.
 
      do 14 l=lmnp,lmx
        cc(kmx+1,l) = cc(kmx,l)
   14 continue
 
c      set up bottom boundary to have same properties as neighbor
c      zone above it.
 
      do 16 k=kmnp,kmx
        cc(k,lmn) = cc(k,lmn+1)
   16 continue
 
c      set up top boundary to have same properties as neighbor
c      zone below it.
 
      do 18 k=kmnp,kmx
        cc(k,lmx+1) = cc(k,lmx)
   18 continue
 
c      coupling constants in the k-direction
 
      do 20 l=lmnp,lmx
        do 20 k=kmn,kmx
          dbb(k,l) = (2.0*cc(k+1,l)*cc(k,l))/(cc(k+1,l)+cc(k,l))
     1     * ( 0.5*(r(k,l-1)+r(k,l))*((r(k,l)-r(k,l-1))**2
     2        +(z(k,l)-z(k,l-1))**2) )
   20 continue
 
c      coupling constants in the l-direction
 
      do 30 k=kmnp,kmx
        do 30 l=lmn,lmx
          cbb(k,l) = (2.0*cc(k,l)*cc(k,l+1))/(cc(k,l)+cc(k,l+1))
     1     * ( 0.5*(r(k-1,l)+r(k,l))*((r(k,l)-r(k-1,l))**2
     2        +(z(k,l)-z(k-1,l))**2) )
   30 continue
 
c      boundary conditions at kmin = source ( left-hand boundary).
c      boundary conditions at kmax = source (right-hand boundary).
 
      do 40 l=lmn,lmx
        a(kmn,l) = 0.0
        b(kmn,l) = temp(kmn,l)
   40 continue
 
c      boundary conditions at lmin = mirror (lower boundary).
c      boundary conditions at lmax = mirror (upper boundary).
 
      do 50 k=kmn,kmx
        a(k,lmn) = 0.0
        b(k,lmn) = temp(k,lmn)
        cbb(k,lmn) = 0.0
        cbb(k,lmx) = 0.0
   50 continue
 
c           alpha,beta forward sweep (l direction)
 
      do 100 k=kmnp,kmx
        do 110 l=lmnp,lmx
          denom  = sig(k,l)+cbb(k,l)+cbb(k,l-1)*(1.-a(k,l-1))
          a(k,l) = cbb(k,l)/denom
          b(k,l) = (sig(k,l)*temp(k,l)+cbb(k,l-1)*b(k,l-1))/denom
  110   continue
 
c          back substitution sweep
 
        ml = lmx+1
        do 120 l=lmnp,lmx
          ml = lmx - l+lmnp
          temp(k,ml) = a(k,ml)*temp(k,ml+1)+b(k,ml)
  120   continue
  100 continue
c$$$      write(3,'(/''End of L-Sweep'')')
c$$$      write(3,'(''   K   L  Alpha          Beta           Temp'')')
c$$$      do k = kmnp, kmx
c$$$         do l=lmnp, lmx
c$$$            write(3,'(i4,i4,3e15.8)') k,l,a(k,l),b(k,l),temp(k,l)
c$$$         enddo
c$$$      enddo
 
c           alpha,beta forward sweep (k direction)
 
      do 200 l=lmnp,lmx
        do 210 k=kmnp,kmx
          denom  = sig(k,l)+dbb(k,l)+dbb(k-1,l)*(1.-a(k-1,l))
          a(k,l) = dbb(k,l)/denom
          b(k,l) = (sig(k,l)*temp(k,l)+dbb(k-1,l)*b(k-1,l))/denom
  210   continue
 
c          back substitution sweep
 
        ml = kmx+1
        do 220 k=kmnp,kmx
          ml = ml-1
          temp(ml,l) = a(ml,l)*temp(ml+1,l)+b(ml,l)
  220   continue
  200 continue
c$$$      write(3,'(/''End of K-Sweep'')')
c$$$      write(3,'(''   K   L  Alpha          Beta           Temp'')')
c$$$      do k = kmnp, kmx
c$$$         do l=lmnp, lmx
c$$$            write(3,'(i4,i4,3e15.8)') k,l,a(k,l),b(k,l),temp(k,l)
c$$$         enddo
c$$$      enddo
 
c      compute dt control for heat conduction
c      compute new zonal energies after heat conduction
 
      ye  = -1.0
CVD$ CNCALL
      do 300 l=lmnp,lmx
CVD$ CNCALL
         do 300 k=kmnp,kmx
            energy(k,l) = eos(temp(k,l) , rho(k,l) , 2)
            temp(k,l) = MAX(temp(k,l) , tflr)
            tempr = abs((temp(k,l)-oldtemp(k,l))/oldtemp(k,l))
            if(tempr .gt. ye) then
               ye = tempr
               ke = k
               le = l
            endif
  300 continue
      if(ye .ne. 0.0) dte = (0.1*dtnph)/ye
      kenn = ke
      lenn = le
 
c      calculate energy flow across boundaries this time step.
 
      sum = 0.0
      do 500 k=kmn,kmx
        sum = sum + cbb(k,lmn)*(temp(k,lmn)-temp(k,lmn+1))
     1            + cbb(k,lmx)*(temp(k,lmx+1)-temp(k,lmx))
  500 continue
 
      do 510 l=lmn,lmx
        sum = sum + dbb(kmn,l)*(temp(kmn,l)-temp(kmn+1,l))
     1            + dbb(kmx,l)*(temp(kmx+1,l)-temp(kmx,l))
  510 continue
      eloss = sum*dtnph
 
      return
      end
