


#include <stdio.h>
#include "SiEList.h"
#include "SiFuns.h"
#include "tag"

int debug = 0;
int show_deps = 0;  /* these two definitions are needed for the library */

int event_id = 0;
FILE *outfile;
#define ENTRY_EVENT 0
#define EXIT_EVENT 1
#define LOOP_ENTRY 2
#define LOOP_EXIT 3
int eventid = 0;
     
main(argc, argv)
int argc;
char **argv;
{

	ELIST e, enames, subname,procs;
	char *proj_name;
	int i, j;
	char *filename;

	if(argc == 1) {
		printf("usage: instr project_name.proj\n");
		exit(0);
		}
	proj_name = argv[1];	
	printf("opening project: %s\n",proj_name);
	if( SiOpenProject(proj_name,proj_name) < 0){
		printf("*** could not open project ***\n");
		exit(0);
		}
	
	if((outfile = fopen("test.f","w")) == NULL){
		printf("could not open file for output\n");
		exit(0);
	      }
#if NEEL
	if(SiDoGlobalAnalysis() == 0)
		printf("global analysis done\n");
	else{
	    printf("global analysis failed\n");
	    exit(0);
	    }
#endif
	procs = SiGetProcIdList(proj_name);
	printf("\n proc list:\n");
	enames =  SiGetProcList(proj_name);
	SiUnparseEList(enames); 
	printf("start analysis\n");
	for(e = procs; e; e = SiECdr(e)){
		filename = SiECadar(enames);
		printf("\nIn source file %s  and doing analysis for-\n",filename);
		subname = SiECddar(enames);
		printf("subroutine =%s=\n",SiECar(subname));
		SiSelectFile(filename);
                instrument(e);
		SiSelectFile(filename);
		enames = SiECdr(enames);
		}
	fclose(outfile);
}

instrument(e)
ELIST e;
{
      ELIST x;
      fix_nested_loops((int) SiECar(e),0); 
      control_traversal_template((int) SiECar(e),0,0, &eventid); 
      x = SiUnparseBlock((int) SiECar(e));
      SiUnparseEList(x); 
     fprintf(outfile, (char *) SiECar(x));
    }
         
int entry_done = 0;  /* used to signal procedure entry mark completion */
int exit_done = 0;

/* insturment will do the following:  
/* first it will decide which type of statement we have
/* then it will decide on the appropriate insturmentation action
/* and then it will insturment all the control children of the statement  */

int executable_stmt(stmt_id)
int stmt_id;
{
  
  if( SiStmtVariant(stmt_id) <= 168 && SiStmtVariant(stmt_id) >=152)
   return 0;
  else return 1;
}

int labels[20];
int loops[20];
int events[20];
int routine_event = 0;

control_traversal_template( stmt_id, level, loop_level, eventid)
int stmt_id;
int level, loop_level;
int *eventid;
{
  ELIST q, q0,branch, q1;
  int child_stmt, saved_ev_no, return_done, last_stmt;
  int first_executable_found = 1;
  printf("variant = %d\n", SiStmtVariant(stmt_id));
  if(SiStmtVariant(stmt_id) == PROG_HEDR || SiStmtVariant(stmt_id) == PROC_HEDR
     || SiStmtVariant(stmt_id) == FUNC_HEDR){
       /* do any header specific action here */
       routine_event = (*eventid)++;
       loops[level] = stmt_id;
       events[level] = routine_event;
       first_executable_found = 0;
       }
  q = SiControlChildren(stmt_id);
  /* q is a list of the form ( (true branch) (false branch) ) */
  /* traverse true branch then false branch                   */
  SiUnparseEList(q); 
  last_stmt = -1;
  if(q != NULL) branch = (ELIST) SiECar(q); else branch = NULL;		     
  
  while(q != NULL){
    branch = (ELIST) SiECar(q);
    /* loop through each child statement on the branch */
    while(branch != NULL ){
      child_stmt = (int) SiECar(branch);
      if(SiStmtVariant(child_stmt) != CONTROL_END) return_done = 0;
      if(first_executable_found == 0 && executable_stmt(child_stmt)){
	  first_executable_found = 1;
	  insert_call_node(ENTRY_EVENT, child_stmt,routine_event);
	  }
      last_stmt = child_stmt;
      switch(SiStmtVariant(child_stmt )){
	  case FOR_NODE:     /* basic loop node: "do" in Fortran */
	  case WHILE_NODE:   /* C while 			 */
	  case CDOALL_NODE:  /* Cedar DoALL node		 */
                      loops[loop_level+1] = child_stmt;
	              insert_call_node(LOOP_ENTRY, child_stmt,
                                        saved_ev_no = (*eventid)++);
		      events[loop_level+1] = saved_ev_no;
		      control_traversal_template(child_stmt,level+1,
                                                 loop_level+1, eventid);
		      insert_call_after_node(LOOP_EXIT, child_stmt,saved_ev_no);
		      break;
	  case IF_NODE:      /* a if node with a "then"        */
	  case ELSEIF_NODE:  /* a matching "else if ... "then" */
		      control_traversal_template(child_stmt,level+1, 
                                                 loop_level, eventid );
		      break;
	  case	LOGIF_NODE:  /* an if without a then */
		      control_traversal_template(child_stmt,level+1, 
                                                 loop_level, eventid);
		      break;
	  case CONTROL_END: break;
	  case GOTO_NODE:
                     check_for_loop_exit(child_stmt, loop_level+1);
		     break;
	  case RETURN_STAT:
	  case RETURN_NODE:
		     insert_call_node(EXIT_EVENT, child_stmt, routine_event);
                     return_done = 1;
		     break;
	  default: 
		  /* the rest of the statement types are here.*/
		  break;
	    } 
     branch = SiECdr(branch);
     } /* go to next stmt on branch  */
  q = SiECdr(q); /*is there another branch? */
  } /* end outer while. switch to new branch */
  if(return_done == 0 && level == 0) 
        insert_call_node(EXIT_EVENT, last_stmt, routine_event);
}

check_for_loop_exit(stmt, level)
int stmt, level;
{
    int goto_lab;
    int target;
    int parent, par_lev, i, ce;
    
    goto_lab = SiGetGotoLabel(stmt);
    if(goto_lab == -1) return -1;
    if((target = SiFindLabelledStmt(stmt,goto_lab)) == -1) return -1;
    parent = SiFindCommonParent(target,stmt);
    for(par_lev = level; par_lev >=0; par_lev--)
       if(loops[par_lev] == parent) break;
    if(par_lev < 0) return -1;
    /* convert to ifthen else if needed */
    for(i = level-1; i > par_lev; i--){
       if(SiStmtVariant(SiControlParent(stmt)) == LOGIF_NODE){
              SiChangeStmtVariant(SiControlParent(stmt), IF_NODE);
              ce = SiNewStmtInt(CONTROL_END, -1, -1, -1, -1);
	      SiInsertStmt(SiEAppend(SiENumber(stmt), 
		        SiEAppend( SiENumber(1), SiENumber(ce))));
	    }
       insert_call_node(LOOP_EXIT, stmt, events[i]);
     }
    return 0;
}
   
insert_call_node(event_type, following_stmt, argument)
    int event_type; 
    int following_stmt;
    int argument;
{
    ELIST arg;
    int arg_id,  arg_list, call_stat, symb;
    
    arg_id = SiNewConstant( SiEAppend(SiENumber(1),SiENumber(argument)));
    arg_list = SiNewExpr(EXPR_LIST,arg_id, -1);
    switch(event_type){
    case ENTRY_EVENT:
             symb = SiMakeSubroutineName("entryevent");
             break;
     case EXIT_EVENT:
             symb = SiMakeSubroutineName("exitevent");
             break;
     case LOOP_ENTRY:
             symb = SiMakeSubroutineName("loopentry");
             break;
     case LOOP_EXIT:
             symb = SiMakeSubroutineName("loopexit");
             break;
	   }
    arg = SiENumber(PROC_STAT);
    arg = SiEAppend(arg, SiENumber(symb));
    arg = SiEAppend(arg, SiENumber(arg_list));
    arg = SiEAppend(arg, SiENumber( -1 ));
    arg = SiEAppend(arg, SiENumber( -1 ));
    call_stat = SiNewStmt(arg);
    SiReplaceStmt(SiEAppend(SiENumber(following_stmt), SiENumber(call_stat)));
    SiInsertStmt(SiEAppend(SiENumber(call_stat), 
		        SiEAppend( SiENumber(1), SiENumber(following_stmt))
               ));
}

insert_call_after_node(event_type, statement, argument)
    int event_type; 
    int statement;
    int argument;
{
    ELIST arg;
    int arg_id,  arg_list, call_stat, symb;
    
    arg_id = SiNewConstant( SiEAppend(SiENumber(1),SiENumber(argument)));
    arg_list = SiNewExpr(EXPR_LIST,arg_id, -1);
    switch(event_type){
    case ENTRY_EVENT:
             symb = SiMakeSubroutineName("entryevent");
             break;
     case EXIT_EVENT:
             symb = SiMakeSubroutineName("exitevent");
             break;
     case LOOP_ENTRY:
             symb = SiMakeSubroutineName("loopentry");
             break;
     case LOOP_EXIT:
             symb = SiMakeSubroutineName("loopexit");
             break;
	   }
    arg = SiENumber(PROC_STAT);
    arg = SiEAppend(arg, SiENumber(symb));
    arg = SiEAppend(arg, SiENumber(arg_list));
    arg = SiEAppend(arg, SiENumber( -1 ));
    arg = SiEAppend(arg, SiENumber( -1 ));
    call_stat = SiNewStmt(arg);
    SiInsertStmt(SiEAppend(SiENumber(statement), 
		        SiEAppend( SiENumber(1), SiENumber(call_stat))
               ));

}


fix_nested_loops( stmt_id, lev)
int stmt_id, lev;
{
  ELIST q, branch, lab_refs;
  int child_stmt, last_stmt,i, found;
  int do_label;
  ELIST SiFindLabelRefs();

  q = SiControlChildren(stmt_id);
  /* q is a list of the form ( (true branch) (false branch) ) */
  /* traverse true branch then false branch                   */
  /* SiUnparseEList(q); */
  last_stmt = -1;
  if(q != NULL) branch = (ELIST) SiECar(q); else branch = NULL;		     
  
  while(q != NULL){
    branch = (ELIST) SiECar(q);
    /* loop through each child statement on the branch */
    while(branch != NULL ){
      child_stmt = (int) SiECar(branch);
      last_stmt = child_stmt;
      switch(SiStmtVariant(child_stmt )){
	  case FOR_NODE:     /* basic loop node: "do" in Fortran */
	  case WHILE_NODE:   /* C while 			 */
	  case CDOALL_NODE:  /* Cedar DoALL node		 */
	              do_label = SiGetDoLabel(child_stmt);
                      loops[lev] = child_stmt;
		      if(do_label > -1){
                          labels[lev] = SiGetFortranLabelNo(do_label);
                          for(found = 0, i = 0; i < lev; i++)
			    if(labels[i] == labels[lev]) found = 1;
                          if(found)
			        replace_the_label(lev, child_stmt,do_label);
			}
       		      fix_nested_loops(child_stmt,lev+1);
		      break;
	  case IF_NODE:      /* a if node with a "then"        */
	  case ELSEIF_NODE:  /* a matching "else if ... "then" */
		      fix_nested_loops(child_stmt, lev);
		      break;
	  case	LOGIF_NODE:  /* an if without a then */
		      fix_nested_loops(child_stmt, lev);
		      break;
	  default: 
		  /* the rest of the statement types are here.*/
		  break;
	    } 
     branch = SiECdr(branch);
     } /* go to next stmt on branch  */
  q = SiECdr(q); /*is there another branch? */
  } /* end outer while. switch to new branch */
}

replace_the_label(lev, child_stmt,do_label)
int lev, child_stmt, do_label;
{
ELIST lab_refs, ref, exps, arg,children;
int new_label, parent_label;
int label_val;
int parent_loop, new_continue;

    new_label = SiMakeUniqueLabel();
    label_val = SiGetFortranLabelNo(new_label);
    labels[lev] = label_val;
    lab_refs = SiFindLabelRefs(child_stmt, do_label);
    SiUnparseEList(lab_refs);
    while(lab_refs){
      ref = (ELIST) SiECar(lab_refs);
      if(SiECdr(ref) == NULL){
	  SiReplaceLabelInStmt((int)SiECar(ref),new_label);
	  exps = NULL;
	}
      else exps = (ELIST) SiECadr(ref);
      while(exps){
         SiReplaceLabelInExpr((int)SiECar(exps), new_label);
	 exps = SiECdr(exps);
       }
      lab_refs = SiECdr(lab_refs);
    }
    SiReplaceLabelInDo(child_stmt, new_label);
    parent_loop = loops[lev-1];
    parent_label = SiGetDoLabel(parent_loop);
    arg = SiENumber(CONT_STAT);
    arg = SiEAppend(arg, SiENumber( -1 ));
    arg = SiEAppend(arg, SiENumber( -1 ));
    arg = SiEAppend(arg, SiENumber( -1 ));
    arg = SiEAppend(arg, SiENumber( -1 ));
    new_continue = SiNewStmt(arg);
    SiReplaceLabelInStmt(new_continue, parent_label);
    children = SiControlChildren(parent_loop);
    children = (ELIST) SiECar(children);
    while(SiECdr(children)) children = SiECdr(children);
    SiInsertStmt(SiEAppend(SiENumber((int)SiECar(children)), 
		        SiEAppend( SiENumber(1), SiENumber(new_continue))
               ));
    
}

