
#include <stdio.h>
#include "SiEList.h"
#include "SiFuns.h"
#include "tag"

struct ref_row
{
  int decidable; /* 0 if not decidable.  if so, goto symbol */
  int c[8];	  /* up to 8 loops deep */
  int offset;
  int expr;	/* this is id of subscript expr */
};

typedef struct dependence *DEPPTR;

struct dependence
{
int refnode;
int future;
DEPPTR next;
};
   
struct reference{
  int SiVarId;  /* id of variable */
  int ref_stmt, ref_expr;
  int region;
  int is_def;  /* 1 if a definition, 0 if a use */
  int dimension; /* 0 for scalar, 1 for vector */
  int depth;	 /* loop nest depth at occurrence */
  int loops[8];  /* loop id's */
  struct ref_row coefs[6];
  DEPPTR defs; 
} reflist[5000];

struct loop{
  int SiVarId;
  int stmt_id;
  int SiLineNumber;
  int low_range, hi_range, stride;
} loops[100];

typedef struct loop * loop_ptr;

int loop_stack[20];

int num_refs = 0;
int num_loops = 0;


int debug = 0;
int show_deps = 0;  /* these two definitions are needed for the library */

int event_id = 0;
FILE *outfile;
#define ENTRY_EVENT 0
#define EXIT_EVENT 1
#define LOOP_ENTRY 2
#define LOOP_EXIT 3
int eventid = 0;
     
int futurecount = 0;
main(argc, argv)
int argc;
char **argv;
{

   ELIST e, enames, subname,procs;
   char *proj_name;
   int i, j;
   char *filename;

   if (argc == 1) 
   {
      printf("usage: instr project_name.proj\n");
      exit(0);
   }
   proj_name = argv[1];	
   printf("opening project: %s\n", proj_name);
   if (SiOpenProject(proj_name,proj_name) < 0)
   {
      printf("*** could not open project ***\n");
      exit(0);
   }
	
   if ((outfile = fopen("test.f","w")) == NULL)
   {
      printf("could not open file for output\n");
      exit(0);
   }
   if (SiDoGlobalAnalysis() == 0)
      printf("global analysis done\n");
   else
   {
      printf("global analysis failed\n");
      exit(0);
   }
   procs = SiGetProcIdList(proj_name);
   printf("proc list:\n");
   enames =  SiGetProcList(proj_name);
   SiUnparseEList(enames); 
   printf("start analysis\n");
  
   for (e = procs; e; e = SiECdr(e))
   {
      filename = SiECadar(enames);
      printf("In source file %s  and doing analysis for-\n",filename);
      subname = SiECddar(enames);
      printf("subroutine =%s=\n",SiECar(subname));
      SiSelectFile(filename);
      count(e);
      enames = SiECdr(enames);
   }
	/* SiCrossReference(); */
   printf("backpatching...\n");
   back_patch();
   dump_table();
   fclose(outfile);
} /* main */


int not_induction_var(depth, var)
int depth, var;
{
   int i;

   for (i = 0; i < depth; i++)
     if (loops[loop_stack[i]].SiVarId == var) 
        return 0;
   return 1;
} /* not_induction_var */

int make_loop_rec(stmt_id)
int stmt_id;
{
   int i; 
   i = num_loops++;
   loops[i].stmt_id = stmt_id;
   loops[i].SiVarId = SiGetStmtSymbol(stmt_id);
   loops[i].SiLineNumber = SiLineNumber(stmt_id);
   return i;
} /* make_loop_rec */

dump_table()
{
   int i,j,k,depth;
   ELIST expr;
   DEPPTR d;
   for (i = 0; i < num_refs; i++)
   {
      depth = reflist[i].depth;
      if (reflist[i].is_def)
      {
         printf("\n%4d DEFINITION:loop nest[",i);
         for (j = 0; j < depth-1; j++) 
            printf("%d,",reflist[i].loops[j]);
         if (depth>0) 
            printf("%d]",reflist[i].loops[depth-1]);
         else printf("]");
         printf (" %s:",SiECar(SiUnparseExpr(reflist[i].ref_expr)));
         if (reflist[i].region)
	    printf("region referenced:[%s]",
                   SiECar(SiUnparseExpr(reflist[i].region)));
         printf("\n       USES");
      }
      else
      {
         printf(" %s.{",SiECar(SiUnparseExpr(reflist[i].ref_expr)));
         d = reflist[i].defs;
         while ((d != NULL) && (d->next != NULL))
         {
	    printf("%d,",d->refnode);
	    d = d->next;
	 }
         if (d != NULL) 
            printf("%d}", d->refnode); 
         else printf("}");
      } 
   }
   printf("\n");
} /* dump_table */

DEPPTR find_defs(var, reaching)
int var;
ELIST reaching;
{
   int stmt, term;
   DEPPTR defs, p;
   ELIST pair;
   int i;

   defs = NULL;
   if (reaching == NULL)
   {
     printf("find_defs: use before definition?\n");
   }
   printf("reaching def list = ");
   SiUnparseEList(reaching);
   while (reaching)
   {
      pair = (ELIST) SiECar(reaching);
      reaching = SiECdr(reaching);
      stmt = (int) SiECar(pair);
      term = (int) SiECadr(pair);
      if (SiVarId(term) == var)
      {
	 for (i = 0; (i < num_refs) && (reflist[i].ref_expr != term); i++);
         p = (DEPPTR) malloc(sizeof(struct dependence));
         p->next = defs;
         defs = p;
	 if (i == num_refs) 
         {
            printf("find_defs:term = %d, stmt = %d, at line %d\n",
                     term, stmt, SiLineNumber(stmt));
            printf("reference not found, will be patched later!\n");
            futurecount++;
            p->future = TRUE;
            p->refnode = (int)pair; /* save the reaching definitions 
					   for later */
         }
	 else 
         {
            p->refnode = i;
            p->future = FALSE;
         }
      }
   }
   return defs;
} /* find_defs */

int trace_defs(var, defn_pair)
int var;
ELIST defn_pair;
{
   int stmt, term;
   stmt = (int) SiECar(defn_pair);
   term = (int) SiECadr(defn_pair);
   if (SiVarId(term) == var)
   {
      int i;

      for (i = 0; (i<num_refs) && (reflist[i].ref_expr != term); i++);
      if (i == num_refs)
         return -1;
      return i;
   }
   else
   { 
     printf("trace_defs: internal bug\n");
     return -1;
   }
} /* trace_defs */
   
int back_patch()
{
  int i;
  DEPPTR p;
  int bpcount = 0;
   
  for (i = 0; i<num_refs; i++)
  {
     if (reflist[i].is_def == 0) /* which means it is a USE */
     {
        for (p = reflist[i].defs; p; p = p->next)
        {
           if (p->future == TRUE)  /* that means it needs backpatching */
           {
              int temp;
              if ((temp = trace_defs(reflist[i].SiVarId, (ELIST)p->refnode))
                   == -1)
              {
		 printf("back_patch: now really there is  no reference\n");
                 p->refnode = -1;
              }
              else 
              {
                 p->refnode = temp;
                 bpcount++;
              }
           }
        }
     }
  }
  printf("should have backpatched %d cases\n", futurecount);
  printf( "backpatched %d cases\n", bpcount);
} /* back_patch */
     
int make_ref(def, depth, stmt, var_ref, reaching)
int def, depth, stmt, var_ref;
ELIST reaching;
{
  int i,j;
  ELIST SiGetSubscript();
  ELIST e, h, coef_lst, offset_expr, term, coefs;
  int expr_list, decidable, offset;
	
  i = num_refs++;
  reflist[i].SiVarId = SiVarId(var_ref);
  reflist[i].ref_stmt = stmt;
  reflist[i].ref_expr = var_ref;
  printf("reflist[%d].ref_expr = %d\n", i, var_ref);
  reflist[i].is_def = def;
  reflist[i].depth = depth;
  reflist[i].region = 0;
  if (def==0)  /* use */
     reflist[i].defs = find_defs(reflist[i].SiVarId, reaching);
  else  /* definition */
  {
     reflist[i].defs = NULL;
     if (SiExprVariant(var_ref) == ARRAY_REF)
	reflist[i].region = SiRegionAccessed(stmt,var_ref);
  }
  for (j = 0; j < depth; j++)
    reflist[i].loops[j] = loop_stack[j];
  reflist[i].dimension = 0;
  if (SiExprVariant(var_ref) != ARRAY_REF)  
     return 0;
  e = SiGetSubscript(stmt,var_ref);
  if (e == NULL)
  {
     printf("bad stmt or expression to get_subscr\n");
     return 0;
  }
  coefs = SiGetLRChild(var_ref);
  expr_list = (int) SiECar(coefs);
  for (j= 0; e; j++, e = SiECdr(e))
  {
    int k;
    h = (ELIST) SiECar(e);
    decidable = (int) SiECar(h);
    reflist[i].coefs[j].decidable = decidable;
    reflist[i].dimension++;
    h = SiECdr(h);
    coefs = (ELIST) SiGetLRChild(expr_list);
    reflist[i].coefs[j].expr = (int) SiECar(coefs);
    expr_list = (int) SiECadr(coefs);

    switch (decidable)
    {
       case 0: /* completely undecidable loop non-invariant*/
	    for (k = 0; k < 8; k++) 
	       reflist[i].coefs[j].c[k] = 0;
            reflist[i].coefs[j].offset = 0;
	    break;
       case 1: /* is decidable */
            coef_lst= (ELIST) SiECadr(h);
	    k = 0;
	    while (k < 8 && coef_lst)
            {
	       reflist[i].coefs[j].c[k++] = (int) SiECar(coef_lst);
	       coef_lst= SiECdr(coef_lst);
            }
	    if (EListP(h)) 
            {
	       reflist[i].coefs[j].expr = (int)SiECaar(h);
	       reflist[i].coefs[j].offset = 0;
	    }
	    else 
               reflist[i].coefs[j].offset = (int) SiECar(h);
	    break;
       case 2: /* a range value like a:b */
	    for (k = 0; k < 8; k++) 
	       reflist[i].coefs[j].c[k] = 0;
	    reflist[i].coefs[j].offset = (int) SiECar(h);
	    break;
    }
  }
  return i;
} /* make_ref */

count(e)
ELIST e;
{
   ELIST x;
   if (SiDoFlowAnalysis() == 0)
      printf("flow analysis also done\n");
   else
   {
      printf("flow analysis failed\n");
      exit(0);
   }
   list_use_mod((int) SiECar(e),0);
   x = SiUnparseBlock((int) SiECar(e));
   /* SiUnparseEList(x); */
   fprintf(outfile, (char *) SiECar(x));
} /* count */

int executable_stmt(stmt_id)
int stmt_id;
{
  
  if (SiStmtVariant(stmt_id) <= 168 && SiStmtVariant(stmt_id) >=152)
     return 0;
  else return 1;
}

print_uses(e)
ELIST e;
{
  ELIST a;
  int stmt, expr;

  while (e)
  {
     a = (ELIST) SiECar(e); e = SiECdr(e);
     stmt = (int) SiECar(a);
     expr = (int) SiECadr(a);
     printf(" s: %d  e: %s\n", stmt, SiECar(SiUnparseExpr(expr)));
  }
}



list_use_mod( stmt_id, depth)
int depth;
int stmt_id;
{
  ELIST q, branch;
  int child_stmt, var, var_ref;
  ELIST e, r;
       
  q = SiControlChildren(stmt_id);
  if (q != NULL) 
     branch = (ELIST) SiECar(q); 
  else 
     branch = NULL;		     
  while (q != NULL)
  {
     child_stmt = -1;
     branch = (ELIST) SiECar(q);
     /* loop through each child statement on the branch */
     while (branch != NULL )
     {
        child_stmt = (int) SiECar(branch);
        switch (SiStmtVariant(child_stmt ))
        {
	  case FOR_NODE:     /* basic loop node: "do" in Fortran */
	  case WHILE_NODE:   /* C while 			 */
	  case CDOALL_NODE:  /* Cedar DoALL node		 */
	       printf("loop:[%d]",child_stmt);
	       SiUnparseEList(SiUnparseStmt(child_stmt));
	       printf("body of loop:[%d]\n",child_stmt);
	       loop_stack[depth] = make_loop_rec(child_stmt);
	       list_use_mod(child_stmt,depth+1);
	       printf("end body of loop:[%d]\n", child_stmt);
	       break;
	  case IF_NODE:      /* a if node with a "then" 	 */
	  case ELSEIF_NODE:  /* a matching "else if ... "then"	 */
	  case LOGIF_NODE:  /* an if without a then		 */
	       printf("conditional:[%d]",child_stmt);
	       SiUnparseEList(SiUnparseStmt(child_stmt));
	       list_use_mod(child_stmt,depth);
	       printf("end of conditional:[%d]\n",child_stmt);
	       break;
	  case CONTINUE_NODE:
	  case CONTROL_END:
	  case RETURN_STAT:
	  case GOTO_NODE:
	       break;
	  default:
	       if (executable_stmt(child_stmt))
               {
		  printf("statment:[%d] at line %d\n",child_stmt, SiLineNumber(child_stmt));
                  SiUnparseEList(SiGetStmtInfo(child_stmt));
		  r = SiReachingDefs(child_stmt);
                  printf("reaching defs=");
                  SiUnparseEList(r);
		  printf("computes value for:");
		  e = SiVarsModInStmt(child_stmt);
		  while (e)
                  {
		    SiUnparseEList(SiUnparseExpr((int) SiECar(e)));
                    printf("id = %d\n", SiECar(e));
		    make_ref(1, depth, child_stmt, (int) SiECar(e), NULL);
		    e = SiECdr(e);
		  }
		  printf("and uses values:");
		  e = SiVarsUsedInStmt(child_stmt);
		  while (e)
                  {
		     var_ref = (int) SiECar(e);
		     var = SiVarId(var_ref);
		     if (not_induction_var(depth, var))
                     {
		        SiUnparseEList(SiUnparseExpr((int) SiECar(e)));
		        make_ref(0, depth, child_stmt, var_ref,r);
		     }
		     e = SiECdr(e);
		  }
		  printf("\n");
		}
		break;
        }  /* switch */
     branch = SiECdr(branch);
     } /* while: go to next stmt on branch  */
  q = SiECdr(q); /*is there another branch? */
  } /* end outer while. switch to new branch */
} /* list_use_mod */
