

#include <stdio.h>
#include "SiEList.h"
#include "SiFuns.h"
#include "tag"

int debug = 0;
int show_deps = 0;  /* these two definitions are needed for the library */

int event_id = 0;
FILE *outfile;
#define ENTRY_EVENT 0
#define EXIT_EVENT 1
#define LOOP_ENTRY 2
#define LOOP_EXIT 3
int eventid = 0;
     
main(argc, argv)
int argc;
char **argv;
{

	ELIST e, enames, subname,procs;
	char *proj_name;
	int i, j;
	char *filename;

	if(argc == 1) {
		printf("usage: %s project_name.proj\n", argv[0]);
		exit(0);
		}
	proj_name = argv[1];	
	printf("opening project: %s\n",proj_name);
	if( SiOpenProject(proj_name,proj_name) < 0){
		printf("*** could not open project ***\n");
		exit(0);
		}
	
	if((outfile = fopen("test.f","w")) == NULL){
		printf("could not open file for output\n");
		exit(0);
	      }
	if (SiDoGlobalAnalysis() == 0)
      	   printf("global analysis done\n");
	else
        {
	    printf("global analysis failed\n");
	    exit(0);
	}
        if (SiDoFlowAnalysis() == 0)
           printf("flow analysis done\n");
        else 
        {
           printf("flow analysis fails\n");
	   exit(0);
        }
	procs = SiGetProcIdList(proj_name);
        printf("unparsing output of SiGetProcIdList\n");
        SiUnparseEList(procs);
        printf("finished unparsing output of SiGetProcIdList\n");
	enames =  SiGetProcList(proj_name);
        printf("unparsing output of SiGetProcList\n");
	SiUnparseEList(enames); 
        printf("finished unparsing output of SiGetProcList\n");
      
	printf("start analysis\n");
	for (e = procs; e; e = SiECdr(e))
        {
           printf("info of all children of procedure %s\n", SiECddar(enames));
           info_of_children(SiECar(e), 0);
#if FALSE
           filename = SiECadar(enames);
	   printf("\nIn source file %s  and doing analysis for-\n",filename);
	   subname = SiECddar(enames);
	   printf("subroutine =%s=\n",SiECar(subname));
	   SiSelectFile(filename);
           instrument(SiECar(e)); /* the current subroutine id */
           enames = SiECdr(enames);
#endif
	}
	fclose(outfile);
}

do_indent(indent)
int indent;
{
   int i;
   for (i= 0; i<indent; i++)
      putchar(' ');
}
info_of_children(stmt_id,indent)
int stmt_id;
int indent;
{
   ELIST e, branch;
   int  flip;
   ELIST varlist;
   int ix;

   e = SiControlChildren(stmt_id);
   do_indent(indent);
   printf("expanding children of stmt with id %d at line %d\n", 
           stmt_id, SiLineNumber(stmt_id));
  
   flip = TRUE;
   while (e)  /* this loop will happen only twice */
   {
      do_indent(indent);
      if (flip) 
      {
         printf("<true>:\n");
         flip = !flip;
      }
      else
         printf("<false>:\n");
      branch = (ELIST)SiECar(e);
      info_of_siblings(branch, indent + 3);
      e = SiECdr(e);
   }
}

char *get_type_string(stmt_id)
int stmt_id;
{
   int type = SiStmtVariant(stmt_id);

   switch (type)
   {
      case FOR_NODE: return "DO";
      case IF_NODE: return "IF";
      case ARITHIF_NODE: printf("neel\n");
                         SiUnparseEList(SiIfLabels(stmt_id));
                         return "ARITH_IF";
      case ELSEIF_NODE: return "IF-ELSE";
      case LOGIF_NODE:  return "LOG_IF";
      case CONTROL_END: return "CONTROL_END";
      case GOTO_NODE:  return "GOTO";
      case RETURN_STAT: return "RETURN";
      default: return "UNKNOWN";
   }
}

info_of_siblings(s, indent)
ELIST s;
int indent;
{
   int child;
   int ix;
   for (ix = 1; s; ix++, s = SiECdr(s))
   {
     child = (int)SiECar(s);
     do_indent(indent);
     printf("child[%d]: stmt_id %d at line number %d, type = %s\n", 
             ix, child, SiLineNumber(child), get_type_string(child));
     info_of_children(child, indent);
   }
} /* info_of_siblings */
