
/* This file tests the usage of the function SiDefsToUses functions */

#include <stdio.h>
#include "SiEList.h"
#include "SiFuns.h"
#include "tag"

int debug = 0;
int show_deps = 0;  /* these two definitions are needed for the library */

int event_id = 0;
FILE *outfile;
#define ENTRY_EVENT 0
#define EXIT_EVENT 1
#define LOOP_ENTRY 2
#define LOOP_EXIT 3
int eventid = 0;
     
main(argc, argv)
int argc;
char **argv;
{

	ELIST e, enames, subname,procs;
	char *proj_name;
	int i, j;
	char *filename;

	if(argc == 1) {
		printf("usage: %s project_name.proj\n", argv[0]);
		exit(0);
		}
	proj_name = argv[1];	
	printf("opening project: %s\n",proj_name);
	if( SiOpenProject(proj_name,proj_name) < 0){
		printf("*** could not open project ***\n");
		exit(0);
		}
	
	if((outfile = fopen("test.f","w")) == NULL){
		printf("could not open file for output\n");
		exit(0);
	      }
	if (SiDoGlobalAnalysis() == 0)
      	   printf("global analysis done\n");
	else
        {
	    printf("global analysis failed\n");
	    exit(0);
	}
	procs = SiGetProcIdList(proj_name);
        printf("unparsing output of SiGetProcIdList\n");
        SiUnparseEList(procs);
        printf("finished unparsing output of SiGetProcIdList\n");
	enames =  SiGetProcList(proj_name);
        printf("unparsing output of SiGetProcList\n");
	SiUnparseEList(enames); 
        printf("finished unparsing output of SiGetProcList\n");
      
	printf("start analysis\n");
	for (e = procs; e; e = SiECdr(e))
        {
           printf("info of all children of procedure %s\n", SiECddar(enames));
           info_of_proc(e, 0);
           filename = SiECadar(enames);
	   printf("\nIn source file %s  and doing analysis for-\n",filename);
	   subname = SiECddar(enames);
	   printf("subroutine =%s=\n",SiECar(subname));
	   SiSelectFile(filename);
           enames = SiECdr(enames);
	}
	fclose(outfile);
}

int executable_stmt(stmt_id)
int stmt_id;
{
  
  if (SiStmtVariant(stmt_id) <= 168 && SiStmtVariant(stmt_id) >=152)
     return 0;
  else return 1;
}

int issimple(stmt_id)
int stmt_id;
{
   int type = SiStmtVariant(stmt_id);
   if (type == FOR_NODE ||
       type == IF_NODE ||
       type == ELSEIF_NODE ||
       type == LOGIF_NODE ||
       type == CONTROL_END ||
       type == GOTO_NODE ||
       type == RETURN_STAT 
      ) return 0; 
    if (type == ARITHIF_NODE)
    {
        printf("arith if node\n");
        SiUnparseEList(SiIfLabels(stmt_id));
    }
    return 1;
}
info_of_proc(e)
ELIST e;
{

   if (SiDoFlowAnalysis() == 0)
      printf("flow analysis done\n");
   else 
   {
      printf("flow analysis fails\n");
      exit(0);
   }
   info_of_stmt(SiECar(e));
}
   
info_of_stmt(stmt_id)
int stmt_id;
{
   ELIST q, branch;
   int child_stmt;
   ELIST  def_to_use_lst;

   q = SiControlChildren(stmt_id);
   while (q)
   {
     branch = (ELIST)SiECar(q);
     while (branch)
     {
     child_stmt = (int)SiECar(branch);
     if (issimple(child_stmt))
     {
        if (executable_stmt(child_stmt))
        {
           printf("expanding children of stmt with id %d at line %d\n", 
                child_stmt, SiLineNumber(child_stmt));
           def_to_use_lst = SiDefsToUses(child_stmt); 
              
           printf("All the uses of the definitions at line %d...\n", 
             SiLineNumber(child_stmt));
           SiUnparseEList(def_to_use_lst);
           DU_Describe(def_to_use_lst);
        }
     }
     else
        info_of_stmt(child_stmt);
     branch = SiECdr(branch);
   }
     q = SiECdr(q);
   }
}

DU_Describe(lst)
ELIST lst;
{
  /* this function describes an elist whose every element is a list of
     (var, use_list) pairs. use_list is a list of elists where each
      element is a (stmt, term) pair of uses of the definition of var */
   ELIST  l, ll, item_lst;
   ELIST st; 
   int stmt, term;

   printf("{*** ");
   if (lst) 
   {
      for (l = lst; l; l = SiECdr(l))
      {
         item_lst = (ELIST)SiECar(l);
         printf("uses of definition: %d are ", SiECar(item_lst));
         for (ll = (ELIST)SiECadr(item_lst); ll; ll = SiECdr(ll)) 
                                       /* traverse the chain */
         {
           st = (ELIST)SiECar(ll);
           stmt = (int)SiECar(st);
           term = (int)SiECadr(st);
           printf("used at line number %d\n", SiLineNumber(stmt));
         }
      }
   }
   printf("***}\n");
} /* DU_Describe */ 
         

