/* file: anal_ind.c */
/**********************************************************************/
/*  This file contains the routines called in sets.c that do all index*/
/*  and subscript analysis.					      */
/**********************************************************************/
#include "db.h"

#define PLUS 2
#define ZPLUS 3
#define MINUS 4
#define ZMINUS 5
#define PLUSMINUS 6
#define NODEP -1

/* extern variables */
extern PTR_SYMB induct_list[MAX_NEST_DEPTH];
extern int stride[MAX_NEST_DEPTH];
extern int language;
extern PTR_FILE cur_file;

extern PCF UnparseBfnd[];
extern PCF UnparseLlnd[];


/* local variables */
struct subscript blank, extra;
int table_generated = 0;
int np = 2 * MAX_NEST_DEPTH;
int tbl_depth = 4 * MAX_NEST_DEPTH + AR_DIM_MAX;
int num_eqn, num_ineq;
int adm = MAX_NEST_DEPTH;
int *table[MAX_NEST_DEPTH * 4 + AR_DIM_MAX];
int upper_bnd[2 * MAX_NEST_DEPTH], lower_bnd[2 * MAX_NEST_DEPTH];
int dist_ub[2 * MAX_NEST_DEPTH], dist_lb[2 * MAX_NEST_DEPTH];
int row_dim=0;/*row for b*/

/* forward references */
char *malloc();
char *calloc();
PTR_SETS alloc_sets();
PTR_REFL alloc_ref();
int disp_refl();
PTR_REFL copy_refl();
PTR_REFL union_refl();

int check_for_indvar(s, d, lis)
PTR_SYMB s, lis[];
int d;
{
  int i;

  for (i = 0; i < d; i++)
    if (s == lis[i])
      return (1);
  return (0);
}

/*************************************************************/
/* find_bounds(b,q,qnew) takes a bifnode-llnd pair (b,q) and */
/*  creates a low level expression that describes the range  */
/* of values that are touched by the reference in the current*/
/* context.  the index expressions are all scalars and ranges*/
/* interms of parameters or constants.	if the index exp is  */
/* undecidable, then the whole range of the index is assumed */
/*  the parameter qnew is a low level list upon which this   */
/*  expression is appended.				     */
/*************************************************************/
PTR_LLND find_bounds(b, q, qnew)
PTR_BFND b;
PTR_LLND q, qnew;
{
  PTR_SYMB ind_list[MAX_NEST_DEPTH];
  PTR_LLND ind_terms[MAX_NEST_DEPTH];
  struct subscript il_lo[MAX_NEST_DEPTH];
  struct subscript il_hi[MAX_NEST_DEPTH];
  struct subscript source[AR_DIM_MAX];	/* a source reference or def. */
  int i, j, count, dumb,sign;
  struct ref sor;
  PTR_LLND qind_list, new_list, q_index, make_llnd(), tmp;
  PTR_LLND exp1, exp2, exp3, append_ll_elist(), build_exp_from_bound();
  PTR_BFND fun;
  PTR_REFL parms;
  PTR_LLND copy_llnd();

  for (i = 0; i < MAX_NEST_DEPTH; i++) {
    ind_list[i] = NULL;
    ind_terms[i] = NULL;
    for (j = 0; j < MAX_NEST_DEPTH; j++) {
      il_lo[i].coefs_symb[j] = NULL;
      il_hi[i].coefs_symb[j] = NULL;
    }
  }

  make_induct_list(b, ind_list, il_lo, il_hi);
  sor.stmt = b;
  sor.refer = q;
  make_subscr(&sor, source);	/* source is an array of */
				/* subscript records that */
				/* shared by all routines */
				/* find the parameter list */
  fun = b;
  while ((fun->variant != PROG_HEDR) &&
	 (fun->variant != FUNC_HEDR) &&
	 (fun->variant != PROC_HEDR))
    fun = fun->control_parent;
  parms = fun->entry.template.sets->in_def;

  qind_list = q->entry.template.ll_ptr1;
  new_list = NULL;
  i = 0;
  while (qind_list != NULL) {
    q_index = qind_list->entry.template.ll_ptr1;
    if (source[i].decidable == 2) {	/* ddot case */
      PTR_LLND low, hi, ar1, ar2, rl1, rl2, ltmp, htmp;
      /* skip stride for now */
      if (q_index->variant == DDOT && q_index->entry.template.ll_ptr1 != NULL 
	  && q_index->entry.template.ll_ptr1->variant == DDOT)
	q_index = q_index->entry.template.ll_ptr1;
      if (q_index->variant == STAR_RANGE) {
	rl1 = make_llnd(cur_file, STAR_RANGE, NULL, NULL, NULL);
      }
      else {
	low = copy_llnd(q_index->entry.template.ll_ptr1);
	hi = copy_llnd(q_index->entry.template.ll_ptr2);

	rl1 = make_llnd(cur_file, EXPR_LIST, low, NULL, NULL);
	rl2 = make_llnd(cur_file, EXPR_LIST, hi, NULL, NULL);
	ar1 = make_llnd(cur_file,ARRAY_REF,rl1,NULL, q->entry.template.symbol);
	ar2 = make_llnd(cur_file,ARRAY_REF,rl2,NULL, q->entry.template.symbol);
	ltmp = find_bounds(b, ar1, NULL);
	htmp = find_bounds(b, ar2, NULL);
	ltmp = ltmp->entry.template.ll_ptr1;
	htmp = htmp->entry.template.ll_ptr1;
	
	if (ltmp!= NULL && (ltmp->variant == EXPR_LIST || ltmp->variant == EXPR_LIST))
	  ltmp = ltmp->entry.template.ll_ptr1;
	if (htmp!= NULL && (htmp->variant == EXPR_LIST || htmp->variant == EXPR_LIST))
	  htmp = htmp->entry.template.ll_ptr1;
	if(ltmp == NULL) low =	make_llnd(cur_file, STAR_RANGE, NULL, NULL, NULL); 
	else if (ltmp->variant == DDOT)
	  low = ltmp->entry.template.ll_ptr1;
	else
	  low = ltmp;
	if(htmp == NULL) hi =  make_llnd(cur_file, STAR_RANGE, NULL, NULL, NULL);
	else if (htmp->variant == DDOT) {
	  hi = htmp->entry.template.ll_ptr2;
	  if (hi->variant == DDOT)
	    hi = hi->entry.template.ll_ptr1;
	}
	else
	  hi = htmp;
	if (low->variant == STAR_RANGE)
	  rl1 = low;
	else if (hi->variant == STAR_RANGE)
	  rl1 = hi;
	else {
	  rl1->variant = DDOT;
	  rl1->entry.template.ll_ptr1 = low;
	  rl1->entry.template.ll_ptr2 = hi;
	}
      }
      new_list = append_ll_elist(new_list, rl1);
    }
    else if (source[i].decidable == 0) {	/* parm */
      if (q_index == NULL || q_index->variant == STAR_RANGE) {
	exp3 = make_llnd(cur_file, STAR_RANGE, NULL, NULL, NULL);
	new_list = append_ll_elist(new_list, exp3);
      }
      else if (reduce_ll_exp(b, parms, ind_list, q_index, &exp2, &dumb) == 0) {
	/* was not able to resolve */
	if (simple_algebraic(q_index)) {
	  sign = 1;
	  exp1 = build_exp_from_bound(il_lo, &(source[i]),&sign);
	  if (exp1 == NULL) {
	    /* this should only happen if the subscript */
	    /* is very strange. 		    */
	  }
	  if (reduce_ll_exp(b, parms, ind_list, exp1, &exp2, &dumb) == 0) {
	    /* was not able to resolve ! */
	    exp3 = make_llnd(cur_file, STAR_RANGE, NULL, NULL, NULL);
	  }
	  else {
	    exp1 = exp2;
	    count = 0;
	    for (j = 0; j < MAX_NEST_DEPTH; j++)
	      if (source[i].coefs[j] != 0)
		count++;
	    if (count == 0)
	      exp3 = exp1;
	    else {
	      sign = 1;
	      exp2 = build_exp_from_bound(il_hi, &(source[i]),&sign);
	       if (reduce_ll_exp(b, parms, ind_list, exp2, &exp3, &dumb) == 0) {
		exp3 = make_llnd(cur_file, STAR_RANGE, NULL, NULL, NULL);
	      }
	      else {
		exp2 = exp3;
		if(sign > 0)
		   exp3 = make_llnd(cur_file, DDOT, exp1, exp2, NULL);
		else
		   exp3 = make_llnd(cur_file, DDOT, exp2, exp1, NULL);
	      }
	    }
	  }
	  new_list = append_ll_elist(new_list, exp3);
	}
	else {
	  tmp = make_llnd(cur_file, STAR_RANGE, NULL, NULL, NULL);
	  new_list = append_ll_elist(new_list, tmp);
	}
      }
      else
	new_list = append_ll_elist(new_list, exp2);
    }
    else if (source[i].decidable == 1) {	/* standard linear */
      sign = 1;
      exp1 = build_exp_from_bound(il_lo, &(source[i]),&sign);
       if (exp1 == NULL) {
	/* fprintf(stderr, "OOPS null!\n"); */
	/* this should only happen if the subscript */
	/* is very strange.  or the low bound is strange   */
      }
      if (reduce_ll_exp(b, parms, ind_list, exp1, &exp2, &dumb) == 0) {
	/* was not able to resolve ! */
	exp3 = make_llnd(cur_file, STAR_RANGE, NULL, NULL, NULL);
      }
      else {
	exp1 = exp2;
	count = 0;
	for (j = 0; j < MAX_NEST_DEPTH; j++)
	  if (source[i].coefs[j] != 0
	      || source[i].coefs_symb[j] != NULL)
	    count++;
	if (count == 0)
	  exp3 = exp1;
	else {
	  sign = 1;
	  exp2 = build_exp_from_bound(il_hi, &(source[i]),&sign);
	   if (reduce_ll_exp(b, parms, ind_list, exp2, &exp3, &dumb) == 0) {
	    exp3 = make_llnd(cur_file, STAR_RANGE, NULL, NULL, NULL);
	  }
	  else {
	    exp2 = exp3;
	    if(sign> 0)
	      exp3 = make_llnd(cur_file, DDOT, exp1, exp2, NULL);
	    else
	      exp3 = make_llnd(cur_file, DDOT, exp2, exp1, NULL);
	  }
	}
      }
      new_list = append_ll_elist(new_list, exp3);
    }
    else {
      fprintf(stderr, "source[i].decidable = %d\n", source[i].decidable);
      fprintf(stderr, "strange brew in find_bounds %s\n",
	      (UnparseLlnd[cur_file->lang])(q_index));
      new_list = append_ll_elist(new_list, q_index);
    }
    qind_list = qind_list->entry.template.ll_ptr2;
    i++;
  }
  if (qnew != NULL)
    qnew->entry.template.ll_ptr1 = new_list;
  else
    qnew = new_list;
  return (qnew);
}

simple_algebraic(p)
PTR_LLND p;
{
  if (p == NULL)
    return (1);
  switch (p->variant) {
   case EXPR_LIST:
   case ADD_OP:
   case DIV_OP:
   case MULT_OP:
   case SUBT_OP:
   case MINUS_OP:
    return (simple_algebraic(p->entry.template.ll_ptr1) *
	    simple_algebraic(p->entry.template.ll_ptr2));
   case VAR_REF:
   case CONST_REF:
   case INT_VAL:
    return (1);
   default:
    return (0);
  }
}

PTR_LLND append_ll_elist(list, item)
PTR_LLND list, item;
{
  PTR_LLND tmp, make_llnd();

  if (list == NULL) {
    tmp = make_llnd(cur_file, EXPR_LIST, item, NULL, NULL);
    return (tmp);
  }
  if (list->variant != EXPR_LIST) {
    fprintf(stderr, "append_ll_elist screw up\n");
    return (list);
  }
  else if (list->entry.list.next == NULL) {
    tmp = append_ll_elist(NULL, item);
    list->entry.list.next = tmp;
    return (list);
  }
  else {
    append_ll_elist(list->entry.list.next, item);
    return (list);
  }
}

PTR_LLND build_exp_from_bound(il, sub, sign)
struct subscript il[MAX_NEST_DEPTH];
struct subscript *sub;
int *sign;
{
  PTR_LLND exp, exp2, exp3, exp4, make_llnd();
  int j;
  
  if (sub->decidable == 2) {	/* ddot case */
    return (sub->vector);
  }
  if (sub->decidable == 0 /* && simple_algebraic(sub->parm_exp) == 0 */ ) {
    /* parameter expression (we hope) */
    /* first we need to check for other vars */
    return (sub->parm_exp);
  }
  if (sub->decidable == 1) {  /* standard linear */
    exp = NULL;
    if (sub->parm_exp == NULL) {
      exp = make_llnd(cur_file, INT_VAL, NULL, NULL, NULL);
      exp->entry.ival = sub->offset;
    }
    else
      exp = sub->parm_exp;
    for (j = 0; j < MAX_NEST_DEPTH; j++) {
      if (sub->coefs_symb[j] != NULL) { /* symbolic case! */
	exp3 = build_exp_from_bound(il, &(il[j]), sign);
	if (exp3 == NULL) {
	  exp4 = NULL;
	  exp = NULL;
	}
	else if (exp3->variant == DDOT) {
	  fprintf(stderr, "DDOT case\n");
	  exp4 = exp3;
	}
	else {	/* exp3 is loop bound which must mult by symbolic coef */
	  exp4 = make_llnd(cur_file, MULT_OP, sub->coefs_symb[j],
			   exp3, NULL);
	}
	if (exp != NULL) {
	  exp3 = make_llnd(cur_file, ADD_OP, exp4, exp, NULL);
	  exp = exp3;
	}
	else
	  exp = exp4;
      }
      else if (sub->coefs[j] != 0) { /* a nice integer coef. */
	exp3 = build_exp_from_bound(il, &(il[j]),sign);
	if (exp3 == NULL) {
	  exp4 = NULL;
	  exp = NULL;
	}
	else if (exp3->variant == DDOT) {
	  fprintf(stderr, "DDOT case\n");
	  exp4 = exp3;
	}
	else if (sub->coefs[j] == 1)
	  exp4 = exp3;
	else {
	  exp2 = make_llnd(cur_file, INT_VAL, NULL, NULL, NULL);
	  exp2->entry.ival = sub->coefs[j];
	  if(sub->coefs[j] < 0) *sign = -1;
	  exp2->type = cur_file->head_type;	/* always INT type */
	  exp4 = make_llnd(cur_file, MULT_OP, exp2, exp3, NULL);
	}
	if (exp != NULL) {
	  exp3 = make_llnd(cur_file, ADD_OP, exp4, exp, NULL);
	  exp = exp3;
	}
	else
	  exp = exp4;
      }
    }
    return (exp);
  }
  else
    return (make_llnd(cur_file, STAR_RANGE, NULL, NULL, NULL));
}

/**************************************************************/
/* compute dist vect.  calculates the distance vector between */
/* two references source and destination.  The vector is an   */
/* array of integers of the form ( len, dist1, dist2, ....)   */
/* trouble is an array which indicates one of several problems*/
/*  if trouble[0] = 1 then there is no intersection!	      */
/*  if trouble[i] = PLUSMINUS then the i-th component is "<=>"*/
/*  if trouble[i] = PLUS then vector is "+" ,i.e. positive    */
/*	  but variable in nature. similar for ZPLUS which     */
/*		  means the vector is "0+" = non-negative     */
/*    other cases are ZMINUS="0-" and MINUS = "-"	      */
/*  if trouble[i] = NODEP then no depend. on this index at all*/
/*    NOTE: trouble[i] = NODEP is the case for scalars.       */
/* the first component of vec is the length of the vector.    */
/* function returns nothing				      */
/**************************************************************/
int comp_dist(vec, trouble, sor, des, lexord)
int vec[], trouble[];
struct ref *sor;
struct ref *des;
int lexord;			/* true if sor precedes des in lex order */
{
  PTR_SYMB sor_ind_l[MAX_NEST_DEPTH], des_ind_l[MAX_NEST_DEPTH];
  struct subscript il_lo[MAX_NEST_DEPTH];
  struct subscript il_hi[MAX_NEST_DEPTH];
  struct subscript source[AR_DIM_MAX];	/* a source reference or def. */
  struct subscript destin[AR_DIM_MAX];	/* a destination ref. or def. */
  int inorder, i, j, sd, dd, depth, step, depfound;
  int eqntbl[AR_DIM_MAX][2 * MAX_NEST_DEPTH + 1];
  PTR_SYMB s;
  PTR_BFND sor_parent,des_parent;
  int tmpdepth;
  PTR_SYMB new_sor_ind_l[MAX_NEST_DEPTH], new_des_ind_l[MAX_NEST_DEPTH];
  struct subscript sor_il_lo[MAX_NEST_DEPTH],des_il_lo[MAX_NEST_DEPTH];
  struct subscript sor_il_hi[MAX_NEST_DEPTH],des_il_hi[MAX_NEST_DEPTH];
  int tmpsd,tmpdd;
  int k; 
  struct subscript new_source[AR_DIM_MAX];  /* a source reference or def. */
  struct subscript new_destin[AR_DIM_MAX];  /* a destination ref. or def. */
  int symb_tag=0;
  int sor_parent_id[MAX_NEST_DEPTH],des_parent_id[MAX_NEST_DEPTH];
  int sor_stride[MAX_NEST_DEPTH];
  int des_stride[MAX_NEST_DEPTH];
  int dep_tag;

  if (table_generated == 0) {
    for (i = 0; i < tbl_depth; i++)
      table[i] = (int *) calloc(2 * MAX_NEST_DEPTH + 1, sizeof(int));
    table_generated = 1;
  }
  for (i = 0; i < tbl_depth; i++)
    for (j = 0; j < np + 1; j++) {
      table[i][j] = 0;
      if (i < AR_DIM_MAX)
	eqntbl[i][j] = 0;
    }

  blank.decidable = 1;
  extra.decidable = 1;
  extra.offset = 0;
  blank.offset = 0;
  for (i = 0; i < MAX_NEST_DEPTH; i++) {
    sor_ind_l[i] = NULL;
    des_ind_l[i] = NULL;
    new_sor_ind_l[i] = NULL;
    new_des_ind_l[i] = NULL;
    blank.coefs[i] = 0;
    il_lo[i].decidable = 1;
    il_hi[i].decidable = 1;
    sor_il_lo[i].decidable = 1;
    sor_il_hi[i].decidable = 1;
    des_il_lo[i].decidable = 1;
    des_il_hi[i].decidable = 1;
    il_lo[i].offset = 0;
    il_hi[i].offset = 0;
    sor_il_lo[i].offset = 0;
    sor_il_hi[i].offset = 0;
    des_il_lo[i].offset = 0;
    des_il_hi[i].offset = 0;
    for (j = 0; j < MAX_NEST_DEPTH; j++) {
      il_lo[i].coefs[j] = 0;
      il_hi[i].coefs[j] = 0;
      sor_il_lo[i].coefs[j] = 0;
      sor_il_hi[i].coefs[j] = 0;
      des_il_hi[i].coefs[j] = 0;
      des_il_lo[i].coefs[j] = 0;
   }
  }

  sd = make_induct_list(sor->stmt, sor_ind_l, il_lo, il_hi);

  dd = make_induct_list(des->stmt, des_ind_l, il_lo, il_hi);

  depth = (sd < dd) ? sd : dd;
  inorder = (sor->stmt->g_line < des->stmt->g_line) ? 1 : 0;

  /*i = 0;
  while ((i < depth) && (des_ind_l[i] == sor_ind_l[i]))
    i++;
  if (i < depth)
    depth = i;*/
  
/*find out the public nest*/ 
  tmpdepth=depth;
  sor_parent=sor->stmt;
  des_parent=des->stmt;
  tmpsd=sd;
  tmpdd=dd;
  k=tmpsd-1;
  while (tmpsd > 0 && sor_parent->variant !=GLOBAL ){
    sor_parent=sor_parent->control_parent;
    if (sor_parent->variant == FOR_NODE){
     sor_parent_id[k]=sor_parent->id;
     k--;
     tmpsd--;
    }
  }
  k=tmpdd-1;
  while (tmpdd > 0 && des_parent->variant !=GLOBAL){
    des_parent=des_parent->control_parent;
    if (des_parent->variant == FOR_NODE){
     des_parent_id[k]=des_parent->id;
     k--;
     tmpdd--;
    }
  }
  i=0;
  for (k=0; k<depth; k++)
    if (sor_parent_id[k] !=des_parent_id[k])
     break;
    else 
     i++;
  depth=i;

  /*if (tmpsd<tmpdd)
    while (tmpsd<tmpdd){
      des_parent=des_parent->control_parent;
      tmpdd--;
    }
   else if (tmpdd<tmpsd)
     while (tmpdd<tmpsd){
       sor_parent=sor_parent->control_parent;
       tmpsd--;
     }

  while ((depth>0) && (sor_parent->control_parent!=des_parent->control_parent)){
    depth--;
    sor_parent=sor_parent->control_parent;
    des_parent=des_parent->control_parent;
  }*/
  tmpsd=depth;
  tmpdd=depth;

 /*The later is for the loosely nested loop*/

  if ((depth != sd ) || (depth != dd)){

   make_subscr(sor, new_source);
   make_subscr(des, new_destin);
  
   sd = make_induct_list(sor->stmt, sor_ind_l, sor_il_lo, sor_il_hi);
   for (i=0;i<sd;i++)
     sor_stride[i]=stride[i];
   dd = make_induct_list(des->stmt, des_ind_l, des_il_lo, des_il_hi);
   for (i=0;i<dd;i++)
     des_stride[i]=stride[i];

   for (i=0; i<sd; i++)
    new_sor_ind_l[i]=sor_ind_l[i];
   for (i=0; i<dd; i++)
    new_des_ind_l[i]=des_ind_l[i];
   for (i=0; i<depth; i++){
    il_lo[i]=sor_il_lo[i];
    il_hi[i]=sor_il_hi[i];
   }
    for (i=0;i<depth;i++)
     stride[i]=sor_stride[i];
    if (dd <= sd){
     j=0;
     for(i=depth;i<dd;i++){
      new_des_ind_l[sd+j]=des_ind_l[i];
      new_sor_ind_l[sd+j]=des_ind_l[i];
      stride[sd+j]=des_stride[i];
      il_lo[sd+j]=des_il_lo[i];
      il_hi[sd+j]=des_il_hi[i];
      for (k=0; k<AR_DIM_MAX; k++){
        new_destin[k].coefs[sd+j]=new_destin[k].coefs[i];
        new_destin[k].coefs_symb[sd+j]=new_destin[k].coefs_symb[i];
        new_source[k].coefs[sd+j]=0;
        new_source[k].coefs_symb[sd+j]=NULL;
      } 
      j++;
     }
     for (i=depth; i<sd; i++){
       il_lo[i]=sor_il_lo[i];
       il_hi[i]=sor_il_hi[i];
       new_des_ind_l[i]=sor_ind_l[i];
       stride[i]=sor_stride[i];
       for (k=0; k<AR_DIM_MAX; k++){
        new_destin[k].coefs[i]=0;
        new_destin[k].coefs_symb[i]=NULL;
      }
     }
     tmpsd=sd+j;
     tmpdd=tmpsd;
    }
    else{
     j=0;
     for(i=depth;i<sd;i++){
       new_sor_ind_l[dd+j]=sor_ind_l[i];
       new_des_ind_l[dd+j]=sor_ind_l[i];
       stride[dd+j]=sor_stride[i];
       il_lo[dd+j]=sor_il_lo[i];
       il_hi[dd+j]=sor_il_hi[i];
       for (k=0; k<AR_DIM_MAX; k++){
        new_source[k].coefs[dd+j]=new_source[k].coefs[i];
        new_source[k].coefs_symb[dd+j]=new_source[k].coefs_symb[i];
        new_destin[k].coefs[dd+j]=0;
        new_destin[k].coefs_symb[dd+j]=NULL;
      }
       j++;
     }
     for (i=depth; i<dd; i++){
       il_lo[i]=des_il_lo[i];
       il_hi[i]=des_il_hi[i];
       new_sor_ind_l[i]=des_ind_l[i];
       stride[i]=des_stride[i];
       for (k=0; k<AR_DIM_MAX; k++){
        new_source[k].coefs[i]=0;
        new_source[k].coefs_symb[i]=NULL;
      }
     }
     tmpdd=dd+j;
     tmpsd=tmpdd;
    } 
    for (i=0; i<AR_DIM_MAX; i++){
     if (new_source[i].parm_exp !=NULL || new_destin[i].parm_exp !=NULL){
      symb_tag=1;
      break;
     }
     for (j=0; j<MAX_NEST_DEPTH; j++)
       if (new_source[i].coefs_symb[j] !=NULL || new_destin[i].coefs_symb[j] !=NULL){
         symb_tag=1;
         break;
       }
     if (symb_tag==1)
       break;
   }
  }
  else{
      make_subscr(sor, source);
      make_subscr(des, destin);
      for (i=0; i<AR_DIM_MAX; i++){
       if (source[i].parm_exp !=NULL || destin[i].parm_exp !=NULL){
        symb_tag=1;
        break;
       }
       for (j=0; j<MAX_NEST_DEPTH; j++)
         if (source[i].coefs_symb[j] !=NULL || destin[i].coefs_symb[j] !=NULL){
           symb_tag=1;
           break;
         }
       if (symb_tag==1)
        break;
     }
 }

  /* for each subscript expression we need to check for */
  /* symbolic references.  if they are the same we are	*/
  /* ok.  if they are different we set the flag to be	*/
  /* undecidable.				       */
  /*if (depth==sd && depth==dd)
   for (j = 0; j < AR_DIM_MAX; j++) {
    if ((source[j].parm_exp != NULL) ||
	(destin[j].parm_exp != NULL)) {
      if (sequiv(source[j].parm_exp, destin[j].parm_exp) == 0) {*/
	/* the following is temporary.	we   */
	/* should do a symbolic subtraction  */
	/*source[j].offset = 1;
	destin[j].offset = 0;
	source[j].decidable = 1;
	destin[j].decidable = 1;
	source[j].parm_exp = NULL;
	destin[j].parm_exp = NULL;
      }
     }
    }*/
  /*else
   for (j = 0; j < AR_DIM_MAX; j++) {
    if ((new_source[j].parm_exp != NULL) ||
        (new_destin[j].parm_exp != NULL)) {
      if (sequiv(new_source[j].parm_exp, new_destin[j].parm_exp) == 0) {*/
        /* the following is temporary.  we   */
        /* should do a symbolic subtraction  */
      /*  new_source[j].offset = 1;
        new_destin[j].offset = 0;
        new_source[j].decidable = 1;
        new_destin[j].decidable = 1;
        new_source[j].parm_exp = NULL;
        new_destin[j].parm_exp = NULL;
      }
     }
    }*/

  s = sor->refer->entry.template.symbol;
  for (i = 1; i < MAX_NEST_DEPTH; i++) {
    vec[i] = 0;
    trouble[i] = NODEP;
  }
  vec[0] = depth;
  trouble[0] = 0;
  /* first check for uniformly generated cases */
  if ((s->type->variant == T_ARRAY || s->type->variant == T_POINTER)
     && depth==sd && depth==dd
     && symb_tag==0
     && unif_gen(sor, des, vec, trouble, source, destin));
  else {
    /* if a scalar ... */
    if (s->type->variant != T_ARRAY && s->type->variant != T_POINTER) {
      for (i = 1; i <= depth; i++) {
	trouble[i] = 0;
	vec[i] = 0;
      }

      if (inorder == 0) {
	vec[depth] = 1;
	trouble[depth] = 0;
      }
      return (1);
    }
    else
      /* if not uniform do generalized shoestak */
     if (depth != sd || depth != dd)
      for (step = 0; step <= depth; step++) {
        if (solve_system(step, depth, tmpsd, new_sor_ind_l,
                         tmpdd, new_des_ind_l, il_lo, il_hi, 
                         new_source, new_destin) != 0) {
          set_troub(step + 1, vec, trouble, PLUS);
        }
        else if (step == 0)
          trouble[0] = 1;
      }
     else
      for (step = 0; step <= depth; step++) {
	if (solve_system(step, depth, sd, sor_ind_l,
			 dd, des_ind_l, il_lo, il_hi, source, destin) != 0) {
	  set_troub(step + 1, vec, trouble, PLUS);
	}
	else if (step == 0)
	  trouble[0] = 1;
      }
  }
  /*dep_tag=0;
  for (step=1;step <=depth+1;step++)
    if (trouble[step]==PLUS)
      dep_tag=1;
  if (dep_tag==0) 
    trouble[0]=1;*/
 
  depfound = 0;

  for (i = 1; i < MAX_NEST_DEPTH; i++) {
    if (vec[i] != 0 || trouble[i] != NODEP)
      depfound = 1;
    if (trouble[i] == -99)
      trouble[i] = 0;
  }

  if (depfound == 0 && !lexord)
    trouble[0] = 1;
  if (!lexord)
    trouble[depth+1]=0;
  return (1);			/* return value means nothing here */

}

solve_system(step,depth,sd,sor_ind_l,dd,des_ind_l,il_lo,il_hi,source,destin)
int step, depth, sd, dd;
PTR_SYMB sor_ind_l[MAX_NEST_DEPTH], des_ind_l[MAX_NEST_DEPTH];
struct subscript il_lo[];
struct subscript il_hi[];
struct subscript source[];	/* a source reference or def. */
struct subscript destin[];	/* a destination ref. or def. */
{
  struct subscript lo, hi;
  int i, j, k, max_depth;
  int num_eqn, num_ineq;
  int symb_tag=0;
  struct subscript tmp_il_lo[MAX_NEST_DEPTH];
  struct subscript tmp_il_hi[MAX_NEST_DEPTH];


  max_depth = (sd > dd) ? sd : dd;

  for (i=0;i<MAX_NEST_DEPTH;i++){
    tmp_il_lo[i]=il_lo[i];
    tmp_il_hi[i]=il_hi[i];
  }

  /* now build equation rows of the table */
  num_eqn = -1;
  /*for (j = 0; j < AR_DIM_MAX; j++){ 
    if (source[j].decidable != -1 || destin[j].decidable != -1)  
      add_eqn(table[j], &source[j], &destin[j]);
    else if (num_eqn == -1)
      num_eqn = j;
  }*/
  
  k=0;
  for (j = 0; j < AR_DIM_MAX; j++) {
    if (source[j].decidable != -1 || destin[j].decidable != -1){
      if (source[j].parm_exp !=NULL || destin[j].parm_exp !=NULL)
        symb_tag=1;
      for (i=0; i<MAX_NEST_DEPTH; i++)
        if (source[j].coefs_symb[i] !=NULL || destin[j].coefs_symb[i] !=NULL)
          symb_tag=1;
      if (symb_tag ==0){
       add_eqn(table[k], &source[j], &destin[j]);
       k++;
      }
      else{
       symb_tag=0;
     }
    }
    else if (num_eqn == -1)
      num_eqn = k;
  }
  /* add step equations */
  for (k = 0; k < step; k++) {
    for (j = 0; j < MAX_NEST_DEPTH; j++) {
      extra.coefs[j] = 0;
      blank.coefs[j] = 0;
    }
    extra.coefs[k] = 1;
    blank.coefs[k] = 1;
    add_eqn(table[num_eqn], &extra, &blank);
    num_eqn++;
    blank.coefs[k] = 0;
  }

  /* fix normalization for stride */
  for (i = 0; i < max_depth; i++) {
    if (stride[i] != 1) {
         if (il_hi[i].decidable==1){
          for (k=0;k<MAX_NEST_DEPTH;k++){
           tmp_il_hi[i].coefs[k]=
             (int)((tmp_il_hi[i].coefs[k]-tmp_il_lo[i].coefs[k])/stride[i]);
           tmp_il_lo[i].coefs[k]=0;
          }
           tmp_il_hi[i].offset=
               (int)((tmp_il_hi[i].offset-tmp_il_lo[i].offset)/stride[i]);
           tmp_il_lo[i].offset=0;
         }
      for (j = 0; j < num_eqn; j++) {
        if (il_lo[i].decidable==1){
         for (k=0;k<MAX_NEST_DEPTH;k++){
          table[j][k]=table[j][k]+table[j][i]*il_lo[i].coefs[k];
          table[j][k+MAX_NEST_DEPTH]=
            table[j][k+MAX_NEST_DEPTH]+table[j][i]*il_lo[i].coefs[k];
         }
          table[j][np]=table[j][np]+table[j][i]*il_lo[i].offset
                       +table[j][MAX_NEST_DEPTH+i]*il_lo[i].offset;
        }
          table[j][i] = table[j][i] * stride[i];
          table[j][MAX_NEST_DEPTH + i] =
            table[j][MAX_NEST_DEPTH + i] * stride[i];
     }
      /*if (stride[i] < 0) {
	for (j = 0; j < num_eqn; j++)
	  if (table[j][i] < 0)
	    for (k = 0; k <= np; k++)
	      table[j][k] = -table[j][k];
      }*/
    
    }
  }

  num_ineq = 0;

  /* now add direction inequality at position step */
  for (j = 0; j < MAX_NEST_DEPTH; j++) {
    extra.coefs[j] = 0;
    blank.coefs[j] = 0;
  }
  extra.coefs[step] = -1;
  blank.coefs[step] = -1;
  extra.offset = -1;
  add_eqn(table[num_eqn], &extra, &blank);
  extra.coefs[step] = 0;
  blank.coefs[step] = 0;
  extra.offset = 0;

  num_ineq = 1;
  /* now add vector range subscript ineq. */
  for (j = 0; j < AR_DIM_MAX; j++) {
    if (source[j].decidable == 2) {
      /* source is vector in component j */
      make_vect_range(sd, source[j].vector, sor_ind_l, &lo, &hi);
      add_eqn(table[num_eqn + num_ineq], &lo, &blank);
      add_eqn(table[num_eqn + num_ineq + 1], &hi, &blank);
      num_ineq = num_ineq + 2;
    }
    if (destin[j].decidable == 2) {
      /* destin is vector in component j */
      make_vect_range(dd, destin[j].vector, des_ind_l, &lo, &hi);
      add_eqn(table[num_eqn + num_ineq], &lo, &blank);
      add_eqn(table[num_eqn + num_ineq + 1], &hi, &blank);
      num_ineq = num_ineq + 2;
    }
  }


  /* now add induction bound inequalities */
  for (j = 0; j < max_depth; j++) {
    /* reverse lo */
    tmp_il_lo[j].offset = -tmp_il_lo[j].offset;
    for (i = 0; i < MAX_NEST_DEPTH; i++)
      tmp_il_lo[j].coefs[i] = -tmp_il_lo[j].coefs[i];
    tmp_il_lo[j].coefs[j] = 1;	/* perhaps repalce by stride ? */
    tmp_il_hi[j].coefs[j] = -1;

    if (il_lo[j].decidable == 1) {
      add_eqn(table[num_eqn + num_ineq], &tmp_il_lo[j], &blank);
      num_ineq = num_ineq + 1;
    }
    if (il_hi[j].decidable == 1) {
      add_eqn(table[num_eqn + num_ineq], &tmp_il_hi[j], &blank);
      num_ineq = num_ineq + 1;
    }
    /* reset lo and reverse hi */
    for (i = 0; i < MAX_NEST_DEPTH; i++) {
      tmp_il_lo[j].coefs[i] = -tmp_il_lo[j].coefs[i];
      tmp_il_hi[j].coefs[i] = -tmp_il_hi[j].coefs[i];
    }
    tmp_il_lo[j].offset = -tmp_il_lo[j].offset;
    tmp_il_hi[j].offset = -tmp_il_hi[j].offset;
    if (il_lo[j].decidable == 1) {
      add_eqn(table[num_eqn + num_ineq], &blank, &tmp_il_lo[j]);
      num_ineq = num_ineq + 1;
    }
    if (il_hi[j].decidable == 1) {
      add_eqn(table[num_eqn + num_ineq], &blank, &tmp_il_hi[j]);
      num_ineq = num_ineq + 1;
    }
    /* reset hi */
    for (i = 0; i < MAX_NEST_DEPTH; i++) {
      tmp_il_hi[j].coefs[i] = -tmp_il_hi[j].coefs[i];
    }
    tmp_il_hi[j].offset = -tmp_il_hi[j].offset;
    tmp_il_lo[j].coefs[j] = 0;
    tmp_il_hi[j].coefs[j] = 0;

  }

  /* table complete.. now put in reduced form */
  if (new_reduce(table, num_eqn, num_eqn + num_ineq,tmp_il_lo,tmp_il_hi,max_depth) == 0)
    return (0);
  else
    return (1);
}

add_eqn(table, source, destin)
struct subscript *source;	/* a source reference or def. */
struct subscript *destin;	/* a destination ref. or def. */
int table[];
{
  int i;

  if (source->decidable < 1 || destin->decidable < 1)
    for (i = 0; i < np + 1; i++)
      table[i] = 0;
  else {
    for (i = 0; i < MAX_NEST_DEPTH; i++) {
      table[i] = source->coefs[i];
      table[i + MAX_NEST_DEPTH] = -(destin->coefs[i]);
    }
    table[np] = source->offset - destin->offset;
  }
}

print_tbl(depth, neqn, neq, tbl)
int depth, neqn, neq;
int *tbl[];
{
  int i, j;

  depth = depth;		/* make lint happy, depth unused */

  fprintf(stderr, "|---------------table----------------------|\n");
  fprintf(stderr, "|  i   j   k   i'   j'	  k'   const   relat|\n");
  fprintf(stderr, "|------------------------------------------|\n");
  j = np / 2;
  for (i = 0; i < neqn; i++)
    fprintf(stderr, "| %2d  %2d  %2d  %2d   %2d   %2d	%4d	     ==  |\n",
	   tbl[i][0], tbl[i][1], tbl[i][2],
	   tbl[i][j], tbl[i][j + 1], tbl[i][j + 2], tbl[i][np]);
  fprintf(stderr, "|------------------------------------------|\n");
  for (i = neqn; i < neqn + neq; i++)
    fprintf(stderr, "| %2d  %2d  %2d  %2d   %2d   %2d	%4d	     >=  |\n",
	   tbl[i][0], tbl[i][1], tbl[i][2],
	   tbl[i][j], tbl[i][j + 1], tbl[i][j + 2], tbl[i][np]);
  fprintf(stderr, "|------------------------------------------|\n");
}

print_etbl(depth, neqn, tbl)
int depth, neqn;
int tbl[AR_DIM_MAX][2 * MAX_NEST_DEPTH + 1];
{
  int i, j;

  depth = depth;		/* make lint happy, depth unused */

  fprintf(stderr, "|---------------table----------------------|\n");
  fprintf(stderr, "|  i   j   k   i'   j'	  k'   const   relat|\n");
  fprintf(stderr, "|------------------------------------------|\n");
  j = np / 2;
  for (i = 0; i < neqn; i++)
    fprintf(stderr, "| %2d  %2d  %2d  %2d   %2d   %2d	%4d	     ==  |\n",
	   tbl[i][0], tbl[i][1], tbl[i][2],
	   tbl[i][j], tbl[i][j + 1], tbl[i][j + 2], tbl[i][np]);
  fprintf(stderr, "|------------------------------------------|\n");
}

int reduce(tbl, num_eqn, tbl_depth,il_lo,il_hi,max_depth)
int *tbl[];
int num_eqn, tbl_depth, max_depth;
struct subscript il_lo[MAX_NEST_DEPTH], il_hi[MAX_NEST_DEPTH];
{
  int j, i, k, t, mgcd, piv, pcol, opc, alf, bet;
  int *tmp;
  int bound[2];
  int tag=0;

  for (i = 0; i < 2 * MAX_NEST_DEPTH; i++) {
    upper_bnd[i] = 32000;
    lower_bnd[i] = -32000;
    if (i < MAX_NEST_DEPTH) {
      dist_lb[i] = -32000;
      dist_ub[i] = 32000;
    }
  }
  /*for (i=0; i<max_depth; i++)
   if (il_lo[i].decidable !=1 || il_hi[i].decidable !=1)
      tag=1;*/

  for (i = 0; i < tbl_depth; i++)
   if (chk_bnds(tbl, i, upper_bnd, lower_bnd) == 0)
      return (0);

 /*for (i = 0; i<num_eqn; i++)
  if (tag==0){
    cmp_bnd(bound,il_lo,il_hi,tbl[i]);
    if (-tbl[i][np] < bound[0] || -tbl[i][np] > bound[1])
      return (0);
   }*/


  pcol = -1;
  /* first eliminate by using the equations */
  for (j = 0; j < num_eqn; j++) {
    /* find leader pivod equation */
    piv = -1;
    opc = pcol;
    for (k = opc + 1; k < MAX_NEST_DEPTH * 2; k++)
      for (t = j; t < num_eqn; t++)
	if (opc == pcol && tbl[t][k] != 0) {
	  pcol = k;
	  piv = t;
	}

    if (piv > -1) {
      /* swap to bring to top */
      tmp = tbl[j];
      tbl[j] = tbl[piv];
      tbl[piv] = tmp;
      /* first reduce by gcd of row */
      if (tbl[j][pcol] < 0)
	for (i = 0; i <= np; i++)
	  tbl[j][i] = -tbl[j][i];
      mgcd = gcd(np - 1, tbl[j]);
      if (mgcd > 1) {
	/* first test for bad congruence class */
	if ((tbl[j][np] % mgcd) != 0)
	  return (0);
	for (i = 0; i <= np; i++)
	  tbl[j][i] = tbl[j][i] / mgcd;
      }
      /* now do elimination on pcol */
      alf = tbl[j][pcol];
      if (alf == 0)
	fprintf(stderr, "reduce error\n");
      else if (alf < 0) {
	alf = -alf;
	for (i = 0; i <= np; i++)
	  tbl[j][i] = -tbl[j][i];
      }
      for (k = j + 1; k < tbl_depth; k++) {
	if ((bet = tbl[k][pcol]) != 0) {
	  /* first reduce row k */
	  for (i = pcol; i <= np; i++)
	    tbl[k][i] = alf * tbl[k][i] - bet * tbl[j][i];
	  /* test for dim 1 or 0 constraint */
	  if (chk_bnds(tbl, k, upper_bnd, lower_bnd) == 0)
	    return (0);
          /*if (tag==0  && k<num_eqn){
           cmp_bnd(bound,il_lo,il_hi,tbl,k);
           if (-tbl[k][np] < bound[0] || -tbl[k][np] > bound[1])
            return (0);
          }*/
	}
      }
    }				/* end of piv found case */
  }				/* end of factorization loop */
  /* second eliminate by adding inequalities */
  for (j = num_eqn; j < tbl_depth; j++) {
    /* find leader pivod equation */
    piv = -1;
    opc = pcol;
    for (k = opc + 1; k < MAX_NEST_DEPTH * 2; k++)
      for (t = j; t < tbl_depth; t++)
	if (opc == pcol && tbl[t][k] > 0) {
	  pcol = k;
	  piv = t;
	}

    if (piv > -1) {
      /* swap to bring to top */
      tmp = tbl[j];
      tbl[j] = tbl[piv];
      tbl[piv] = tmp;
      /* now do elimination on pcol */
      alf = tbl[j][pcol];
      if (alf <= 0)
	fprintf(stderr, "reduce error\n");
      for (k = j + 1; k < tbl_depth; k++) {
	if ((bet = tbl[k][pcol]) < 0) {
	  /* first do the ellimination */
	  for (i = 0; i <= np; i++)
	    tbl[k][i] = alf * tbl[k][i] - bet * tbl[j][i];
	  /* now check for constraint errors */
	  if (chk_bnds(tbl, k, upper_bnd, lower_bnd) == 0)
	    return (0);
	}
      }
    }				/* end of piv found case */
  }				/* end of factorization loop */

  /* now look for contradictions in eqnations */
  for (j = 0; j < tbl_depth; j++)
    if (chk_bnds(tbl, j, upper_bnd, lower_bnd) == 0)
      return (0);
  return (1);
}

int chk_bnds(tbl, k, upper_bnd, lower_bnd)
int *tbl[];
int k;
int upper_bnd[], lower_bnd[];
{
  int i, first, second, third, gama;

  third = -1;
  first = -1;
  second = -1;
  for (i = 0; i < np; i++)
    if (tbl[k][i] != 0) {
      if (first == -1)
	first = i;
      else if (second == -1)
	second = i;
      else if (third == -1)
	third = i;
    }
  if (first == -1) {		/* this is a dimension 0 constraint */
    if ((k < num_eqn) & (tbl[k][np] != 0))
      return (0);
    if ((k >= num_eqn) & (tbl[k][np] < 0))
      return (0);
  }
  else if (second == -1) {	/* this is a dimension 1 constraint */
    if (k < num_eqn) {
      gama = -tbl[k][np] / tbl[k][first];
      /* var first has lower bound gama and upper bound gama */
      if (gama < lower_bnd[first])
	return (0);
      lower_bnd[first] = gama;
      if (gama > upper_bnd[first])
	return (0);
      upper_bnd[first] = gama;
    }
    else {			/* this is an inequality */
      if (tbl[k][first] > 0) {	/* the inequality is > */
	gama = -tbl[k][np] / tbl[k][first];
	/* gama is a new lower bound */
	if (gama > upper_bnd[first])
	  return (0);
	if (gama > lower_bnd[first])
	  lower_bnd[first] = gama;
      }
      else {			/* the inequality is < */
	gama = -tbl[k][np] / tbl[k][first];
	/* gama is a new upper bound */
	if (gama < lower_bnd[first])
	  return (0);
	if (gama < upper_bnd[first])
	  upper_bnd[first] = gama;
      }
    }
  }				/* end dim 1 case */
  else if (third == -1 && (second - first) == MAX_NEST_DEPTH) { 
  
  /* dimension 2 case involving i and i' look for i' - i > k forms */
    if (tbl[k][first] == -tbl[k][second]) {
      if (k < num_eqn) {
	dist_ub[first] = -tbl[k][np] / tbl[k][second];
	dist_lb[first] = dist_ub[first];
      }
      else if (tbl[k][second] < 0
	       && dist_ub[first] > tbl[k][np] / tbl[k][first])
	dist_ub[first] = tbl[k][np] / tbl[k][first];
      else if (tbl[k][second] > 0
	       /*&& dist_lb[first] < tbl[k][np] / tbl[k][second])*/
               && dist_lb[first] < tbl[k][np] / tbl[k][second])
	dist_lb[first] = -tbl[k][np] / tbl[k][second];
      if (dist_ub[first] < dist_lb[first])
	return (0);
    }
  }				/* end dim 2 case */
  return (1);
}

/* calculate the low bound and high bound of the equtions and inequations */

cmp_bnd (bound,temp_il_lo,temp_il_hi,tbl)
int bound[];
struct coef_off{
                int offset;
                int coefs[2*MAX_NEST_DEPTH];
 };
struct coef_off temp_il_lo[],temp_il_hi[];
int tbl[];
/*int level;*/
{
 int low_coef[2*MAX_NEST_DEPTH]; /* coef for computing low bound */
 int up_coef[2*MAX_NEST_DEPTH];  /* coef for computing up bound */
 /*struct subscript temp_il_lo[2*MAX_NEST_DEPTH],temp_il_hi[2*MAX_NEST_DEPTH];*/
 int i,j,k;

 for (i=0; i<MAX_NEST_DEPTH; i++){
  /*temp_il_lo[i]=il_lo[i];
  temp_il_lo[i+MAX_NEST_DEPTH]=il_lo[i];
  temp_il_hi[i]=il_hi[i];
  temp_il_hi[i+MAX_NEST_DEPTH]=il_hi[i];*/
  low_coef[i]=tbl[i];
  low_coef[i+MAX_NEST_DEPTH]=tbl[i+MAX_NEST_DEPTH];
  up_coef[i]=tbl[i];
  up_coef[i+MAX_NEST_DEPTH]=tbl[i+MAX_NEST_DEPTH];
 }

 bound[0]=0;
 bound[1]=0;

 k=np-1;
 while (k >= 0){
  bound[0]=bound[0]+plus_part(low_coef[k])*temp_il_lo[k].offset
                   -minus_part(low_coef[k])*temp_il_hi[k].offset;
  bound[1]=bound[1]+plus_part(up_coef[k])*temp_il_hi[k].offset
                   -minus_part(up_coef[k])*temp_il_lo[k].offset;
  for (j=0; j<k; j++){
    low_coef[j]=low_coef[j]+plus_part(low_coef[k])*temp_il_lo[k].coefs[j]
                           -minus_part(low_coef[k])*temp_il_hi[k].coefs[j];
    up_coef[j]=up_coef[j]+plus_part(up_coef[k])*temp_il_hi[k].coefs[j]
                         -minus_part(up_coef[k])*temp_il_lo[k].coefs[j];
   }
  k=k-1;
 }
}



/*search for the max intersect sub_equations*/
int search_intersect(tbl,num_eqn,set,intersect_set)
int *tbl[];
int num_eqn;
int set[];
int intersect_set[];
{
int i,j,k,m,n,l;
int break_tag;
for (i=0;i<100;i++)
 intersect_set[i]=-1;
k=0;
for (i=0;i<num_eqn;i++)
 if (set[i] !=-1){
   intersect_set[0]=set[i];
   set[i]=-1;
   k++;
   break;
 }
if (intersect_set[0]==-1)
   return(0);

for (l=0;l<k;l++){
 m=intersect_set[l];
 for(i=1;i<num_eqn;i++){
   n=set[i];
   if (n==-1)
    continue;
   for(j=0;j<np;j++){
    if(tbl[m][j]==0 || tbl[n][j]==0)
     continue;
    intersect_set[k]=set[i];
    set[i]=-1;
    k++;
    break;
  }
 }
}
/*for(i=1;i<num_eqn;i++)
  if(set[i]!=-1){
   m=set[i];
   break_tag=0;
   for(j=0;j<k;j++){
    n=intersect_set[j];
    for(l=0;l<np;l++){
     if(tbl[m][l]==0 || tbl[n][l]==0)
      continue;
     intersect_set[k]=set[i];
     set[i]=-1;
     k++;
     break_tag=1;
     break;
   }
   if (break_tag)
     break;
  }
 }*/
 return(k);
}

/*build the intersect sub_equation tbl*/
build_tbl(tbl,new_tbl,num,num_eqn,depth,intersect_set)
int *tbl[];
/*int new_tbl[][];*/
int *new_tbl[];
int num;
int num_eqn;
int depth;
int intersect_set[];
{
 int i,m,j;
 for (i=0;i<num;i++){
   m=intersect_set[i];
   new_tbl[i]=tbl[m];
 }
 m=i;
 for (i=num_eqn;i<depth;i++){
   new_tbl[m]=tbl[i];
   m++;
 }
}


int new_reduce(tbl,num_eqn,tbl_depth,il_lo,il_hi,max_depth)
int *tbl[];
int num_eqn,tbl_depth,max_depth;
struct subscript il_lo[MAX_NEST_DEPTH],il_hi[MAX_NEST_DEPTH];
{
 int i,j,k,t,mgcd,piv,pcol,opc,alf,bet;
 int *tmp;
 int bound[2];
 int tag=0;
 struct coef_off{
                 int offset;
                 int coefs[2*MAX_NEST_DEPTH];
 };
 struct coef_off temp_il_lo[2*MAX_NEST_DEPTH],temp_il_hi[2*MAX_NEST_DEPTH];
 int var_tbl[2*MAX_NEST_DEPTH];
 int *b[1000],*p;
 int a[40];
 int m,n,l;
 int *temp_tbl[MAX_NEST_DEPTH * 4 + AR_DIM_MAX];
 int *new_tbl[MAX_NEST_DEPTH * 4 + AR_DIM_MAX];
 int temp_new_tbl[MAX_NEST_DEPTH * 4 + AR_DIM_MAX][2*MAX_NEST_DEPTH+1];
 int num;
 int set[100],intersect_set[100];
 int new_tbl_depth;

for (i=0; i<max_depth; i++)
  if (il_lo[i].decidable !=1 || il_hi[i].decidable !=1){
     tag=1;
     break;
}

if (tag==1)
  for (i = 0; i < 2 * MAX_NEST_DEPTH; i++) {
    upper_bnd[i] = 32000;
    lower_bnd[i] = -32000;
    if (i < MAX_NEST_DEPTH) {
      dist_lb[i] = -32000;
      dist_ub[i] = 32000;
    }
  }

else{
 for (i=0; i<2*MAX_NEST_DEPTH; i++){
  if (i < MAX_NEST_DEPTH){
   temp_il_lo[i].offset=il_lo[i].offset;
   temp_il_hi[i].offset=il_hi[i].offset;
   for (j=0; j<MAX_NEST_DEPTH; j++){
    temp_il_lo[i].coefs[j]=il_lo[i].coefs[j];
    temp_il_lo[i].coefs[j+MAX_NEST_DEPTH]=0;
    temp_il_hi[i].coefs[j]=il_hi[i].coefs[j];
    temp_il_hi[i].coefs[j+MAX_NEST_DEPTH]=0;
   }
  }
  else{
   temp_il_lo[i].offset=il_lo[i-MAX_NEST_DEPTH].offset;
   temp_il_hi[i].offset=il_hi[i-MAX_NEST_DEPTH].offset;
   for (j=0; j<MAX_NEST_DEPTH; j++){
    temp_il_lo[i].coefs[j+MAX_NEST_DEPTH]=il_lo[i-MAX_NEST_DEPTH].coefs[j];
    temp_il_lo[i].coefs[j]=0;
    temp_il_hi[i].coefs[j+MAX_NEST_DEPTH]=il_hi[i-MAX_NEST_DEPTH].coefs[j];
    temp_il_hi[i].coefs[j]=0;
   }
  }
 } 

 for (i =0; i < 2 * MAX_NEST_DEPTH; i++){
   for (j=0; j<np; j++)
    var_tbl[j]=0;
   var_tbl[i]=1;
   cmp_bnd(bound,temp_il_lo,temp_il_hi,var_tbl);
   upper_bnd[i]=bound[1];
   lower_bnd[i]=bound[0];
   if (i < MAX_NEST_DEPTH){
     comp_bound(bound,il_lo,il_hi,i);
     dist_lb[i]=bound[0];
     dist_ub[i]=bound[1];
   }
 }
}

for (i=0; i<tbl_depth; i++){
  if (chk_bnds(tbl,i,upper_bnd,lower_bnd)==0)
    return (0);
  if (tag==0){
   cmp_bnd(bound,temp_il_lo,temp_il_hi,tbl[i]);
   if (i<num_eqn){
     if (-tbl[i][np] < bound[0] || -tbl[i][np] > bound[1])
      return (0);
   }
   else 
     if (bound[1]<-tbl[i][np])
       return(0);
     else if (-tbl[i][np] < bound[0])
            tbl[i][np]=-bound[0];
  }
}

for (i=0; i<40; i++)
  a[i]=32000;

for (i=0;i<num_eqn;i++)
  set[i]=i;
num=search_intersect(tbl,num_eqn,set,intersect_set);
while (num!=0){
 build_tbl(tbl,new_tbl,num,num_eqn,tbl_depth,intersect_set);
 m=0;
 for (i=0; i<np; i++)
  for (j=0; j<num; j++)
   if (new_tbl[j][i] !=0){
    a[m]=i;
    m++;/*represent the number of the variables*/
    break;
   }
 if (m==0)
  return(1);
 /*fprintf(stderr,"%d,%d\n",m,num_eqn);*/
 if (num > m){
  /*if (m > 1){
   b[0]=(int *)malloc((m-1)*sizeof(int));*/
   b[0]=(int *)malloc(m*sizeof(int));
   if (!b[0])
    fprintf(stderr,"space distribute failed\n");
  /*}
  n=m-1;*/
  n=m;
 }
 else{
  if (num > 1){ 
   b[0]=(int *)malloc((num-1)*sizeof(int));
   if (!b[0])
    fprintf(stderr,"space distribute failed\n");
  }
  n=num-1;
 }
 if (n > 0){
  row_dim=0;    
  permute(a,0,b,0,m,n);

 new_tbl_depth=num+tbl_depth-num_eqn;
 for (i=0;i<=row_dim;i++){
  for (t=0;t<new_tbl_depth;t++){
   for (l=0;l<=np;l++)
    temp_new_tbl[t][l]=new_tbl[t][l];
   temp_tbl[t]=temp_new_tbl[t];
  }
 /* for (t=0;t<num;t++){
   fprintf(stderr,"\n");
   for (l=0;l<=np;l++)
    fprintf(stderr,"%d  ",temp_tbl[t][l]);
  }*/
  p=b[i];
  for (j=0;j<n;j++){
   pcol=*p;
   piv=-1;
  /*first eliminate by using the equations*/
   for (l=j;l<num_eqn;l++) 
    if (temp_tbl[l][pcol] !=0){
     piv=l;
     break;
    }
  if (piv >-1){
  /* swap to bring to top */
    tmp=temp_tbl[j];
    temp_tbl[j]=temp_tbl[piv];
    temp_tbl[piv]=tmp;
  /* first reduce by gcd of row */
   if (tbl[j][pcol] < 0)
    for (t=0; t<=np;t++)
     temp_tbl[j][t]= -temp_tbl[j][t];
   mgcd = gcd(np-1,temp_tbl[j]); 
   if (mgcd > 1){
   /* first test for bad congruence class */
   /*fprintf(stderr,"%d\n",temp_tbl[j][np]);*/
    if ((temp_tbl[j][np] % mgcd) !=0){
     for (m=0; m<=row_dim; m++)
      free(b[m]);
      return (0);
   }
   for (t = 0; t <=np; t++)
    temp_tbl[j][t]=temp_tbl[j][t]/mgcd;
  }
  /* now do elimination on pcol */
  alf=temp_tbl[j][pcol];
  if (alf ==0)
   fprintf(stderr,"reduce error\n");
  else if (alf < 0){
   alf=-alf;
   for (t=0; t<=np; t++)
     temp_tbl[j][t]=-temp_tbl[j][t];
  }

 for (k=j+1;k<new_tbl_depth;k++){
   if ((bet=temp_tbl[k][pcol])!=0){
   /*first reduce row k*/
   for (t=0; t<=np;t++)
    temp_tbl[k][t]=alf * temp_tbl[k][t] -bet * temp_tbl[j][t];
  /*test for dim 1 or 0 constraint */
  if (chk_bnds(temp_tbl,k,upper_bnd,lower_bnd)==0){
    for (m=0; m<=row_dim; m++)
     free(b[m]);
     return(0);
  }
  if (tag==0){
    cmp_bnd(bound,temp_il_lo,temp_il_hi,temp_tbl[k]);
    if ((k < num)&&
      (-temp_tbl[k][np]<bound[0] || -temp_tbl[k][np] > bound[1])){
      for (m=0; m<=row_dim; m++)
       free(b[m]);
       return(0);
     }
   else{
     if (bound[1]<-temp_tbl[k][np]){
      for (m=0; m<=row_dim; m++)
       free(b[m]);
       return(0);
     }
    else if (-temp_tbl[k][np] < bound[0])
           temp_tbl[k][np]=-bound[0];
     }
    } 
   }
  }
 }
 p++;
 }
/*eliminate (n+1)th variable by adding the last equation to the inequation*/
for (t=0; t<=np; t++)
  if (temp_tbl[j][t]!=0){
     pcol=t;
     break;
  }
alf=temp_tbl[j][pcol];
for(k=num;k<new_tbl_depth;k++){
  if ((bet=temp_tbl[k][pcol])!=0){
     /*first reduce row k*/
   for (t=0; t<=np;t++)
    temp_tbl[k][t]=alf * temp_tbl[k][t] -bet * temp_tbl[j][t];
  /*test for dim 1 or 0 constraint */
  if (chk_bnds(temp_tbl,k,upper_bnd,lower_bnd)==0){
    for (m=0; m<=row_dim; m++)
     free(b[m]);
     return(0);
  }
  if (tag==0){
    cmp_bnd(bound,temp_il_lo,temp_il_hi,temp_tbl[k]);
    if (bound[1]<-temp_tbl[k][np]){
      for (m=0; m<=row_dim; m++)
       free(b[m]);
       return(0);
     }
    else if (-temp_tbl[k][np] < bound[0])
           temp_tbl[k][np]=-bound[0];
  }
 }
}
}
 for (i=0; i<=row_dim; i++)
  free(b[i]);
}
else{
  new_tbl_depth=num+tbl_depth-num_eqn;
  for (t=0;t<new_tbl_depth;t++)
    temp_tbl[t]=new_tbl[t];
  /*for (t=0;t<new_tbl_depth;t++){
    fprintf(stderr,"\n");
    for (l=0;l<=np;l++)
     fprintf(stderr,"%d  ",temp_tbl[t][l]);
  }*/
  for (j=0;j<num;j++){
   for (t=0;t<np;t++)
    if (temp_tbl[j][t]!=0){
      pcol=t;
      break;
    }
 /* first reduce by gcd of row */
   if (tbl[j][pcol] < 0)
    for (t=0; t<=np;t++)
     temp_tbl[j][t]= -temp_tbl[j][t];
   mgcd = gcd(np-1,temp_tbl[j]);
   if (mgcd > 1){
   /* first test for bad congruence class */
   /*fprintf(stderr,"%d\n",temp_tbl[j][np]);*/
    if ((temp_tbl[j][np] % mgcd) !=0)
      return (0);
   for (t = 0; t <=np; t++)
    temp_tbl[j][t]=temp_tbl[j][t]/mgcd;
  }
  /* now do elimination on pcol */
  alf=temp_tbl[j][pcol];
  if (alf ==0)
   fprintf(stderr,"reduce error\n");
  else if (alf < 0){
   alf=-alf;
   for (t=0; t<=np; t++)
     temp_tbl[j][t]=-temp_tbl[j][t];
  }

   for (k=num;k<new_tbl_depth;k++){
     if ((bet=temp_tbl[k][pcol])!=0){
     for (t=0; t<=np;t++)
      temp_tbl[k][t]=alf * temp_tbl[k][t] -bet * temp_tbl[j][t];
     if (chk_bnds(temp_tbl,k,upper_bnd,lower_bnd)==0)
      return(0);
     if (tag==0){
      cmp_bnd(bound,temp_il_lo,temp_il_hi,temp_tbl[k]);
      if (bound[1]<-temp_tbl[k][np])
       return(0);
      else if (-temp_tbl[k][np] < bound[0])
           temp_tbl[k][np]=-bound[0];
     }
    }
   }
  }
 }
num=search_intersect(tbl,num_eqn,set,intersect_set);
}

 /* second eliminate by adding inequalities */
  for (j = num_eqn; j < tbl_depth; j++) {
    /* find leader pivod equation */
    piv = -1;
    opc = pcol;
    for (k = opc + 1; k < MAX_NEST_DEPTH * 2; k++)
      for (t = j; t < tbl_depth; t++)
        if (opc == pcol && tbl[t][k] > 0) {
          pcol = k;
          piv = t;
        }

    if (piv > -1) {
      /* swap to bring to top */
       tmp = tbl[j];
       tbl[j] = tbl[piv];
       tbl[piv] = tmp;
      /* now do elimination on pcol */
      alf = tbl[j][pcol];
      if (alf <= 0)
        fprintf(stderr, "reduce error\n");
      for (k = j + 1; k < tbl_depth; k++) {
        if ((bet = tbl[k][pcol]) < 0) {
          /* first do the ellimination */
          for (i = 0; i <= np; i++)
            tbl[k][i] = alf * tbl[k][i] - bet * tbl[j][i];
          tbl[k][np]=tbl[k][np]-abs(bet*alf)+alf+abs(bet)-1;
          /* now check for constraint errors */
          if (chk_bnds(tbl, k, upper_bnd, lower_bnd) == 0)
            return (0);
          if (tag==0){
           cmp_bnd(bound,temp_il_lo,temp_il_hi,tbl[k]);
           if (bound[1] < -tbl[k][np])
             return(0);
           else
            if (bound[0] > -tbl[k][np])
             tbl[k][np]=-bound[0];
         }
        }
      }
    }                           /* end of piv found case */
  }                             /* end of factorization loop */

  /* now look for contradictions in eqnations */
  for (j = 0; j < tbl_depth; j++)
    if (chk_bnds(tbl, j, upper_bnd, lower_bnd) == 0)
      return (0);
  return (1);
}




  
permute(a,pointa,b,pointb,m,n)
int a[];/*for the initiate value*/
int pointa;/*point to a*/
int *b[];/*for the result*/
int pointb;/*col for b*/
int m;/*total number of the a*/
int n;/*num of each unit*/
{
 int i,temp_dim;
 if (pointa>m-1 && pointb<=n-1){
   /*b[row_dim][0]=-32000;*/
   free(b[row_dim]);
   row_dim--;
   return;
 }
 if (pointb>n-1)
   return;
 b[row_dim][pointb]=a[pointa];
 temp_dim=row_dim;
 /*fprintf(stderr,"%d\n",row_dim);*/
 permute(a,pointa+1,b,pointb+1,m,n);
 row_dim++;
 b[row_dim]=(int *)malloc(n*sizeof(int));
 if (!b[row_dim]){
  fprintf(stderr,"space distribute failed\n");
  fprintf(stderr,"%d\n",row_dim);
 }
 for(i=0;i<pointb;i++)
  b[row_dim][i]=b[temp_dim][i];
 /*fprintf(stderr,":%d\n",row_dim);*/
 permute(a,pointa+1,b,pointb,m,n);
}





/*****************************************************************/
/* set_vec check the previous state of the troub and val vectors */
/* to see if a previous index computation has determined values  */
/* for the i-th induction var that differ from the current one.  */
/* if a val of zero is set troub[i] is set to -99 as a reminder. */
/*****************************************************************/
set_vec(i, vec, troub, val)
int i;
int vec[], troub[];
int val;
{
  if ((vec[i] != 0) || (troub[i] == -99)) {
    if (vec[i] != val)
      troub[0] = 1;
    if (val == 0)
      troub[i] = -99;
  }
  else if (((val < 0) && (troub[i] == ZPLUS)) ||
	   ((val > 0) && (troub[i] == ZMINUS)) ||
	   ((val == 0) && ((troub[i] == PLUS) || (troub[i] == MINUS)))
    )
    troub[0] = 1;
  else {
    vec[i] = val;
    if (val == 0)
      troub[i] = -99;
    else
      troub[i] = 0;
  }
}

set_troub(i, vec, troub, val)
int i;
int vec[], troub[];
int val;
{
  switch (val) {
   case PLUS:
    if ((vec[i] < 0) || (troub[i] == -99) ||
	(troub[i] == ZMINUS))
      troub[0] = 1;
    break;
   case MINUS:
    if ((vec[i] > 0) || (troub[i] = -99) ||
	(troub[i] == ZPLUS))
      troub[0] = 1;
    break;
   case ZPLUS:
    if ((vec[i] < 0) || (troub[i] == MINUS))
      troub[0] = 1;
    break;
   case ZMINUS:
    if ((vec[i] > 0) || (troub[i] == PLUS))
      troub[0] = 1;
    break;
   case PLUSMINUS:		/* does not invalidate anything! */
    break;
   default:
    troub[i] = val;
  }
  if ((troub[i] == NODEP) && (vec[i] == 0))
    troub[i] = val;
}



