#include <strings.h>
#include "db.h"
#include "elist.h"


/* the library version of the ELIST functions
 *
 */
ELIST SiENew(etype)
int etype;
{
  return ENew(etype);
} /* SiENew */

void SiEFree(e)
ELIST e;
{
   EFree(e);
} /* SiEFree */


/* SiESetType: sets the type of an ELIST e to type.
   returns 0 on success. -1 if type is outofrange or if
   e is null */
int SiESetType(e, type)
ELIST e;
int type;
{
  
  if (type != TEINT && type != TELIST && type != TESTRING)
     return -1;
  if (e)
  {
    e->type = type;
    return 0;
  }
  return -1;
} /* SiESetType */

/* SiESetCar: sets the car of an ELIST e to value.
   returns 0 on success. -1 if e is null 
*/
int SiESetCar(e, value)
ELIST e;
char *value;
{
  if (e) 
  {
    e->car = value;
    return 0;
  }
  return -1;
} /* SiESetCar */

/* SiESetCdr: sets the cdr of an ELIST e to ptr.
   returns 0 on success. -1 if e is null 
*/
int SiESetCdr(e,ptr)
ELIST e;
ELIST ptr;
{
  if (e) 
  {
     e->cdr = ptr;
     return 0;
  }
  return -1;
} /* SiESetCdr */

ELIST SiECopy(e)
ELIST e;
{
   return ECopy(e);
} /* SiECopy */

ELIST SiEAppend(e1, e2)
ELIST e1, e2;
{
   return EAppend(e1, e2);
} /* SiEAppend */

ELIST SiEString(s)
char *s;
{
   return EString(s);
} /* SiEString */

ELIST SiENumber(num)
int num;
{
   return ENumber(num);
} /* SiENumber */

/* 
/* special ELIST manipulation functions
*/
static ELIST elistFreeList = NULL;

ELIST ENew(etype)
int etype;
{
   ELIST e;
   if (elistFreeList == NULL)
      e = (ELIST) Allocate(sizeof(struct ELIST_rec));
   else
   {
      e = elistFreeList;
      elistFreeList = e->cdr;
   }
   e->cdr = NULL;
   e->car = NULL;
   e->type = etype;
   return e;
} /* ENew */

void EFree(e)
ELIST e;
{
   if (e == NULL) 
      return;
   if (e->type == TELIST) 
      EFree((ELIST) e->car);
   EFree(e->cdr);
   e->cdr = elistFreeList;
   elistFreeList = e;
   return;
} /* EFree */

ELIST ECopy(e)
ELIST e;
{
   ELIST cpy;
   if (e == NULL) 
      return NULL;

   cpy = ENew(e->type);
   switch (e->type)
   {
      case 0: 
           cpy->car = e->car;
	   break;
      case 1: 
           cpy->car = (char *) Allocate(strlen(e->car)+1);
           strcpy(cpy->car, e->car);
           break;
      case 2: 
           cpy->car = (char *) ECopy(e->car);
           break;
   } /* switch */
   cpy->cdr = ECopy(e->cdr);
   return cpy;
} /* ECopy */


ELIST EAppend(e1, e2)
ELIST e1, e2;
{
   ELIST p;

   p = e1;
   if (p == NULL) 
      return e2;
   while (p->cdr != NULL) 
      p = p->cdr;
   p->cdr = e2;
   return e1;
} /* EAppend */

ELIST EString(s)
char *s;
{
   ELIST e;
   e = ENew(1);
   e->car = Allocate(strlen(s)+1);
   strcpy(e->car, s);
   e->cdr = NULL;
   return e;
} /* EString */

ELIST ENumber(n)
int n;
{
   ELIST e;

   e = ENew(0);
   e->car = (char *) n;
   return e;
} /* ENumber */

ELIST SiECons(e1, e2)
ELIST e1, e2;
{
   return ECons(e1, e2);
} /* SiECons */

int SiENumP(e)
ELIST e;
{
   return ENumP(e);
} /* SiENump */

int SiEStringP(e)
ELIST e;
{
   return EStringP(e);
} /* SiEStringP */

int SiEListP(e)
ELIST e;
{
   return EListP(e);
} /* SiEListP */


/*
 * version of ELIST functions used by the cbaselibrary functions 
 */
ELIST ECons(e1, e2)
ELIST e1, e2;
{
   ELIST e;

   e = ENew(2);
   e->car = (char *) EAppend(e1,e2);
   e->cdr = NULL;
   return e;
} /* ECons */


int ENumP(e)
ELIST e;
{
    if (e == NULL) 
       return FALSE;
    if (e->type == TEINT) 
       return TRUE;
    return FALSE;
} /* ENumP */

int EStringP(e)
ELIST e;
{
   if (e == NULL) 
      return FALSE;
   if (e->type == TESTRING) 
      return TRUE;
   else 
      return FALSE;
} /* EStringP */

int EListP(e)
ELIST e;
{
   if (e == NULL) 
      return FALSE;
   if (e->type == TELIST) 
      return TRUE;
   else 
      return FALSE;
} /* EListP */

