#include <strings.h>
#include "db.h"
#include "elist.h"


/*
 * functions used from libdb.a 
 */
extern char *copys();         /* from libdb.a */
extern void normal_form();    /* from libdb.a */
extern PTR_LLND make_llnd();
extern PTR_LLND replace_ref();
extern void collect_garbage();

/*
 * externs from libdb.a
 */
extern PCF UnparseLlnd[];
extern PCF UnparseType[];
/*
 * External references
 */
extern PTR_FILE currentFile;	/* point to the current active file */


/*
 * external functions from elsewhere in the cbaselib
 */
extern PTR_LLND FindLLNode();

/* unparses expression tree rooted at id */
/****************************************************************
 *								*
 * SiUnparseExpr - unparses expression tree rooted at id        *
 *								*
 * Input -							*
 *      id - id of the expression node.				*
 *								*
 * Output -							*
 *   elist of unparsed expr. NULL is failure.			*
 *								*
 ****************************************************************/
ELIST SiUnparseExpr(id)
int id;
{
   PTR_LLND ll;
   ELIST e;
   char * s, *t;

   if (ll = FindLLNode(id)) 
   {
      s = t = (UnparseLlnd[currentFile->lang])(ll);
      while (*t) 
         if (*t == '\n') 
            *t = '\0'; 
         else 
            t++;
      return EString(s);
   } 
   else
      return NULL;
} /* SiUnparseExpr */


/****************************************************************
 *								*
 *  SiExprVariant -- get the low level node type 		*
 *								*
 *  Input:							*
 *	id - the low level node id				*
 *								*
 *  Output:							*
 *      the node variant, -1 indicates error			*
 *								*
 ****************************************************************/
int SiExprVariant(id)
int id;
{
   PTR_LLND ll;

   if (ll = FindLLNode(id))
      return ll->variant;
   else
      return -1;
} /*  SiExprVariant */

 /* 
  * (static) expand_expr1 : help function for SiExpandExpr 
  */
static ELIST expand_expr1(ll)
PTR_LLND ll;
{
   ELIST e;
   if (ll) 
   {
      e = ENumber(ll->id);
      e = EAppend(e, ECons( expand_expr1(ll->entry.template.ll_ptr1), NULL));
      e = EAppend(e, ECons( expand_expr1(ll->entry.template.ll_ptr2), NULL));
      return e;
   }
   else 
      return NULL;
} /* expand_expr1 */


/****************************************************************
 *								*
 * SiExpandExpr - returns list of ids for the expression        *
 *   The form of the tree is (root (left_child_tree)            *
 *                                 (right_child_tree))          *
 *								*
 * Input -							*
 *   the id for the expression node.                            *
 *								*
 * Output -							*
 *    ELIST of expanded expression nodes, NULL indicates no     *
 *      such expression						*
 *								*
 ****************************************************************/
ELIST SiExpandExpr(id)
int id;
{
   PTR_LLND ll;

   if (ll = FindLLNode(id)) 
   {
      return expand_expr1(ll);
   } 
   else
      return NULL;
} /* SiExpandExpr */


/****************************************************************
 *								*
 * SiGetLRChild - Similar to SiExpandExpr  But only at the      *
 *   toppest level .ie. returns only the left and right child   *
 *   of the first level.                           		*
 *								*
 * Input -							*
 *   id - the id of the expression node                         *
 *								*
 * Output -							*
 *    The ELIST consisting of the left and right child		*
 *    NULL indicates no such node.				*
 *								*
 ****************************************************************/
ELIST SiGetLRChild(id)
int id;
{
   PTR_LLND ll;

   if (ll = FindLLNode(id))
      return  EAppend(  
	   ENumber(ll->entry.template.ll_ptr1 
                    ? ll->entry.template.ll_ptr1->id
                    : -1),
	   ENumber(ll->entry.template.ll_ptr2
                    ? ll->entry.template.ll_ptr2->id
                    : -1));
   else
      return NULL;
}


/****************************************************************
 * the following can be used for extracting field entries of    *
 * expression nodes.                                            *
 ****************************************************************/
/****************************************************************
 *								*
 * SiGetLeftChild - returns the left child of an expression     *
 *								*
 * Input -							*
 *   id - the id of the expression node                         *
 *								*
 * Output -							*
 *    id of the left child, -1 in case of error.		*
 *								*
 ****************************************************************/
 
int SiGetLeftChild(id)
int id;
{
	PTR_LLND p;
	if((p = FindLLNode(id)) && p->entry.template.ll_ptr1)
		return p->entry.template.ll_ptr1->id;
	return -1;
}
/****************************************************************
 *								*
 * SiGetRightChild - returns the right child of an expression   *
 *								*
 * Input -							*
 *   id - the id of the expression node                         *
 *								*
 * Output -							*
 *   id of the right child, -1 in case of error.		*
 *								*
 ****************************************************************/
int SiGetRightChild(id)
int id;
{
	PTR_LLND p;
	if((p = FindLLNode(id)) && p->entry.template.ll_ptr2)
		return p->entry.template.ll_ptr2->id;
	return -1;
}

/****************************************************************
 * (support function for other functions in cbaselib)           *
 * CopyLLNode -- recursively make a copy of a given low level   *
 *    node         -						*
 *								*
 * Input:							*
 *	p - the pointer of the low level node to be duplicated	*
 *								*
 * Output:							*
 *	the pointer of the new low level node			*
 *								*
 ****************************************************************/
PTR_LLND CopyLLNode(p)
PTR_LLND p;
{
   PTR_LLND q;

    if (p == NULL) 
       return NULL;

    q = make_llnd(currentFile,
                  p->variant,LLNULL,LLNULL,p->entry.template.symbol);
    q->type = p->type;
    q->entry.template.ll_ptr1 = CopyLLNode(p->entry.template.ll_ptr1);
    q->entry.template.ll_ptr2 = CopyLLNode(p->entry.template.ll_ptr2);
    return q;
} /* CopyLLNode */


/****************************************************************
 *								*
 *  SiCopyExpr -- make a copy of the given expression		*
 *								*
 *  Input:							*
 *	id - the expression node id.				*
 *								*
 *  Output:							*
 *	id of the newly created node, -1 on error               *
 *								*
 ****************************************************************/
int SiCopyExpr(id)
int id;
{
   PTR_LLND l1, l2;

    if (l1 = FindLLNode(id))
       if (l2 = CopyLLNode(l1)) 
	  return(l2->id);
    return -1;
} /* SiCopyExpr */


/* VarId: help function for SiVarId and some flow functions in csvr_flow.c */
int VarId(id)
{
   PTR_LLND ll;

   if (ll = FindLLNode(id)) 
   {
      if (ll->variant == VAR_REF || ll->variant == ARRAY_REF)
	 return ll->entry.template.symbol->id;
      else
	 return -1; /* signal not a VAR_REF error */
    }
    return -1;
} /* VarId */

/****************************************************************
 *								*
 *  SiVarId -- gets symbol table id for VAR_REF node		*
 *								*
 *  Input:							*
 *	id -- the low level node id 				*
 *								*
 *  Output:							*
 *	id of the node if success, -1 if failure .		*
 *								*
 ****************************************************************/
int SiVarId(id)
int id;
{
   return VarId(id);
} /* SiVarId */

/****************************************************************
 *								*
 *  SiConstantValue --  get the value of constant type		*
 *								*
 *  Input:							*
 *	id -- the symbol table id of the type interested	*
 *								*
 *  Output:							*
 *      returns an ELIST of value, coz floats are returned as   *
 *     strings							*
 *								*
 ****************************************************************/
ELIST SiConstantValue(id)
int id;
{
   PTR_SYMB sym;
   PTR_LLND ll;
   ELIST e;

   for (sym = currentFile->head_symb; sym; sym = sym->thread)
      if (sym->id == id) 
      {
	 ll = sym->entry.const_value;
	 switch (ll->variant) 
         {
	    case INT_VAL:
	    case BOOL_VAL:
		 e = EAppend(ENumber(ll->variant), ENumber(ll->entry.ival));
		 break;
	    case FLOAT_VAL:
	    case DOUBLE_VAL:
	    case STRING_VAL:
		 e = EAppend(ENumber(ll->variant), 
                             ENumber(ll->entry.string_val));
		 break;
	    case CHAR_VAL:
		 e = EAppend(ENumber(ll->variant), ENumber(ll->entry.cval));
		 break;
	    default:
		 return NULL;
	    }
	    return e;
	}
    return NULL;
} /* SiConstantValue */

/****************************************************************
 * the following can be used for extracting the values of       *
 * constant expression nodes.  not that SiGetConstantValue()    *
 * is used only for symbols.                                    *
 ****************************************************************/
/****************************************************************
 *								*
 *  SiGetCharConstExpr --  get the value of a constant 	expr.   *
 *      node.							*
 *								*
 *  Input:							*
 *	id -- expression node id 				*
 *      pchar -- pointer to the character where the value is to *
 *          be stored						*
 *								*
 *  Output:							*
 *     saves the character value in *pchar, and returns the     *
 *     status of the computation, 0 means success , -1 means    *
 *     failure.							*
 *								*
 ****************************************************************/
int SiGetCharConstExpr(id, pchar)
int id;
char *pchar;
{
   PTR_LLND p;
   if ((p = FindLLNode(id)) && (p->variant == CHAR_VAL))
   {
      *pchar = p->entry.cval;
      return 0;
   }
   return -1;
} /* SiGetCharConstExpr */

/****************************************************************
 *								*
 *  SiGetCharConstExpr --  get the value of a constant 	expr.   *
 *      node.							*
 *								*
 *  Input:							*
 *	id -- expression node id 				*
 *      pint -- pointer to the integer where the value is to    *
 *          be stored						*
 *								*
 *  Output:							*
 *     saves the integer value in *pint, and returns the        *
 *     status of the computation, 0 means success , -1 means    *
 *     failure.							*
 *								*
 ****************************************************************/
int SiGetIntConstExpr(id, pint)
int id;
int *pint;
{
   PTR_LLND p;
   if ((p = FindLLNode(id)) && (p->variant == INT_VAL))
   {
      *pint =  p->entry.ival;
      return 0; 
       
   }
   return -1;
} /* SiGetIntConstExpr */

/****************************************************************
 *								*
 *  SiGetFloatConstExpr --  get the value of a float constant   *
 *      expr  node.						*
 *								*
 *  Input:							*
 *	id -- expression node id 				*
 *      pstr -- pointer to the string which stores the float    *
 *         constant as a string. Note that it doesnot return    *
 *         the value itself for portability reasons.		*
 *								*
 *  Output:							*
 *     saves the string value in *pstr, and returns the         *
 *     status of the computation, 0 means success , -1 means    *
 *     failure.							*
 *								*
 ****************************************************************/
int SiGetFloatConstExpr(id, pstr)
int id;
char **pstr;
{
   PTR_LLND p;
   if ((p = FindLLNode(id)) && (p->variant == FLOAT_VAL))
   {
      *pstr =  p->entry.string_val;
      return 0;
   }
   return -1;
} /* SiGetFloatConstExpr */

/****************************************************************
 *								*
 *  SiGetStringConstExpr --  get the value of a string constant *
 *      expr node.						*
 *								*
 *  Input:							*
 *	id -- expression node id 				*
 *      pstr -- pointer to the character where the value is to  *
 *          be stored						*
 *								*
 *  Output:							*
 *     saves the pointer to the string in *pstr, and returns    *
 *     status of the computation, 0 means success , -1 means    *
 *     failure.							*
 *								*
 ****************************************************************/
int SiGetStringConstExpr(id, pstr)
int id;
char **pstr;
{
   PTR_LLND p;
   if ((p = FindLLNode(id)) && (p->variant == STRING_VAL))
   {
      *pstr =  p->entry.string_val;
      return 0;
   }
   return -1;
}

/****************************************************************
 *								*
 *  SiGetSymbolField -- Get the id of the variable, given       *
 *    node id of the variable of an array expression.		* 
 *								*
 *  Input:							*
 *	id -- expression node id 				*
 *								*
 *  Output:							*
 *      returns the symbol table id of the variable name	*
 *      -1 is returned in case of error.			*
 *      for an array reference, the base array name id is       *
 *      returned						*
 *								*
 ****************************************************************/
int SiGetSymbolField(id)
int id;
{
	PTR_LLND p;
	if((p = FindLLNode(id)) && p->entry.template.symbol)
		return p->entry.template.symbol->id;
	return -1;
}

/****************************************************************
 *								*
 *  SiGetSymbolFieldIdent -- Get the name of the variable, given*
 *    node id of the variable of an array expression.		* 
 *								*
 *  Input:							*
 *	id -- expression node id 				*
 *								*
 *  Output:							*
 *      returns the string name of the variable.		*
 *      NULL is returned in case of error.			*
 *      for an array reference, the array name is               *
 *      returned						*
 *								*
 ****************************************************************/
char * SiGetSymbolFieldIdent(id)
int id;
{
	PTR_LLND p;
	if((p = FindLLNode(id)) && p->entry.template.symbol)
		return p->entry.template.symbol->ident;
	return NULL;
}	

/****************************************************************
 *								*
 *  SiSimplifyExpr -- calls symbolic algebra package and 	*
 *		     simplifies the expression			*
 *								*
 *  Input:							*
 *	id -- the id of the low level node to be simplified	*
 *								*
 *  Output:							*
 *	id of the new low level node 				*
 *	-1 indicates error					*
 *								*
 ****************************************************************/
int SiSimplifyExpr (id)
int id;
{
   PTR_LLND ll;

   if (ll = FindLLNode(id)) 
   {
      normal_form(&ll);
      return ll->id;
   } 
   else
      return -1;
} /* SiSimplifyExpr */


/****************************************************************
 *								*
 *  SiCollectExprGarbage -- clean up the temporary low level	*
 *		node created in the middle of transformations	*
 *		or the kind					*
 *								*
 *  Input:							*
 *	none							*
 *								*
 *  Output:							*
 *	0                                          		*
 *								*
 ****************************************************************/
void SiCollectExprGarbage()
{
   collect_garbage(currentFile);
} /* SiCollectExprGarbage */



/****************************************************************
 *								*
 *  SiNewExpr -- create a new expression tree with given type   *
 *		   and children 				*
 *								*
 *  Input:							*
 *	type   - integer, the variant of the new low level node *
 *	lchild - id of the left tree (low level node)		*
 *	rchild - id of the right tree (low level node)		*
 *								*
 *  Output:							*
 *	id of the new low level node, -1 indicating failure     *
 *								*
 ****************************************************************/
int SiNewExpr(type, lchild, rchild)
int type, lchild, rchild;
{
    PTR_LLND l1 = LLNULL, l2 = LLNULL, l3;

    if (lchild > 0) l1 = FindLLNode(lchild);
    if (rchild > 0) l2 = FindLLNode(rchild);
	  if ((l3 = make_llnd(currentFile, type, l1, l2, SMNULL))) 
          {
	     return l3->id;
	  }
    return -1;
} /* SiNewExpr */


/****************************************************************
 *								*
 *  SiNewCFuncCall-- create a new expression tree with given type*
 *		   and children 				*
 *								*
 *  Input:							*
 *	function symbol id (see SiMakeFunctionName)
 *	arglist a EXPR_LIST of arguments
 *  Output:							*
 *	id of the new low level node, -1 indicating failure     *
 *								*
 ****************************************************************/
int SiNewCFuncCall(symb_id, arg_id)
int symb_id, arg_id;
{
    PTR_LLND l1 = LLNULL, l2 = LLNULL, l3;
    PTR_SYMB s1;

    if ((l1 = FindLLNode(arg_id)) && (s1 = FindSymbolNode(symb_id)))
       if ((l3 = make_llnd(currentFile, FUNC_CALL, l1, l2, s1))) 
       {
	  return l3->id;
       }
    return -1;
} /* SiNewCFuncCall */


/****************************************************************
 *								*
 *  SiNewSymbolRef -- create a new symbol reference node or     *
 *		    array reference node			*
 *								*
 *  Input:							*
 *	type - integer, the variant of the new low level node	*
 *	symb - id of the symbol table node			*
 *	expr - "-1" if it's simple variable. Otherwise the id	*
 *	       of the low level node of the array reference	*
 *								*
 *  Output:							*
 *	A status code and id of the new low level node were	*
 *	written to the stdout.					*
 *								*
 ****************************************************************/
int SiNewSymbolRef(type, symb, expr)
int type, symb, expr;
{
   PTR_SYMB s;
   PTR_LLND ll = LLNULL, l1;

   if (s = FindSymbolNode(symb))
      if (expr < 0 || (expr > 0 && (ll = FindLLNode(expr))))
	 if ((l1 = make_llnd(currentFile, type, ll, LLNULL, s))) 
         {
	    return l1->id;
	 }
   return -1;
} /* SiNewSymbolRef */


/****************************************************************
 *								*
 *  SiSubChildren -- change the left or right child of a given	*
 *		    expression node				*
 *								*
 *  Inputs :							*
 *	root   - the id of the root of the tree to be changed	*
 *	lchild - the id of the new left child tree		*
 *	rchild - the id of the new right child tree		*
 *      A child id of -1 will cause that child to be not 	*
 *      modified.						*
 *								*
 *  Output:							*
 *     0 for success, -1 for failure				*
 *								*
 ****************************************************************/
int SiSubChildren(root, lchild, rchild)
int root, lchild, rchild;
{
    PTR_LLND lr, ll;

    if (lr = FindLLNode(root)) 
    {
       if (lchild != -1 && (ll = FindLLNode(lchild)))
	  lr->entry.template.ll_ptr1 = ll;
       if (rchild != -1 && (ll = FindLLNode(rchild)))
	  lr->entry.template.ll_ptr2 = ll;
       return 0;
    }
    return -1;
} /* SiSubChildren */


/****************************************************************
 *								*
 *  SiReplaceExpr(old_exp, new_exp) can be used replace a	*
 *    a subexpression in an expression with a different 	*
 *    different sub_expression. 				*
 *  NOTE: destroys the old_exp values.				*
 *  Inputs							*
 *	old_exp   - the id of the sub_expression to be modified *
 *	new_exp   - the id of the new sub_expression.		*
 *								*
 *  Outputs							*
 *	0 for success, -1 for failure. 				*
 *								*
 ****************************************************************/

int SiReplaceExpr(old_llnd, new_llnd)
int old_llnd, new_llnd;
{
  PTR_LLND ond, nnd;

  if (( ond = FindLLNode(old_llnd)) &&
      ( nnd = FindLLNode(new_llnd)))
  {
     ond->variant = nnd->variant;
     ond->entry.template.symbol = nnd->entry.template.symbol;
     ond->entry.template.ll_ptr1 = nnd->entry.template.ll_ptr1;
     ond->entry.template.ll_ptr2 = nnd->entry.template.ll_ptr2;
     return 0;
  }
  return -1;
} /* SiReplaceExpr */
	  

/****************************************************************
 *								*
 *  SiNewConstant -- creates new node with constant value.	*
 *								*
 * Input -							*
 *    ELIST containing the constant value 			*
 *								*
 * Output -							*
 *    id of the newly created node, -1 for failure 		*
 *								*
 ****************************************************************/
int SiNewConstant(e)
ELIST e;
{
    PTR_LLND ll = LLNULL;
    PTR_LLND l1,l2;
    int type, value;
    char tmps[100], *tmp;

    type = (int) e->car;
    e = e->cdr;
    tmp = &(tmps[100]);
    switch (type) 
    {
       case 1:			/* integer */
	    value = (int) e->car;
	    if (ll = make_llnd(currentFile, INT_VAL, LLNULL, LLNULL, SMNULL))
	       ll->entry.ival = value;
	    break;
       case 2:			/* real */
	    tmp = (char *) e->car;
	    if (ll = make_llnd(currentFile, FLOAT_VAL, LLNULL, LLNULL, SMNULL))
	       ll->entry.string_val = copys(tmp);
	    break;
       case 3:			/* double precision */
	    tmp = (char *) e->car;
	    if (ll = make_llnd(currentFile, DOUBLE_VAL, LLNULL, LLNULL, SMNULL))
	       ll->entry.string_val = copys(tmp);
	    break;
       case 6:			/* boolean */
	    value = (int) e->car;
	    if (ll = make_llnd(currentFile, BOOL_VAL, LLNULL, LLNULL, SMNULL))
	       ll->entry.bval = value;
	    break;
       case 4:			/* character constant */
	    value = (int) e->car;
	    if (ll = make_llnd(currentFile, CHAR_VAL, LLNULL, LLNULL, SMNULL))
	       ll->entry.cval = value;
	    break;
       case 5:			/* string constant */
	    tmp = (char *) e->car;
	    if (ll = make_llnd(currentFile, STRING_VAL, LLNULL, LLNULL, SMNULL))
	       ll->entry.string_val = copys(tmp);
	    break;
	case 8:		      /*  KEYWORD_VAL */
	    tmp = (char *) e->car;
	    if (ll = make_llnd(currentFile,KEYWORD_VAL, LLNULL, LLNULL, SMNULL))
	       ll->entry.string_val = copys(tmp);
	    break;
	case 7:		     /* complex val */
	    tmp = (char *) e->car; e = e->cdr;
	    if (l1 = make_llnd(currentFile, FLOAT_VAL, LLNULL, LLNULL, SMNULL))
	       l1->entry.string_val = copys(tmp);
	    tmp = (char *) e->car; e = e->cdr;
	    if (l2 = make_llnd(currentFile, FLOAT_VAL, LLNULL, LLNULL, SMNULL))
	       l2->entry.string_val = copys(tmp);
	    ll = make_llnd(currentFile, FLOAT_VAL, l1, l2, SMNULL);
	    break;
	    
    } /* switch */
    if (ll)
       return ll->id;
    else
       return -1;
} /* SiNewConstant */


/****************************************************************
 *								*
 * SiReplaceVar -- replace all occurrence of variable of "id"	*
 *		  in expr1 with expr2.				*
 *								*
 * Input -							*
 *   expr1 - id of the source expr				*
 *   id    - id of the variable in expr1			*
 *   expr2 - the replacing expression				*
 *								*
 * Output -							*
 *   id of the newly created node, -1 for failure		*
 *								*
 ****************************************************************/
int SiReplaceVar(expr1, id, expr2)
int expr1, id, expr2;
{
   PTR_LLND ll1, ll2, ll3;
   PTR_SYMB s;

    if ((ll1 = FindLLNode(expr1)) && (ll2 = FindLLNode(expr2)) &&
	(s = FindSymbolNode(id)))
       if (ll3 = replace_ref(ll1, s, ll2)) 
       {
	    return ll3->id;
       }
    return -1;
} /* SiReplaceVar */


/*
 * (static) make_vec_expr - help function for SiVectorExpr 
 */
static PTR_LLND make_vec_expr(low, hi, inc)
    PTR_LLND low, hi, inc;
{
    PTR_LLND l1;

    l1 = make_llnd(currentFile, DDOT, low, hi, SMNULL);
    return  make_llnd(currentFile, DDOT, l1, inc, SMNULL);
} /* make_vec_expr */


/****************************************************************
 *								*
 *  SiVectorExpr -- create a vector expression			*
 *								*
 *  Input:							*
 *    e - elist (low hi inc)					*
 *	low  - id of the expression tree (low level node) that	*
 *	       represents the lower bound of the loop		*
 *	hi   - id of the expression tree (low level node) that	*
 *	       represents the upper bound of the loop		*
 *	inc  - id of the expression tree (low level node) that	*
 *	       represents the increment of the loop.  Should be *
 *	       0 if no increment specified			*
 * 								*
 *  Output:							*
 *     id of the newly created expression tree			*
 *								*
 *								*
 ****************************************************************/
int SiVectorExpr(e)
ELIST e;
{
    int  n, l1[10], l2[10], l3[10], i, j, *ip1, *ip2, *ip3, ch;
    PTR_LLND low[10], hi[10], inc[10], *lp1, *lp2, *lp3, ll, ll1, ll2;
    ELIST b;

    n = (int) e->car;
    e = e->cdr;
    ip1 = l1;  ip2 = l2;  ip3 = l3;
    for (i = 0; i < n; i++) 
    {
       b = (ELIST) e->car;
       *ip1++ = (int) b->car; 
       b = b->cdr;
       *ip2++ = (int) b->car; 
       b = b->cdr;
       *ip3++ = (int) b->car; 
       b = b->cdr;
       e = e->cdr;
    }

    ip1 = l1;  ip2 = l2;  ip3 = l3;
    lp1 = low; lp2 = hi;  lp3 = inc;

    /* Make sure all nodes exist before doing any changes */
    for (i = 0; i < n; i++) 
    {
       if (!((*lp1++ = FindLLNode(*ip1++)) && (*lp2++ = FindLLNode(*ip2++))))
	  goto vect_err;
       *lp3 = LLNULL;
       if (j = *ip3++)	/* if inc specified, find its llnd */
	  if (! (*lp3 = FindLLNode(j))) goto vect_err;
       lp3++;
    }
    
    lp1 = low; lp2 = hi;  lp3 = inc;
    ll = LLNULL;		/* keep track the first node for return */
    for (i = 0; i < n; i++) 
    {
       ll2 = make_vec_expr(*lp1++, *lp2++, *lp3++);
       ll2 = make_llnd(currentFile, EXPR_LIST, ll2, LLNULL, SMNULL);
       if (ll) 
       {
	  ll1->entry.template.ll_ptr2 = ll2; /* after 1st dim */
	  ll1 = ll2;
       } 
       else
          ll = ll1 = ll2; /* the first dimension */
    }
    return ll->id;

vect_err:
    return -1;
} /* SiVectorExpr */

/****************************************************************
 *								*
 * SiAppendExprList - Will append the expression t to the end   *
 *   of the expression list identified by e. If e = -1 then a   *
 *   new expression list is created.				*
 *								*
 * Input -							*
 *   e - expression list           				*
 *   t - expression to be appended				*
 *								*
 * Output -							*
 *   The id of the newly created expression                     *
 *								*
 ****************************************************************/
int SiAppendExprList(e, t)
int e, t;
{
   PTR_LLND l,nt, lback, ne1, ne2;

   if ((nt = FindLLNode(t)) == NULL) 
      return -1;
   ne2 = make_llnd(currentFile, EXPR_LIST, nt, LLNULL, SMNULL);
   if (e == -1)	
      return ne2->id;
   if ((l = FindLLNode(e)) == NULL) 
      return -1;
   
   lback == NULL;
   while (l->variant == EXPR_LIST && l->entry.template.ll_ptr2)
   {
      lback = l;
      l = l->entry.template.ll_ptr2;
   }
   if (l->variant != EXPR_LIST)
   {
      if (lback)
      {
	 ne1 = make_llnd(currentFile, EXPR_LIST, l, ne2, SMNULL);
	 lback->entry.template.ll_ptr2 = ne1;
	 return e;
      }
      else
      {
	 ne1 = make_llnd(currentFile, EXPR_LIST, l, ne2, SMNULL);
	 return ne1->id;
      }
   }
   else
   {
      l->entry.template.ll_ptr2 = ne2;
      return e;
   }
} /* SiAppendExprList */



/****************************************************************
 *  SiConstExprValue -- get the value of a constant expression  *
 *                                                              *
 *  Input:                                                      *
 *      id -- the exression id (this should be a constant-valued*
 *            expression)                                       *
 *  Output:							*
 *      returns an ELIST of value, coz floats are returned as   *
 *      strings							*
 *                                                              *
 ****************************************************************/
ELIST
SiConstExprValue(id)
int id;
{
    PTR_LLND ll;

    if ((ll = FindLLNode(id)))
       switch (ll->variant) 
       {
	  case INT_VAL:
	  case BOOL_VAL:
	       return ENumber(ll->entry.ival);
	  case CHAR_VAL:
	       return ENumber((int) ll->entry.cval);
	  case FLOAT_VAL:
	  case DOUBLE_VAL:
	       return EString(ll->entry.string_val);
	  case STRING_VAL:	/* emacs needs the quotes */
	       return EString(ll->entry.string_val);
	  default:
	       break;
       } /* switch */
    return NULL;
} /* SiConstExprValue */


/* helper function for SiBaseType */
static int
int_basetype (llnd)
PTR_LLND llnd;
{
   PTR_SYMB sym;
   static int type_2_basetype();

   switch (llnd->variant) 
   {
      case INT_VAL: 
           return T_INT;
           break;
    
      case LABEL_REF:
           return T_UNKNOWN;
           break;
    
      case FLOAT_VAL:
           return T_FLOAT;
           break;
    
      case DOUBLE_VAL:
           return T_DOUBLE;
           break;
    
      case STMT_STR:
      case STRING_VAL:
           return T_STRING;
            break;
    
      case BOOL_VAL:
           return T_BOOL;
           break;
    
      case CHAR_VAL:
           return T_CHAR;
           break;
    
      case CONST_REF:
      case VAR_REF:
      case ENUM_REF:
      case ARRAY_REF:
      case PROC_CALL:
      case FUNC_CALL:
            sym = llnd->entry.template.symbol;
            return type_2_basetype(sym->type);
            break;
    
      case CONSTRUCTOR_REF:
      case ACCESS_REF:
      case CONS:
      case ACCESS:
      case IOACCESS:
           return T_UNKNOWN;
           break;
    
      case EXPR_LIST:
      case EQUI_LIST:
      case COMM_LIST:
      case NAMELIST_LIST:
      case VAR_LIST:
      case RANGE_LIST:
      case CONTROL_LIST:
           /* return the type of the first thing in the list???? */
           return int_basetype(llnd->entry.list.item);
           break;
    
      case DDOT:
      case SEQ:
      case STAR_RANGE:
            return T_SUBRANGE;
            break;
    
      case DEF_CHOICE: /* ???? */
      case SPEC_PAIR:
           return T_UNKNOWN;
           break;
    
      case EQ_OP:
      case LT_OP:
      case GT_OP:
      case NOTEQL_OP:
      case LTEQL_OP:
      case GTEQL_OP:
      case NOT_OP:
           return T_BOOL;
           break;
    
      case ADD_OP:
      case SUBT_OP:
      case OR_OP:
      case MULT_OP:
      case DIV_OP:
      case MOD_OP:
      case AND_OP:
      case EXP_OP:
      case CONCAT_OP:
            /* I assume that both operands will be of the same type,
               so I can just check the left child */
           return int_basetype(llnd->entry.binary_op.l_operand);
           break;
    
      case MINUS_OP:
           return int_basetype(llnd->entry.unary_op.operand);
           break;
    
  } /* switch */
} /* int_basetype */


/* return the basetype of a type (which may be an array type) */
/* another helper for SiBaseType */
static int type_2_basetype(typ)
PTR_TYPE typ;
{
  switch (typ->variant) 
  {
     case T_INT: 
     case T_CHAR:
     case T_FLOAT: 
     case T_DOUBLE: 
     case T_VOID: 
     case T_BOOL:
          return typ->variant;
     default:
          return type_2_basetype(typ->entry.template.base_type);
  } /* switch */
} /* type_2_basetype */

/****************************************************************
 *								*
 *  SiBaseType -- returns base type of an expression node       *
 *                if the exression is an array, this is the     *
 *                base type of the array. if it is a binary     *
 *                expression, the basetype is the basetype      *
 *                of the operands                               *
 *                                                              *
 *  Input:                                                      *
 *      id -- an expression id                                  *
 *                                                              *
 *  Output:                                                     *
 *      type -- an integer type specifier                       *
 *								*
 ****************************************************************/
int SiBaseType(id)
int id; /* an llnd id */
{
  return int_basetype(FindLLNode(id));
}


