#include <strings.h>
#include "db.h"
#include "elist.h"


/*
 * functions used from libdb.a 
 */
extern PTR_SYMB make_label();

/*
 * External references
 */
extern PTR_FILE currentFile;	/* point to the current active file */

/*
 * external functions from elsewhere in the cbaselib
 */
extern PTR_BFND FindCurrBifNode();
extern PTR_LLND FindLLNode();
extern PTR_LABEL FindLabNode();

/****************************************************************
 * SiGetFortranLabelNo - return the integer value which		* 
 *    is the name of the label					*
 *								*
 *  Input:							*
 *	id - label id                                           *
 *                                                              *
 *  Output:							*
 *	integer value of the label                              *
 *								*
 ****************************************************************/
int SiGetFortranLabelNo(id)
int id;
{
   PTR_LABEL l;
   if ((l = FindLabNode(id)) == 0) 
      return -1;
   else
      return l->stateno;
} /* SiGetFortranLabelNo */

/****************************************************************
 * SiMakeUniqueLabel - creates a unique label and returns       * 
 *    the id.                 					*
 *								*
 *  Input:							*
 *                                                              *
 *  Output:							*
 *	id of the newly created label                           *
 *								*
 ****************************************************************/
int SiMakeUniqueLabel()
{
   PTR_LABEL ll;
   long maxname = 1;

   for (ll = currentFile->head_lab; ll; ll = ll->next)
      if (ll->stateno > maxname) 
         maxname = ll->stateno;
    maxname++;
    ll = (PTR_LABEL) make_label(currentFile, maxname);
    return ll->id;
} /* SiMakeUniqueLabel */

/****************************************************************
 * SiReplaceLabelInStmt - modifies the label in statement sid to*
 * be that identified by lid  					*
 *								*
 *  Input:							*
 *     sid - source statement id                                *
 *     lid - target statement id                                *
 *                                                              *
 *  Output:							*
 *	0 for success, -1 for failure                           *
 *								*
 ****************************************************************/
int SiReplaceLabelInStmt(sid,lid)
int sid;
int lid;
{
   PTR_LABEL lab;
   PTR_BFND s;

   if ((s = FindCurrBifNode(sid)) == NULL) 
      return -1;
   if ((lab = FindLabNode(lid)) == NULL) 
      return -1;
   s->label = lab;
   lab->statbody = s;
   return 0;
} /* SiReplaceLabelInStmt */


/****************************************************************
 * SiReplaceLabelInDo   - if sid is the id of a do loop this    *
 * function will replace the label with one identified by lid   *
 *								*
 *  Input:							*
 *     sid - source statement id                                *
 *     lid - target statement id                                *
 *                                                              *
 *  Output:							*
 *	0 for success, -1 for failure                           *
 *								*
 ****************************************************************/
int SiReplaceLabelInDo(sid, lid)
int sid;
int lid;
{
   PTR_LABEL lab;
   PTR_BFND s;

   if ((s = FindCurrBifNode(sid)) == NULL) 
      return -1;
   if ((lab = FindLabNode(lid)) == NULL) 
      return -1;
   s->entry.for_node.doend = lab;
   return 0;
} /* SiReplaceLabelInDo */

/****************************************************************
 * SiReplaceLabelInExpr   - If there is a label in the expr     *
 * llid (such as a goto) this function will replace it with the *
 * label identified by lid.                                     *
 * 								*
 *  Input:							*
 *     llid - source label id                                   *
 *     lid - target id                                          *
 *                                                              *
 *  Output:							*
 *	0 for success, -1 for failure                           *
 *								*
 ****************************************************************/
int SiReplaceLabelInExpr(llid,lid)
int llid;
int lid;

{
   PTR_LABEL lab;
   PTR_LLND s;

   if ((s = FindLLNode(llid)) == NULL) 
      return -1;
   if ((lab = FindLabNode(lid)) == NULL) 
      return -1;
   if (s->variant != LABEL_REF) 
      return -1;
   s->entry.label_list.lab_ptr = lab;
   return 0;
} /* SiReplaceLabelInExpr */


/****************************************************************
 *								*
 * stmt_label - get the label of this statement 		*
 * Input: id of the statement					*
 * Output: id of the label node 				*
 ****************************************************************/

/****************************************************************
 *								*
 *  SiStmtLabel - Get the label of this statement               *
 *								*
 *  Input:							*
 *	id - id of the statement                                *
 *                                                              *
 *  Output:							*
 *	label id of the statement, -1 indicates failure         *
 *								*
 ****************************************************************/
int SiStmtLabel(id)
int id;
{
  PTR_BFND bf;

  if ((bf = FindCurrBifNode(id)) && bf->label)
     return bf->label->id;
  else
     return -1;
} /* SiStmtLabel */

/****************************************************************
 *								*
 *  SiGetDoLabel - Get the label of a DO loop                   *
 *								*
 *  Input:							*
 *	id - id of the statement                                *
 *                                                              *
 *  Output:							*
 *	id of the label node, -1 indicates failure              *
 *								*
 ****************************************************************/

int SiGetDoLabel(id)
int id;
{
  PTR_BFND bf;

  if ((bf = FindCurrBifNode(id))&& bf->entry.for_node.doend)
      return bf->entry.for_node.doend->id;
  else 
     return -1;
} /* SiGetDoLabel */

/****************************************************************
 *								*
 *  SiGetGotoLabel - Get the label id for the goto label        *
 *								*
 *  Input:							*
 *	id - id of the statement                                *
 *                                                              *
 *  Output:							*
 *	id of the label node, -1 indicates failure              *
 *								*
 ****************************************************************/
int SiGetGotoLabel(id)
int id;
{
   PTR_BFND bf;

   if ((bf = FindCurrBifNode(id)) && 
       (bf->variant == GOTO_NODE) &&
       (bf->entry.template.ll_ptr3) &&
       (bf->entry.template.ll_ptr3->variant  == LABEL_REF))
      return bf->entry.template.ll_ptr3->entry.label_list.lab_ptr->id;
   else 
      return -1;
} /* SiGetGotoLabel */


/******************************************************************
 * (static)        						  *
 * find_next_loop_up  - finds the enclosing loop for the statement*
 *    given by bif node s					  *
 *     								  *
 * Input: s - bifnode of the statement                            *
 * Output: bifnode ptr to the enclosing loop. NULL if global or   *
 *   no enclosing loop.						  *
 *								  *
 ******************************************************************/
static PTR_BFND  find_next_loop_up(s)
PTR_BFND s;
{
   while (s!= NULL && s->control_parent != NULL)
   {
      s = s->control_parent;
      if (s->variant == FOR_NODE || s->variant == WHILE_NODE ||
	  s->variant == CDOALL_NODE || s->variant == PROC_HEDR ||
	  s->variant == PROG_HEDR || s->variant == FUNC_HEDR)
         return s;
      if (s->variant == GLOBAL) 
         return NULL;
   }
   return NULL;
} /* find_next_loop_up */

/******************************************************************
 * (static)							  *
 * find_comm_parent   - finds the loop or procedure node that is  *
 *    the common control parent of the two statements given by    *
 *    nodes stnd1 stnd2				                  * 
 *     								  *
 * Input: stnd1, stnd2						  *
 * Output: node pointer to the enclosing common parent            *
 *         NULL if there is no enclosing common parent            *
 *                           					  *
 ******************************************************************/
PTR_BFND  find_comm_parent(stnd1, stnd2)
PTR_BFND stnd1, stnd2;
{
   PTR_BFND up1[20], up2[20];
   int ns1, ns2;

   if (stnd1 == stnd2) 
      return find_next_loop_up(stnd1);
   ns1 = 1; up1[0] = stnd1;
   while (stnd1 = find_next_loop_up(stnd1)) 
      up1[ns1++] = stnd1;
   ns2 = 1; up2[0] = stnd2;
   while (stnd2 = find_next_loop_up(stnd2)) 
      up2[ns2++] = stnd2;
   up1[ns1] = NULL; up2[ns2] = NULL;
   ns1--; ns2--;
   while (up1[ns1] == up2[ns2])
   { 
       ns1--; ns2--;
   }
   return up1[ns1+1];
} /* find_comm_parent */

/******************************************************************
 * SiFindCommonParent - finds the loop or procedure node that is  *
 *    the common control parent of the two statements given by    *
 *    stmt1 and stmt2	 				          * 
 *     								  *
 * Input: stat1, stat2						  *
 * Output: the id of the deepest loop or proc that contains both  *
 *  -1 if no common ancestor.					  *
 ******************************************************************/
int SiFindCommonParent(stat1, stat2)
int stat1, stat2;
{
  PTR_BFND bf1, bf2;
  if ((bf1 = FindCurrBifNode(stat1)) && (bf2 = FindCurrBifNode(stat2)))
  {
     bf1 = find_comm_parent(bf1, bf2);
     if (bf1) 
        return bf1->id;
     else 
        return -1;
  }
  else 
     return -1;
} /* SiFindCommonParent */

      
 /******************************************************************
 * SiFindLabelledStmt - find the statement in the routine          *
 *  with id id, the label lab.		           		   *
 * Input: 							   *
 *     id of statment, id of label				   *
 * Output: 							   *
 *      id of statement.					   *
 * *****************************************************************/

int SiFindLabelledStmt(id, lab)
int id, lab;
{
   PTR_LABEL lp, ll;
   PTR_BFND  b;
   long num;

   if ((lp = FindLabNode(lab)) == NULL) 
      return -1;
   if (lp->statbody != NULL) 
      return lp->statbody->id;
   num = lp->stateno;
   if ((b = FindCurrBifNode(id)) == NULL) 
      return -1;
   for (ll = currentFile->head_lab; ll; ll = ll->next)
      if (ll->stateno == num && ll->statbody != NULL)
	 if (find_comm_parent(b,ll->statbody)) 
            return ll->statbody->id;
    return -1; 
} /* SiFindLabelledStmt */



/****************************************************************
 *  (static)							*
 *  search_lab_ref -                                            *
 *      Find all label references                               * 
 *								*
 ****************************************************************/
static search_lab_ref(lp,lab,el)
PTR_LLND lp;
PTR_LABEL lab;
ELIST *el;
{
  if (lp == NULL) 
     return;
  if (lp->variant == LABEL_REF &&
     (lab == NULL || (lp->entry.label_list.lab_ptr->stateno == lab->stateno)))
  {
     if (*el == NULL) 
        *el = ENumber(lp->id); 
     else 
        *el = EAppend (*el, ENumber(lp->id));
  }
  if(lp->variant != VAR_REF){
     search_lab_ref(lp->entry.template.ll_ptr1,lab,el);
     search_lab_ref(lp->entry.template.ll_ptr2,lab,el);
   }
} /* search_lab_ref */

/****************************************************************
 *								*
 * SiIfLabels  - finds all the labels in a statement that are   *
 * buried in the expressions of an arithmetic IF statement	*
 *								*
 * Input - id of the arithmetic IF statement			*
 *								*
 * Output -							*
 *   ELIST containing labels					*
 *								*
 *								*
 ****************************************************************/
ELIST SiIfLabels(id)
int id;
{
      PTR_BFND bf;
      ELIST el;

   el = NULL;
   if (bf = FindCurrBifNode(id)) 
   {
      search_lab_ref(bf->entry.template.ll_ptr1,NULL, &el);
      search_lab_ref(bf->entry.template.ll_ptr2,NULL, &el);
      return el;
   }
   else 
      return NULL;
} /* SiIfLabels */

/*
 * (static) - find_label_refs used by SiFindLabelRefs
 */
static ELIST find_label_refs(scope_id, label_id)
int scope_id, label_id;
{
  PTR_BFND bf, stmt;
  ELIST srefs, erefs;
  PTR_LLND l;
  PTR_BLOB bl;
  PTR_LABEL lab_ptr;

  srefs = NULL;
  if ((bf = FindCurrBifNode(scope_id)) == NULL) 
     return NULL;
  lab_ptr = FindLabNode(label_id);
  bl = bf->entry.template.bl_ptr1;

  while (bl)
  {
     erefs = NULL;
     stmt = bl->ref;
     search_lab_ref(stmt->entry.template.ll_ptr1,lab_ptr, &erefs);     
     search_lab_ref(stmt->entry.template.ll_ptr2,lab_ptr, &erefs);
     search_lab_ref(stmt->entry.template.ll_ptr3,lab_ptr, &erefs);
     if (erefs != NULL)
     {
	srefs = EAppend(srefs, ECons(ENumber(stmt->id), ECons(erefs, NULL)));
     }
     else if (stmt->label && stmt->label->stateno == lab_ptr->stateno)
          {
	     srefs = EAppend(srefs, ECons(ENumber(stmt->id), NULL));
          }
     if (stmt->entry.template.bl_ptr1 || stmt->entry.template.bl_ptr2)
	srefs = EAppend(srefs, find_label_refs(stmt->id, label_id));
     bl = bl->next;
  } /* while */
  bl = bf->entry.template.bl_ptr2;
  while (bl)
  {
     erefs = NULL;
     stmt = bl->ref;
     search_lab_ref(stmt->entry.template.ll_ptr1,lab_ptr, &erefs);     
     search_lab_ref(stmt->entry.template.ll_ptr2,lab_ptr, &erefs);
     search_lab_ref(stmt->entry.template.ll_ptr3,lab_ptr, &erefs);
     if (erefs != NULL)
     {
	srefs = EAppend(srefs, ECons(ENumber(stmt->id), ECons(erefs, NULL)));
      }
      else if (stmt->label && stmt->label->stateno == lab_ptr->stateno)
           {
	      srefs = EAppend(srefs, ECons(ENumber(stmt->id), NULL));
           }
     if (stmt->entry.template.bl_ptr1 || stmt->entry.template.bl_ptr2)
	srefs = EAppend(srefs, find_label_refs(stmt->id, label_id));
     bl = bl->next;
   } /* while */
   return srefs;
} /* find_label_refs */

/****************************************************************
 *								*
 * SiFindLabelRefs - given a label id and the id of a control   *
 * node find all statements and expressions that reference the  *
 * label 							*
 * It uses a recursive help function find_label_refs which does *
 * all the work 						*
 *								*
 * Input: 							*
 *    scope_id  - id of the control statement			*
 *    label_id  - label id                                      *
 *								*
 *  Output:							*
 *      ((stmt (label_ref label_ref)) (stmt (label_ref ..))..)  *
 *								*
 ****************************************************************/
ELIST SiFindLabelRefs(scope_id, label_id)
int scope_id, label_id;
{
   return find_label_refs(scope_id, label_id);
} /* SiFindLabelRefs */

