/****************************************************************
 *								*
 *   csvr_flow.c:						*
 *								*
 *   This file contains a set of low level interfaces to the	*
 *   kernel of the program data base				*
 *								*
 ****************************************************************/
#include <strings.h>
#include "db.h"
#include "elist.h"


/*
 * functions used from libdb.a 
 */
extern void normal_form();    /* from libdb.a */
extern PTR_SETS build_sets();
extern void reset_scalar_propogation();
extern void gendeps();
extern void relink();
extern void GlobalAnal();
extern PTR_LLND find_bounds();

/*
 * External references
 */
extern PTR_PROJ currentProject;	/* point to the current active project */
extern PTR_FILE currentFile;	/* point to the current active file */


/*
 * external functions from elsewhere in the cbaselib
 */
extern PTR_BFND FindCurrBifNode();
extern PTR_LLND FindLLNode();

/* (static)
 * myvars_mod_in_stmt: help function for  SiDissReachingUses.
 */
/*
static PTR_REFL  myvars_mod_in_stmt(id)
int id;
{
   PTR_BFND bf;
   PTR_REFL defs, used, tmp, tmp2, tmp3;
   PTR_REFL gather_refl(), union_refl();
   PTR_LLND lp1, lp2, lp3;

   defs = used = tmp = tmp2 = tmp3 = (PTR_REFL) NULL;
   if (bf = FindCurrBifNode(id))
   {
      lp1 = bf->entry.template.ll_ptr1;
      lp2 = bf->entry.template.ll_ptr2;
      lp3 = bf->entry.template.ll_ptr3;
      switch (bf->variant)
      {
         case PROC_HEDR:
         case PROG_HEDR:
         case FUNC_HEDR:
              return NULL;
         case ASSIGN_STAT: 
              defs = gather_refl(1, &tmp, bf, lp1);
              if (defs)
              {
                 tmp2 = defs->next;
                 defs->next = (PTR_REFL) NULL;
              }
              used = union_refl(defs,tmp);
              disp_refl(defs); disp_refl(tmp); disp_refl(tmp2);
              defs = gather_refl(1, &tmp, bf, lp2);
              tmp = union_refl(tmp, used);
              disp_refl(defs);
              disp_refl(used);
              used = tmp;
              break;
         case IF_NODE:
         case ELSEIF_NODE:
         case LOGIF_NODE:
              used = gather_refl(1, &tmp, bf, lp1);
              disp_refl(used);
              used = tmp;

              break;
         case PROC_STAT:
              used = gather_refl(1, &defs,bf, (PTR_LLND) 1);
              disp_refl(used);
              used = defs;
              break;
         default:
              defs = gather_refl(1, &tmp, bf, lp1);
              used = gather_refl(1, &tmp2, bf, lp2);
              disp_refl(defs); disp_refl(used);
              used = union_refl(tmp, tmp2);
              disp_refl(tmp); disp_refl(tmp2);
              break;
       } 
       tmp = used;
       return(tmp);
   }
   return NULL;
}
*/
 /* myvars_mod_in_stmt */




/* (static)
 * vars_mod_in_stmt: help function for SiVarsModInStmt, SiDefsToUses
 *                                 and SiUsesReachingDefs.
 */
static ELIST vars_mod_in_stmt(id)
int id;
{
   PTR_BFND bf;
   PTR_REFL defs, used, tmp, tmp2, tmp3;
   PTR_REFL gather_refl(), union_refl();
   ELIST u, t;
   PTR_LLND lp1, lp2, lp3;

   defs = used = tmp = tmp2 = tmp3 = (PTR_REFL) NULL;
   if (bf = FindCurrBifNode(id)) 
   {
      lp1 = bf->entry.template.ll_ptr1;
      lp2 = bf->entry.template.ll_ptr2;
      lp3 = bf->entry.template.ll_ptr3;
      switch (bf->variant)
      {
	 case PROC_HEDR:
	 case PROG_HEDR:
	 case FUNC_HEDR:
	      return NULL;
	 case ASSIGN_STAT: /* must be fortran */
	      defs = gather_refl(1, &tmp, bf, lp1);
	      if (defs)
              { 
                 tmp2 = defs->next;
		 defs->next = (PTR_REFL) NULL;
	      }
	      used = union_refl(defs,tmp);
	      disp_refl(defs); disp_refl(tmp); disp_refl(tmp2);
	      defs = gather_refl(1, &tmp, bf, lp2);
	      tmp = union_refl(tmp, used);
	      disp_refl(defs);
	      disp_refl(used);
	      used = tmp;
	      break;
	 case IF_NODE:
	 case ELSEIF_NODE:
	 case LOGIF_NODE:
	      used = gather_refl(1, &tmp, bf, lp1);
	      disp_refl(used);
	      used = tmp;
	      break;
	 case PROC_STAT: 
	      used = gather_refl(1, &defs,bf, (PTR_LLND) 1);
	      disp_refl(used);
	      used = defs;
	      break;
	 default:
	      defs = gather_refl(1, &tmp, bf, lp1);
	      used = gather_refl(1, &tmp2, bf, lp2);
	      disp_refl(defs); disp_refl(used);
	      used = union_refl(tmp, tmp2);
	      disp_refl(tmp); disp_refl(tmp2);
	      break;
       } /* switch */
       u = NULL; 
       tmp = used;
       while (used)
       {
	  u = EAppend(u,ENumber(used->node->refer->id));
	  used = used->next;
       }
       disp_refl(tmp);
       return u;
   }
   return NULL;
} /* vars_mod_in_stmt */

/****************************************************************
 *								*
 *  SiVarsModInStmt - returns a list of VAR_REF and ARRAY_REF   *
 *    expression nodeids in the statement that represents defns *
 *    of variables                                              *
 *  It uses the help function vars_mod_in_stmt to perform all   *
 *  its task							*
 *								*
 *  Input:							*
 *	id - statement id                                       *
 *                                                              *
 *  Output:							*
 *	ELIST of list of expression nodeids                     *
 *								*
 ****************************************************************/
ELIST SiVarsModInStmt(id)
int id;
{
   return vars_mod_in_stmt(id);
} /* SiVarsModInStmt */



/* (static)
 * vars_used_in_stmt: help function for SiVarsUsedInStmt
 */
static ELIST vars_used_in_stmt(id)
int id;
{
   PTR_BFND bf;
   PTR_REFL defs, used, tmp, tmp2, tmp3;
   PTR_REFL gather_refl(), union_refl();
   ELIST u, t;
   PTR_LLND lp1, lp2, lp3;

   defs = used = tmp = tmp2 = tmp3  = (PTR_REFL) NULL;
   if ((bf = FindCurrBifNode(id))) 
   {
      lp1 = bf->entry.template.ll_ptr1;
      lp2 = bf->entry.template.ll_ptr2;
      lp3 = bf->entry.template.ll_ptr3;
      switch (bf->variant)
      {
	 case PROC_HEDR:
	 case PROG_HEDR:
	 case FUNC_HEDR:
	      return NULL;
	 case ASSIGN_STAT: /* must be fortran */
	      defs = gather_refl(1, &tmp, bf, lp1);
	      if (defs) 
                 used = defs->next;
	      defs = gather_refl(1, &tmp, bf, lp2);
	      tmp = union_refl(defs, used);
	      disp_refl(defs);
	      disp_refl(used);
	      used = tmp;
	      break;
	 case PROC_STAT: 
	      used = gather_refl(1, &defs,bf, (PTR_LLND) 1);
	      break;
	 default:
	      defs = gather_refl(1, &tmp, bf, lp1);
	      used = gather_refl(1, &tmp, bf, lp2);
	      if (bf->variant == FOR_NODE)
              {
		 tmp2 = gather_refl(1, &tmp, bf, lp3);
		 tmp3 = union_refl(used,tmp2);
		 disp_refl(used, tmp2);
		 used = tmp3;
	      }
	      tmp = union_refl(defs, used);
	      disp_refl(defs);
	      disp_refl(used);
	      used = tmp;
	      break;
      } /* switch */
      u = NULL; tmp = used;
      while (used)
      {
	 u = EAppend(u,ENumber(used->node->refer->id));
	 used = used->next;
      }
      disp_refl(tmp);
      return u;
   }
   return NULL;
} /* vars_used_in_stmt */

/* (static)
 * myvars_used_in_stmt: help function for SiDissReachingUses
 */
/*
static PTR_REFL myvars_used_in_stmt(id)
int id;
{
   PTR_BFND bf;
   PTR_REFL defs, used, tmp, tmp2, tmp3;
   PTR_REFL gather_refl(), union_refl();
   PTR_LLND lp1, lp2, lp3;

   defs = used = tmp = tmp2 = tmp3  = (PTR_REFL) NULL;
   if ((bf = FindCurrBifNode(id)))
   {
      lp1 = bf->entry.template.ll_ptr1;
      lp2 = bf->entry.template.ll_ptr2;
      lp3 = bf->entry.template.ll_ptr3;
      switch (bf->variant)
      {
         case PROC_HEDR:
         case PROG_HEDR:
         case FUNC_HEDR:
              return NULL;
         case ASSIGN_STAT: 
              defs = gather_refl(1, &tmp, bf, lp1);
              if (defs)
                 used = defs->next;
              defs = gather_refl(1, &tmp, bf, lp2);
              tmp = union_refl(defs, used);
              disp_refl(defs);
              disp_refl(used);
              used = tmp;
              break;
         case PROC_STAT:
              used = gather_refl(1, &defs,bf, (PTR_LLND) 1);
              break;
         default:
              defs = gather_refl(1, &tmp, bf, lp1);
              used = gather_refl(1, &tmp, bf, lp2);
              if (bf->variant == FOR_NODE)
              {
                 tmp2 = gather_refl(1, &tmp, bf, lp3);
                 tmp3 = union_refl(used,tmp2);

                 disp_refl(used, tmp2);
                 used = tmp3;
              }
              tmp = union_refl(defs, used);
              disp_refl(defs);
              disp_refl(used);
              used = tmp;
              break;
      } 
      tmp = used;
      return (tmp);
   }
   return NULL;
}
*/
 /* myvars_used_in_stmt */


/****************************************************************
 *								*
 *  SiVarsUsedInStmt - returns a list of VAR_REF and ARRAY_REF  *
 *    expression nodeids in the statement that represents uses  *
 *    variables                                                 *
 *								*
 *  Input:							*
 *	id - statement id                                       *
 *                                                              *
 *  Output:							*
 *	ELIST of list of references                             *
 *								*
 ****************************************************************/
ELIST SiVarsUsedInStmt(id)
int id;
{
   return vars_used_in_stmt(id);
} /* SiVarsUsedInStmt */

                 
/****************************************************************
 * (static)							*
 * make_mref_pairs -                                            *
 *	used by SiWhereDefined					*
 *  Input:							*
 *	                                                        *
 *                                                              *
 *  Output:							*
 *	                                                        *
 *								*
 ****************************************************************/
static ELIST make_mref_pairs(stmt, s)
int stmt;
PTR_SYMB s;
{
   ELIST result, e, t;
   int r;
   PTR_LLND ll;
   e = vars_mod_in_stmt(stmt);
   result = NULL;
   while(e){
      r = (int) e->car; e = e->cdr; 
      ll = FindLLNode(r);
      if(ll && (ll->entry.template.symbol == s)){
	    t = EAppend(ENumber(stmt), ENumber(r));
	    result = EAppend(result, ECons(t,NULL));
	    }
      }
    return result;
}
/* static where_defined: help function for SiWhereDefined */
static ELIST where_defined(var)
int var;
{
   PTR_SYMB s, sym;
   PTR_BLOB b;
   ELIST e,t;
   int stmt;

   e = NULL;
   if (currentFile && currentFile->global_bfnd && 
      (sym = FindSymbolNode(var)))
   {
      for (s = currentFile->head_symb; s; s = s->thread) 
	 if ((sym == s) && (s->ud_chain))
            for (b = s->ud_chain; b; b = b->next)
            {
	       stmt = b->ref->id;
	       e = EAppend(e,make_mref_pairs(stmt,s));
	    } 
      return e;
   }
   else 
      return NULL;
} /* where_defined */

/****************************************************************
 *								*
 *  SiWhereDefined - looks for all places in the current file   *
 *    where a variable  is defined.    				*
 *    Note that symbols that are in common areas are treates as *
 *    different symbols in each routine and symbols passed as   *
 *    parameters are also different across procedure boundaries *
 *								*
 *  Input:							*
 *	var - symbol table id                                   *
 *       ...                                                    *
 *  Output:							*
 *	list of (stmt term) pairs                               *
 *								*
 ****************************************************************/
ELIST SiWhereDefined(var)
int var;
{
   return where_defined(var);
} /* SiWhereDefined */

/****************************************************************
 * (static)							*
 * make_muse_pairs -                                            *
 *	used by SiWhereUsed					*
 *  Input:							*
 *	                                                        *
 *                                                              *
 *  Output:							*
 *	                                                        *
 *								*
 ****************************************************************/
static ELIST make_muse_pairs(stmt, s)
int stmt;
PTR_SYMB s;
{
   ELIST result, e, t;
   int r;
   PTR_LLND ll;
   e = vars_used_in_stmt(stmt);
   result = NULL;
   while (e)
   {
      r = (int) e->car; e = e->cdr; 
      ll = FindLLNode(r);
      if (ll && (ll->entry.template.symbol == s))
      {
	 t = EAppend(ENumber(stmt), ENumber(r));
	 result = EAppend(result, ECons(t,NULL));
      }
   }
   return result;
} /* make_muse_pairs */
      
/* static where_used: help function for SiWhereUsed, SiDefsToUses */
static
ELIST where_used(var)
int var;
{
   PTR_SYMB s, sym;
   PTR_BLOB b;
   ELIST e,t;
   int stmt;
   e = NULL;
   if (currentFile && currentFile->global_bfnd && (sym = FindSymbolNode(var)))
   {
      for (s = currentFile->head_symb; s; s = s->thread) 
	 if ((sym == s) && (s->ud_chain))
	    for (b = s->ud_chain; b; b = b->next)
            {
	       stmt = b->ref->id;
	       e = EAppend(e,make_muse_pairs(stmt,s));
	    } 
      return e;
   }
   else 
      return NULL;
} /* where_used */

/****************************************************************
 *								*
 *  SiWhereUsed    - looks for all places in the current file   *
 *    where the given symbol table value for var is used.       *
 *    Note that symbols that are in common areas are treated as *
 *    different symbols in each routine and symbols passed as   *
 *    parameters are also different accross procedure boundaries*
 *								*
 *  Input:							*
 *	var - symbol table id                                   *
 *       ...                                                    *
 *  Output:							*
 *	list of (stmt term) pairs                               *
 *								*
 ****************************************************************/
ELIST SiWhereUsed(var)
int var;
{
   return where_used(var);
} /* SiWhereUsed */

/*
 static reaching_defs:
       help function for SiReachingDefs, SiDefsReachingUses
			 SiDefsToUses.
*/
static ELIST reaching_defs(stmt)
int stmt;
{
   PTR_BFND bf;
   PTR_REFL r;
   ELIST e, t;
   if ((bf = FindCurrBifNode(stmt))) 
   {
      r = bf->entry.template.sets->in_def;
      e = NULL;
      while (r)
      {
	 t = EAppend(ENumber(r->node->stmt->id),
		     ENumber(r->node->refer->id));
	 e = EAppend(e, ECons(t,NULL));
	 r = r->next;
      }
      return e;
    }
   else 
      return NULL;
} /* reaching_defs */ 

/****************************************************************
 *								*
 *  SiReachingDefs  - returns a list of ELIST of pairs (stmt,   *
 *     ref) that are the definitions reaching a given stmt.     *
 *  Input:							*
 *	stmt - id of the statement                              *
 *                                                              *
 *  Output:							*
 *	ELIST of reaching definitions                           *
 *      NULL indicates no definiton reaches                     *
 *								*
 ****************************************************************/
ELIST  SiReachingDefs(stmt)
int stmt;
{
   return reaching_defs(stmt);
} /* SiReachingDefs */

/*
 static reaching_aligns:
       help function for SiReachingAligns.
*/
static ELIST reaching_aligns(stmt)
int stmt;
{
   PTR_BFND bf;
   PTR_REFL r;
   ELIST e, t;
   if ((bf = FindCurrBifNode(stmt)))
   {
      r = bf->entry.template.sets->in_al;
      e = NULL;
      while (r)
      {
         t = EAppend(ENumber(r->node->stmt->id),
                     ENumber(r->node->refer->id));
         e = EAppend(e, ECons(t,NULL));
         r = r->next;
      }
      return e;
    }
   else
      return NULL;
} /* reaching_aligns*/

/****************************************************************
 *                                                              *
 *  SiReachingAligns  - returns a list of ELIST of pairs (stmt, *
 *     ref) that are the aligns reaching a given stmt.          *
 *  Input:                                                      *
 *      stmt - id of the statement                              *
 *                                                              *
 *  Output:                                                     *
 *      ELIST of reaching   ailgns                              *
 *      NULL indicates no aligns reaches                        *
 *                                                              *
 ****************************************************************/
ELIST  SiReachingAligns(stmt)
int stmt;
{
   return reaching_aligns(stmt);
} /* SiReachingAligs*/

/****************************************************************
 *                                                              *
 *  SiAlignsReachingUses -                                      *
 *    This function, given a statement id, returns an ELIST     *
 *    of the following form. Each element of the list is an     *
 *    ELIST, with car of the ELIST being ref-id of the variable *
 *    which occurs in the statement at USE position. The cdr of *
 *    the ELIST is a ELIST of ELISTs, Where each of the ELISTs  *
 *    is a (Stmt, Ref) pair ie. the statement at which the      *
 *    variable was aligned(Stmt) and the align point(Ref)  *
 *    in that statement.                                        *
 *                                                              *
 * Input -                                                      *
 *   stmt_id - id of thst statement                             *
 *                                                              *
 * Output -                                                     *
 *   ELIST of the above description                             *
 *                                                              *
 ****************************************************************/
ELIST SiAlsReachingUses(stmt_id)
int stmt_id;
{
   ELIST tmp,varlist; /* variables used in the statement */
   ELIST deflist; /* reaching definition ELIST */
   ELIST rdlist; /* my reaching definition list to be returned */
   ELIST vlst, dlst, dvlst;
   int ref, vid, rid;
   varlist=NULL;
   varlist = vars_used_in_stmt(stmt_id);
   deflist = reaching_aligns(stmt_id);
   rdlist = NULL;

   for (vlst = varlist; vlst; vlst = ECdr(vlst))
   {
      vid = VarId(ECar(vlst));
      dvlst = NULL;
      for (dlst = deflist; dlst; dlst = ECdr(dlst))
      {
         ref = (int)ECadar(dlst);
         rid = VarId(ref);
         if (vid == rid)
         {
            dvlst = EAppend(dvlst, ECons(ECar(dlst), NULL));
         }
      }

      if (dvlst) /* that means there has been some reaching stuff */
      {
         ELIST tn;
         tn = ENew(0);
         tn->car = ECar(vlst);
         rdlist = EAppend(rdlist,
                        ECons(EAppend(tn, ECons(dvlst, NULL)), NULL)
);
      }
   }
   EFree(varlist);
   return rdlist;
} /* SiAlsReachingUses */


ELIST SiAlsReachingMods(stmt_id)
int stmt_id;
{
   ELIST tmp,varlist; /* variables used in the statement */
   ELIST deflist; /* reaching definition ELIST */
   ELIST rdlist; /* my reaching definition list to be returned */
   ELIST vlst, dlst, dvlst;
   int ref, vid, rid;
   varlist=NULL;
   varlist = vars_mod_in_stmt(stmt_id); 
   deflist = reaching_aligns(stmt_id);
   rdlist = NULL;

   for (vlst = varlist; vlst; vlst = ECdr(vlst))
   {
      vid = VarId(ECar(vlst));
      dvlst = NULL;
      for (dlst = deflist; dlst; dlst = ECdr(dlst))
      {
         ref = (int)ECadar(dlst);
         rid = VarId(ref);
         if (vid == rid)
         {
            dvlst = EAppend(dvlst, ECons(ECar(dlst), NULL));
         }
      }

      if (dvlst) /* that means there has been some reaching stuff */

      {
         ELIST tn;
         tn = ENew(0);
         tn->car = ECar(vlst);
         rdlist = EAppend(rdlist,
                        ECons(EAppend(tn, ECons(dvlst, NULL)), NULL)
);
      }
   }
   EFree(varlist);
   return rdlist;
} /* SiAlsReachingUses */




/*
 static reaching_distributes:
       help function for SiReachingDistributes.
*/
static ELIST reaching_distributes(stmt)
int stmt;
{
   PTR_BFND bf;
   PTR_REFL r;
   ELIST e, t;
   if ((bf = FindCurrBifNode(stmt)))
   {
      r = bf->entry.template.sets->in_dis;
      e = NULL;
      while (r)
      {
         t = EAppend(ENumber(r->node->stmt->id),
                     ENumber(r->node->refer->id));
         e = EAppend(e, ECons(t,NULL));
         r = r->next;
      }
      return e;
    }
   else
      return NULL;
} /* reaching_distributes*/

/****************************************************************
 *                                                              *
 *  SiReachingDistributes  - returns a list of ELIST of pairs   *
 * (stmt,ref) that are the aligns reaching a given stmt.        *
 *  Input:                                                      *
 *      stmt - id of the statement                              *
 *                                                              *
 *  Output:                                                     *
 *      ELIST of reaching   Distributes                         *
 *      NULL indicates no Distributes reaches                   *
 *                                                              *
 ****************************************************************/
ELIST  SiReachingDistributes(stmt)
int stmt;
{
   return reaching_distributes(stmt);
} /* SiReachingDistributes*/


/* 
   static reaching_uses: 
   help function for SiReachingUses, SiUsesReachingDefs
*/

static ELIST reaching_uses(stmt)
int stmt;
{
   PTR_BFND bf;
   PTR_REFL r;
   ELIST e, t;
   if (bf = FindCurrBifNode(stmt)) 
   {
      r = bf->entry.template.sets->in_use;
      e = NULL;
      while (r)
      {
	 t = EAppend(ENumber(r->node->stmt->id),
		     ENumber(r->node->refer->id));
	 e = EAppend(e, ECons(t,NULL));
	 r = r->next;
      }
      return e;
    }
   else 
      return NULL;
} /* SiReachingUses */ 
/****************************************************************
 *								*
 *  SiReachingUses  - returns a list of ELIST of pairs (stmt,   *
 *     ref) that are the uses reaching a given stmt.            *
 *  Input:							*
 *	stmt - id of the statement                              *
 *                                                              *
 *  Output:							*
 *	ELIST of reaching uses                                  *
 *      NULL indicates no use  reaches                          *
 *								*
 ****************************************************************/
ELIST SiReachingUses(stmt)
int stmt;
{
   return reaching_uses(stmt);
} /* SiReachingUses */

/****************************************************************
 *								*
 *  SiCrossReference - Prints the list of cross references      * 
 *  (uses and definitions) for the current file                 *
 *								*
 *  Input:							*
 *                                                              *
 *  Output:							*
 *								*
 ****************************************************************/
void SiCrossReference()
{
  dump_udchain(currentFile);
} /* SiCrossReference */





/****************************************************************
 *                                                              *
 *  SiDissReachingUses-                                        *
 *    This function, given a statement id, returns an ELIST     *
 *    of the following form. Each element of the list is an     *
 *    ELIST, with car of the ELIST being ref-id of the variable *
 *    which occurs in the statement at USE position. The cdr of *
 *    the ELIST is a ELIST of ELISTs, Where each of the ELISTs  *
 *    is a (Stmt, Ref) pair ie. the statement at which the      *
 *    variable was distribute(Stmt) and the distribute point(Ref)  *
 *    in that statement.                                        *
 *                                                              *
 * Input -                                                      *
 *   stmt_id - id of thst statement                             *
 *                                                              *
 * Output -                                                     *
 *   ELIST of the above description                             *
 *                                                              *
 ****************************************************************/
/*
ELIST SiDissReachingUses(stmt_id)
int stmt_id;
{
   ELIST varlist; 
   ELIST dislist; 
   ELIST allist;
   ELIST rdlist; 
   ELIST vlst, dlst, dvlst;
   int ref, vid, rid;
     
   PTR_BFND bf;
   PTR_REFL inal, indis;
   ELIST e, t;
   if ((bf = FindCurrBifNode(stmt_id)))
   {
     inal = bf->entry.template.sets->in_al;
     indis =  bf->entry.template.sets->in_dis;
     e = NULL;
     while (r)
         {
           
   varlist = vars_used_in_stmt(stmt_id);
   dislist = reaching_distributes(stmt_id);
   allist = reaching_aligns(stmt_id);

   rdlist = NULL;

   for (vlst = varlist; vlst; vlst = ECdr(vlst))
   {
      vid = VarId(ECar(vlst));
      dvlst = NULL;
      for (dlst = dislist; dlst; dlst = ECdr(dlst))
      {
         ref = (int)ECadar(dlst);
         rid = VarId(ref);
         if (vid == rid)
         {
            dvlst = EAppend(dvlst, ECons(ECar(dlst), NULL));
         }
      }
      if (dvlst) 
      {
         ELIST tn;
         tn = ENew(0);
         tn->car = ECar(vlst);
         rdlist = EAppend(rdlist,
                        ECons(EAppend(tn, ECons(dvlst, NULL)), NULL));
      }
   }
   EFree(varlist);
   return rdlist;
} 

*/

/*
PTR_REFL mymerge(refl1,refl2)
PTR_REFL refl1,refl2;
{
  PTR_REFL t;
  t=refl1;
  if ( t=NULL )
   {
    return ( refl2);
   }
  while (t->next)
  {
   t=t->next;
  }
  t->next=refl2;
  return (refl1); 
}


static PTR_REFL *mydissreachinguses(stmt)
int stmt;

{
  PTR_REFL al, dis, use, ddis, aal;
  PTR_REFL uuse,*uad,uuad,mod,*t;
  PTR_LLND llp1, llp2=NULL;
  PTR_BFND bf,bf1;
  PTR_SYMB s;
  PTR_SETS sets;
   
  uad=NULL;
  t=uad;
  if ((bf=FindCurrBifNode(stmt)))
   {
     sets=bf->entry.template.sets;
     al=sets->in_al;
     dis=sets->in_dis;
     use= myvars_used_in_stmt(stmt);
     mod= myvars_mod_in_stmt(stmt);
     use= mymerge(use,mod);
     
     for (uuse=use; uuse; uuse=uuse->next)
      {
       *uad=uuse;
       (*uad)->next=NULL;
       uuad=*uad;
       for (aal=al;aal;aal=aal->next)
         {
           if (uuse->id==aal->id)
     	     {
              bf1=aal->node->stmt;
 	      llp2=bf1->entry.template.ll_ptr2;
              llp1=llp2->entry.template.ll_ptr1;		
              s=llp1->entry.template.symbol;
              for(ddis=dis;ddis;ddis=ddis->next)
                {
  		  if(s==ddis->id)
     		    {
		      uuad->next=ddis;
  		      uuad=uuad->next;
 		    }
  	      	}  				 	
              }
 	 }
         uuad=NULL;
      }
      uad++;
     }
   uad=NULL; 
   return(t);          
 }  
*/

/****************************************************************
 *								*
 *  SiUsesReachingDefs - This function, given a statement id,   *
 *    returns an ELIST of the following form.  Each element of  *
 *    the list is an ELIST, with the car of the ELIST being     *
 *    ref-id of the variable which occurs in the statement in   *
 *    the DEFINE position.  The cdr of the ELIST is an ELIST of *
 *    ELISTs, where each of the ELISTs is a (Stmt, Ref) pair ie.*
 *    the statement at which the variable is used(Stmt) and the *
 *    use point(Ref) in that statement.				*
 *								*
 * Input -							*
 *   stmt_id - id of the statement				*
 *								*
 * Output -							*
 *   ELIST of the above description.
 *								*
 ****************************************************************/
ELIST SiUsesReachingDefs(stmt_id)
int stmt_id;
{
   ELIST varlist; /* variables defined in the statement */
   ELIST uselist; /* reaching use list */ 
   ELIST rulist; /* my reaching use list to be returned */
   ELIST vlst, ulst, uvlst;
   int ref, vid, rid;
   
   varlist = vars_mod_in_stmt(stmt_id); 
   uselist = reaching_uses(stmt_id);
   rulist = NULL;

   for (vlst = varlist; vlst; vlst = ECdr(vlst))
   {
      vid = VarId(ECar(vlst));
      uvlst = NULL;
      for (ulst = uselist; ulst; ulst = ECdr(ulst))
      {
         ref = (int)ECadar(ulst);
         rid = VarId(ref);
         if (vid == rid)
         {
            uvlst = EAppend(uvlst, ECons(ECar(ulst), NULL));
         }
      }
      if (uvlst) /* that means there has been some reaching stuff */
      {
         ELIST tn;
         tn = ENew(0);
         tn->car = ECar(vlst);
         rulist = EAppend(rulist,
                        ECons(EAppend(tn, ECons(uvlst, NULL)), NULL));
      }
   }
   EFree(varlist);
   return rulist;
} /* SiUsesReachingDefs */


/****************************************************************
 *								*
 *  SiDefsReachingUses - 					*
 *    This function, given a statement id, returns an ELIST     *
 *    of the following form. Each element of the list is an     *
 *    ELIST, with car of the ELIST being ref-id of the variable *
 *    which occurs in the statement at USE position. The cdr of *
 *    the ELIST is a ELIST of ELISTs, Where each of the ELISTs  *
 *    is a (Stmt, Ref) pair ie. the statement at which the      *
 *    variable was defined(Stmt) and the definition point(Ref)  *
 *    in that statement.					*
 *								*
 * Input -							*
 *   stmt_id - id of thst statement				*
 *								*
 * Output -							*
 *   ELIST of the above description				*
 *								*
 ****************************************************************/
ELIST SiDefsReachingUses(stmt_id)
int stmt_id;
{
   ELIST varlist; /* variables used in the statement */
   ELIST deflist; /* reaching definition ELIST */
   ELIST rdlist; /* my reaching definition list to be returned */
   ELIST vlst, dlst, dvlst;
   int ref, vid, rid;

   varlist = vars_used_in_stmt(stmt_id);
   deflist = reaching_defs(stmt_id);
   rdlist = NULL;
 
   for (vlst = varlist; vlst; vlst = ECdr(vlst))
   {
      vid = VarId(ECar(vlst));
      dvlst = NULL;
      for (dlst = deflist; dlst; dlst = ECdr(dlst))
      {
         ref = (int)ECadar(dlst);
         rid = VarId(ref); 
         if (vid == rid)
         {
            dvlst = EAppend(dvlst, ECons(ECar(dlst), NULL));
         }
      }
      if (dvlst) /* that means there has been some reaching stuff */
      {
         ELIST tn;   
         tn = ENew(0);
         tn->car = ECar(vlst);
         rdlist = EAppend(rdlist, 
			ECons(EAppend(tn, ECons(dvlst, NULL)), NULL));
      }
   }
   EFree(varlist);
   return rdlist;
} /* SiDefsReachingUses */

         
/* 
    static is_term_in_stmt_term_pair : help function for SiDefsToUses 
 */
static int is_term_in_stmt_term_pair(pair,var)
ELIST pair;
int var;
{
  return var == (int)ECadr(pair);  
} /* is_term_in_stmt_term_pair */

/****************************************************************
 *								*
 *  SiDefsToUses -- 						*
 *     Given a statement id, it returns an ELIST where each     *
 *     element of the ELIST corresponds to each variable        *
 *     modified in the statement and is an ELIST itself whose   *
 *     CAR consists of the VAR id which is defined and whose	*
 *     CDR consists of an ELIST of (stmt term) pairs where      *
 *     "stmt" stands for the id of the statement which uses this*
 *     definition, and "term" stands for the point of use in    *
 *     that statement and variable name				*
 *								*
 * Input -							*
 *     stmt_id - id of the statement				*
 *								*
 * Output -							*
 *   ELIST of the above description				*
 *								*
 ****************************************************************/
ELIST SiDefsToUses(stmt_id)
int stmt_id;
{
   ELIST modlist, m;
   ELIST whereUsed, w;
   ELIST all_du_list, du_list;
   ELIST rd_list, rd;
   ELIST st_pair;
   ELIST tn1, tn2, tn;
   int stmt, term;
   int var, varid;

   modlist = vars_mod_in_stmt(stmt_id); 
   /* for each of the variables modified in a statement, find
      all the (statement, ref) pairs where the definition is
      used */
   all_du_list = NULL;
   for (m = modlist; m; m = ECdr(m))
   {
      var = (int)ECar(m);
      varid = VarId(var);
      whereUsed = where_used(varid);
      du_list = NULL;
      for (w = whereUsed; w; w = ECdr(w))
      {
         stmt = (int)ECaar(w);
         term = (int)ECadar(w);
         /* check if the definition var reaches this point */
         rd_list = reaching_defs(stmt);
         for (rd = rd_list; rd; rd = ECdr(rd))
         {
            st_pair = (ELIST)ECar(rd);
            if (is_term_in_stmt_term_pair(st_pair, var)) 
            {
               /* the defn has reached this use */
               tn1 = ENew(0); tn2 = ENew(0);
               tn1->car = (char *)stmt; tn2->car = (char *)term;
               du_list = EAppend(du_list,
        			   ECons(EAppend(tn1, tn2), NULL));
            } /* if */
         } /* for rd */
         EFree(rd);
      } /* for w */
      EFree(whereUsed);
      /* append the info found for the variable to the global list */
      tn = ENew(0);
      tn->car = (char *)var;
      all_du_list = EAppend(all_du_list,
                               ECons(EAppend(tn,
                                         ECons(du_list, NULL)), NULL));
   } /* for m */
   EFree(modlist);
   return all_du_list;
} /* SiDefsToUses */
/****************************************************************
 *								*
 *  SiDoGlobalAnalysis - Do the global flow analysis only.      *
 *    This will propogate information about variable modific.   *
 *    and use around the call tree so that at call sites we will*
 *    have all summary information about the way the call uses  *
 *    the data structures.					*
 *  Input:							*
 *	none - will work for the current project                *
 *                                                              *
 *  Output:							*
 *	0 if ok -1 if failure                                   *
 *								*
 ****************************************************************/
int SiDoGlobalAnalysis()
{
   if (currentProject)
   {
      GlobalAnal(currentProject);
      return 0;
   }
   else 
      return 1;
} /* SiDoGlobalAnalysis */



/****************************************************************
 *								*
 * SiDoFlowAnalysis - generate reaching definitions and reaching*
 * uses for the current file.            		   	*
 *	       NOTE: assumes that GlobalAnal() has been done.	*
 * Input:							*
 *	none							*
 *	(always use the current active file in current project) *
 *								*
 * Output:							*
 *	status number 						*
 *								*
 ****************************************************************/
int SiDoFlowAnalysis()
{
   PTR_BFND p;

   if (currentProject && currentFile) 
   {
      /* GlobalAnal(currentProject); */
      p = currentFile->global_bfnd;
      reset_scalar_propogation(p);
      (void) build_sets(1, p, (PTR_REFL)NULL, (PTR_REFL)NULL, (PTR_REFL)NULL, (PTR_REFL)NULL, 1);
      (void) build_sets(1, p, (PTR_REFL)NULL, (PTR_REFL)NULL, (PTR_REFL)NULL, (PTR_REFL)NULL, 2);
       relink(currentFile);
       return 0;
    } 
    else
       return -1;
} /* SiDoFlowAnalysis */


/****************************************************************
 *								*
 *   SiGetCallInfo - returns information about the structures	*
 *     that are used and modified by this call to a function or	*
 *     or procedure.						*
 *								*
 *   Input:  the id of the call statement			*
 *								*
 *   Output: a list ( used mod ) where used and mod are the ids *
 *    of expression nodes which are the lists of used and	*
 *    and modified versions of the formal arguments.		*
 *								*
 ****************************************************************/
ELIST SiGetCallInfo(id)
int id;
{
    PTR_BFND bf;
    ELIST e1, e2;
    PTR_LLND used, modified;

    used = NULL; 
    modified = NULL;
    if (bf = FindCurrBifNode(id)) 
    {
       bind_call_site_info(bf,&used, &modified);
       if (used != NULL) 
	  e1 = ENumber(used->id);
       else 
          e1 = NULL;
       if (modified != NULL) 
	  e2 = ENumber(modified->id);
       else 
          e2 = NULL;
       return EAppend(e1,e2);
    } 
    else
       return NULL;
} /* SiGetCallInfo */


/*
 * for loops
 */

/*
 * returns list of symbol table id's for induction variables.
 */
/****************************************************************
 *								*
 *  SiInductionVars - returns a list of symbol table ids' for   *
 *		induction variables   				*
 *   Currently only returns one induction variable id for       *
 *   FORTRAN loops and it doesnot work for C programs. To be    *
 *   fixed							*
 *								*
 *  Input -  							*
 *      loop_id - id of the loop				*
 ****************************************************************/
ELIST SiInductionVars(loop_id)
int loop_id;
{
   PTR_BFND bf;
   ELIST e;

   e = NULL;
   if (bf = FindCurrBifNode(loop_id))
      return EAppend(e,ENumber(bf->entry.for_node.control_var->id));
   else
      return NULL;
} /* SiInductionVars */


/****************************************************************
 *								*
 *  SiLoopBounds -- get the loop bounds				*
 *		return a list of pairs				*
 *								*
 *  Input -  							*
 *      loop_id - id of the loop				*
 *  Output -							*
 *	(low_bound  hi_bound stride)				*
 *	where low-bound, hi-bound and stride are id's for	*
 *	low level nodes.  The stride could be empty if none.	*
 *								*
 ****************************************************************/
ELIST SiLoopBounds(loop_id)
int loop_id;
{
   PTR_BFND bf;
   ELIST e;

   e = NULL;
   if (bf = FindCurrBifNode(loop_id)) 
   {
      if (bf->entry.for_node.range)
	 e = EAppend(
		ENumber(bf->entry.for_node.range->entry.binary_op.l_operand->id),
		ENumber(bf->entry.for_node.range->entry.binary_op.r_operand->id));
      else
	 e = EAppend(ENumber(-1),ENumber(-1));
      if (bf->entry.for_node.increment)
	 e = EAppend(e,ENumber( bf->entry.for_node.increment->id));
      return e;
   }
   return NULL;
} /* SiLoopBounds */


/****************************************************************
 *								*
 *  SiArraysUsed -- get a list of low level expr ids describing	*
 *		  use of arrays in loops			*
 *	NOT DONE YET...						*
 * Input -							*
 *								* 
 * Output -							*
 *								*  
 *								*
 ****************************************************************/
ELIST SiArraysUsed (loop_id)
int loop_id;
{
  return NULL;
} /* SiArraysUsed */


/****************************************************************
 *								*
 *  SiArraysModified -- get a list of low level expr ids        *
 *      describing modifications of arrays in loops		*
 *	NOT DONE YET...						*
 * Input -							*
 *								* 
 * Output -							*
 *								*  
 *								*
 ****************************************************************/
ELIST SiArraysModified (loop_id)
int loop_id;
{
  return NULL;
} /* SiArraysModified */

/****************************************************************
 *								*
 * SiModInRoutine-- returns list of variables modified by this	*
 *      function /subroutine call.				*
 *      stmt_id is the id of a PROC_HEDR FUNC_HEDR this returns *
 *      the summary of the variables external to the routine    *
 *      that are modified by a call. The result is expressed in *
 *      terms of formal parameter and common block names.       *
 *      NOTE: this assumes that the parser was called without   *
 *        the -nodep flag. Also assumes that do_global_analysis *
 *        function has been called otherwise it will only give  *
 *        the information about structures modified in the body *
 *        of the function or subroutine.			*
 *								*
 * Input:							*
 *	stmt_id - the bif node id of the statement interested	*
 *								*
 * Output:							*
 *								*
 ****************************************************************/
int SiModInRoutine(stmt_id)
int stmt_id;
{
   PTR_BFND bf;

   if (bf = FindCurrBifNode(stmt_id)) 
   {
      PTR_LLND lp;

      switch (bf->variant) 
      {
	case PROC_HEDR:
	case FUNC_HEDR:
	case PROC_STAT:
	     lp = bf->entry.template.ll_ptr2;
	     if ((lp != NULL) && (lp->entry.template.ll_ptr2 != NULL))
		return lp->entry.template.ll_ptr2->id;
	     else
		return -1; /* no dependence analysis yet */
	default:
	     break;
      }
   }
   return -1;
} /* SiModInRoutine */


/****************************************************************
 *								*
 * SiUsedInRoutine-- returns list of variables used by this	*
 *		  function call.				*
 *								*
 * Input:							*
 *	stmt_id - the bif node id of the statement interested	*
 *								*
 * Output:							*
 *								*
 ****************************************************************/
int SiUsedInRoutine(stmt_id)
int stmt_id;
{
   PTR_BFND bf;

   if (bf = FindCurrBifNode(stmt_id)) 
   {
      PTR_LLND lp;

      switch (bf->variant) 
      {
	 case PROC_HEDR:
	 case FUNC_HEDR:
	 case PROC_STAT:
	      lp = bf->entry.template.ll_ptr3;
	      if ((lp != NULL) && (lp->entry.template.ll_ptr2 != NULL))
	         return lp->entry.template.ll_ptr2->id;
	      else
		 return -1; /* no dependence analysis yet */
	 default:
	      break;
	}/* switch */
    }
    return -1;
} /* SiUsedInRoutine */


int
lift_deps(id)
    int id;
{
    PTR_BFND bf;

    if ((bf = FindCurrBifNode(id))) {
	LiftDepInfo(bf->entry.template.bl_ptr1);
	return 0;
    }
    return -1;
}
/****************************************************************
*  induct_list -- returns the list of induction variables that	*
* control the execution of the given statement within the given *
* procedure. It also returns the loop bounds for those variables*
* experssed as subscript lists ( loop bounds can be functions	*
* of other induction variables).				*/
/*** not yet written ***/

/****************************************************************
 *          							*
 * SiGetSubscript - 						*
 *  A subscript list for an array reference takes the following	*
 *  form.  suppose we have a nested sequence of loops over ind.	*
 *  variables i1, i2, ... ik.  and an array reference		*
 *  x(h1(i1..ik),h2(i1..ik),...hn(i1..ik))			*
 *  The subscript list is a list of lists, one for each h expr	*
 *  of the form							*
 *     ( decidable offset  coef-list)				*
 *  where							*
 *    decidable = 1 means that h was linear in i1...ik plus	*
 *		      or minus a loop invariant.		*
 *	if this is the case, offset is the loop invariant and	*
 *	coef-list is the list of coefficients of i1 .. ik.	*
 *	the entries in this list and the value of offset is	*
 *	either an integer or a list.  If it is an integer	*
 *	then that is the exact coef. or offset constant.	*
 *	if it is a list ( n ) then n is the id of the expression*
 *	that describes this loop invariant coef. or const.	*
 *    decidable = 0 means that h was not analizable		*
 *    decidable = 2 means that h was a range value like 1..n.	*
 *      if this is the case, then the coef-list is empty and	*
 *      offset is set to the expr id				*
 *      of the range expression. 				*
 *								*
 *  Input -							*
 *     stmt_id - id of the loop statement			*
 *     expr - subscripted exprssion				*
 *  Output -							*
 ****************************************************************/
ELIST SiGetSubscript(stmt_id, expr)
int stmt_id, expr;
{
   struct ref sor;
   struct subscript source[AR_DIM_MAX];
   PTR_SYMB induct_list[MAX_NEST_DEPTH];
   struct subscript lo_bnd[MAX_NEST_DEPTH];
   struct subscript hi_bnd[MAX_NEST_DEPTH];
   int dim, i,j, depth;
   PTR_BFND b;
   PTR_LLND a;
   ELIST e, subsc, coefs;

   if ((b = FindCurrBifNode(stmt_id)) == NULL) 
      return NULL;
   if ((a = FindLLNode(expr)) == NULL) 
      return NULL;
   sor.stmt = b;
   sor.refer = a;
   make_subscr(&sor,source);
   for (i = 0; i < MAX_NEST_DEPTH; i++) 
      induct_list[i] = NULL;
   depth = make_induct_list(b,induct_list,lo_bnd,hi_bnd);
   /* print_subscr(&sor,source,induct_list); */
   e = NULL; 
   i = 0;
   while (source[i].decidable != -1)
   {
      subsc = NULL;
      subsc = EAppend(subsc,ENumber(source[i].decidable));
      /* linear subscript case */
      if (source[i].decidable == 1)
      {
         if (source[i].parm_exp != NULL)
	    subsc = EAppend(subsc, ECons(ENumber(source[i].parm_exp->id),NULL));
	 else
	    subsc = EAppend(subsc, ENumber(source[i].offset));
	 coefs = NULL;
	 for (j = 0; j < depth; j++)
	    if (source[i].coefs_symb[j] != NULL)
	       coefs = EAppend(coefs, ECons(ENumber(source[i].coefs_symb[j]->id),NULL));
	    else
	       coefs = EAppend(coefs, ENumber(source[i].coefs[j]));
	 subsc = EAppend(subsc,ECons(coefs,NULL));
	 e = EAppend(e,ECons(subsc,NULL));
     }
    /* vector range subscript case */
     else if (source[i].decidable == 2)
          {
	     subsc = EAppend(subsc, ENumber(source[i].vector->id));
	     subsc = EAppend(subsc, ECons(NULL,NULL));
	     e = EAppend(e,ECons(subsc,NULL));
	  }
     /* give up case */
     else
     {
        subsc = EAppend(subsc, ENumber(0));
        subsc = EAppend(subsc, ECons(NULL,NULL));
        e = EAppend(e,ECons(subsc,NULL));
     }
     i++;
   } /* while */
   return e;
} /* SiGetSubscript */

/***************************************************************
 *                                                             *
 * SiRegionAccessed -                                          *
 *                                                             *
 ***************************************************************/
int SiRegionAccessed(stmt_id, array_ref_id)
int stmt_id, array_ref_id;
{
   PTR_LLND arr, reslt;
   PTR_BFND stmt;

   if ((stmt = FindCurrBifNode(stmt_id)) == NULL) 
      return -1;
   if ((arr = FindLLNode(array_ref_id)) == NULL) 
      return -1;

   reslt = find_bounds(stmt, arr, NULL);
   normal_form(&reslt);
   if (reslt == NULL) 
      return -1;
   else 
      return reslt->id;
} /* SiRegionAccessed */



/****************************************************************
 *								*
 * SiGenerateDeps -- generate deps. for function containing the	*
 *		    statement with id "id"			*
 *								*
 * Input:							*
 *	id - identifies the function or the subroutine          *
 *								*
 * Output:							*
 *	status number, 0 for success       			*
 *								*
 ****************************************************************/
int SiGenerateDeps (id)
int id;
{
   PTR_BFND bf;

   if (bf = FindCurrBifNode(id)) 
   {
      while (bf->variant !=  PROC_HEDR && bf->variant !=  FUNC_HEDR &&
	      bf->variant !=  PROG_HEDR && bf->variant !=  GLOBAL)
	 bf = bf->control_parent;
      if (bf->variant != GLOBAL) 
      {
	 /*
	 reset_scalar_propogation(bf);
	 (void) build_sets(1,bf,(PTR_REFL)NULL,(PTR_REFL)NULL,1);
	 (void) build_sets(1,bf,(PTR_REFL)NULL,(PTR_REFL)NULL,2);
	 */
	 gendeps(bf);
	 /* relink(currentFile); */
	 return 0;
      }
  }
  return 0;
} /* SiGenerateDeps */

      
static char *depstrs[] = {"flow","anti","output","huh??","got me?" };
static char *dirstrs[] = { "  ", "= ", "- ", "0-", "+ ", "0+", ". ", "+-"};

static PTR_BFND current_par_loop = NULL;

static int
same_loop(from, to)
	PTR_BFND from, to;
{
    PTR_BFND c;
    c = from;
    while(c != NULL && c->variant != GLOBAL && c->variant != PROG_HEDR
	  && c->variant != PROC_HEDR && c->variant != FUNC_HEDR
	  && c != current_par_loop)
	c = c->control_parent;
    if(c != current_par_loop) return(0);
    c = to;
    while(c != NULL && c->variant != GLOBAL  && c->variant != PROG_HEDR
	  && c->variant != PROC_HEDR && c->variant != FUNC_HEDR
	  && c != current_par_loop)
	c = c->control_parent;
    if(c != current_par_loop) return(0);
    return(1);
}
static
ELIST
esearch_deps(nb,q,depth)
    ELIST	nb;
    PTR_BLOB	q;
    int 	depth;
{
    PTR_BFND  bchild;
    PTR_DEP   d;
    ELIST lb,e,t;
    int i, outer;
	
    while(q != NULL){
	bchild = q->ref;
	q = q->next;
	d = bchild->entry.template.dep_ptr1;
	e = NULL;
	while(d != NULL){
	    for(i = 1, outer = 0; i < depth; i++) if (d->direct[i] > 1) outer = 1;
	    if(outer == 0) 
              if((d->symbol->type->variant == T_ARRAY && d->direct[depth] > 1) ||
                ((d->type == 0 || d->type == 3) && d->direct[depth] > 1) )
		  if(same_loop(d->from.stmt,d->to.stmt)){
		    e = ENumber(d->symbol->id); /* name of symbol */
		    e = EAppend(e, ENumber(d->type)); /* type of dependence  */
		    t = EAppend(ENumber(d->from.stmt->id), ENumber(d->from.refer->id));
		    e = EAppend(e, ECons(t,NULL)); /* from_id, reference_id pair */
		    t = EAppend(ENumber(d->to.stmt->id), ENumber(d->to.refer->id));
		    e = EAppend(e, ECons(t,NULL)); /* from_id, reference_id pair */
		    t = NULL;
		    for(i = 1; i < 6; i++) t = EAppend(t,ENumber(d->direct[i]));
		    e = EAppend(e, ECons(t,NULL));
		    nb = EAppend(nb,ECons(e,NULL));
		 }
	    d = d->from_fwd;
	}
	if(bchild->entry.template.bl_ptr1 != NULL){
	    nb = esearch_deps(nb, bchild->entry.template.bl_ptr1,depth);
	}
	if(bchild->entry.template.bl_ptr2 != NULL){
	    nb = esearch_deps(nb, bchild->entry.template.bl_ptr2,depth);
	}
    }
    return(nb);
}

/* osearch_deps looks for carried output deps indicating temporaries */
/* of smaller dimension than the loop.  These are candidates for     */
/* expansion or localization */

static
ELIST
osearch_deps(nb,q,depth)
    ELIST	nb;
    PTR_BLOB	q;
    int 	depth;
{
    PTR_BFND  bchild;
    PTR_DEP   d;
    ELIST lb,e,t;
    int i, outer;
	
    while(q != NULL){
	bchild = q->ref;
	q = q->next;
	d = bchild->entry.template.dep_ptr1;
	e = NULL;
	while(d != NULL){
	    for(i = 1, outer = 0; i < depth; i++) if (d->direct[i] > 1) outer = 1;
	    if((outer == 0) && (d->type == 2) && (d->direct[depth] > 1)) 
		  if(same_loop(d->from.stmt,d->to.stmt)){
		    e = ENumber(d->symbol->id); /* name of symbol */
		    e = EAppend(e, ENumber(d->type)); /* type of dependence  */
		    t = EAppend(ENumber(d->from.stmt->id), ENumber(d->from.refer->id));
		    e = EAppend(e, ECons(t,NULL)); /* from_id, reference_id pair */
		    t = EAppend(ENumber(d->to.stmt->id), ENumber(d->to.refer->id));
		    e = EAppend(e, ECons(t,NULL)); /* from_id, reference_id pair */
		    t = NULL;
		    for(i = 1; i < 6; i++) t = EAppend(t,ENumber(d->direct[i]));
		    e = EAppend(e, ECons(t,NULL));
		    nb = EAppend(nb,ECons(e,NULL));
		 }
	    d = d->from_fwd;
	}
	if(bchild->entry.template.bl_ptr1 != NULL){
	    nb = osearch_deps(nb, bchild->entry.template.bl_ptr1,depth);
	}
	if(bchild->entry.template.bl_ptr2 != NULL){
	    nb = osearch_deps(nb, bchild->entry.template.bl_ptr2,depth);
	}
    }
    return(nb);
}

/******************************************************************
 *								  *
 *  SiGetCarriedDeps -                                            *
 *                       return the ELIST of dependences assoc.   *
 * with a statement or a loop.  If stmt is a loop, the function	  *
 * returns the list of loop carried deps only.  Otherwise you	  *
 * get only the dependences that originate at the given statement *
 * the dependences are encoded as follows:			  *
 *( symbol_id  type (from_stmt from_ref) (to_stmt to_ref)(vector))*
 * where symbol_id is the id of the symbol involved.		  *
 * type is 0="flow", 1="anti", 2="output", 3="huh??", 4="got me?" *
 * (from_stmt from_ref) is the source of the dependence. 	  *
 * (to_stmt to_ref) is the destination of the dependenc. 	  *
 * vector is a list (d1, d2, d3, d4, d5) where each d is a	  *
 * direction  1="=", 2="-", 3="0-", 4="+", 5="0+", 6=".", 7="+-"  *
 * see also get_dep_info and other functions			  *
 * Assumes that generate_deps() has been called for the contain-  *
 * ing function or subroutine.					  *
 *								  *
 * Input -							  *
 *    id of the statement					  *
 *								  *
 * Output -							  *
 *    ELIST of dependences explained above...			  *
 *								  *
 *******************************************************************/
ELIST SiGetCarriedDeps(stmt)
int stmt;
{
    PTR_BFND b,bpar;
    PTR_DEP d;
    int depth,i;
    PTR_BLOB q;
    ELIST nb, e, t;

    if ((b = FindCurrBifNode(stmt))== NULL) 
       return NULL;
    if (b == NULL) 
       return(NULL);
    /* if b is a loop, we look for all loop carried deps for */
    /* this loop.  otherwise just list dependence going out */
    if (b->variant == FOR_NODE)
    {
      depth = 0;
      bpar = b;
      current_par_loop = b;
      while (bpar != NULL && bpar->variant != GLOBAL)
      {
	 if (bpar->variant == FOR_NODE ||
	     bpar->variant == CDOALL_NODE ||
	     bpar->variant == WHILE_NODE ||
	     bpar->variant == FORALL_NODE) 
            depth++;
	 bpar = bpar->control_parent;
      }	 
      q = b->entry.template.bl_ptr1;
      nb = NULL;
      nb = esearch_deps(nb,q,depth);
      return nb;
   }				/* if loop case */
   d = b->entry.template.dep_ptr1;
   nb = NULL;
   while (d != NULL)
   {
      e = ENumber(d->symbol->id); /* name of symbol */
      e = EAppend(e, ENumber(d->type)); /* type of dependence  */
      t = EAppend(ENumber(d->from.stmt->id), ENumber(d->from.refer->id));
      e = EAppend(e, ECons(t,NULL)); /* from_id, reference_id pair */
      t = EAppend(ENumber(d->to.stmt->id), ENumber(d->to.refer->id));
      e = EAppend(e, ECons(t,NULL)); /* from_id, reference_id pair */
      t = NULL;
      for (i = 1; i < 6; i++) 
         t = EAppend(t,ENumber(d->direct[i]));
      e = EAppend(e, ECons(t,NULL));
      nb = EAppend(nb,ECons(e,NULL));
      d = d->from_fwd;
   }
   return nb;
} /* SiGetCarriedDeps */

/****************************************************************
 * SiGetOutputCarriedDeps -					*
 *    If stmt is a loop then look for loop carried deps else    *
 * just look for outgoing dependences                           *
 *								*
 *								*
 *								*
 *								*
 *								*
 * Input -							*
 *								* 
 *								*
 *								*
 *								*
 * Output -							*
 *								*
 *								*
 ****************************************************************/
ELIST SiGetOutputCarriedDeps(stmt)
int stmt;
{
   PTR_BFND b,bpar;
   PTR_DEP d;
   int depth,i;
   PTR_BLOB q;
   ELIST nb, e, t;

    if ((b = FindCurrBifNode(stmt))== NULL) 
       return NULL;
    /* if b is a loop, we look for all loop carried deps for */
    /* this loop.  otherwise just list dependence going out */
    if (b->variant == FOR_NODE)
    {
       depth = 0;
       bpar = b;
       current_par_loop = b;
       while (bpar != NULL && bpar->variant != GLOBAL)
       {
	  if (bpar->variant == FOR_NODE ||
	      bpar->variant == CDOALL_NODE ||
	      bpar->variant == WHILE_NODE ||
	      bpar->variant == FORALL_NODE) 
            depth++;
	  bpar = bpar->control_parent;
      }	 
      q = b->entry.template.bl_ptr1;
      nb = NULL;
      nb = osearch_deps(nb,q,depth);
      return nb;
   }				/* if loop case */
   d = b->entry.template.dep_ptr1;
   nb = NULL;
   while (d != NULL)
   {
      e = ENumber(d->symbol->id); /* name of symbol */
      e = EAppend(e, ENumber(d->type)); /* type of dependence  */
      t = EAppend(ENumber(d->from.stmt->id), ENumber(d->from.refer->id));
      e = EAppend(e, ECons(t,NULL)); /* from_id, reference_id pair */
      t = EAppend(ENumber(d->to.stmt->id), ENumber(d->to.refer->id));
      e = EAppend(e, ECons(t,NULL)); /* from_id, reference_id pair */
      t = NULL;
      for (i = 1; i < 6; i++) t = EAppend(t,ENumber(d->direct[i]));
      e = EAppend(e, ECons(t,NULL));
      nb = EAppend(nb,ECons(e,NULL));
      d = d->from_fwd;
  }
  return nb;
} /* SiGetOutputCarriedDeps */
