
#include <strings.h>
#include "db.h"
#include "elist.h"

/*
 * External references
 */
extern PTR_PROJ currentProject;
extern PTR_FILE currentFile;
PTR_BFND *SiStmtTable[20];
int SiStmtTable_pages = 0;

static void LoadStmtTable(file)
PTR_FILE file;
{
	int i, count, page, item;
	PTR_BFND b;
	PTR_BFND *t;
	for(page = 0; page < SiStmtTable_pages; page++)
	   for(item = 0; item < 1024; item++) 
		SiStmtTable[page][item] = (PTR_BFND) NULL;
	page = item = 0;
	for( b = file->head_bfnd; b; b = b->thread){
	    count = b->id;
            if(count >= SiStmtTable_pages*1024){
	       SiStmtTable[SiStmtTable_pages++]= t= (PTR_BFND *) malloc(1024*sizeof(PTR_BFND));
	       for(i = 0; i < 1024; i++) t[i] = (PTR_BFND) NULL;
	       }
	    page = count/1024;
	    item = count - page*1024;
	    SiStmtTable[page][item] = b;
	    }
}

void SiAddToStmtTable(b)
PTR_BFND b;
{
	int i, count, page, item;
	PTR_BFND * t;
	count = b->id;
        if(count >= SiStmtTable_pages*1024){
	    SiStmtTable[SiStmtTable_pages++] = t = (PTR_BFND *) malloc(1024*sizeof(PTR_BFND));
	    for(i = 0; i < 1024; i++) t[i] = (PTR_BFND) NULL;
	    }
	page = count/1024;
	item = count - page*1024;
	SiStmtTable[page][item] = b;
}

PTR_BFND SiFindInStmtTable(id)
int id;
{
	int page, item;
	page = id/1024;
	item = id - page*1024;
	if(page >= SiStmtTable_pages) return NULL;
	return SiStmtTable[page][item];
}

/* the following routines are used by parse_string
*/
static char *bufPtr;

static char get_buf_char()
{
  char c;

  while (*bufPtr == ' ' || *bufPtr == '\n' || *bufPtr == ' ') bufPtr++;
  return *bufPtr; 
} /* get_buf_char */

static int get_buf_int()
{
   int i,sign;

   i = 0;
   while (*bufPtr == ' ' || *bufPtr == '\n' || *bufPtr == ' ') 
      bufPtr++; 
   if (*bufPtr == '-')
   { 
      sign = -1; 
      bufPtr++;
   } 
   else 
      sign = 1;
   while (*bufPtr >= '0' && *bufPtr <= '9')
   { 
      i = 10*i+(*bufPtr - '0');
      bufPtr++;
   }
   return sign*i;
} /* get_buf_int */

static char *get_buf_string()
{
   char *p;
   char *q, *lp;

   while (*bufPtr == ' ' || *bufPtr == '\n' || *bufPtr == ' ') 
      bufPtr++;
   if (*bufPtr == '"') /* this is a quoted string */
   { 
      bufPtr++;
      p = bufPtr; 
      lp = p;
      while (*p != '"' && *p != '\0') 
      {
         lp = p; 
         p++;
      }
      q = Allocate(p-bufPtr+1);
      p = q; 
      lp = p;
      while (*bufPtr != '"' && *bufPtr != '\0')
      {
	 lp = p;
	 *p= *bufPtr;
	 p++; 
         bufPtr++;
      }
      *p = '\0';
      bufPtr++;
      return q;
   }
   else
   {
      p = bufPtr; 
      lp = p;
      while ((*lp != '\n' || *p != '.') && *p != '\0')
      {
         lp = p; 
         p++;
      }
      q = Allocate(p-bufPtr+1);
      p = q; 
      lp = p;
      while ((*lp != '\n' || *bufPtr != '.') && *bufPtr != '\0')
      {
	 lp = p;
	 *p= *bufPtr;
	 p++; 
         bufPtr++;
      }
      *p = '\0';
      if (*bufPtr == '.')
         bufPtr++;
      return q;
  }
} /* char_buf_string */

/* (static)
 * parse_out_buf: help function for SiParseString
 */
static ELIST parse_out_buf()
{
   int i;
   char c;
   ELIST e;

   c = get_buf_char();
   e = ENew(0);
   e->cdr = NULL;
   switch (c)
   {
      case '0': 
      case '1': 
      case '2': 
      case '3': 
      case '4': 
      case '5':
      case '6': 
      case '7': 
      case '8': 
      case '9': 
      case '-':
           e->type = 0;
           e->car = (char *) get_buf_int();
           break;
      case '(':
           e->type = 2;
	   bufPtr++;
           e->car = (char *) parse_out_buf();
           break;
      case '\0':
	   return(e);
           break;
      default:
           e->type = 1;
           e->car = get_buf_string();
           break;
    } /* switch */
    while (*bufPtr == ' ' || *bufPtr == '\n' || *bufPtr == ' ') 
       bufPtr++;
    if (*bufPtr == ')')
    { 
       bufPtr++;  
       return e;
    }
    if (*bufPtr != '\0' ) 
       e->cdr = parse_out_buf(); 
    return e;
} /* parse_out_buf */
/****************************************************************
 *								*
 * SiParseString   -						*
 *								*
 *								*
 *								*
 *								*
 *								*
 *								*
 * Input -							*
 *								* 
 *								*
 *								*
 *								*
 * Output -							*
 *								*
 *								*
 ****************************************************************/
ELIST SiParseString(s)
char *s;
{
   int i;

   bufPtr = s;
   if ((i =get_buf_int()) < 0) 
      return NULL;
   else 
      return parse_out_buf();
} /* SiParseString */


/* (static)
 *  unparse_elist: help function for SiUnparseEList 
 */
static unparse_elist(e)
ELIST e;
{
   if (e == NULL) 
      return;
   switch (e->type)
   {
      case 0: 
           printf(" %d ",(int) e->car);
	   break;
      case 1: 
           printf(" \"%s\" ",(char *) e->car);
	   break;
      case 2: 
           printf("   ( ");
	   unparse_elist(e->car);
	   printf("   )\n ");
	   break;
   } /* switch */
   unparse_elist(e->cdr);
} /* unparse_elist */

void SiUnparseEList(e)
ELIST e;
{
   printf("(");
   unparse_elist(e);
   printf(")\n");
} /* SiUnparseEList */


/* (static)
 * sunparse_elist: help function for SiSUnparseEList
 */
static void sunparse_elist(s,e)
char *s;
ELIST e;
{
   char *p;
   char q[10000];
   int i;
   *s = '\0';
   while (e != NULL)
   {
      switch (e->type)
      {
         case 0: 
              sprintf(s," %d ",(int) e->car);
	      break;
         case 1: 
              sprintf(s," %s",(char *) e->car);
	      break;
	 case 2: 
              sprintf(s,"   ( ");
	      while (*s != '\0') 
                 s++;
	      sunparse_elist(q,e->car);
	      i = 0;
	      while (q[i] != '\0') 
                 *s++ = q[i++];
	      sprintf(s,"	) ");
	      break;
      }
      e = e->cdr;
      while (*s != '\0') 
         s++;
  }
} /* sunparse_elist */

void SiSUnparseEList(s,e)
char *s;
ELIST e;
{
   sunparse_elist(s, e);
} /* SiSUnparseEList */

/****************************************************************
 *								*
 *   SiOpenProject -- handles the open project primitives	*
 *								*
 *   Inputs:							*
 *	Expects two strings:            			*
 *	    proj_name -- the project name 			*
 *	    proj_file -- file name that contains the name of 	*
 *		       file that constitutes the project	*
 *   Output:							*
 *      0 indicates success					*
 *      negative values indicate failure			*
 *								*
 ****************************************************************/
int SiOpenProject(proj_name, proj_file)
char * proj_name;
char * proj_file;
{
   char *mem[500]; 	/* for file in the project */
   int   no = 0;	/* number of file in the project */
   int   c;
   FILE *fd;		/* file descriptor for project */
   char **p, *t;
   char *tmp, tmpa[20000];

   tmp = &(tmpa[0]);
   if ((fd = fopen(proj_file, "r")) == NULL)
   {
      printf("SiOpenProject: cannot open project file %s\n", proj_file);
      return -1;
   }

   p = mem;
   t = tmp;

   while ((c = getc(fd)) != EOF)
   {
     if (c != ' ')	/* assum no blanks in filename */
     {
	if (c == '\n') 
        {
	   if (t != tmp) 
           {
	      /* not a blank line */
	      *t = '\0';
	      *p = Allocate((unsigned) (strlen(tmp) + 1));
	      strcpy(*p++, tmp);
	      t = tmp;
	   }
	} else	*t++ = c;
      }
   } /* while */

   fclose(fd);
   no = p - mem;
   if (no > 0) /* Now make it the active project */
   {
      if ((currentProject = OpenProj(proj_name, no, mem))) 
      {
	 currentFile = (PTR_FILE)currentProject->file_chain->ref;
	 LoadStmtTable(currentFile);
	  return 0;
      } 
      else
      {
         printf("SiOpenProject: Cannot open project\n");
	 return -2;
      }
   }
   else
   {
      printf("SiOpenProject: No files in the project\n");
      return -3;
   }
 	 
} /* SiOpenProject */


/****************************************************************
 *								*
 *  SiCloseProject -- handles the close project primitives	*
 *								*
 *  Input: ???~                                                 *
 *								*
 *  Output:							*
 *								*
 ****************************************************************/
int SiCloseProject()
{
    return 1;
} /* SiCloseProject */


/****************************************************************
 *								*
 *  SiAddToProject -- handles the primitive that adds new       *
 *		   file(s) to a given project			*
 *								*
 *  Inputs:							*
 *	    str -- a string that specifies the project name as	*
 *		   in open_proj 				*
 *	    count -- an integer that specifies the number of	*
 *		   files to be added				*
 *	    (str-1, ... , str-n) -- n strings that		*
 *		   specify the list of files to be added to	*
 *		   the project					*
 *								*
 *  Output:							*
 *	an integer status was written to stdout with a minus	*
 *	number indicating that there was something wrong	*
 *								*
 ****************************************************************/
int SiAddToProject(proj, count, e)
char *proj;
int count;
ELIST e;
{
    int   no, okay = 1, m = 0;

    if (SelectProj(proj) == NULL)
       okay = -1;	/* so that -2 means "project not found" */
			/* still need to flush the rest of inputs */
    if (e == NULL) 
       return -1;
    no = count;

    for (; count > 0; count--, m++) 
    {
	if (okay > 0)
	   okay = AddToProj(e->car); /* okay is 0 if something wrong */
	if (e->cdr == NULL) 
           break;
	else 
           e = e->cdr;
    } /* for */
    if (m < no) 
       return m-count;
    else if (okay == 0) return -1;
	    else return 0; 
} /* SiAddToProject */


/****************************************************************
 *								*
 * SiDeleteFromProject -- handles the primitive that deletes    *
 *		    file(s) to a given project			*
 *								*
 *  Inputs:							*
 *	   str  -- a string that specifies the project name as	*
 *		   in open_proj 				*
 *	   count	-- an integer that specifies the number *
 *		    of files to be deleted			*
 *	   (str-1, ... , str-n) -- followed by n strings that	*
 *		   specify the list of files to be deleted from *
 *		   the project					*
 *								*
 *  Output:							*
 *	status was written to stdout with a minus number	*
 *	indicating that there was something wrong		*
 *								*
 ****************************************************************/
int SiDeleteFromProject(proj, count, e)
char *proj;
int count;
ELIST e;
{
    int   no, m = 0, okay = 1;

    if (SelectProj(proj) == NULL)
	okay = -1;	/* so that -2 means "project not found" */
    no = count;
    if (e == NULL) 
       return -1;
    for (; count > 0; count--, m++) 
    {
       if (okay)
	  okay = DelFromProj(e->cdr);
	if (e->cdr == NULL) 
           break;
    }
    if (m - no < 0) 
       return m-no;
    else 
    if (okay == 0) 
       return -1;
    else 
       return 0;
} /* SiDeleteFromProject */


/****************************************************************
 *								*
 *  SiSelectProject -- handles the primitive that select a      *
 *       project to be the current active project		*
 *								*
 *  Input:							*
 *	   s: name of the project                               *
 *								*
 *  Output:							*
 *        0 indicates success and -1 indicates failure.		*
 *								*
 ****************************************************************/
int SiSelectProject(s)
char *s;
{
    return SelectProj(s)? 0: -1;
} /* SiSelectProject */


/****************************************************************
 *								*
 *  SiSelectFile -- handles the primitive that select a file to	*
 *		   be the current active file			*
 *								*
 *  Input:							*
 *	   s: file name                                         *
 *								*
 *  Output:							*
 *         0 indicates success 					*
 *	   -1 indicates that there was something wrong		*
 *								*
 ****************************************************************/
int SiSelectFile(s)
char *s;
{
    PTR_BLOB b;

    for (b = currentProject->file_chain; b; b = b->next)
       if (!strcmp(((PTR_FILE)b->ref)->filename, s)) 
       {
	  currentFile = (PTR_FILE) b->ref;
	  LoadStmtTable(currentFile);
	  return 0;
       }
    return -1;
} /* SiSelectFile */


/* (static)
 * get_project_info:  help function for SiGetProjectInfo, SiUnresolvedRef
 */
static ELIST get_project_info(projname,info)
char *projname;
int info;
{
    PTR_BLOB1 b;
    ELIST e;

    b = GetProjInfo(projname, info);
    e = NULL;
    if (b) 
    {
	for (; b; b = b->next) 
        {
	    e = EAppend(e,EString( ((PTR_INFO)b->ref)->source));
	}
	return e;
    }
    else
       return NULL;
} /* get_project_info */
#if FALSE

/****************************************************************
 *								*
 *  SiGetProjectInfo -- real routine to handle the inquiry to   *
 *		     the project-related stuff			*
 *  Uses get_project_info as a help function to do all the work *
 *								*
 *  Input:							*
 *         projname -- name of the project                      *
 *	   info -- an integer indicating the info inquired	*
 *								*
 *  Output:							*
 *         NULL if something is wrong.                          *
 *         Otherwise list of strings with the project info.     *
 *         (this comment should be clearer - Neel)              *
 *								*
 ****************************************************************/
ELIST SiGetProjectInfo(projname, info)
char *projname;
int info;
{
   return get_project_info(projname, info);
} /* SiGetProjectInfo */
#endif



/****************************************************************
 *								*
 *  SiGetFileList -- handles the primitive that enquires the	*
 *		     list of files that constitutes a given	*
 *		     project					*
 *								*
 *  Input:							*
 *	   projname: the name of the project                    *
 *								*
 *  Output:							*
 *	   same as SiGetProjectInfo				*
 *								*
 ****************************************************************/
ELIST SiGetFileList(projname)
char *projname;
{
    PTR_BLOB1 b;
    ELIST e;

    b = GetProjInfo(projname, ProjSrc);
    e = NULL;
    if (b) 
    {
       for (; b; b = b->next) 
       {
	  e = EAppend(e, EString(((PTR_INFO)b->ref)->filename));
       }
       return e;
    }
    else
       return NULL;
} /* SiGetFileList */


/****************************************************************
 *								*
 *  SiGetProcList -- handles the primitive that enquires the	*
 *		     list of procedures/functions that were	*
 *		     declared in a given project		*
 *								*
 *  Input:							*
 *	   projname: name of the project                        *
 *								*
 *  Output:							*
 *	   same as SiGetProjectInfo				*
 *								*
 ****************************************************************/
ELIST SiGetProcList(projname)
char * projname;
{
    PTR_BLOB1 b;
    PTR_INFO  info;
    char     *cp;
    ELIST e;

    b = GetProjInfo(projname, ProjNames);
    if (b) 
    {
       e = NULL;
       for (; b; b = b->next) 
       {
	  info = (PTR_INFO) b->ref;
	  cp = info->source;
	  cp[strlen(cp) - 1] = '\0'; /* don't want the '\n' at the end */
	  e = EAppend(e, ECons(EAppend( EAppend(ENumber(info->g_line), 
			EString(info->filename)), EString(cp)), NULL));
       }
       return e;
    }
    else
       return NULL;
} /* SiGetProcList */


/****************************************************************
 *								*
 *  SiGetProcIdList -- handles the primitive that enquires	*
 *			 the list of bif node ids of		*
 *			 procedures/functions that were 	*
 *			 declared in a given project		*
 *								*
 *  Input:							*
 *        projname: project name                                *
 *								*
 *  Output:							*
 *	   a list of the form (id1 id2 ...) 			*
 *								*
 ****************************************************************/
ELIST SiGetProcIdList(projname)
char *projname;
{
    PTR_BLOB b, bl;
    PTR_FILE f;
    ELIST e;

    e = NULL;
    for (b = currentProject->file_chain; b; b = b->next) 
    {
	f = (PTR_FILE) b->ref;
	if (f) 
        {
	   for (bl = f->global_bfnd->entry.template.bl_ptr1; bl; bl = bl->next)
	     if (f->lang == ForSrc ||
		 f->lang == CSrc && bl->ref->variant == FUNC_HEDR)
	        e = EAppend(e, ENumber( bl->ref->id ) );
        }
    }
    return e;
} /* SiGetProcIdList */


/****************************************************************
 *								*
 * SiIsRoutineDefined						*
 *    - checks if a routine given by its id is defined in the   *
 * current project.						*
 *								*
 * Input - routine_id - id of the routine in question		*
 *								*
 * Output -							*
 *   TRUE if defined, FALSE otherwise.				*
 *								*
 ****************************************************************/

int
SiIsRoutineDefined(routine_id)
int routine_id;
{
    PTR_BLOB b, bl;
    PTR_FILE f;
    ELIST e;

    e = NULL;
    for (b = cur_proj->file_chain; b; b = b->next) 
    {
       f = (PTR_FILE) b->ref;
       if (f) 
       {
	  for (bl = f->global_bfnd->entry.template.bl_ptr1; bl; bl = bl->next)
	      if (f->lang == ForSrc ||
		  f->lang == CSrc && bl->ref->variant == FUNC_HEDR)
		 if (bl->ref->entry.template.symbol->id == routine_id) 
                    return TRUE;
       }
    }	
    return FALSE;
} /* SiIsRoutineDefined */

/****************************************************************
 *								*
 *  SiGetGlobalDecl -- handles the primitive that enquires the	*
 *		       list of global declarations in the given *
 *		       project					*
 *								*
 *  Input:							*
 *       projname - project name                                *
 *								*
 *  Output:							*
 *	   same as SiGetProjectInfo				*
 *								*
 ****************************************************************/
ELIST SiGetGlobalDecl(projname)
char *projname;
{
    PTR_BLOB1 b;
    ELIST e;

    b = GetProjInfo(projname, ProjGlobals);
    if (b) 
    {
       e = NULL;
       for (; b; b = b->next) 
       {
	  e = EAppend(e, EString( ((PTR_INFO)b->ref)->source) );
       }
       return e;
    }
    else
       return NULL;
} /* SiGetGlobalDecl */


/****************************************************************
 *								*
 *  SiGetUnresolvedRef -- handles the primitive that enquires   *
 *		  the list of unsolved global references in the *
 *		      given project				*
 *								*
 *  Input:							*
 *                                                              *
 *								*
 *  Output:							*
 *	   same as SiGetProjectInfo				*
 *								*
 ****************************************************************/
ELIST SiGetUnresolvedRef(proj_name)
char *proj_name;
{
    return get_project_info(proj_name, UnsolvRef); 
}


/****************************************************************
 *								*
 *  SiWriteDepFile -- write the dep information of the current	*
 *		      active file to a file using the given	*
 *		      name (or use the one stored in the file	*
 *		      object as default)			*
 *								*
 *  Input:							*
 *	  fname: is the file to be written into.                *
 *        If null then will be written to currentFile->filename * 
 *								*
 *  Output:							*
 *	   a status number 0 sent to stdout if everything was	*
 *	   okay.  A -1 if something wrong.			*
 *								*
 ****************************************************************/
int SiWriteDepFile(fname)
char *fname;
{
    char tmp[100];

    if (fname == NULL)
	fname = currentFile->filename;
    else 
    {
       while ((*fname == ' ') || (*fname == '\t'))
	  fname++;
     if(*(fname+(strlen(fname)-1)) == '\n')
          *(fname+(strlen(fname)-1)) = '\0'; 
    }
    relink(currentFile);
    return rewrite_depfile(currentFile, fname);
} /* SiWriteDepFile */
