#include <strings.h>
#include "../toolbox/include/macro.h"
#include "../toolbox/include/dependence.h"
#include "../toolbox/include/ext_lib.h"
#include "../toolbox/include/ext_low.h"
#include "../toolbox/include/ext_mid.h"
#include "../toolbox/include/ext_high.h"
#include "elist.h"


/*******************************************************************/
/*  These are the new routines based on the irisa library          */
/*   they are been designed by Francois Bodin, Univ. at Rennes     */
/*******************************************************************/

  

/*
        This procedures builds a LL node corresponding to constant
integer {\sf i} and returns  a pointer to the LL node.
*/

int SiBuildInt(i)
int i;
{
   return (bif_build_int(i))->id;
}


/*
        This function builds a VAR_REF node. The parameters must be given
in this order~: node code, node type, node symbol.
*/

int SiLLBuildVarRef(type, sym)
int type, sym;
{
     PTR_SYMB s;
     PTR_TYPE t;
     PTR_LLND l;
     if((s = FindSymbolNode(sym)) == NULL) return 0;
     t = FindTypeNode(type);
     l = bif_build(VAR_REF, t, s);
     return (l == NULL)? -1: l->id;
}

/*
        This procedure builds and returns a array reference
(ARRAY\_REF LOW LEVEL node), with the following parameters in the
given order~: first the array symbol {\sf symb}, then the number of
dimension {\sf n}, and ${\sf expr_{1},\ldots{},expr_{n}}$ the index
expressions of the array reference.
*/

int SiLLBuildArrayRef(symb, n, expr)
int symb, n;
int expr[];
{
      PTR_SYMB s;
      PTR_LLND l, *exp;
      int i;
      if((s = FindSymbolNode(symb)) == NULL) return 0;
      for(i = 0; i < n; i++)
          if((exp[i] = FindLLNode(expr[i])) == NULL) return 0;
      switch(n){
         case 1:
              l = Make_Array_Ref(s,n,exp[0]);
              break; 
         case 2:
              l = Make_Array_Ref(s,n,exp[0],exp[1]);
              break; 
         case 3:
              l = Make_Array_Ref(s,n,exp[0],exp[1],exp[2]);
              break; 
         case 4:
              l = Make_Array_Ref(s,n,exp[0],exp[1],exp[2],exp[3]);
              break; 
         case 5:
              l = Make_Array_Ref(s,n,exp[0],exp[1],exp[2],exp[3],exp[4]);
              break; 
         case 6:
              l = Make_Array_Ref(s,n,exp[0],exp[1],exp[2],exp[3],exp[4], exp[5]);
              break;
         default:
              printf(" array dimension too big (>= 7)\n");
	             l = NULL;
         }
      return (l == NULL)? -1: l->id;
} 
      
/*
        This function builds a LL operator node. The parameters must be given
in this order~: node code, node type and two  node operands.
It returns a pointer to the new LL node.
*/

int SiLLBuild(code, type, low1, low2)
int code, type, low1, low2;
{
     PTR_TYPE t;
     PTR_LLND l, l1, l2;
     l1 = FindLLNode(low1);
     l2 = FindLLNode(low2); 
     t = FindTypeNode(type);
     l = bif_build(code,type, l1, l2);
     return(l = NULL)? -1: l->id;
}
 
/*
        This function builds an entry in the symbol table for a symbol
whose code is  code, whose name  is name and whose type is
type. It returns a pointer on the newly created symbol entry.
*/

int SiBuildSymb(code, name, type)
int code;
char *name;
int type;
{
   PTR_TYPE ptype;
   PTR_SYMB s;
   if((ptype = FindTypeNode(type)) == NULL) return -1;

   s = bif_build_decl (code, name, ptype);
   return (s == NULL)? -1: s->id;
}


/*
        This procedure returns a pointer on the SYMB node
corresponding to the function whose  name is {\sf name}.
*/

int SiGetFunctionSymb(name)
char * name;
{
   PTR_SYMB s;

   if((s = get_function_symb(name)) == NULL) return -1;
   else return s->id;
}


/*
        This procedure returns the Statement node corresponding to the definition
of the function whose SYMB node is  {\sf symb}. This Statement node is
called header Statement node.
*/

int SiGetFunctionHeader(symb)
int symb;
{
  PTR_SYMB s;

  if((s = FindSymbolNode(symb)) == NULL) return -1;

  return get_function_header(s)->id;
}


/*
        This procedure returns TRUE if the symbol {\sf symb}
is in the scope of the function whose header Statement node is {\sf statement},
FALSE otherwise. The symbol may be either a local, a global or a parameter symbo
*/


int SiInScope(stmt, symb)
int stmt, symb;
{
	PTR_BFND b;
	PTR_SYMB s;
	if((b = FindCurrBifNode(stmt)) == NULL) return 0;
        if((s = FindSymbolNode(symb)) == NULL) return 0;

        return in_scope(b,s);
}



/*
        This procedure returns the TYPE node of the expression low
*/

int SiGetTypeForExpression(low)
int low;
{
   PTR_LLND l;
   PTR_TYPE t;
   if((l = FindLLNode(low)) == NULL) return -1;
	
   t = Get_type_for_expression(l);
   return (t == NULL)? -1: t->id;
}



/*
        This procedure inserts a list of Statement nodees beginning with
${\sf statement_{1}}$ after ${\sf statement_{2}}$, the control parent ${\sf
statement_{3}}$, of the Statement nodes which are inserted, can be omitted.
*/

int  SiInsertListOfStmts(statement1, statement2, statement3)
int statement1, statement2, statement3;
{
	
   PTR_BFND b1, b2, b3;
   if((b1 = FindCurrBifNode(statement1)) == NULL) return -1;
   if((b2 = FindCurrBifNode(statement2)) == NULL) return -1;
   if(statement3 <= 0) b3 = NULL;
   else b3 = FindCurrBifNode(statement3);

   return Insert_List_of_BifNode(b1,b2,b3);
}


/*
        This procedure inserts a list of Statement nodes beginning with
${\sf statement_{1 }}$ before ${\sf statement_{2}}$, the control parent ${\sf
statement_{3}}$, of the Statement nodes which are inserted, can be omitted.
*/

int SiInsertListOfStmtsBefore(statement1, statement2, statement3)
int statement1, statement2, statement3;
{
	
   PTR_BFND b1, b2, b3;
   if((b1 = FindCurrBifNode(statement1)) == NULL) return -1;
   if((b2 = FindCurrBifNode(statement2)) == NULL) return -1;
   if(statement3 <= 0) b3 = NULL;
   else b3 = FindCurrBifNode(statement3);

   return Insert_Bifnode_Before(b1, b2, b3);
}

/*
        This procedure extracts from th thread the Statement nodes between
${\sf statement_{1}}$ and ${\sf statement_{2}}$ (including themselves) 
and updates the blob lists of their control parent, returns the 
first Statement node ${\sf statement_{1}}$.
*/

int SiExtractStmntSection(s1, s2){
	PTR_BFND statement1, statement2, b;
	
   if((statement1 = FindCurrBifNode(s1)) == NULL) return -1;
   if((statement2 = FindCurrBifNode(s2)) == NULL) return -1;
   b = Extract_Bif_Section (statement1,statement2);
   if(b == NULL) return -1;
   else return b->id;
}

/*
This procedure returns the Statement node preceding {\sf statement} in the
thread
*/

int SiGetStmtBefore(id)
int id;
{
	PTR_BFND statement, b;

        if((b = Get_Node_Before(statement)) == NULL) return -1;
	else return b->id;
}


/*
        This procedure returns TRUE if ${\sf statement_{3}}$ is the list of
Statement nodes beginning with ${\sf statement_{1}}$ and ending with ${\sf
statement_{2}}$ (including themselves), FALSE otherwise.
*/

int SiInSection(s1,s2, s3)
int s1, s2, s3;
{
  PTR_BFND statement1, statement2, statement3;

   if((statement1 = FindCurrBifNode(s1)) == NULL) return -1;
   if((statement2 = FindCurrBifNode(s2)) == NULL) return -1;
   if((statement3 = FindCurrBifNode(s3)) == NULL) return -1;
   return In_Section( statement1, statement2, statement3);
}


/*
        This procedure returns the last Statement node of a statement beginning
with {\sf statement} (himself or the last node of his blob lists).
*/

int SiGetLastStatmentInList(s)
int s;
{
    PTR_BFND statement,s;
    if((statement = FindCurrBifNode(s)) == NULL) return -1;
    s = Get_Last_Node_Of_Stmt(statement);
    if(s == NULL) return -1;
    else return s->id;
}



/*  
        This procedure replaces all the references to the symbol
{\sf symb} by the expression {\sf low} in the section beginning with
${\sf statement_{1}}$ and ending with ${\sf statement_{2}}$.
*/

int SiReplaceVarInSection(s1, s2, sym, exp)
int s1, s2, sym, exp;
{
  PTR_SYMB p;
  PTR_LLND l;
  PTR_BFND statement1, statement2;

   if((p = FindSymbolNode(sym)) == NULL) return -1;
   if((l = FindLLNode(exp)) == NULL) return -1;
   if((statement1 = FindCurrBifNode(s1)) == NULL) return -1;
   if((statement2 = FindCurrBifNode(s2)) == NULL) return -1;
   Replace_Var_In_Section(statement1, statement2, p, l);
}


/*
        This procedure copies a code section beginning with {\sf statement},
the control parent and th blob lists of these new nodes are updated,
returns the beginning of this new section.
*/

SiDuplicateStmts(s)
int s;
{
    PTR_BFND statement;
    if((statement = FindCurrBifNode(s)) == NULL) return -1;
    Duplicate_stmts(statement);
}

/*
    This procedure deletes the Statement node {\sf statement} from the thread,
updates the blob lists of his control parent, and returns th Statement node
which was before {\sf statement}.
*/

SiIDeleteStmt(s)
int s;
{
  PTR_BFND statement, s;

  
    if((statement = FindCurrBifNode(s)) == NULL) return -1;
    s = Delete_Bif_Node(statement);
    if(s  == NULL) return -1;
    else return s->id;
}


/*
This procedure returns a copy of the expression beginning with
the LOW LEVEL node {\sf low}.
*/


int SiICopyExpr(l)
int l;
{
	PTR_LLND low, p;
	if((low = FindLLNode(l)) == NULL) return -1;
        p = bif_copy_expr(low);
	if(p == NULL) return -1;
	else return p->id;
}


/*
        This procedure tries to evaluate an expression given by {\sf
low}. It returns an array of two elements, the first is set to -1 if
the evaluation failed, the second one contains the integer which is
result of the evaluation, if it suceeded.  If the low level node
is not found it returns NULL;
*/

int *SiEvaluateExpr(l)
int l;
{
	PTR_LLND low;
	if((low = FindLLNode(l)) == NULL) return NULL;
        return Evaluate_Expression(low);
}


/*
        This procedure replaces the symbol {\sf symb} by the
expression ${\sf low_{2}}$ in the expression ${\sf low_{1}}$, take
care that the ${\sf low_{1}}$ isn't a VAR REF. Returns ???
*/

int SiReplaceVarInExpression(l1, s, l2)
int l1, s, l2;
{
	PTR_LLND low1, low2;
	PTR_SYMB symb;	
	if((low1  = FindLLNode(l1)) == NULL) return -1;
	if((low2  = FindLLNode(l2)) == NULL) return -1;
	if((symb  = FindSymbolNode(s)) == NULL) return -1;
	if(low1->variant == VAR_REF) return l2;
	return Replace_Var_In_Expression(low1, symb, low2);
}


/*
        This procedure replaces the symbol ${\sf symb_{1}}$ by the
symbol of an array ${\sf symb_{2}}$ in the expression {\sf low},
returns the new expression {\sf low}.
*/

int SiReplaceSymbolRef(l, s1, s2)
int l, s1, s2;
{
	PTR_LLND low, l;
	PTR_SYMB symb1, symb2;
	if((low  = FindLLNode(l)) == NULL) return -1;
	if((symb1  = FindSymbolNode(s1)) == NULL) return -1;
	if((symb2  = FindSymbolNode(s2)) == NULL) return -1;

        l = Replace_Symbol_Ref(low, symb1, symb2);
	if(l == NULL) return -1;
	else return l->id;
}


/*
        This procedure builds a linear expression...
*/

void SiBuildLinearRep(l, i1, s, i2, i3)
int l;
int *i1;
int s;
int i2;
int *i3;
{
	PTR_SYMB symb;
	PTR_LLND low;
	if((low  = FindLLNode(l)) == NULL) return;
	if((symb  = FindSymbolNode(s)) == NULL) return;
        Build_Linear_Rep(low, i1, symb, i2, i3);
}


/*
        This procedure returns the first parameter of a function call
given by {\sf low}.
*/

int SiGetFirstParameterForCall(l)
int l;
{
	PTR_LLND low, p;
	if((low  = FindLLNode(l)) == NULL) return;
	p = Get_First_Parameter_For_Call(low);
	if(p == NULL) return -1;
	else return p->id;
}
/*
        This procedure returns the second parameter of a function call
given by {\sf low}.
*/

int SiGetSecondParameterForCall(l)
int l;
{
	PTR_LLND low, p;
	if((low  = FindLLNode(l)) == NULL) return;
	p = Get_Second_Parameter_For_Call(low);
	if(p == NULL) return -1;
	else return p->id;
}

/*
        This procedure returns the ${\sf i}^{th}$ parameter of a function call
given by {\sf low}.
*/

int SiGetNthParameterForCall(l, i)
int l;
{
	PTR_LLND low, p;
	PTR_LLND Get_Nieme_Parameter_For_Call();
	if((low  = FindLLNode(l)) == NULL) return;
	p = Get_Nieme_Parameter_For_Call(low,i);
	if(p == NULL) return -1;
	else return p->id;
}


/*
        This procedure returns an expression list of LOW LEVEL nodes containing
all the references to the array, which is given by the  SYMB node {\sf
symb}, found in the expression {\sf low}.
*/

int SiLookForArrayRefInExpr(l, s)
int l, s;
{
	PTR_SYMB symb;
	PTR_LLND low, p;
	if((low  = FindLLNode(l)) == NULL) return -1;
	if((symb  = FindSymbolNode(s)) == NULL) return -1;
        p = Look_for_Arrayref_In_Expr( low, symb);
	if(p == NULL) return -1;
	else return p->id;
}


/*
        This procedure returns a expression list of LOW LEVEL nodes containing
all the references to the array, which is given by the  SYMB node {\sf
symb}, found in the section which the Statement node {\sf statement} is the first element.
*/

int SiLookForArrayRefInSection(st, s)
int st, s;
{
	PTR_SYMB symb;
	PTR_LLND p;
	PTR_BFND stmt;
	if((stmt = FindCurrBifNode(st)) == NULL) return -1;
	if((symb  = FindSymbolNode(s)) == NULL) return -1;
        p = Look_for_Arrayref_In_Section(stmt, symb);
	if(p == NULL) return -1;
	else return p->id;
}

/**************************************************************/
/*  Mid Level Fortran Primatives ******************************/
/**************************************************************/

/*
        This procedure builds a new loop where the index variable is {\sf
symb}, the lower bound is ${\sf low_{1}}$, the upper bound is ${\sf
low_{2}}$, the step is ${\sf low_{3}}$ and wher the body of this new
loop begins with {\sf bif}. The control parent of the loop body are
updated, and it returns the new FOR\_NODE.
*/

int SiMakeForLoop(s, l1, l2, l3, stmt)
int s, l1, l2, l3, stmt;
{
	PTR_BFND bif,b;
	PTR_LLND low1, low2, low3; 
	PTR_SYMB symb;

        if((bif = FindCurrBifNode(stmt)) == NULL) return -1;
        if((symb  = FindSymbolNode(s)) == NULL) return -1;
	if((low1  = FindLLNode(l1)) == NULL) return -1;
	if((low2  = FindLLNode(l2)) == NULL) return -1;
	if((low3  = FindLLNode(l3)) == NULL) return -1;
	b = Make_For_Loop(symb,low1, low2, low3, bif);
	return (b==NULL)? -1: b->id;
}

/*
        This procedure updates the lower bound of the loop given by
the BIF node {\sf bif} with the expression {\sf low}. It returns the
former expression of this bound, it can also be used if the bound had
no value  in this case the result returned is NULL.
*/

int SiLoopSetBoundInf(s, l)
int s, l;
{
	PTR_BFND bif;
	PTR_LLND low, p;
        if((bif = FindCurrBifNode(s)) == NULL) return -1;
        if((low  = FindLLNode(l)) == NULL) return -1;
	p = Loop_Set_Borne_Inf(bif,low);
	return (p == NULL)? -1: p->id;
}


/*
        This procedure updates the upper bound of the loop given by
the BIF node {\sf bif} with the expression {\sf low}. It returns the
former expression of this bound, it can also be used if the bound had
no value  in this case the result returned is NULL.
*/

int SiLoopSetBoundSup(s, l)
int s, l;
{
        PTR_BFND bif;
        PTR_LLND low, p;
        if((bif = FindCurrBifNode(s)) == NULL) return -1;
        if((low  = FindLLNode(l)) == NULL) return -1;
        p = Loop_Set_Borne_Sup(bif,low);
        return (p == NULL)? -1: p->id;
}


/*
        This procedure updates the step of the loop given by
the BIF node {\sf bif} with the expression {\sf low}. It returns the
former expression of this step, it can also be used if the step had
no value  in this case the result returned is NULL.
*/

int SiLoopSetStep(s, l)
int s, l;
{
        PTR_BFND bif;
        PTR_LLND low, p;
        if((bif = FindCurrBifNode(s)) == NULL) return -1;
        if((low  = FindLLNode(l)) == NULL) return -1;
        p = Loop_Set_Step(bif,low);
        return (p == NULL)? -1: p->id;
}

/*
        This procedure updates the index variable of the loop given by
the BIF node {\sf bif} with the symbol {\sf symb}. It returns the
former symbol of this index variable, it can also be used if the index had
no value  in this case the result returned is NULL.
*/

int SiLoopSetIndex(s, l)
int s, l;
{
        PTR_BFND bif;
        PTR_SYMB symb, p;
        if((bif = FindCurrBifNode(s)) == NULL) return -1;
        if((symb  = FindSymbolNode(l)) == NULL) return -1;
        p = Loop_Set_Index(bif,symb);
        return (p == NULL)? -1: p->id;
}

/*
        This procedure returns the expression of the lower bound of
the loop given by{\sf bif}.
*/


int SiLoopGetBoundInf(s)
int s;
{
        PTR_BFND bif;
        PTR_LLND p;
        if((bif = FindCurrBifNode(s)) == NULL) return -1;
        p = Loop_Get_Borne_Inf(bif);
        return (p == NULL)? -1: p->id;
}


/*
        This procedure returns the expression of the upper bound of
the loop given by{\sf bif}.
*/

int SiLoopGetBoundSup(s)
int s;
{
        PTR_BFND bif;
        PTR_LLND p;
        if((bif = FindCurrBifNode(s)) == NULL) return -1;
        p = Loop_Get_Borne_Sup(bif);
        return (p == NULL)? -1: p->id;
}


/*
        This procedure returns the expression of the step of
the loop given by{\sf bif}.
*/

int SiLoopGetStep(s)
int s;
{
        PTR_BFND bif;
        PTR_LLND p;
        if((bif = FindCurrBifNode(s)) == NULL) return -1;
        p = Loop_Get_Step(bif);
        return (p == NULL)? -1: p->id;
}

/*
        This procedure returns the symbol of the index variable of
the loop given by{\sf bif}.
*/

int SiLoopGetIndex(s)
int s;
{
        PTR_BFND bif;
        PTR_SYMB p;
        if((bif = FindCurrBifNode(s)) == NULL) return -1;
        p = Loop_Get_Index(bif);
        return (p == NULL)? -1: p->id;
}


/*
        This procedure builds an array declaration where {\sf bif} is
his scope, {\sf name} is his name, {\sf type} is the type of his
elements, ${\sf i_{1}}$ is the number of dimension and ${\sf i_{2}}$
is an array containing the number of element of each dimension. The
symbol table is updated and it returns this new BIF node.
*/

SiDeclareArray(b, name, t, i1, i2)
int b;
char *name;
int t;
int i1;
int *i2;
{
	PTR_BFND bif, b;
	PTR_TYPE type;
        if((bif = FindCurrBifNode(b)) == NULL) return -1;
	if((type = FindTypeNode(t)) == NULL) return -1;
	b = Declare_Array(bif, name, type, i1, i2);
	return (b == NULL)? -1: b->id;
}
	



/*
        This procedure returns the scope of the BIF node {\sf bif}.
*/

int SiGetScopeForDecl(s)
int s;
{
        PTR_BFND bif, b;
        if((bif = FindCurrBifNode(s)) == NULL) return -1;
        b = Get_Scope_For_Declare(bif);
        return (b == NULL)? -1: b->id;
}

/*
        This procedure returns the scope of the label-calling BIF node
{\sf bif}, which the first preceding node which is a function, procedure or
program declaration.
*/

int SiGetScopeForLabel(s)
int s;
{
        PTR_BFND bif, b;
        if((bif = FindCurrBifNode(s)) == NULL) return -1;
        b = Get_Scope_For_Label(bif);
        return (b == NULL)? -1: b->id;
}

/*
        This procedure builds and returns a procedure declaration
thanks to the following parameters in the given order~: first {\sf
name} the name of the procedure, {\sf type} the type of the procedure,
{\sf b} the first BIF node of the procedure body, {\sf i} the number
of parameters of this new procedure, and ${\sf symb_{1},type_{1}, \ldots{},
symb_{n}, type_{n}}$ the list of the parameters symbols and their type
(${\sf type_{1}}$ is the type of the parameter which symbol is ${\sf
symb_{1}}$).
*/


int SiMakeProcedure(name, type, b, i, symb, type2)
char * name;
int type;
int b, i;
int *symb;
int *type2;
{
	PTR_SYMB s[7];
	PTR_TYPE pt, t[7];
	PTR_BFND bif, n;
	int k;
	if(i > 7){
		printf("too many arguments to MakeProcedure... sorry\n");
		return -1;
		}
	if((pt = FindTypeNode(type)) == NULL) return -1;
	if((bif = FindCurrBifNode(b)) == NULL) return -1;
	for(k = 0; k < i; k++){
		s[k] = FindSymbolNode(symb[k]);
		t[k] = FindTypeNode(type2[k]);
		}
	switch(i){
	   case 0:
		n = Make_Procedure(name,pt,bif,i);
		break;
	   case 1:
		n = Make_Procedure(name,pt,bif,i,s[0],t[0]);
		break;
	   case 2:
		n = Make_Procedure(name,pt,bif,i,s[0],t[0], s[1], t[1]);
		break;
	   case 3:
	        n = Make_Procedure(name,pt,bif,i,s[0],t[0],s[1],t[1],
                s[2],t[2]);
		break;
	   case 4:
	        n = Make_Procedure(name,pt,bif,i,s[0],t[0],s[1],t[1],
                     s[2],t[2],s[3],t[3]);
		break;
           case 5:	     
                 n = Make_Procedure(name,pt,bif,i,s[0],t[0],s[1],t[1], 
                     s[2],t[2],s[3],t[3],s[4],t[4]);
		break;
           case 6:	     
                 n = Make_Procedure(name,pt,bif,i,s[0],t[0],s[1],t[1], 
                     s[2],t[2],s[3],t[3],s[4],t[4],s[5],t[5]);
		break;
           case 7:	     
                 n = Make_Procedure(name,pt,bif,i,s[0],t[0],s[1],t[1], 
                     s[2],t[2],s[3],t[3],s[4],t[4],s[5],t[5],s[6],t[6]);
		break;
	}
	return (n == NULL)? -1: n->id;
}
	
	

/*
        This procedure builds and returns a procedure call thanks to
the following  parameters in the given order~: first {\sf symb} the
symbol of the called procedure, {\sf i} his number of parameters,
${\sf low_{1},\ldots{},\;low_{n}}$ the parameters of the called procedure.
*/

int SiMakeProcedureCall(symb,i,low)
int symb;
int i;
int *low;
{
	PTR_LLND l[7];
	PTR_BFND n;
	PTR_SYMB s;
	int k;
	if(i > 7){
		printf("too many arguments to MakeProcedure... sorry\n");
		return -1;
		}
	if((s = FindSymbolNode(s)) == NULL) return -1;
	for(k = 0; k < i; k++){
		l[k] = FindLLNode(low[k]);
		}
	switch(k){
	   case 0:
		n = Make_Procedure_Call(s, i);
		break;
	   case 1:
		n = Make_Procedure_Call(s, i,l[0]);
		break;
	   case 2:
		n = Make_Procedure_Call(s, i,l[0], l[1]);
		break;
	   case 3:
	        n = Make_Procedure_Call(s, i,l[0],l[1],l[2]);
		break;
	   case 4:
	        n = Make_Procedure_Call(s, i,l[0],l[1],l[2],l[3]);
		break;
           case 5:	     
                 n = Make_Procedure_Call(s, i,l[0],l[1],l[2],l[3],l[4]);
		break;
           case 6:	     
                 n = Make_Procedure_Call(s, i,l[0],l[1],l[2],l[3],l[4],l[5]);
		break;
           case 7:	     
                 n = Make_Procedure_Call(s, i,l[0],l[1],
                                              l[2],l[3],l[4],l[5],l[6]);
		break;
	}
	return (n == NULL)? -1: n->id;
}



/*
         This procedure builds and returns a function call thanks to
the following  parameters in the given order~: first {\sf s } the
symbol of the called function, {\sf t } his type, {\sf i} his number
of parameters, {\sf PTR\_LLND* l } an array of LOW LEVEL nodes containing the
parameter expressions passed to the called function.
*/

int SiMakeFunctionCall(s, t, i, l)
int s, t, i;
int *l;
{
	PTR_LLND low[200];
	PTR_LLND l2;
	PTR_SYMB sym;
	PTR_TYPE typ;
	int k;
	if((sym = FindSymbolNode(s)) == NULL) return -1;
	if((typ = FindTypeNode(t)) == NULL) return -1;
	if(i > 200){
		printf("too many parameters in MakeFunctionCall. sorry\n");
		return -1;
		}
	for(k = 0; k < i; k++)
		low[k] = FindLLNode(l[k]);
	l2 = Make_Function_Call_bis(sym, typ, i, low);
	return (l2 == NULL)? -1: l2->id;
}

/*
        This procedure builds and returns a linear expression, the
symbols of this expression are in the SYMB node array {\sf symb[\,]} and the
coefficients in the LOW LEVEL node array ${\sf low_{2}}$ {\sf [\,]},
the integer constant is represented by the LOW LEVEL node ${\sf
low_{1}}$ and {\sf i} is the length of the arrays.
*/

int SiMakeLinearExpression(i1, l1, symb, l2)
int i1, l1;
int *symb, *l2;
{
	int i;
	PTR_LLND low2[200];
	PTR_SYMB s[200];
	PTR_LLND low1, l;
	if (i1 > 200){
		printf("too many expressions in MakeLinearExpression.\n");
		return -1;
		}
	if(l1 > -1) low1 = FindLLNode(l1);
	else low1 = NULL;
	for(i = 0; i < i1; i++){
		if((s[i] = FindSymbolNode(symb[i])) == NULL) return -1;
		if(l2[i]!=-1) low2[i] = FindLLNode(l2[i]);
		else low2[i] = NULL;
		}
	l = Make_Linear_Expression(i1, low1, s, low2);
	return (l == NULL)? -1: l->id;
}

/*
        This procedure builds and returns a linear expression, the
symbols of this expression are in the SYMB node array {\sf symb [\,]} and the
coefficients in the integer array ${\sf i_{3}}$  {\sf [\,]}, ${\sf
i_{2}}$ is the integer constant and ${\sf i_{1}}$ the length of the
arrays.
*/

int SiMakeLinearExpressionFromInt(i1, i2, s, i3)
int i1, i2;
int *s, *i3;
{
	int i;
	PTR_SYMB symb[200];
	PTR_LLND l;
	if(i > 200){
		printf("too many expr. in MakeLinearExpressionFromInt\n");
		return -1;
		}
	for(i = 0; i < i1; i++){
		if((symb[i] = FindSymbolNode(s[i])) == NULL) return -1;
		}
	l = Make_Linear_Expression_From_Int(i1, i2, symb, i3);
	return (l == NULL)? -1: l->id;
}


/*
        This procedure builds and returns an assignment in which ${\sf
low_{1}}$ is the right part and ${\sf low_{2}}$ the left of part of
the new instruction.
*/

int SiMakeAssign(l1, l2)
int l1, l2;
{
	PTR_LLND low1, low2;
	PTR_BFND b;
	if((low1 = FindLLNode(l1)) == NULL) return -1;
	if((low2 = FindLLNode(l2)) == NULL) return -1;
	b = Make_Assign(low1, low2);
	return(b == NULL)? -1: b->id;
}
	


/*
        This procedure builds and returns a conditional statement in
which {\sf l} is the condition expression, ${\sf b_{1}}$ is the
beginning of the TRUE case, ${\sf b_{2}}$ of the FALSE case. The
node control parents of the two bodies are updated.
*/

int SiIfThenElse(b1, b2, l)
int b1, b2, l;
{
	PTR_LLND low;
	PTR_BFND b, bif1, bif2;
	if((low = FindLLNode(l)) == NULL) return -1;
        bif1 = FindCurrBifNode(b1);
        bif2 = FindCurrBifNode(b2);
	b = Make_If_Then_Else(bif1, bif2, low);
	return(b == NULL)? -1: b->id;
}

/*
        This procedure returns TRUE if it managed to build the
declaration of {\sf s} and insert it in the code, {\sf b} is our
location in the code at this moment.
*/

int SiDeclareScalar(s, b)
int s, b;
{
	PTR_SYMB symb;
	PTR_BFND bif;
        if((bif = FindCurrBifNode(b)) == NULL) return -1;
	if((symb = FindSymbolNode(s)) == NULL) return -1;
	return Declare_Scalar(symb, bif);
}
	
/*
        This procedure extracts the loop body of the loop given by
{\sf bif}, returns the first BIF node of the body.
*/

int SiExtractLoopBody(b)
int b;
{
	PTR_BFND bif, p;
        if((bif = FindCurrBifNode(b)) == NULL) return -1;
	p = Extract_Loop_Body(bif);
	return (p == NULL)? -1: p->id;
}

/*
        This procedure returns the next perfectly nested loop after
the loop given by {\sf bif}.
*/

int SiGetNextNestedLoop(b)
int b;
{
	PTR_BFND bif, p;
        if((bif = FindCurrBifNode(b)) == NULL) return -1;
	p = Get_Next_Nested_Loop(bif);
	return (p == NULL)? -1: p->id;
}

/*
        This procedure returns the previous perfectly nested loop before
the loop given by {\sf bif}.
*/

int SiGetPreviousNestedLoop(b)
int b;
{
	PTR_BFND bif, p;
        if((bif = FindCurrBifNode(b)) == NULL) return -1;
	p = Get_Previous_Nested_Loop(bif);
	return (p == NULL)? -1: p->id;
}


/*
        This procedure returns the most internal loop in the loop nest
given by {\sf bif}.
*/

int SiGetInternalLoop(b)
int b;
{
	PTR_BFND bif, p;
        if((bif = FindCurrBifNode(b)) == NULL) return -1;
	p = Get_Internal_Loop(bif);
	return (p == NULL)? -1: p->id;
}


/*
        This procedure returns the number of parfectly nested loops
contained in the nest beginning with {\sf bif}.
*/

int SiPerfectlyNested(b)
int b;
{
	PTR_BFND bif;
        if((bif = FindCurrBifNode(b)) == NULL) return -1;
	return Perfectly_Nested(bif);
}


/*
        This procedure returns TRUE  if the loop given by {\sf bif} ends with
a CONTROL\_END (BIF node), FALSE otherwise.
*/

int SiIsGoodLoop(b)
int b;
{
        PTR_BFND bif;
        if((bif = FindCurrBifNode(b)) == NULL) return -1;
        return Is_Good_Loop(bif);
}


/*
        This procedure returns TRUE if there is a function call
(FUNC\_CALL LOW LEVEL node) in the expression given by {\sf low},
FALSE otherwise.
*/

int SiVerifFunc(l)
int l;
{
        PTR_LLND low;
        if((low = FindLLNode(l)) == NULL) return 0;
        return Verif_Func(low);
}

/*
        This procedure returns TRUE if the body of the loop nest beginning with
{\sf bif} contains only assignments, FALSE otherwise.
*/

int SiVerifAssign(b)
int b;
{
        PTR_BFND bif;
        if((bif = FindCurrBifNode(b)) == NULL) return -1;
        return Verif_Assign(bif);
}



/*
        This procedure returns TRUE if the body of the loop nest beginning with
{\sf bif} contains only assignments and conditional statements, FALSE
otherwise.
*/

int SiVerifAssignIf(b)
int b;
{
        PTR_BFND bif;
        if((bif = FindCurrBifNode(b)) == NULL) return -1;
        return Verif_Assign_If(bif);
}

/* 
The procedure converts nested loops so that they  have
nice "enddo" instead of shared continue statements.
*/

int SiConvertLoop(b)
int b;
{
        PTR_BFND bif, n;
        if((bif = FindCurrBifNode(b)) == NULL) return -1;
        Loop_Conversion(bif);
}


/*
        This procedure generates a new name and symbol for a variable of type
{\sf t }, which should appear in the code in the same scope as {\sf
b }. The symbol table is updated, it returns the new created symbol.
*/

int SiGenerateVariableName(b, t)
int b;
int t;
{
        PTR_BFND bif;
        PTR_TYPE typ;
        PTR_SYMB s;
        if((bif = FindCurrBifNode(b)) == NULL) return -1;
	if((typ = FindTypeNode(t)) == NULL) return -1;
        s = Generate_Variable_Name(bif, typ);
	return (s == NULL)? -1: s->id;
}


/*
        This procedure generates a new symbol of type {\sf type} and
inserts a declaration in the code in the same scope as {\sf bif},
which is our location in the code at this moment.
*/


int SiInstallVariable(b, t)
int b;
int t;
{
        PTR_BFND bif;
        PTR_TYPE typ;
        PTR_SYMB s;
        if((bif = FindCurrBifNode(b)) == NULL) return -1;
	if((typ = FindTypeNode(t)) == NULL) return -1;
        s = Install_Variable(bif, typ);
	return (s == NULL)? -1: s->id;
}


/*
        This procedure returns TRUE if it managed to build a
conditional statement with {\sf l } as the expression of the
condition and the body of the loop given by {\sf b } as the true and
false bodies, FALSE otherwise.
*/

int SiGenerateAlternativeCode(l, b)
int b;
int l;
{
        PTR_BFND bif;
        PTR_LLND low;
        if((bif = FindCurrBifNode(b)) == NULL) return -1;
	if((low = FindLLNode(l)) == NULL) return -1;
        return Generate_Alternative_Code(low, bif);
}

/*  find doc!!! */

void SiLocalizeArraySection(b, l, i)
int b, l, i;
{
	PTR_BFND bif;
	PTR_LLND low;
        if((bif = FindCurrBifNode(b)) == NULL) return ;
	if((low = FindLLNode(l)) == NULL) return ;
	Localize_Array_Section(bif, low, i);
}
	
/*
        This procedure returns TRUE if the expressions of the bounds
and step of the loop given by {\sf bif} use the index variables of the
previous loops of the nest, FALSE otherwise.
*/

int SiCheckIndex(b)
int b;
{
        PTR_BFND bif;
        if((bif = FindCurrBifNode(b)) == NULL) return -1;
        return Check_Index(bif);
}



/*
        This procedure returns TRUE if the right hand side of the
assignment given by {\sf bif} uses an index variable of the loop nest,
FALSE otherwise.
*/

int SiCheckRightAssign(b)
int b;
{
        PTR_BFND bif;
        if((bif = FindCurrBifNode(b)) == NULL) return -1;
        return Check_Right_Assign(bif);
}


/*
        This procedure returns TRUE if the left hand side of the
assignment given by {\sf bif} is an index variable of the loop nest,
FALSE otherwise.
*/

int SiCheckLeftAssign(b)
int b;
{
        PTR_BFND bif;
        if((bif = FindCurrBifNode(b)) == NULL) return -1;
        return Check_Left_Assign(bif);
}


/*
        This procedure returns TRUE if the expressions of bounds and
steps of the loops of the loop nest given by {\sf bif} don't use any
index variable of the loop nest, FALSE otherwise.
*/

int SiNoDependentIndex(b)
int b;
{
        PTR_BFND bif;
        if((bif = FindCurrBifNode(b)) == NULL) return -1;
        return No_Dependent_Index(bif);
}

/*
        This procedure returns TRUE if the assignments in the loop
body of the loop nest given by {\sf bif} don't use an index variable
of the loop nest, FALSE otherwise.
*/

int SiNoBasicInduction(b)
int b;
{
        PTR_BFND bif;
        if((bif = FindCurrBifNode(b)) == NULL) return -1;
        return No_Basic_Induction(bif);
}


/*
        This procedure returns TRUE if an index variable of the loop
nest given by {\sf bif} is set in the body, FALSE otherwise.
*/

int SiNoDefOfInduction(b)
int b;
{
        PTR_BFND bif;
        if((bif = FindCurrBifNode(b)) == NULL) return -1;
        return No_Def_Of_Induction(bif);
}


/***************************************************************/
/******** high level Fortran Transformations *******************/
/***************************************************************/

/*
        This procedure will return TRUE if the loop nest given by {\sf
bif} can be tiled.
*/

int SiTilingP(b,i)
int b;
int *i;
{
        PTR_BFND bif;
        if((bif = FindCurrBifNode(b)) == NULL) return -1;
        return tiling_p(bif,i);
}



/*
        This procedure tile the loop nest in which {\sf bif} is the
most external loop, {\sf i} is the number of loops in the nest and the
array {\sf tab [\,]} contains the size of the tiles. The loops are
supposed to be normalized and perfectly nested.
*/

int SiTiling(b, i, tab )
int b;
int i;
int * tab;
{

        PTR_BFND bif, n;
        if((bif = FindCurrBifNode(b)) == NULL) return -1;
        tiling(bif, i, tab);
}


/*
        This procedure tiles the loop given by {\sf bif}, the size
of the tile is {\sf i}.
*/

void SiStripMining(b, i )
int b;
int i;
{

        PTR_BFND bif, n;
        if((bif = FindCurrBifNode(b)) == NULL) return;
        strip_mining(bif, i);
}


/*
        This procedure distributes the statements of the loop given by
{\sf bif}, {\sf i} should be set to {\sf 1} for the moment. It returns
the list of the new loops. (now only returns first, must be fixed);
*/
int SiDistributeLoop(b, i )
int b;
int i;
{

        PTR_BFND bif;
	PTR_BLOB n;
        if((bif = FindCurrBifNode(b)) == NULL) return -1;
        n = Distribute_Loop(bif, i);
        return (n == NULL)? -1: n->ref->id;
}

/*
        This procedure distributes also the statements of the loop
given by {\sf b }, but the statements are distributed in the loops
following the strongly connected component of the dependence graph.
*/

int SiDistributeLoopSCC(b )
int b;
{

        PTR_BFND bif;
	PTR_BLOB n;
        if((bif = FindCurrBifNode(b)) == NULL) return -1;
        n = Distribute_Loop_SCC(bif);
        return (n == NULL)? -1: n->ref->id;
}


/*
        This procedure merges the two loops given by ${\sf statement_{1}}$
and ${\sf statement_{2}}$, the index variable of this new loop is the index
variable of the first loop (${\sf statement_{1}}$).
*/

int SiLoop_Fusion(b1,b2)
int b1,b2;
{

        PTR_BFND bif1, bif2, n;
        if((bif1 = FindCurrBifNode(b1)) == NULL) return -1;
        if((bif2 = FindCurrBifNode(b2)) == NULL) return -1;
        return Loop_Fusion(bif1, bif2);
}



/*
        This procedure unrolls the loop given by {sf b } {\sf i}
times, the loops contained in it must be perfectly nested and must not
have a step, compensation code is inserted.
*/

int SiUnrollLoop(b,i)
int b;
int i;
{

        PTR_BFND bif, n;
        if((bif = FindCurrBifNode(b)) == NULL) return -1;
        return Unroll_Loop(bif, i);
}


/*
        This procedure interchanges the loops of the loop nest given
by {\sf b} thanks to the permutation contained in the array ${\sf
*i_{1}}$, ${\sf i_{2}}$ is the number of loops which must be
interchanged. The loop nest must be perfectly nested.
*/

int SiInterchangeLoops(b,i1, i2)
int b;
int *i1;
int i2;
{

        PTR_BFND bif, n;
        if((bif = FindCurrBifNode(b)) == NULL) return -1;
        return Interchange_Loops(bif, i1, i2);
}

/*
        This procedure replaces the variable which symbol is {\sf
s} by an array reference in the loop body of the loop given by {\sf
bb}, which is supposed to be the most internal loop if there is a
loop nest. The declaration of this new array is inserted in the code,
and if {\sf i} is set to {\sf 1}, code is inserted after the loop in
order to set the varible to the value it should have had after the
execution of the loop.
*/
int SiExpandScalar(b,s, i)
int b;
{

        PTR_BFND bif, n;
        PTR_SYMB sym;
        if((bif = FindCurrBifNode(b)) == NULL) return -1;
	if((sym = FindSymbolNode(s)) == NULL) return -1;
        return Expand_Scalar(bif,sym,i);
}



/*
        This procedure returns a modified loop by given {\sf b},
where the variable {\sf s} is replaced in the loop body by the
expression it was set, if this variable was set once in the body.
*/

int SiScalarForwardSubstitution(b,s)
int b;
{

        PTR_BFND bif, n;
        PTR_SYMB sym;
        if((bif = FindCurrBifNode(b)) == NULL) return -1;
	if((sym = FindSymbolNode(s)) == NULL) return -1;
        n = Scalar_Forward_Substitution(bif,sym);
        return (n == NULL)? -1: n->id;
}


/*
        This procedure returns TRUE if the loop given by the stmt node
{\sf b} is normalized, FALSE otherwise.
*/
int SiNormalized(b)
int b;
{

        PTR_BFND bif;
        if((bif = FindCurrBifNode(b)) == NULL) return 0;
        return Normalized(bif);
}


/*
        This procedure normalize the loop given by {\sf bif}, the step
and the lower bound are set to {\sf 1},  therefore the instructions of the body
and the upper bound are modified.
*/

int SiNormalizeLoop(b)
int b;
{

        PTR_BFND bif, n;
        if((bif = FindCurrBifNode(b)) == NULL) return -1;
        return Normalize_Loop(bif);
}


/*
        This procedure replaces the array expressions of the loop body
of the loop given by {sf bif}, by an expression ``lower bound:upper
bound''.
*/

int SiVectorize(b)
int b;
{

        PTR_BFND bif;
        if((bif = FindCurrBifNode(b)) == NULL) return -1;
        return Vectorize(bif);
}


/*
        This procedures vectorizes the loop nest given by {\sf bif}.
/*

int SiVectorizeNest(b)
int b;
{

	PTR_BFND bif;
        if((bif = FindCurrBifNode(b)) == NULL) return -1;
	return Vectorize_Nest(bif);
}

/*
        This procedure prints in the file {\sf file} all the
properties of the loop nest given by {\sf bif} (normalized,
vectorizablee, induction variables\ldots{}), {\sf i} must be set to 1.
*/

void SiPrintPropertyForLoop(file, b, i)
FILE *file;
int b, i;
{
	PTR_BFND bif;
        if((bif = FindCurrBifNode(b)) == NULL) return;
	Print_Property_For_Loop(file, bif, i);
}


