/****************************************************************
 *								*
 *   csvr_stcn.c						*
 *								*
 *   This file contains a set of low level interfaces to the	*
 *   kernel of the program data base				*
 *								*
 ****************************************************************/
#include <strings.h>
#include "db.h"
#include "elist.h"


extern void free();
/*
 * functions used from libdb.a 
 */
extern int GetLangType();     /* from libdb.a */
extern char *bget_prop();     /* from libdb.a */
extern char *copys();         /* from libdb.a */
extern void normal_form();    /* from libdb.a */
extern PTR_BFND	make_bfndnt();
extern PTR_LLND make_llnd();
extern PTR_TYPE make_type();
extern PTR_SYMB make_label();
extern PTR_BLOB make_blob();
extern PTR_SYMB make_symb();
extern PTR_SETS build_sets();
extern PTR_LLND replace_ref();
extern void reset_scalar_propogation();
extern void gendeps();
extern void relink();
extern void collect_garbage();
extern void GlobalAnal();
extern PTR_LLND find_bounds();   
extern PTR_BFND FindBifNode();
extern void SiAddToStmtTable();

/*
 * externs from libdb.a
 */
extern PCF UnparseBfnd[];
extern PCF UnparseLlnd[];
extern PCF UnparseSymb[];
extern PCF UnparseType[];
extern PCF UnparseBlock[];
/*
 * External references
 */
extern PTR_PROJ currentProject;	/* point to the current active project */
extern PTR_FILE currentFile;	/* point to the current active file */

/*
 * external functions from elsewhere in the cbaselib
 */
extern char *Allocate();
extern PTR_BFND FindCurrBifNode();
extern PTR_LLND FindLLNode();
extern PTR_SYMB FindSymbolNode();
extern void LiftDepInfo();
extern void ReNumAfterDeletion();
extern void ReNumAfterInsertion();
extern void CountLines();
extern PTR_LLND CopyLLNode();

static PTR_BFND last_thread(p)
PTR_BFND p;
{
	PTR_BLOB b;
	PTR_BFND c;
	c = p;
	b = p->entry.template.bl_ptr2;
	if(b == NULL) b = p->entry.template.bl_ptr1;
	if(b == NULL) return c;
	while(b && b->next) b = b->next;
	return last_thread( b->ref );
}

/*
* returns either FORTRAN_SRC or C_SRC 
*/
int SiLanguage()
{
  return currentFile->lang;
}

/* (static)
 * make_subroutine_name: help function for SiMakeSubroutineName,
 *         SiMakeNewSubroutine etc.
 */
int make_subroutine_name(vname)
char *vname;
{
   PTR_SYMB s;
   s =  make_symb(currentFile, PROCEDURE_NAME, vname);
   return s->id;
} /* make_subroutine_name */
/****************************************************************
 *								*
 * SiMakeSubroutineName - adds a new subroutine name to the     *
 *    current file with the given string identifier and returns *
 *    symbol id.						*
 * Uses help routine make_subroutine_name for all the work      *
 *								*
 * Input -							*
 *   vname - name of the subroutine				*
 *								*
 * Output -							*
 *   id of the newly created subroutine				*
 *								*
 ****************************************************************/
int SiMakeSubroutineName(vname)
char *vname;
{
   return make_subroutine_name(vname);
} /* SiMakeSubroutineName */



/*
 * (static)
 * make_function_name: used by SiMakeFunctionName SiMakeNewFunction etc.
 */
static int make_function_name(vname, type)
char *vname;
int type;
{
   PTR_SYMB s;

   s = make_symb(currentFile, FUNCTION_NAME, vname);
   s->type = FindTypeNode(type);

   return s->id;
} /* make_function_name */
/****************************************************************
 *								*
 * SiMakeFunctionName - adds a new function name to the current *
 *     file with the given string identifier and returns the    *
 *     symbol id.  The type of the function is identified by    *
 *     type.							*
 *     All the work is done through the help routine 		*
 *       make_function_name					*
 *								*
 * Input -							*
 *   vname - name of the new function to be added		*
 *   type - type of the new function to be added		*
 *								*
 * Output -							*
 *   id of the newly added function.				*
 *								*
 ****************************************************************/
int SiMakeFunctionName(vname, type)
char *vname;
int type;
{
   return make_function_name(vname, type);
} /* SiMakeFunctionName */


/****************************************************************
 *								*
 * SiVarName -- Returns elist string identifier for symbol 	*
 *   with integer identifier id. 				*
 *								*
 * Input -							*
 *   id of the symbol						*
 *								*
 * Output -							*
 *   ELIST with the corresponding string identifier		*
 *								*
 ****************************************************************/
ELIST SiVarName(id)
int id;
{
    PTR_SYMB sym;

    for (sym = currentFile->head_symb; sym; sym = sym->thread)
       if (sym->id == id) 
       {
	  return EString(sym->ident);
	}
    return NULL;
} /* SiVarName */


/****************************************************************
 *								*
 *  SiStmtAtLine -- find the bif node id of the statement at	*
 *		    line "line_no" in the current active file	*
 *								*
 *  Input:							*
 *	line_no -- the line number				*
 *								*
 *  Output:							*
 *	bif node id if the stmt was found or -1 for no          *
 *      statement at line_no.			                *
 *								*
 ****************************************************************/
int SiStmtAtLine(line_no)
int line_no;
{
    PTR_BFND bf;

    if ((bf = FindBifNode(currentFile->filename, line_no)))
       return bf->id;
    else
       return -1;
} /* SiStmtAtLine */


/****************************************************************
 *								*
 *  SiUnparseStmt -- get the unparsed string of the bif node	*
 *								*
 *  Input:							*
 *	the statement id                                        *
 *								*
 *  Output:							*
 *      NULL if bif node is not found                           *
 *      ELIST ptr otherwise                                     *
 *                                                              *
 ****************************************************************/
ELIST SiUnparseStmt(id)
    int      id;
{
    PTR_BFND bf;
    char    *ch;
    ELIST e;

    if ((bf = FindCurrBifNode(id))) 
    {
       ch = (UnparseBfnd[currentFile->lang])(bf);
       e = EString(ch);
       free(ch);
       return e;
    } 
    else
       return NULL;
} /* SiUnparseStmt */


/****************************************************************
 *								*
 * SiGetStmtSymbol -- get the symbol field in this statement	*
 *   could be do loop id or name a subroutine in call statment	*
 *                      					* 
 *   Input  							*
 *       stmt_id - statement id 				*
 *								*
 *   Output							*
 *       id of the symbol field                                 *       
 *       -1 if stmt_id is invalid or if symbol is not present   *
 *								*
 ****************************************************************/
int SiGetStmtSymbol(stmt_id)
int stmt_id;
{
   PTR_BFND bf;
   ELIST e1, e2;

   if (bf = FindCurrBifNode(stmt_id)) 
   {
      if (bf->entry.template.symbol)
	 return bf->entry.template.symbol->id;
      else 
         return -1;
   } 
   else
      return -1;
} /* SiGetStmtSymbol */
	

/****************************************************************
 *								*
 * SiGetBothSides -- get the left-hand-side and right-hand-side *
 *           expressions of a statement                         *
 *								*
 *  Input:							*
 *	stmt_id - id of the statment                            *
 *     								*
 *  Output:							*
 *	ELIST with both  the expressions in it                  *
 *      NULL if stmt_id is not valid                            *
 *								*
 ****************************************************************/
ELIST SiGetBothSides(stmt_id)
int stmt_id;
{
   PTR_BFND bf;
   ELIST e1, e2;

   if (bf = FindCurrBifNode(stmt_id)) 
   {
      if (bf->entry.template.ll_ptr1 != NULL) 
	 e1 = ENumber(bf->entry.template.ll_ptr1->id);
      else 
         e1 = ENumber(-1);
      if (bf->entry.template.ll_ptr2 != NULL) 
	 e2 = ENumber(bf->entry.template.ll_ptr2->id);
      else 
         e2 = ENumber(-1);
      return EAppend(e1,e2);
   } 
   else
      return NULL;
} /* SiGetBothSides */

/****************************************************************
 *								*
 *  SiReplaceLeftSide -  Replace the left hand expression in    *
 *    a statement by a new expression                           * 
 *								*
 *  Input:							*
 *	stmt_id - id of the statement in question               *
 *	expr    - the expression which replaces the left hand   *
 *                                                              *
 *  Output:							*
 *	0 if success, -1 if failure                             *
 *								*
 ****************************************************************/
int SiReplaceLeftSide(stmt_id, expr)
int stmt_id, expr;
{
   PTR_BFND bf;
   PTR_LLND ln;
   if ((bf = FindCurrBifNode(stmt_id)) == NULL) 
      return -1;
   if ((ln = FindLLNode(expr)) == NULL) 
      return -1;
   bf->entry.template.ll_ptr1 = ln;
   return 0;
} /* SiReplaceLeftSide */

/****************************************************************
 *								*
 *  SiReplaceRightSide -  Replace the right hand expression in  *
 *    a statement by a new expression                           * 
 *								*
 *  Input:							*
 *	stmt_id - id of the statement in question               *
 *	expr    - the expression which replaces the right hand  *
 *                                                              *
 *  Output:							*
 *	0 if success, -1 if failure                             *
 *								*
 ****************************************************************/
int SiReplaceRightSide(stmt_id, expr)
int stmt_id, expr;
{
   PTR_BFND bf;
   PTR_LLND ln;

   if ((bf = FindCurrBifNode(stmt_id)) == NULL) 
      return -1;
   if ((ln = FindLLNode(expr)) == NULL) 
      return -1;
   bf->entry.template.ll_ptr2 = ln;
   return 0;
} /* SiReplaceRightSide */


/****************************************************************
 *								*
 *  SiUnparseBlock -- get the unparsed string of the bif nodes	*
 *		     in the block starting with the given one	*
 *								*
 *  Input:							*
 *	id - the statement id                                   *
 *								*
 *  Output:							*
 *      the ELIST ptr if the node is found otherwise NULL       *
 *								*
 ****************************************************************/
ELIST SiUnparseBlock(id)
int id;
{
    PTR_BFND bf;
    ELIST e;
    char    *ch;

    if (bf = FindCurrBifNode(id)) 
    {
       ch = (UnparseBlock[currentFile->lang])(bf);
       e = EString(ch);
	free(ch);
	return e;
    } 
    else
       return NULL;
} /* SiUnparseBlock */


/****************************************************************
 *								*
 *  SiCopyStmt -- make a copy of existing statement		*
 *								*
 *  Input:							*
 *	stmt-id -- the id of the statement to be copied 	*
 *								*
 *  Output:							*
 *	the id of the newly created statement			*
 *								*
 ****************************************************************/
int SiCopyStmt(id)
int      id;
{
   PTR_BFND b1, b2;
   PTR_LLND ll1, ll2, ll3;

   if ((b1 = FindCurrBifNode(id))) 
   {
      ll1 = CopyLLNode(b1->entry.template.ll_ptr1);
      ll2 = CopyLLNode(b1->entry.template.ll_ptr2);
      ll3 = CopyLLNode(b1->entry.template.ll_ptr3);
      b2 = make_bfndnt(currentFile, b1->variant, b1->entry.template.symbol, 
                     ll1, ll2, ll3);
      b2->filename = b1->filename;
      SiAddToStmtTable(b2);
      return b2->id;
    }
    return -1;
} /* SiCopyStmt */


/****************************************************************
 *								*
 *  SiGetStmtInfo -- get information about a statement          *
 *								*
 *  Input:							*
 *	id  --  id of the stmt                                  *
 *                                                              *
 *  Output:							*
 *	the list ( control_parent symbol expr1 expr2 expr3 )    *
 *      of information in the statement                         *
 *      if the statement node is not found NULL is returned     *
 *								*
 ****************************************************************/
ELIST SiGetStmtInfo(id)
int id;
{
    PTR_BFND bf;
    ELIST e;

    if (bf = FindCurrBifNode(id))
    {
	e = NULL;
	e = EAppend(e, 
                    ENumber( bf->control_parent? bf->control_parent->id: -1));
	e = EAppend(e, ENumber(
	       bf->entry.template.symbol? bf->entry.template.symbol->id: -1));
	e = EAppend(e, ENumber(
	       bf->entry.template.ll_ptr1? bf->entry.template.ll_ptr1->id: -1));
	e = EAppend(e, ENumber(
	       bf->entry.template.ll_ptr2? bf->entry.template.ll_ptr2->id: -1));
	e = EAppend(e, ENumber(
	       bf->entry.template.ll_ptr3? bf->entry.template.ll_ptr3->id: -1));
	return e;
    }
    else
       return NULL;	
} /* SiGetStmtInfo */


/****************************************************************
 *								*
 *  SiStmtVariant -- inquire the variant of a given statement	*
 *								*
 *  Input:							*
 *	id -- the bif node id of the statement			*
 *								*
 *  Output:							*
 *      variant type of the statement                           *
 *								*
 ****************************************************************/
int SiStmtVariant (id)
int id;
{
   PTR_BFND b;

   if (b = FindCurrBifNode(id))
      return b->variant;
   else
      return -1;
} /* SiStmtVariant */


/****************************************************************
 *								*
 *  SiPreviousStmt -- get the bif node id for the previous	*
 *		     statement					*
 *								*
 *  Input:							*
 *	stmt_id -- bif node id of the current statement		*
 *								*
 *  Output:							*
 *      id of the textually previous statement                  *
 *	-1 if failure                                           *
 *								*
 ****************************************************************/
int SiPreviousStmt(id)
int id;
{
   PTR_BFND cp, bf;
   PTR_BLOB bl1, bl2;

   bf = FindCurrBifNode(id);
   cp = bf->control_parent;
   for (bl1 = bl2 = cp->entry.template.bl_ptr1; 
        bl1 && bl1->ref != bf; bl1 = bl1->next)
      bl2 = bl1;
   if (bl1->ref == bf) 
   {
      return bl2->ref->id;
   }
    for (bl1 = bl2 = cp->entry.template.bl_ptr2; 
         bl1 && bl1->ref != bf; bl1 = bl1->next)
       bl2 = bl1;
    if (bl1->ref == bf)
       return bl2->ref->id;
    else
       return -1;
} /* SiPreviousStmt */

/****************************************************************
 *								*
 *  SiChangeStmtVariant -- change the variant of a statement    *
 *    to a value given by var					*
 *								*
 *  Input:							*
 *	stmt_id - id of the statement                           *
 *      var - is the value to which variant is to be changed to *
 *  Output:							*
 *	0 for success, -1 for failure                           *
 *								*
 ****************************************************************/

int SiChangeStmtVariant(stmt_id, var)
int stmt_id, var;
{
   PTR_BFND bf;
   if (bf = FindCurrBifNode(stmt_id))
   {
      bf->variant = var;
      return 0;
   }
   else
      return -1;
} /* SiChangeStmtVariant */


/****************************************************************
 *								*
 *  SiControlParent -- get the bif node id for the control	*
 *		      parent of the current statement		*
 *								*
 *  Input:							*
 *	id -- bif node id of the current statement		*
 *								*
 *  Output:							*
 *	 bif node id of the control parent                      *
 *	 -1  if anything wrong				        *
 *								*
 ****************************************************************/
int SiControlParent (id)
int id;
{
    PTR_BFND bf;

    if ((bf = FindCurrBifNode(id)) && bf->control_parent)
       return bf->control_parent->id;
    else
       return -1;
} /* SiControlParent */


/* static functions used by SiExpandChildren */
static ELIST rec_print_children  ();
ELIST static rec_print_child1(bf)
PTR_BFND bf;
{
   PTR_BLOB b;
   ELIST e1, e2;

   e1 = NULL;
   for (b = bf->entry.template.bl_ptr1; b; b = b->next)
      if (b->ref->entry.template.bl_ptr1 || b->ref->entry.template.bl_ptr2)
	 e1 = EAppend(e1,rec_print_children(b->ref));
      else
	 e1 = EAppend(e1, ENumber( b->ref->id));
   e1 = EAppend(ENumber(bf->id), ECons(e1, NULL));
   e2 = NULL;
   for (b = bf->entry.template.bl_ptr2; b; b = b->next)
      if (b->ref->entry.template.bl_ptr1 || b->ref->entry.template.bl_ptr2)
	 e2 = EAppend(e2,rec_print_children(b->ref));
      else
	 e2 = EAppend(e2, ENumber( b->ref->id));
   e1 = EAppend(e1, ECons(e2, NULL));
   return e1;
} /* rec_print_child1 */

static ELIST rec_print_children(bf)
PTR_BFND bf;
{
    return rec_print_child1(bf);
} /* rec_print_children */


/****************************************************************
 *								*
 *  SiExpandChildren -- get a list of control children of	*
 *			current statement			*
 *								*
 *  Input:							*
 *	stmt_id -- bif node id of the current statement	        *
 *								*
 *  Output:							*
 *	a list of id's for control children	                *
 *	This is a tree represented as a	                        *
 *	list as in						*
 *								*
 *	  ((3 5 ((6  7	8) ()) 9  10 11) (12 13 14))		*
 *								*
 *	where the first list corresponds to the first branch	*
 *	while the second list corresponds to the second branch	*
 *								*
 ****************************************************************/
ELIST SiExpandChildren (id)
int id;
{
   PTR_BFND bf;

   if (!(bf = FindCurrBifNode(id))) 
   {
      return NULL;
   }
   return rec_print_children(bf);
} /* SiExpandChildren */


/****************************************************************
 *								*
 *  SiControlChildren - returns the list of the toplevel nodes  *
 *       control children of stmt_id                            *
 *								*
 *  Input:							*
 *      stmt_id - statement id of the node                      *
 *                                                              *
 *  Output:							*
 *	description of the return expression ....               *
 *								*
 ****************************************************************/
ELIST SiControlChildren(stmt_id)
int stmt_id;
{
    PTR_BFND bf;
    register PTR_BLOB b;
    ELIST e1, e2;

    if (!(bf = FindCurrBifNode(stmt_id))) 
    {
       return NULL;
    }
    e1 = NULL;
    for (b = bf->entry.template.bl_ptr1; b; b = b->next)
       e1 = EAppend(e1, ENumber(b->ref->id));
    e2 = NULL;
    for (b = bf->entry.template.bl_ptr2; b; b = b->next)
       e2 = EAppend(e2, ENumber( b->ref->id));
    return EAppend(ECons(e1,NULL), ECons(e2,NULL));
} /* SiControlChildren */


/****************************************************************
 *								*
 *  SiLineNumber -- get the line number of a given big node	*
 *								*
 *  Input:							*
 *	id - the id of the bif node                       	*
 *								*
 *  Output:							*
 *	source line number of this node                         *
 *	-1 in case of error. 				        *
 *								*
 ****************************************************************/
int SiLineNumber(id)
int id;
{
   PTR_BFND bf;

   if (bf = FindCurrBifNode(id))
      return bf->g_line;
   else
      return -1;
} /* SiLineNumber */


/****************************************************************
 *								*
 *  SiScopeLineNumbers -- find the span of a given bif node	*
 *								*
 *  Input:							*
 *	id -- bif node id					*
 *								*
 *  Output:							*
 *	An ELIST of two numbers that give the			*
 *	scope of this node, i.e. the first is the id of the	*
 *	first bif node in this scope and the second is the	*
 *	id of the last bif node in the scope.  			*
 *								*
 ****************************************************************/
ELIST SiScopeLineNumbers(id)
int id;
{
   PTR_BFND bf, bf1;
   PTR_BLOB bl1, bl2;
   int found = 0;
   ELIST e;

   if (bf = FindCurrBifNode(id)) 
   {
      bf1 = (bf->variant == GLOBAL)? bf: bf->control_parent;
      /* Check to see if it's in the first branch */
      for (bl1 = bl2 = bf1->entry.template.bl_ptr1; bl1 ; bl1 = bl1->next) 
      {
	 if (bl1->ref->id == id)
            found = 1;
	 bl2 = bl1;
      }
      if (found) 
      {
	 e = EAppend(ENumber( bf1->id), ENumber(bl2->ref->id));
	 return e;
      }

	/* Check to see if it's in the second branch */
      for (bl1 = bl2 = bf1->entry.template.bl_ptr2; bl1 ; bl1 = bl1->next) 
      {
	 if (bl1->ref->id == id)
	    found = 1;
	 bl2 = bl1;
      }
      if (found) 
      {
	 e = EAppend(ENumber(bf1->id), ENumber(bl2->ref->id));
	 return e;
      }
    } 
    return NULL;     
} /* SiScopeLineNumbers */



/****************************************************************
 *  (static)
 *  make_new_routine - make a new routine node. used by         *
 *      SiMakeNewSubroutine etc.                                *
 *								*
 *  Input:							*
 *	symbol is the symbol for the new routine                *
 *      ftype is type: subroutine, function ...                 *
 *								*
 *  Output:							*
 *	pointer to the newly made bif node                      *
 *								*
 ****************************************************************/
static PTR_BFND make_new_routine(symbol, ftype)
PTR_SYMB symbol;
int ftype;
{
   PTR_BLOB bb, make_blob();
   PTR_BFND bf, cntl_end, make_bfnd(); /* bfnd is added to end so thread is o.k. */

   if ((bf = make_bfnd(currentFile,ftype,symbol,LLNULL, LLNULL, LLNULL)) &&
      (cntl_end = 
        make_bfnd(currentFile,CONTROL_END, SMNULL, LLNULL, LLNULL, LLNULL)))
   {
      SiAddToStmtTable(bf); SiAddToStmtTable(cntl_end);

      if (currentFile->global_bfnd->entry.template.bl_ptr1 == NULL)
      {
	 currentFile->global_bfnd->entry.template.bl_ptr1 =
 	    make_blob(currentFile, bf, (PTR_BLOB)NULL);
      }
      else
      {
	 bb = currentFile->global_bfnd->entry.template.bl_ptr1;
	 while(bb->next != NULL) bb = bb->next;
	 bb->next = make_blob(currentFile, bf, (PTR_BLOB)NULL);
      }
      bf->control_parent = currentFile->global_bfnd;
      bf->filename = cntl_end->filename = currentFile->global_bfnd->filename;
      bf->entry.template.bl_ptr1 
        = make_blob(currentFile,cntl_end,(PTR_BLOB) NULL);
      cntl_end->control_parent = bf;
      symbol->entry.proc_decl.proc_hedr = bf;
      return bf;
  }
  return NULL;
} /* make_new_routine */

/****************************************************************
 *								*
 *  SiMakeNewSubroutine                                         *
 *     Creates new subroutine with the given name               *
 *     It has no parameters but it does have a CONTROL_END(end) *
 *     statement. The new subroutine is logically added to the  *
 *     end of the current file but the linenumber is 0.         *
 *								*
 *  Input:							*
 *	name - name of the new subroutine                       *
 *								*
 *  Output:							*
 *	id of the newly created node, -1 indicating failure     *
 *								*
 ****************************************************************/
int SiMakeNewSubroutine(name)
char *name;
{
  PTR_BFND b;
  PTR_SYMB s;
  int symb_id;
  
  if ((s = FindSymbolNode(make_subroutine_name(name))) == NULL) 
     return -1;
  b = make_new_routine(s,PROC_HEDR);
  if (b == NULL) 
     return -1;
  else 
     return b->id;
} /* SiMakeNewSubroutine */


/****************************************************************
 *								*
 *  SiMakeNewFunction                                           *
 *     Creates new function with the given name                 *
 *     It has no parameters but it does have a CONTROL_END(end) *
 *     statement. The new function is logically added to the    *
 *     end of the current file but the linenumber is 0.         *
 *								*
 *  Input:							*
 *	name - name of the new subroutine                       *
 *      type - return type of the new function                  *
 *								*
 *  Output:							*
 *	id of the newly creatd node, -1 indicating failure      *
 *								*
 ****************************************************************/
int SiMakeNewFunction(name, type)
char *name;
int type;
{
  PTR_BFND b;
  PTR_SYMB s, FindSymbolNode();
  int symb_id;
  
  if ((s = FindSymbolNode(make_function_name(name,type))) == NULL) 
     return -1;
  b = make_new_routine(s,FUNC_HEDR);
  if (b == NULL) 
     return -1;
  else 
     return b->id;
} /* SiMakeNewFunction */


/****************************************************************
 *  (static)                                  			*
 *  make_loop_node - This is a static routine used by SiNewLoop *
 *								*
 *  Input:							*
 *	var - indicates the type of the loop to be made         *
 *      index - symbol table ptr of the index variable          *
 *      low, hi, inc - range of the loop                        *
 *      body - loop body                                        *
 *                                                              *
 *  Output:							*
 *	returns a pointer to the created node, NULL is failure  *
 *								*
 ****************************************************************/
static PTR_BFND make_loop_node(var, index, low, hi, inc, body)
    int      var;
    PTR_SYMB index;
    PTR_LLND low, hi, inc;
    int body[];
{
    PTR_LLND range;
    PTR_BFND bf, b2,thptr;
    PTR_BLOB bl, bb;
    int *bp;
  
    bb = (PTR_BLOB) NULL;
    if ((range = make_llnd(currentFile, DDOT, low, hi, SMNULL)) &&
	(bf = make_bfndnt(currentFile, var, index, range, inc, LLNULL))) 
    {
       bf->filename = currentFile->global_bfnd->filename;
       SiAddToStmtTable(bf);
       bp = body;
       thptr = (PTR_BFND) NULL; 
       while (*bp != -1)
           if (b2 = FindCurrBifNode(*bp)) 
          {
             b2->control_parent = bf;
	     if (bp == body){ 	/* The first one */
		bf->thread = b2; thptr = last_thread(b2);
	        bb = bl = make_blob(currentFile, b2, (PTR_BLOB)NULL);
		}
	     else 
             {
	        thptr->thread = b2; thptr = last_thread(b2);
	        bl->next = make_blob(currentFile, b2, (PTR_BLOB)NULL);
	        bl = bl->next;
	     }
             bp++;

          }

       if (b2 = make_bfndnt(currentFile,CONTROL_END,SMNULL,LLNULL,LLNULL,LLNULL))
       {
	  SiAddToStmtTable(b2);
	  if(thptr){ 
		bf->thread = b2;
		}
	  else 
		thptr->thread = b2;
	  b2->filename = currentFile->global_bfnd->filename;
	  bl->next = make_blob(currentFile, b2, (PTR_BLOB)NULL);
	  bf->entry.template.bl_ptr1 = bb;
	  return bf;
       }
     }
     return NULL;
} /* make_loop_node */


/****************************************************************
 *								*
 *  SiNewLoop-- make a new loop 				*
 *								*
 *  Input:							*
 *	type - integer specifying the variant of the loop	*
 *	ivar - symbol table id of the induction variable	*
 *	low  - id of the expression tree (low level node) that	*
 *	       represents the lower bound of the loop		*
 *	hi   - id of the expression tree (low level node) that	*
 *	       represents the upper bound of the loop		*
 *	inc  - id of the expression tree (low level node) that	*
 *	       represents the increment of the loop.  Should be *
 *	       0 if no increment specified			*
 *	body - list of id's of the body 			*
 *								*
 *  Output:							*
 *     id of the newly created node, -1 indicates failure       *
 *								*
 ****************************************************************/
int SiNewLoop(e)
ELIST e;
{
    int type, ivar, low, hi, inc, err;
    PTR_BFND bf;
    PTR_SYMB sym;
    PTR_LLND l1, l2, l3;
    int body[200], *bp;
    char ch;

   /* below three line is added by zjd in 95.8.9 in order to correct the false in this function  */
    int i;
    for(i=0;i<200;i++)
	body[i]=-1;
    

    type = (int) e->car; e = e->cdr;
    ivar = (int) e->car; e = e->cdr;
    low  = (int) e->car; e = e->cdr;
    hi	 = (int) e->car; e = e->cdr;
    inc  = (int) e->car; e = e->cdr;

 /* Now read in the list that represents body */
    err= 0;
    if (! err) 
    {
       for (bp = body; e != NULL; e = e->cdr)
	  *bp++ = (int) e->car;

       if ((sym = FindSymbolNode(ivar)) && (l1  = FindLLNode(low)) &&
	    (l2  = FindLLNode(hi)) && (bp != body)) 
       {
	  l3 = inc? FindLLNode(inc): NULL;
	  bf	= make_loop_node(type, sym, l1, l2, l3, body);
	  if (bf) 
          {
		return bf->id;
	  }
       }
    }
    return -1;
} /* SiNewLoop */

/****************************************************************
 *								*
 *  SiFormalArgs -- get list of ll expr ids for the formal args	*
 *		   for a function				*
 *   Input -    						*
 *       stmt_id - id of the the subroutine/function call site  *
 *           or header						*
 *								*
 *   Output -							*
 *       This returns an ELIST of symbol pointer ids. 		*
 *       NULL indicates failure.				*
 ****************************************************************/
ELIST SiFormalArgs (stmt_id)
int stmt_id;
{
   PTR_BFND bf;
   PTR_SYMB s;
   ELIST e;
	
    e = NULL;
    if (bf = FindCurrBifNode(stmt_id))
    {
       s = bf->entry.template.symbol;
       s = s->entry.proc_decl.in_list;
       while (s != NULL)
       {  /* gather formal args in formal_args */
	  e = EAppend(e,ENumber(s->id));
	  s = s->entry.var_decl.next_in;
       }
       return e;
    }
    else
       return NULL;
} /* SiFormalArgs */


/****************************************************************
 *								*
 * SiPassedArgs -- finds exprs id for passed arguments	for	*
 *       a call site.                                           *
 *								*
 *  Input:							*
 *	stmt_id  --  id of the call site                        *
 *                                                              *
 *  Output:							*
 *	returns the EXPR_LIST id of the actual arguments        *
 *								*
 ****************************************************************/

int SiPassedArgs (stmt_id)
int stmt_id;
{
    PTR_BFND bf;

    if (bf = FindCurrBifNode(stmt_id))
       return bf->entry.template.ll_ptr1? bf->entry.template.ll_ptr1->id: 0;
    else
       return -1;
} /* SiPassedArgs */


/****************************************************************
 *								*
 * SiRemoveControlEnd -- removes the CONTROL_END node from the  *
 *		   end of the block.				*
 *								*
 * Input:							*
 *	stmt_id - the bif node id of the statement interested	*
 *								*
 * Output:							*
 *								*
 ****************************************************************/
int SiRemoveControlEnd (stmt_id)
int stmt_id;
{
    PTR_BFND bf;
    PTR_BLOB bl, bp;

    if (bf = FindCurrBifNode(stmt_id)) 
    {
       bp = NULL;
       for (bl = bf->entry.template.bl_ptr1; bl; bl = bl->next) 
       {
	  if (bp != NULL && bl->ref->variant == CONTROL_END) 
          {
	     last_thread(bp->ref)->thread = bl->ref->thread;
	     bp->next = bl->next;
	     break;
	  }
	  bp = bl;
       }
       return 0;
    } 
    else 
       return -1;
} /* SiRemoveControlEnd */


/****************************************************************
 *								*
 *  SiNewStmt -- create new statement with given information	*
 *								*
 *  Inputs:                                                     *
 *       e - (variant, symbol, ll1, ll2, ll3) all ids as elist  *
 *								*
 *  Output:							*
 *       id of the newly created statement                      *
 *	 -1 if failure 						*
 *								*
 ****************************************************************/
int SiNewStmt(e)
ELIST e;
{
   int var, sym, id1, id2, id3;
   PTR_BFND new;
   PTR_LLND l1, l2, l3 = LLNULL;
   PTR_SYMB s;

   var = (int) e->car; e = e->cdr;
   sym = (int) e->car; e = e->cdr;
   id1 = (int) e->car; e = e->cdr;
   id2 = (int) e->car; e = e->cdr;
   id3 = (int) e->car; e = e->cdr;
   s = (sym == -1)? NULL: FindSymbolNode(sym);
   l1= (id1 == -1)? NULL: FindLLNode(id1);
   l2= (id2 == -1)? NULL: FindLLNode(id2);
   l3= (id3 == -1)? NULL: FindLLNode(id3);    
   if ((new = make_bfndnt(currentFile, var, s, l1, l2, l3))) {
      SiAddToStmtTable(new);
      return new->id;
      }
   else 
      return -1;
} /* SiNewStmt */


/****************************************************************
 *								*
 *  SiNewStmtInt -- create new statement with given information *
 *								*
 *  Inputs: variant, symbol, ll1, ll2, ll3 all ids as integers	*
 *								*
 *  Output:							*
 *       id of the newly created statement                      *
 *	 -1 if failure 						*
 *								*
 ****************************************************************/
int SiNewStmtInt(var, sym, id1, id2, id3)
int var, sym, id1, id2, id3;
{

   PTR_BFND new;
   PTR_LLND l1, l2, l3 = LLNULL;
   PTR_SYMB s;

   s = (sym == -1)? NULL: FindSymbolNode(sym);
   l1= (id1 == -1)? NULL: FindLLNode(id1);
   l2= (id2 == -1)? NULL: FindLLNode(id2);
   l3= (id3 == -1)? NULL: FindLLNode(id3);    
   if (new = make_bfndnt(currentFile, var, s, l1, l2, l3)) {
      SiAddToStmtTable(new);
      return new->id;
      }
   else 
      return -1;
} /* SiNewStmtInt */

int SiAddTrueChild(stmt, chld)
int stmt, chld;
{
	PTR_BLOB b,nb;
	PTR_BFND s, c;
	if((s = FindCurrBifNode(stmt)) == NULL) return -1;
	if((c = FindCurrBifNode(chld)) == NULL) return -1;
	b = s->entry.template.bl_ptr1;
	while(b && b->ref->variant != CONTROL_END &&
              b->next != NULL && b->next->ref->variant != CONTROL_END)
		{ 
		 b  = b->next;
		 }
	if(b && b->ref->variant == CONTROL_END){
	  nb = make_blob(currentFile, c, b);
	  s->entry.template.bl_ptr1 = nb;
	  last_thread(c)->thread = s->thread;
	  s->thread = c;
	  return chld;
          }	
	/* otherwise b is null ... */
	if(b == NULL){
	  nb = make_blob(currentFile, c, NULL);
          s->entry.template.bl_ptr1 = nb; 
	  last_thread(c)->thread = s->thread;
          s->thread = c;
	  return chld;
	  }
        /* or b is pointing to the node that precedes the cntl end */
        /* (if there is one) and c goes just after b */
	else {
	  nb = make_blob(currentFile, c, b->next);
          last_thread(c)->thread = last_thread(b->ref)->thread;
	  last_thread(b->ref)->thread = c;
          b->next = nb; 
	  return chld;
	}
}

int SiAddToCntlEnd(stmt,child)
int stmt, child;
{
	return SiAddTrueChild(stmt, child);
}
	
int SiAddFalseChild(stmt, chld)
int stmt, chld;
{
	PTR_BLOB tb, b,nb;
	PTR_BFND pt, s, c;
	if((s = FindCurrBifNode(stmt)) == NULL) return -1;
	if((c = FindCurrBifNode(chld)) == NULL) return -1;
	b = s->entry.template.bl_ptr2;
	while(b && b->ref->variant != CONTROL_END &&
              b->next != NULL && b->next->ref->variant != CONTROL_END)
		{ 
		 b  = b->next;
		 }
	tb = s->entry.template.bl_ptr1;
	while(tb && tb->next != NULL) tb = tb->next;
	if(tb) pt = last_thread(tb->ref);
	else pt = s;

	if(b && b->ref->variant == CONTROL_END){
	  nb = make_blob(currentFile, c, b);
	  s->entry.template.bl_ptr2 = nb;
	  last_thread(c)->thread = pt->thread;
	  pt->thread = c;
	  return chld;
          }	
	/* otherwise b is null ... */
	if(b == NULL){
	  nb = make_blob(currentFile, c, NULL);
          s->entry.template.bl_ptr2 = nb; 
	  last_thread(c)->thread = pt->thread;
          pt->thread = c;
	  return chld;
	  }
        /* or b is pointing to the node that precedes the cntl end */
        /* (if there is one) and c goes just after b */
	else {
	  nb = make_blob(currentFile, c, b->next);
          last_thread(c)->thread = last_thread(b->ref)->thread;
	  last_thread(b->ref)->thread = c;
          b->next = nb; 
	  return chld;
	}
}
	

/****************************************************************
 *								*
 *  SiModifyLoopStmt - given an elist (id start end inc), the   *
 *    loop identified by id is modified so that it has new range*
 *    given by start end and inc.                               *
 *								*
 *  Input:							*
 *	e - the elist (id start end inc)                        *
 *								*
 *  Output:							*
 *	0 for success, -1 for failure                           *
 *								*
 ****************************************************************/
int SiModifyLoopStmt(e)
ELIST e;
{
   int id, start, end, inc, error=0;
   PTR_BFND bf;
   PTR_LLND ls, le, li;

   id	  = (int) e->car; e = e->cdr;
   start = (int) e->car; e = e->cdr;
   end   = (int) e->car; e = e->cdr;
   inc   = (int) e->car; e = e->cdr;
   if (bf = FindCurrBifNode(id)) 
   {
      if (start > 0)
	 if (ls = FindLLNode(start))
	    bf->entry.template.ll_ptr1->entry.template.ll_ptr1 = ls;
	 else
	    error = -1;
      if (end > 0)
	 if (le = FindLLNode(end))
	    bf->entry.template.ll_ptr1->entry.template.ll_ptr2 = le;
	 else
	    error = -1;
      if (inc > 0)
	 if ((le = FindLLNode(inc)))
	    bf->entry.template.ll_ptr2 = li;
	 else
	    error = -1;
    } 
    else
       error = -1;
    return error;
} /* SiModifyLoopStmt */

static PTR_BFND find_previous_thread(b)
PTR_BFND b;
{
	PTR_BFND p;
	p = currentFile->global_bfnd;
	while(p && p->thread != b) p = p->thread;
	if(b != NULL) return p;
	else return NULL;
}

static PTR_BFND find_next_thread(b)
PTR_BFND b;
{
	PTR_BFND p;
	PTR_BLOB bl;
	if(b->entry.template.bl_ptr2){
		bl = b->entry.template.bl_ptr2;
		while(bl && bl->next != NULL) bl = bl->next;
		p = find_next_thread(bl->ref);
		return p->thread;
		}
	if(b->entry.template.bl_ptr1){
		bl = b->entry.template.bl_ptr1;
		while(bl && bl->next != NULL) bl = bl->next;
		p = find_next_thread(bl->ref);
		return p->thread;
		}
	return b->thread;
}
      		 
/****************************************************************
 *								*
 * SiDeleteStmt -- removes the bif node from the list		*
 *								*
 * Input:							*
 *	stmt_id - the bif node id of the statement interested	*
 *								*
 * Output:							*
 *     id of statement preceding deleted item or control parent *
 * if the first one on list.                                    *
 *								*
 ****************************************************************/
int SiDeleteStmt(stmt_id)
int stmt_id;
{
    PTR_BFND cp, bf, prevt;
    PTR_BLOB bl1, bl2;

    if ((bf = FindCurrBifNode(stmt_id))) 
    {
       cp = bf->control_parent;

       /* search the first branch */
       bl2 = NULL;
       for (bl1 = cp->entry.template.bl_ptr1; bl1; bl1 = bl1->next) 
       {
	  if (bl1->ref == bf) 
          {
	     if(bl2) bl2->next = bl1->next;
	     else cp->entry.template.bl_ptr1 = bl1->next;

	     prevt = find_previous_thread(bf);
	     prevt->thread = find_next_thread(bf);	
	     ReNumAfterDeletion(bf);
	     if(bl2) return bl2->ref->id;
	     else return bf->control_parent->id;
	  }
	  bl2 = bl1;
       }

       /* search the second branch */
 
	bl2 = NULL;
	for (bl1 = cp->entry.template.bl_ptr2; bl1; bl1 = bl1->next) 
        {
	   if (bl1->ref == bf) 
           {
	     if(bl2) bl2->next = bl1->next;
	     else cp->entry.template.bl_ptr1 = bl1->next;
	     prevt = find_previous_thread(bf);
	     prevt->thread = find_next_thread(bf);	
	      ReNumAfterDeletion(bf);
	     if(bl2) return bl2->ref->id;
	     else return bf->control_parent->id;
	   }
	   bl2 = bl1;
	}
     }
     return -1;
} /* SiDeleteStmt */


/****************************************************************
 *								*
 *  SiInsertStmt -- insert a list of bif nodes			*
 *								*
 * Input: e - elist consisting of,				*
 *	ido - the bif node id of the statement after which	*
 *	      the new nodes will be inserted			*
 *	no  - number of nodes to be inserted			*
 *	id1 ... idn - the new bif nodes to be inserted		*
 *								*
 * Output:							*
 *	0 for success, -1 for failure                           *
 *								*
 ****************************************************************/
int SiInsertStmt (e)
ELIST e;
{
    int id1, id2, i, n, count = 1;
    PTR_BFND endth, last, bf1, bf2, cp, *bp = NULL, *bp1;
    PTR_BLOB bl1, bl2;

    id1 = (int) e->car; e = e->cdr;
    n	= (int) e->car; e = e->cdr;
    if (bf1 = FindCurrBifNode(id1)) 
    {
       cp = (bf1->variant == GLOBAL)? bf1: bf1->control_parent;

       /* search the first branch */
       bl2 = NULL;	
       for (bl1 = cp->entry.template.bl_ptr1; bl1; bl1 = bl1->next) 
       {
	  if (bl1->ref == bf1)
	     break;
	  bl2 = bl1;
       }

	/* search the second branch */
	if (bl1 == NULL)
	   for (bl1 = cp->entry.template.bl_ptr2; bl1; bl1 = bl1->next)
           {
	      if (bl1->ref == bf1)
		 break;
	      bl2 = bl1;
	   }

	if (bl1) 
        {
	    bp = bp1 =
		 (PTR_BFND *)Allocate((unsigned)n*(unsigned)sizeof(PTR_BFND *));
	    /*
	     * check all the given bif nodes exist before making changes
	     */
	    for (i = n; i > 0; --i) 
            {
	       id2 = (int) e->car; e = e->cdr;
	       if (!(*bp1++ = FindCurrBifNode(id2)))
		  goto error;
	    }

	    /*
	     * Now link them together
	     */
	    for (i = n, bp1 = bp; i > 0; --i) 
            {
	        last = last_thread(bf1);
		bf2 = *bp1++;
		bf2->control_parent = cp;
		bl2 = make_blob(currentFile, bf2, bl1->next);
		bl1->next = bl2;
		last_thread(bf2)->thread = last->thread;
		last->thread = bf2;
		CountLines(bf2, &count);
		bl1 = bl2;
		bf1 = bf2;
	    }
	    ReNumAfterDeletion(bf1, --count, bf1->g_line);
	    free((char *)bp);
	    return 0;
	}
    }

error:
    if (bp)
       free((char *)bp);
    return -1;
} /* SiInsertStmt */

/*********************************************************************/
/*  SiMakeCommonStat( function or sub header id, common block        */
/*     name string, id of expr list of variable references)          */
/* makes a common statement for the subroutine with the given common */
/* block name and the associated list of variable identifiers.       */
/*the common stat is inserted as the first statement in the subroutine*/
/**********************************************************************/

int SiMakeCommonStat(sub, name, vars)
int sub;
char * name;
int vars;
{
        PTR_BFND f, c;
        PTR_SYMB cb;
        PTR_LLND varlist, cmlst;
        PTR_BLOB b;

        if((f = FindCurrBifNode(sub)) == NULL) return -1;
        if((varlist = FindLLNode(vars)) == NULL) return -1;
        cb = make_symb(currentFile, DEFAULT, name);
	cb->scope = currentFile->global_bfnd;
        cb->next_symb = f->entry.template.symbol->next_symb;
        f->entry.template.symbol->next_symb = cb;
        cmlst = make_llnd(currentFile, COMM_LIST, varlist, LLNULL, cb);
        c = make_bfndnt(currentFile, COMM_STAT, NULL, cmlst, NULL, NULL, NULL);
        SiAddToStmtTable(c);
        b = make_blob(currentFile, c, f->entry.template.bl_ptr1);
	c->thread = f->thread;
	f->thread = c;
        f->entry.template.bl_ptr1 = b;
        c->control_parent = f;
        return c->id;
}
        


/****************************************************************
 *								*
 *  SiReplaceStmt -- replace the bif node stmt_id1 by        	*
 *      stmt_id2 						*
 *								*
 * Input: e - elist consisting of                               *
 *	stmt_id1 - the bif node id of the statement to be	*
 *		   replaced					*
 *	stmt_id2 - the new bif node to be used			*
 *								*
 * Output:							*
 *     id of the replaced node if success, -1 if  failure       *
 *								*
 ****************************************************************/
int SiReplaceStmt(e)
ELIST e;
{
   int stmt_id1, stmt_id2;
   PTR_BFND th, bf1, bf2, cp;
   PTR_BLOB bl1, bl2;

   stmt_id1 = (int) e->car; 
   stmt_id2 = (int) e->cdr->car; 
   if ((bf1 = FindCurrBifNode(stmt_id1)) &&
       (bf2 = FindCurrBifNode(stmt_id2))) 
   {
      cp = (bf1->variant == GLOBAL)? bf1: bf1->control_parent;

      /* search the first branch */
      bl2 = NULL;
      for (bl1 = cp->entry.template.bl_ptr1; bl1; bl1 = bl1->next) 
      {
	 if(bl2) th = last_thread(bl2->ref);
	 else th = cp;
	 if (bl1->ref == bf1) 
         {
	    th->thread = bf2;
	    bl1->ref = bf2;
	    last_thread(bf2)->thread = last_thread(bf1)->thread;
	    last_thread(bf1)->thread = NULL;
	    bf2->control_parent = cp;
	    if(bl2) return bl2->ref->id;
	    else return cp->id;
	 }
	 bl2 = bl1;
      }

      /* search the second branch */
      for (bl1 = cp->entry.template.bl_ptr2; bl1; bl1 = bl1->next) 
      {
	 if(bl2) th = last_thread(bl2->ref);
	 else th = cp;
	 if (bl1->ref == bf1) 
         {
	    th->thread = bf2;
	    bl1->ref = bf2;
	    last_thread(bf2)->thread = last_thread(bf1)->thread;
	    last_thread(bf1)->thread = NULL;
	    bf2->control_parent = cp;
	    if(bl2) return bl2->ref->id;
	    else return cp->id;
	 }
	 bl2 = bl1;
      }
    }
    return -1;
} /* SiReplaceStmt */


/****************************************************************
 *								*
 *  SiGetProperty -- obtain a property given by "prop" assoc.   *
 *      with statement with id "stmt_id"                        *
 *								*
 *  Input:							*
 *	stmt_id - id of the statement                           *
 *      prop - name of the property                             *
 *  Output:							*
 *	returns value of the property                           *
 *      0 indicates no property                                 *
 *      -1 indicates error                                      *
 *								*
 ****************************************************************/
int SiGetProperty(stmt_id,prop)
int stmt_id;
char *prop;
{
   int prop_val;
   PTR_BFND bf;

   if (bf = FindCurrBifNode(stmt_id))  /* bif node found */
   {
      if (prop_val = (int) bget_prop(bf, prop))
	 return prop_val; /* property there */
      else
	 return 0;	/* no property */
   }
   return -1;
} /* SiGetProperty */


/****************************************************************
 *								*
 *  SiPutProperty -- install a property given by "prop" whose.  *
 *      value is "val" with statement with id "stmt_id"         *
 *								*
 *  Input:							*
 *	stmt_id - id of the statement                           *
 *      val - is the value of the property                      *
 *      prop - name of the property                             *
 *  Output:							*
 *      0 indicates success                                     *
 *      -1 indicates error                                      *
 *								*
 ****************************************************************/
int SiPutProperty(stmt_id, val, prop_name)
int stmt_id, val; 
char *prop_name;
{
    char prop[100], *pp;
    PTR_BFND bf;
    PTR_PLNK pr;

    strcpy(prop,prop_name);
    if (bf = FindBifNode(currentFile->filename, stmt_id))
       if ((pr = (PTR_PLNK) Allocate(sizeof(struct prop_link))) &&
	    (pp = Allocate(strlen(prop) + 1))) 
       {
	  pr->prop_name = strcpy(pp, prop);
	  pr->prop_val  = (char *) val;
	  pr->next = bf->prop_list;
	  bf->prop_list = pr;
	  return 0;
       }
    return -1;
} /* SiPutProperty */


/****  below function(make_if_node) if added by zjd in 95.8.10 with the template
 of  make_loop_node  ********/


/****************************************************************
 *  (static)                                                    *
 *  make_if_node - This is a static routine used by SiNewIf     *
 *                                                              *
 *  Input:                                                      *
 *      var - indicates the type of the loop to be made         *
 *      low - condition expression of the if                    *
 *      body - if body                                          *
 *                                                              *
 *  Output:                                                     *
 *      returns a pointer to the created node, NULL is failure  *
 *                                                              *
 ****************************************************************/
static PTR_BFND make_if_node(var, low, body)
    int      var;
    PTR_LLND low;
    int body[];
{
    PTR_LLND range;
    PTR_BFND bf, b2,thptr;
    PTR_BLOB bl, bb;
    int *bp;

    bb = (PTR_BLOB) NULL;


    if(bf = make_bfndnt(currentFile, var, NULL, low, NULL, LLNULL))
    {
       bf->filename = currentFile->global_bfnd->filename;
       SiAddToStmtTable(bf);
       bp = body;
       thptr = (PTR_BFND) NULL;
       while (*bp != -1)
           if (b2 = FindCurrBifNode(*bp))
          {
             b2->control_parent = bf;
             if (bp == body){   /* The first one */
                bf->thread = b2; thptr = last_thread(b2);
                bb = bl = make_blob(currentFile, b2, (PTR_BLOB)NULL);
                }
             else
             {
                thptr->thread = b2; thptr = last_thread(b2);
                bl->next = make_blob(currentFile, b2, (PTR_BLOB)NULL);
                bl = bl->next;
             }
             bp++;

          }

       if(b2 = make_bfndnt(currentFile,CONTROL_END,SMNULL,LLNULL,LLNULL,LLNULL))
       {
          SiAddToStmtTable(b2);
          if(thptr){
                bf->thread = b2;
                }
          else
                thptr->thread = b2;
          b2->filename = currentFile->global_bfnd->filename;
          bl->next = make_blob(currentFile, b2, (PTR_BLOB)NULL);
          bf->entry.template.bl_ptr1 = bb;
          return bf;
       }
     }
     return NULL;
} /* make_if_node */





/****  below function(SiNewIf) if added by zjd in 95.8.10 with the template of SiNewLoop  *****/
/****************************************************************
 *                                                              *
 *  SiNewIf-- make a new if                                     *
 *                                                              *
 *  Input:                                                      *
 *      type - integer specifying the variant of the loop       *
 *      low  - id of the expression tree  that  represents the  *
 *             condition expression of the if                   *
 *      body - list of id's of the body                         *
 *                                                              *
 *  Output:                                                     *
 *     id of the newly created node, -1 indicates failure       *
 *                                                              *
 ****************************************************************/

int SiNewIf(e)
ELIST e;
{
    int type, low, err;
    PTR_BFND bf;
    PTR_LLND l1;
    int body[200], *bp;
    char ch;

   /* below three line is added by zjd in 95.8.9 in order to correct the false in this function  */
    int i;
    for(i=0;i<200;i++)
        body[i]=-1;


    type = (int) e->car; e = e->cdr;
    low  = (int) e->car; e = e->cdr;

 /* Now read in the list that represents body */
    err= 0;
    if (! err)
    {
       for (bp = body; e != NULL; e = e->cdr)
          *bp++ = (int) e->car;

       if ((l1  = FindLLNode(low))&& (bp != body)) 
       {
          bf    = make_if_node(type, l1, body);
          if (bf)
          {
                return bf->id;
          }
       }
    }
    return -1;
} /* SiNewIf */


