/****************************************************************
 *								*
 *			     svr-lines.c			*
 *								*
 *   This file contains routines that handle the line number	*
 *   adjusting routines.  They are used when we delete or	*
 *   insert bif	nodes to the graph				*
 *								*
 ****************************************************************/

#include "db.h"

static int delta, line_base;
static PTR_FNAME fi;

extern int figure_tabs();


/****************************************************************
 *  (static)							*
 *  adj_num -- adjust the line number after node deletion	*
 *								*
 *  Input:							*
 *	b - the bif node pointer				*
 *	(implicitly)						*
 *	delta	  - the difference to be adjusted		*
 *	line_base - the line base after which the line number	*
 *		    needs to be adjust				*
 *								*
 *  Output:							*
 *	none							*
 *								*
 ****************************************************************/
static void adj_num(b)
PTR_BFND b;
{
   PTR_BLOB bl;

   if (b == NULL) return;

   if (b->g_line > line_base)
       b->g_line -= delta;

   for (bl = b->entry.template.bl_ptr1; bl; bl = bl->next)
      adj_num(bl->ref);

   for (bl = b->entry.template.bl_ptr2; bl; bl = bl->next)
      adj_num(bl->ref);
} /* adj_num */


/****************************************************************
 *								*
 *  ReNumAfterDeletion -- adjust the line number after node     *
 *			  deletion				*
 *								*
 *  Input:							*
 *	b - the bif node that was deleted			*
 *								*
 *  Output:							*
 *	none							*
 *								*
 *  Side Effects:						*
 *	set up the static variable "delta" and "line_base"	*
 *	before calling adj_num					*
 *								*
 ****************************************************************/
void ReNumAfterDeletion(bf)
PTR_BFND bf;
{
  /* the following is being deleted because we wish to
     reconsider the wisdom of changing line numbers while
     the code is being restructured.
   PTR_BFND b;
   PTR_BLOB bl1, bl2;
   int i1, i2;

   line_base = i1 = bf->g_line;
   if (bf->variant == CDOALL_NODE || bf->variant == SDOALL_NODE) 
   {
      bl1 = bf->entry.template.bl_ptr2;
      bl2 = bf->entry.template.bl_ptr1;
   }
   else 
   {
      bl1 = bf->entry.template.bl_ptr1;
      bl2 = bf->entry.template.bl_ptr2;
   }

   for (; bl1; bl1 = bl1->next)
      i2 = bl1->ref->g_line;

   for (; bl2; bl2 = bl2->next)
      i2 = bl2->ref->g_line;

   delta = i2 - i1 + 1;
   for (b = bf; b && b->variant != GLOBAL; b = b->control_parent)
	;
   if (b)
      adj_num(b);
  */
} /* ReNumAfterDeletion */


/****************************************************************
 *								*
 *  adj_num2 -- adjust the line number after node insertion	*
 *								*
 *  Input:							*
 *	b - the bif node pointer				*
 *	(implicitly)						*
 *	delta	  - the difference to be adjusted		*
 *	line_base - the line base after which the line number	*
 *		    needs to be adjust				*
 *								*
 *  Output:							*
 *	none							*
 *								*
 ****************************************************************/
static void adj_num2(b)
PTR_BFND b;
{
   PTR_BLOB bl;

   if (b == NULL) 
      return;

   if ((b->g_line <= 0) && (b->variant != GLOBAL)) 
   {
      b->g_line = line_base - b->g_line;
      b->filename = fi;
   } 
   else if (b->g_line > line_base)
	   b->g_line += delta;

   for (bl = b->entry.template.bl_ptr1; bl; bl = bl->next)
      adj_num2(bl->ref);

   for (bl = b->entry.template.bl_ptr2; bl; bl = bl->next)
      adj_num2(bl->ref);
} /* adj_num2 */


/****************************************************************
 *								*
 *  ReNumAfterInsertion -- adjust the line number after node    *
 *                         insertion				*
 *								*
 *  Input:							*
 *	b - the bif node that is to be inserted			*
 *								*
 *  Output:							*
 *	none							*
 *								*
 *  Side Effects:						*
 *	set up the static variable "delta" and "line_base"	*
 *	before calling adj_num2					*
 *								*
 ****************************************************************/
void ReNumAfterInsertion(bf, d, base)
PTR_BFND bf;
int	 d;
int	 base;
{
   PTR_BFND b;

   line_base = base;
   delta = d;

   for (b = bf; b && b->variant != GLOBAL; b = b->control_parent)
		;
   if (b) 
   {
      fi = b->filename;
      adj_num2(b);
   }
} /* ReNumAfterInsertion */


/****************************************************************
 *								*
 *  CountLines -- figures the number of lines of a given node	*
 *		   The heuristic is that each bif node is one	*
 *		   line long					*
 *								*
 *  Inputs:							*
 *	bf  - the bif node pointer to be counted		*
 *	cnt - the starting line number for this bif node	*
 *								*
 *  Output:							*
 *	cnt is set to one greater than the line number of the	*
 *		last bif node					*
 *								*
 *  Side Effect:						*
 *	The g_line field of each bif node travered was set to	*
 *	-(*cnt) so that later it can be used by renum2 to	*
 *	re-calculate the line number for the whole tree		*
 *								*
 ****************************************************************/
void CountLines(bf,cnt)
PTR_BFND bf;
int *cnt;
{
   PTR_CMNT cmnt;
   PTR_BLOB b;

   if (bf == NULL)
      return;

   if (bf->entry.template.cmnt_ptr != NULL)
      for (cmnt=bf->entry.template.cmnt_ptr; cmnt; cmnt = cmnt->next)
	 (*cnt)++;
	
   bf->g_line = -(*cnt);
   if (bf->variant == CONTROL_END)
      return;
   (*cnt)++;

   switch (bf->variant) 
   {
      case PROG_HEDR:
      case PROC_HEDR:
      case FUNC_HEDR:
	   for (b = bf->entry.template.bl_ptr1; b; b = b->next)
	      CountLines(b->ref,cnt);
	   (*cnt)++;	/* for "end" */
	   break;
      case SDOALL_NODE:
      case CDOALL_NODE:
	   for (b = bf->entry.template.bl_ptr2; b; b = b->next)
	      CountLines(b->ref, cnt);
	   if (bf->entry.template.bl_ptr2)
	      (*cnt)++;	/* for the line of "loop" */
	   for (b = bf->entry.template.bl_ptr1; b; b = b->next)
	      CountLines(b->ref, cnt);
	   (*cnt)++;	/* for "end {c|s}doall" */
	   break;
      case FOR_NODE:
	   for (b = bf->entry.template.bl_ptr1; b; b = b->next)
	      CountLines(b->ref,cnt);
	   (*cnt)++;	/* for "end do" */
	   break;
      case IF_NODE:
	   for (b = bf->entry.template.bl_ptr1; b; b = b->next)
	      CountLines(b->ref,cnt);
           b = bf->entry.template.bl_ptr2;
           if (b != NULL)
           {
	      (*cnt)++; /* for "else" */
	      for (; b; b = b->next)
		 CountLines(b->ref,cnt);
	   }
	   (*cnt)++;	/* for "end if" */
	   break;
      default:
	   break;
   } /* switch */
} /* CountLines */
