#include "db.h"
#include "elist.h"

/*
 * External references
 */
extern PTR_PROJ currentProject;
extern PTR_FILE currentFile;


PTR_BFND make_bfndnt();
PTR_LLND make_llnd();
extern void SiAddToStmtTable();
extern PTR_BFND SiFindInStmtTable();

int strcmp();

/****************************************************************
 *								*
 *   FindCurrBifNode -- find the bif node in the current active	*
 *		    file with given id number			*
 *								*
 *   Input:							*
 *	id -- the bif node id number				*
 *								*
 *   Output:							*
 *	the corresponding bif node pointer if found		*
 *	nil otherwise						*
 *								*
 ****************************************************************/
PTR_BFND
FindCurrBifNode(id)
    int id;
{
	return SiFindInStmtTable(id);

} /* FindCurrBifNode */


/****************************************************************
 *								*
 *   FindLLNode -- find the low level node in the current	*
 *		   active file with given id number		*
 *								*
 *   Input:							*
 *	id -- the low level node id number			*
 *								*
 *   Output:							*
 *	the corresponding low level node pointer if found	*
 *	nil otherwise						*
 *								*
 ****************************************************************/
PTR_LLND
FindLLNode(id)
    int id;
{
    PTR_LLND ll;

    for (ll = cur_file->head_llnd; ll; ll = ll->thread)
       if (ll->id == id)
	  return ll;
    return NULL;
} /* FindLLNode */

/****************************************************************
 *								*
 *   FindLabNode -- find the label   node in the current	*
 *		   active file with given id number		*
 *								*
 *   Input:							*
 *	id -- the label node id number				*
 *								*
 *   Output:							*
 *	the corresponding label node pointer if found		*
 *	nil otherwise						*
 *								*
 ****************************************************************/
PTR_LABEL
FindLabNode(id)
    int id;
{
    PTR_LABEL ll;

    for (ll = cur_file->head_lab; ll; ll = ll->next)
       if (ll->id == id)
	  return ll;
    return NULL;
} /* FindLabNode */


/****************************************************************
 *								*
 *   FindSymbolNode -- find the symbol node in the current      *
 *		    active file with given id number		*
 *								*
 *   Input:							*
 *	id -- the symbol node id number 			*
 *								*
 *   Output:							*
 *	the corresponding bif node pointer if found		*
 *	nil otherwise						*
 *								*
 ****************************************************************/
PTR_SYMB FindSymbolNode(id)
    int id;
{
    PTR_SYMB s;

    for (s = cur_file->head_symb; s; s = s->thread)
       if (s->id == id)
	  return s;
    return NULL;
} /* FindSymbolNode */

/****************************************************************
 *								*
 *   FindTypeNode -- find the type node in the current          *
 *		    active file with given id number		*
 *								*
 *   Input:							*
 *	id -- the type node id number 				*
 *								*
 *   Output:							*
 *	the corresponding bif node pointer if found		*
 *	nil otherwise						*
 *								*
 ****************************************************************/
PTR_TYPE FindTypeNode(id)
int id;
{
  PTR_TYPE t;
  for (t = cur_file->head_type; t; t = t->thread)
     if (t->id == id) 
        return t;
  return NULL;
} /* FindTypeNode */

/****************************************************************
 *								*
 *   FindFileObj   -- find the corresponding file object of a	*
 *		      given "filename"				*
 *								* 
 *   Input:							*
 *	  filename -- a string					*
 *								*
 *   Output:							*
 *	  the corresponding file object in the current project, *
 *	  NULL if not found					*
 *								*
 ****************************************************************/
PTR_FILE FindFileObj(filename)
char *filename;
{
    PTR_BLOB b;

    for (b = cur_proj->file_chain; b; b = b->next)
       if (!strcmp(filename, ((PTR_FILE)b->ref)->filename))
	  return (PTR_FILE)b->ref;
    return NULL;
} /* FindFileObj */


/****************************************************************
 *								*
 *   FindDepNode -- find the dep node in the current active	*
 *		    file with given id number			*
 *								*
 *   Input:							*
 *	id -- the dep node id number				*
 *								*
 *   Output:							*
 *	the corresponding dep node pointer if found		*
 *	nil otherwise						*
 *								*
 ****************************************************************/
PTR_DEP
FindDepNode(id)
    int id;
{
    PTR_DEP d;

    for (d = cur_file->head_dep; d; d = d->thread)
       if (d->id == id)
	  return d;
    return NULL;
} /* FindDepNode */


/****************************************************************
 *								*
 * MakeDeclStmt    -						*
 *								*
 *								*
 *								*
 *								*
 *								*
 *								*
 * Input -							*
 *								* 
 *								*
 *								*
 *								*
 * Output -							*
 *								*
 *								*
 ****************************************************************/
PTR_BFND MakeDeclStmt(s)
    PTR_SYMB s;
{
    PTR_LLND l1, l2, l3, exp;
    PTR_TYPE t;
    PTR_BFND b;

    t = s->type;
    if (t->variant == T_ARRAY)
    {
       exp = t->entry.ar_decl.ranges;
       l2 = make_llnd(cur_file, ARRAY_REF, exp, LLNULL, s);
    }
    else
       l2 = make_llnd(cur_file, VAR_REF, LLNULL, LLNULL, s);
    l3 = make_llnd(cur_file, TYPE_OP, LLNULL, LLNULL, SMNULL);
    l3->type = t;
    l1 = make_llnd(cur_file, EXPR_LIST, l2, LLNULL, SMNULL);

    b = make_bfndnt(cur_file, VAR_DECL, SMNULL, l1, l3, LLNULL);
    SiAddToStmtTable(b);
    return b;
} /* MakeDeclStmt */

char *Allocate(size)
int size;
{
   char *ptr, *malloc();
   if (size <= 0)
   {
      printf("Allocate: cannot allocate %d bytes\n", size);
      return NULL;
   }
   if ((ptr = malloc(size)) == NULL)
   {
      printf("Allocate: out of memory\n");
      return NULL;
   }
   return ptr;
} /* Allocate */
