C
C     BECAUSE Benchmark Set
C
C     TEST: BBS.2.5.1
C
C     VERSION: 1.0
C
C     DATE: 02/02/92
C
C Benchmark approximation to assembly of Poisson's equation on an
C artificial mixed mesh of tetrahedra and hexahedra.
C
C Origin: RAL Nov. 1991
C         Added ELAPSE and std header 2/2/92
C
      PROGRAM BM1
C
C MAXDIV is the maximum number of nodes on each edge of the test mesh.
C MAXDIV=41 requires about 12Mbytes of memory, if possible MAXDIV should
C go up to 101 for a million node mesh size. The actual sizes used by
C default are those in the data array ISIZE, subject to MAXDIV being
C large enough
C
C     .. Parameters ..
      INTEGER MAXDIV,MAXNOD,MAXELE
      PARAMETER (MAXDIV=41,MAXNOD=MAXDIV**3,MAXELE=3* (MAXDIV-1)**3)
C     ..
C     .. Local Scalars ..
      DOUBLE PRECISION ERR,FLOPS,SUM,TIME
      INTEGER I,IC,ND,NR,NSIZE,TOTELS,TOTNOD
C     ..
C     .. Local Arrays ..
      DOUBLE PRECISION COORD(3,MAXNOD),RHS(MAXNOD),SOL(3,MAXNOD)
      INTEGER ELTOP(10,MAXELE),ISIZES(11)
C     ..
C     .. External Subroutines ..
C      EXTERNAL ASSTST,MSHGEN
C     ..
C     .. Intrinsic Functions ..
      INTRINSIC ABS
C     ..
C     .. Data statements ..
      DATA NR/1/,FLOPS/206.35D0/,NSIZE/11/,ISIZES/11,14,17,21,26,33,41,
     +     51,64,80,101/ 
C     ..
      PRINT 9030
   10 PRINT *,' Input number of edge divisions, or 0 for set values:'
      READ *,ND
      IF (ND.NE.0) THEN
          IF (ND.GT.MAXDIV) THEN
              PRINT *,' Error: maximum divisions=',MAXDIV
              GO TO 10

          ELSE IF (ND.LE.2) THEN
              PRINT *,' Error: Minimum divisions=3'
              GO TO 10

          ELSE
              PRINT *,'Using ',ND**3,' nodes'
              NSIZE = 1
              ISIZES(1) = ND
          END IF

      END IF
C
C Loop over sizes:
C
      IC = 0
   20 IC = IC + 1
      ND = ISIZES(IC)
      IF (ND.GT.MAXDIV .OR. IC.GT.NSIZE) THEN
          GO TO 40

      END IF
C
C Generate a simple test mesh described by coordinates and an element
C topology array. This is set-up step is not timed.
C
      CALL MSHGEN(ELTOP,10,MAXELE,COORD,3,MAXNOD,ND,TOTNOD,TOTELS,
     +            .TRUE.)
C
C Perform timing tests on a simplifed element assembly process
C which attempts to perform a similar order of magnitude of
C computational work as the real assembly.
C
      CALL ASSTST(ELTOP,10,MAXELE,COORD,3,MAXNOD,TOTNOD,TOTELS,SOL,RHS,
     +            NR,TIME)
C
C Check the answer
C
      SUM = 0.D0
      DO 30 I = 1,TOTNOD
          SUM = SUM + RHS(I)
   30 CONTINUE
      ERR = ABS(SUM/ (ND-1)**3-4.D0)
C      IF (ERR.GT.1D-9) THEN
          WRITE (*,FMT=9000) ERR
C      END IF
      
      WRITE (*,FMT=9020) TIME/NR,TOTELS,NR*TOTELS*FLOPS/TIME*1D-6

 9040 FORMAT (' res',E14.6)
C     DO I = 1,TOTNOD
C         WRITE (*,FMT=9040) RHS(I)
C     ENDDO

      GO TO 20

   40 CONTINUE
      STOP
 9000 FORMAT (' Error: Tolerance criteria not satisfied: Error=',E14.6)
 9020 FORMAT (' Time: ',F9.3,' secs.   Elements:',I7,'  approx. MFLOP/',
     +       's:',E14.6)  
 9030 FORMAT (/,' BECAUSE benchmark BBS.2.5.1 (matrix assembly)',/)

      END

      SUBROUTINE ASSTST(ELTOP,IELTOP,JELTOP,COORD,ICOORD,JCOORD,TOTNOD,
     +                  TOTELS,SOL,RHS,NR,TIME)
C
C Time assembly of system matrix/RHS terms.
C
C     .. Scalar Arguments ..
      DOUBLE PRECISION TIME
      INTEGER ICOORD,IELTOP,JCOORD,JELTOP,NR,TOTELS,TOTNOD
C     ..
C     .. Array Arguments ..
C
      DOUBLE PRECISION COORD(ICOORD,JCOORD),RHS(JCOORD),SOL(3,JCOORD)
      INTEGER ELTOP(IELTOP,JELTOP)
C     ..
C
C
C     .. Local Scalars ..
      REAL XDUM1, XDUM2
      INTEGER I
C     ..
C     .. External Functions ..
      REAL ELAPSE
      EXTERNAL ELAPSE
C     ..
C     .. External Subroutines ..
C      EXTERNAL FRMDIS,SETSOL
C     ..
      CALL SETSOL(SOL,TOTNOD,COORD)
      TIME = ELAPSE(XDUM1)
      DO 10 I = 1,NR
C
          CALL FRMDIS(SOL,COORD,TOTNOD,ELTOP,TOTELS,RHS)
C
   10 CONTINUE
      TIME = ELAPSE(XDUM2)
      END

      SUBROUTINE FRMDIS(SOL,COORD,TOTNOD,ELTOP,TOTELS,RHS)
C
C This subroutine forms the discrete system by looping through all elements
C in the mesh evaluating the local contribution and adding it to the RHS
C vector.
C
C     .. Scalar Arguments ..
      INTEGER TOTELS,TOTNOD
C     ..
C     .. Array Arguments ..
      DOUBLE PRECISION COORD(3,TOTNOD),RHS(TOTNOD),SOL(3,TOTNOD)
      INTEGER ELTOP(10,TOTELS)
C     ..
C     .. Local Scalars ..
      DOUBLE PRECISION EPSR
      INTEGER I,J,NELE,NODEL
C     ..
C     .. Local Arrays ..
      DOUBLE PRECISION AREAS(6),ELAREA(6),ELCORD(3,8),ELLENG(6),
     +                 ELSOL(3,8),ELVOL(8),LENGTH(6),PJACOB(8,8),
     +                 PRHS(8),VOLS(4)
      INTEGER STEER(8)
C     ..
C     .. External Subroutines ..
C      EXTERNAL FILL,POSDIV
C     ..
C     .. Data statements ..
      DATA AREAS/6*0.0625/,LENGTH/6*1D0/,VOLS/4*0.05D0/
C     ..
C
      EPSR = 12.3D0
C
      DO 10 I = 1,TOTNOD
          RHS(I) = 0D0
   10 CONTINUE
*
C *** LOOP AROUND ALL ELEMENTS
C
      DO 100 NELE = 1,TOTELS
*
          NODEL = ELTOP(1,NELE)
*
          DO 30 I = 1,NODEL
              STEER(I) = ELTOP(I+1,NELE)
              DO 20 J = 1,3
                  ELSOL(J,I) = SOL(J,STEER(I))
                  ELCORD(J,I) = COORD(J,STEER(I))
   20         CONTINUE
   30     CONTINUE
C
          DO 50 I = 1,NODEL
              PRHS(I) = 0D0
              DO 40 J = 1,NODEL
                  PJACOB(I,J) = 0D0
   40         CONTINUE
   50     CONTINUE
C
          IF (NODEL.EQ.4) THEN
              DO 60 I = 1,6
                  ELLENG(I) = LENGTH(I)
                  ELAREA(I) = AREAS(I)
   60         CONTINUE
              DO 70 I = 1,4
                  ELVOL(I) = VOLS(I)
   70         CONTINUE

          ELSE
              ELLENG(1) = ELCORD(1,2) - ELCORD(1,1)
              ELLENG(2) = ELCORD(2,4) - ELCORD(2,1)
              ELLENG(3) = ELCORD(3,5) - ELCORD(3,1)
              ELAREA(1) = 0.25*ELLENG(2)*ELLENG(3)
              ELAREA(2) = 0.25*ELLENG(3)*ELLENG(1)
              ELAREA(3) = 0.25*ELLENG(1)*ELLENG(2)
              ELVOL(1) = 0.125*ELLENG(1)*ELLENG(2)*ELLENG(3)
              DO 80 I = 2,8
                  ELVOL(I) = ELVOL(1)
   80         CONTINUE
          END IF

*
          CALL POSDIV(ELSOL,EPSR,ELAREA,ELLENG,PRHS,PJACOB,NODEL)
*
          DO 90 I = 1,NODEL
              PRHS(I) = PRHS(I) + ELVOL(I)* (ELSOL(3,I)-ELSOL(2,I))
              PJACOB(I,I) = PJACOB(I,I) +
     +                      ELVOL(I)* (ELSOL(2,I)+ELSOL(3,I))
   90     CONTINUE
*
          CALL FILL(PRHS,PJACOB,RHS,TOTNOD,NODEL,STEER)
*
  100 CONTINUE
C
      END
C
      SUBROUTINE POSDIV(ELSOL,EPSR,ELAREA,ELLENG,PRHS,PJACOB,NODEL)
C
C For a given element (tetrahedra or hexahedra) evaluate terms in the
C the discrete formulation for Poissons equation.
C
C     .. Scalar Arguments ..
      DOUBLE PRECISION EPSR
      INTEGER NODEL
C     ..
C     .. Array Arguments ..
      DOUBLE PRECISION ELAREA(6),ELLENG(6),ELSOL(3,8),PJACOB(8,8),
     +                 PRHS(8)
C     ..
C     .. Local Scalars ..
      DOUBLE PRECISION X
      INTEGER L1,M1,N1,N2
C     ..
C     .. Local Arrays ..
      INTEGER IHE(3,12),ITE(3,6)
C     ..
C     .. Data statements ..
      DATA IHE/1,2,1,2,3,2,3,4,1,4,1,2,5,6,1,6,7,2,7,8,1,8,5,2,1,5,3,2,
     +     6,3,3,7,3,4,8,3/ 
      DATA ITE/1,2,1,2,3,2,3,1,3,1,4,4,2,4,5,3,4,6/
C     ..
C
      IF (NODEL.EQ.8) THEN
          DO 10 L1 = 1,12
              N1 = IHE(1,L1)
              N2 = IHE(2,L1)
              M1 = IHE(3,L1)
              X = -EPSR*ELAREA(M1)/ELLENG(M1)
              PJACOB(N1,N1) = PJACOB(N1,N1) - X
              PJACOB(N2,N1) = PJACOB(N2,N1) + X
              PJACOB(N1,N2) = PJACOB(N1,N2) + X
              PJACOB(N2,N2) = PJACOB(N2,N2) - X
              X = X* (ELSOL(1,N2)-ELSOL(1,N1))

              PRHS(N1) = PRHS(N1) + X
              PRHS(N2) = PRHS(N2) - X
   10     CONTINUE

      ELSE
          DO 20 L1 = 1,6
              N1 = ITE(1,L1)
              N2 = ITE(2,L1)
              M1 = ITE(3,L1)
              X = -EPSR*ELAREA(M1)/ELLENG(M1)
              PJACOB(N1,N1) = PJACOB(N1,N1) - X
              PJACOB(N2,N1) = PJACOB(N2,N1) + X
              PJACOB(N1,N2) = PJACOB(N1,N2) + X
              PJACOB(N2,N2) = PJACOB(N2,N2) - X
              X = X* (ELSOL(1,N2)-ELSOL(1,N1))
              PRHS(N1) = PRHS(N1) + X
              PRHS(N2) = PRHS(N2) - X
   20     CONTINUE
      END IF
C
      END
      SUBROUTINE FILL(PRHS,PJACOB,RHS,TOTNOD,NODEL,STEER)
C
C Transfer element contribution to RHS vector. Also add element contribution
C to system matrix into the RHS vector (simplification of problem, omits
C the sparse matrix addressing).
C
C     .. Scalar Arguments ..
      INTEGER NODEL,TOTNOD
C     ..
C     .. Array Arguments ..

      DOUBLE PRECISION PJACOB(8,8),PRHS(8),RHS(*)
      INTEGER STEER(8)
C     ..
C     .. Local Scalars ..
      INTEGER I,INOD,J
C     ..
      DO 20 I = 1,NODEL
          INOD = STEER(I)
          RHS(INOD) = RHS(INOD) + PRHS(I)
          DO 10 J = 1,NODEL
              RHS(INOD) = RHS(INOD) + PJACOB(I,J)
   10     CONTINUE
   20 CONTINUE
C
      END
C
      SUBROUTINE SETSOL(SOL,N,COORD)
C
C Set simple values for SOL array (not timed - can run on host).
C
C     .. Scalar Arguments ..
      INTEGER N
C     ..
C     .. Array Arguments ..
      DOUBLE PRECISION COORD(3,N),SOL(3,N)
C     ..
C     .. Local Scalars ..
      INTEGER I
C     ..
C     .. Intrinsic Functions ..
      INTRINSIC MOD
C     ..
      DO 10 I = 1,N
          SOL(1,I) = 0.1* (MOD(COORD(1,I),2D0))
          SOL(2,I) = 1.D0
          SOL(3,I) = 2.D0
   10 CONTINUE
      END
