/* Copyright (C) 1992, Irisa, Indiana University, 
   subroutines of the Sigma toolbox */
/**************************************************************************
 *                                                                        *
 *   Library for dependences for toolbox                                  *
 *                                                                        *
 *                                                                        *
 **************************************************************************/

#include <stdio.h>
#include <malloc.h>
#include "gvarargs.h"  /* for special arg passing */
#include "macro.h"
#include "dependence.h"
#include "ext_lib.h"
#include "ext_low.h"
#include "ext_mid.h"

int NUMBER_OF_NODE_IN_GRAPH = 0; /* to make id of graph node*/

/***************************************************************************/
/* quelques primitives pour la gestion memoire, graphe de dependence etc ... */



/* list of allocated data */
ptstack_chaining Current_Allocated_Data = NULL;
ptstack_chaining First_STACK= NULL;

make_a_malloc_stack()
{
  ptstack_chaining pt;
  
    
  pt = (ptstack_chaining) malloc(sizeof(struct stack_chaining));
  if (!pt)
    {
      Message("sorry : out of memory\n",0);
      exit(1);
    }
  
  if (Current_Allocated_Data)
    Current_Allocated_Data->next = pt;
  pt->first = NULL;
  pt->last = NULL;
  pt->prev = Current_Allocated_Data;
  if (Current_Allocated_Data)
    pt->level = Current_Allocated_Data->level +1;
  else
    pt->level = 0;
/*  printf("make_a_malloc_stack %d \n",pt->level);*/
  Current_Allocated_Data = pt;
  if (First_STACK == NULL)
    First_STACK = pt;
}

myfree()
{
  ptstack_chaining pt;
  ptchaining pt1, pt2;
  if (!Current_Allocated_Data)
     {
      Message("Stack not defined\n",0);
      exit(1);
    }

  pt2 = Current_Allocated_Data->first;

/*  printf("myfree %d \n", Current_Allocated_Data->level);*/
  while (pt2)
    {
      free(pt2->zone);
      pt2->zone = 0;
      pt2 = pt2->list;
    }

  pt2 = Current_Allocated_Data->first;
  while (pt2)
    {
      pt1 = pt2;
      pt2 = pt2->list;
      free(pt1);
    } 
  pt = Current_Allocated_Data;
  Current_Allocated_Data = pt->prev;
  Current_Allocated_Data->next = NULL;
  free(pt);
}


char * mymalloc(size)
int size;
{
  char *pt1;
  ptchaining pt2;
  if (!Current_Allocated_Data)
    {
      Message("Allocated Stack not defined\n",0);
      exit(1);
    }

/*  if (Current_Allocated_Data->level > 0)
    printf("mymalloc  %d \n", Current_Allocated_Data->level); */
  pt1 = malloc(size);
  if (!pt1)
    {
      Message("sorry : out of memory\n",0);
      exit(1);
    }
 
  pt2 = (ptchaining) malloc(sizeof(struct chaining));
  if (!pt2 )
    {
      Message("sorry : out of memory\n",0);
      exit(1);
    }
  
  pt2->zone = pt1;
  pt2->list = NULL;

  if (Current_Allocated_Data->first == NULL)
    Current_Allocated_Data->first = pt2;
  
  if (Current_Allocated_Data->last == NULL)
    Current_Allocated_Data->last = pt2;
  else
    {
      Current_Allocated_Data->last->list = pt2;
      Current_Allocated_Data->last = pt2;
    }
  return pt1;
}

/***************************************************************************/

/* quelques primitives pour le debug */
Print_Dep_Vector(f,dep)
     FILE *f;
     PTR_DEP dep;
{
  char *vect;
  int i;
  vect = DEP_DIRECTION(dep);
  fprintf(f,"( ");
  for (i = 1 ; i < MAX_DEP; i++)
    {
      switch (vect[i])
        {
          case 1 :
            fprintf(f,"=   ");
            break;
          case 2 :
            fprintf(f,"<   ");
            break;
          case 3 :
            fprintf(f,"<=  ");
            break;
          case 4 :
            fprintf(f,">   ");
            break;
          case 5 :
            fprintf(f,">=  ");
            break;
          case 6 :
            fprintf(f,".   ");
            break;
          case 7 :
            fprintf(f,"*   ");
            break;
        }
    }
  
  fprintf(f,")\n");

  
}

Print_Dep(f,dep)
     FILE *f;
     PTR_DEP dep;
{
  PTR_BFND to;
  PTR_BFND from;
  
  from = DEP_FROM_BIF(dep);
  to = DEP_TO_BIF(dep);
  fprintf(f,"------\n");
  switch(DEP_TYPE(dep))
    {
    case 0 :
      fprintf(f,"flow-dependence");
      break;
    case 1 :
      fprintf(f,"anti-dependence");
      break;
    case 2 :
      fprintf(f,"output-dependence");
      break;
    default:
      Message("unknown dependence",0);      
    }
  fprintf(f,"  between line  %d and %d on symbol %s\n",
         BIF_LINE(from),
         BIF_LINE(to),
         SYMB_IDENT(DEP_SYMB(dep)));
  /* on affiche les expressions associe */
  fprintf(f,"id dep %d, (id %d) %s",DEP_ID(dep),NODE_ID(DEP_FROM_LL(dep)),funparse_llnd(DEP_FROM_LL(dep)));
  
  fprintf(f,"(id %d) %s",NODE_ID(DEP_TO_LL(dep)), funparse_llnd(DEP_TO_LL(dep)));
   
  fprintf(f,"Direction vector : ");
  Print_Dep_Vector(f,dep);
}

Print_Dep_For_Loop(f,b)
     FILE *f;
     PTR_BFND b;
{
  PTR_BFND body, body_end;
  PTR_DEP dep;
  int Dep_Found = 0;
  
  
  if (!f)
    f = stdout;

  /* pour chaque instruction dans la boucle va
     chercher la liste des dependances et les affiches */
  /* internal_loop = Get_Internal_Loop(b);
  body = BIF_NEXT(internal_loop);
  body_end = Get_Last_Node_Of_Stmt(internal_loop); */
  body  = b;
  body_end = Get_Last_Node_Of_Stmt(b);
  
  for(dep = FIRST_DEP_IN_PROJ(cur_file); dep; dep = DEP_NEXT(dep))      
    {
      if (In_Section(body, body_end, DEP_TO_BIF(dep)) &&
          In_Section(body, body_end, DEP_FROM_BIF(dep)))
        {
          Print_Dep(f,dep);
          Dep_Found = 1;
        }
    }

  if (!Dep_Found)
    fprintf(f,"No dependencies found\n");
}

/*************************************************************************/
/*                       primitive de gestion de graphe                  */

PTR_GRAPH
Get_node_bif(graph,bif)
     PTR_BFND bif;
     PTR_GRAPH graph;
{
  PTR_GRAPH pt;

  pt = graph;
  while (pt)
    {
      if (GRAPH_BIF(pt) == bif)
	return(pt);
      pt = CHAIN_LIST(pt);
    }
  return(NULL);
}


Create_Arc(graph,dep)
     PTR_GRAPH graph;
     PTR_DEP dep;     
{
  PTR_GRAPH pt1,pt2;
  int i;
  
  if (!dep)
    return;
  pt1 = Get_node_bif(graph,DEP_FROM_BIF(dep));
  pt2 = Get_node_bif(graph,DEP_TO_BIF(dep));
  
  if (!pt1 || !pt2)
    {
      Message("Pb in Create_Arc ",0);      
      return;
    }
  
  
  for (i = 0; (GRAPH_SUC(pt1)[i] != NULL) && (i<MAXSUC); i++);

  GRAPH_SUC(pt1)[i] = pt2;
  GRAPH_VECT(pt1)[i] = DEP_DIRECTION(dep);
  GRAPH_DEP (pt1) [i] = dep; /*added by S.L. 18/8/92 14h05 to modify graph */
                             /*  without modifying dependences*/
  GRAPH_TYPE(pt1)[i] = DEP_TYPE(dep);
  GRAPH_LL_FROM(pt1)[i] = DEP_FROM_LL(dep);
  GRAPH_LL_TO(pt1)[i] = DEP_TO_LL(dep);
  for (i = 0; (GRAPH_PRED(pt2)[i] != NULL) && (i<MAXSUC); i++);

  GRAPH_PRED(pt2)[i] = pt1;

}


PTR_GRAPH
Create_node(graphe, bif)
     PTR_GRAPH graphe;
     PTR_BFND bif;     
{
  PTR_GRAPH pt, last;

  pt = (PTR_GRAPH) mymalloc(sizeof(struct graph));
  bzero(pt, sizeof(struct graph));
  GRAPH_ID(pt) = NUMBER_OF_NODE_IN_GRAPH++;
  GRAPH_BIF(pt) = bif;
  GRAPH_LINE(pt) = BIF_LINE(bif);
  
  for(last = graphe; last; last = CHAIN_LIST(last))
    {
      if (!CHAIN_LIST(last))
        {
          CHAIN_LIST(last) = pt;
          break;
        }
    }
return pt;
}


PTR_GRAPH 
Dep_Graph_For_Loop(b)
     PTR_BFND b;
{
  PTR_BFND body, internal_loop, body_end;
  PTR_DEP dep;
  PTR_GRAPH first;
  PTR_BLOB temp;
  
  /* allocate stack for the graph */
  make_a_malloc_stack();
  
  /* pour chaque instruction dans la boucle va
     chercher la liste des dependances et les affiches */
  internal_loop = Get_Internal_Loop(b);
  body = BIF_NEXT(internal_loop);
  body_end = Get_Last_Node_Of_Stmt(internal_loop);

  /* Attention on veut les noeuds dans le graphe, dans l'ordre
     lexicographique */

  /* creation des noeuds pour le grapph */
  first = NULL;
  for (temp =BIF_BLOB1(internal_loop); temp ; temp = BLOB_NEXT(temp))
    {
      if (BIF_CODE(BLOB_VALUE(temp)) != CONTROL_END)
        {          
          if (!first)
            first = Create_node(NULL, BLOB_VALUE(temp));
          else
            Create_node(first, BLOB_VALUE(temp));
        }
    }
  
  for(dep = FIRST_DEP_IN_PROJ(cur_file); dep; dep = DEP_NEXT(dep))      
    {
      if (In_Section(body, body_end, DEP_TO_BIF(dep)) &&
          In_Section(body, body_end, DEP_FROM_BIF(dep)))
        {
          Create_Arc(first,dep);
        }
    }

  return first;
}


Depth_Traversal(graphe,mark)
     PTR_GRAPH graphe;
     int mark;
{
  int i;
  
  if (!graphe)
    return;
  
  if (GRAPH_MARK(graphe) != 0)
     return;
  
  GRAPH_MARK(graphe) = mark;
  
  for (i = 0; i < MAXSUC ; i++)
    {
      if (GRAPH_SUC(graphe)[i])
        Depth_Traversal(GRAPH_SUC(graphe)[i],mark+1);
      else
        break;
    }
    return;
}



Depth_Traversal_Reverse(graphe,mark)
     PTR_GRAPH graphe;
     int mark;
{
  int i;
  
  if (!graphe)
    return;
  
  if (GRAPH_MARK(graphe) <= 0)
     return;
  
  GRAPH_MARK(graphe) = mark;
  
  for (i = 0; i < MAXSUC ; i++)
    {
      if (GRAPH_PRED(graphe)[i])
        Depth_Traversal_Reverse(GRAPH_SUC(graphe)[i],mark);
      else
        break;
    }
    return;
}

PTR_GRAPH 
Get_Higher_Mark(graphe)
     PTR_GRAPH graphe;
{
  PTR_GRAPH last,node = NULL;
  int mark = -1;
    
  
  for(last = graphe; last; last = CHAIN_LIST(last))
    {
      if (GRAPH_MARK(last) >= mark)
        {
          node = last;
          mark = GRAPH_MARK(last);          
        }
    }
  return node; 

}

Reset_Mark(graphe)
     PTR_GRAPH graphe;
{
  PTR_GRAPH last;
    
  for(last = graphe; last; last = CHAIN_LIST(last))
    {
       GRAPH_MARK(last) = 0;          
    }
}


Reset_order(graphe)
     PTR_GRAPH graphe;
{
  PTR_GRAPH last;
    
  for(last = graphe; last; last = CHAIN_LIST(last))
    {
       GRAPH_ORDER(last) = NOT_ORDERED;          
    }
}

Print_Mark(graphe)
     PTR_GRAPH graphe;
{
  PTR_GRAPH last;
    
  for(last = graphe; last; last = CHAIN_LIST(last))
    {
      printf("Noeud id %d (line %d) avec Mark %d date %d\n",
             BIF_ID(GRAPH_BIF(last)),
             GRAPH_LINE(last),
             GRAPH_MARK(last),
             GRAPH_ORDER(last)
             );          
    }
}


/* on calcul les composantes fortement connexe du graphe 
   Le graphe est marques par les numero de composantes fortement 
   connnexe (GRAPH_MARK = 1 premiere composante, 2 la deuxieme etc )
*/

PTR_GRAPH
Strongly_Connected(graphe)
     PTR_GRAPH graphe;
{
  PTR_GRAPH last,node;
  int first_componante = -1;
  int mark;
  
  Reset_Mark(graphe);
  
  /* on marque  d'abord le graphe avec  */
  Depth_Traversal(graphe,1);
  node = Get_Higher_Mark(graphe);
  Depth_Traversal_Reverse(node,first_componante); /* premiere composante */

  /* 0, les noeud non visite -x les composantes connexes +x les noeud visites*/
  
  node = Get_Higher_Mark(graphe);
  mark = GRAPH_MARK(node);
  while ((mark >= 0) && node)
    {
      if ( mark > 0)
        {          
          first_componante--;          
          Depth_Traversal_Reverse(node,first_componante);
        } else
          if ( mark== 0)
            {
              Depth_Traversal(node,1);
              node = Get_Higher_Mark(graphe);
              if (node)
                {                  
                  first_componante--; 
                  Depth_Traversal_Reverse(node,first_componante); 
                }
            }
      node = Get_Higher_Mark(graphe);
      mark = GRAPH_MARK(node);
    }

  /* reverse marque */
  for(last = graphe; last; last = CHAIN_LIST(last))
    {
       GRAPH_MARK(last) = - GRAPH_MARK(last);          
    }
  return graphe;
}


/* check if strongly connected component have 
   predecessor order */

Predecessor_Ordered_and_not_Ordered(graph,num)
     PTR_GRAPH graph;
     int num;
{
  PTR_GRAPH pt;
  int i;
  
  for (pt = graph; pt; pt = CHAIN_LIST(pt))
    {
      if (GRAPH_MARK(pt) == num)
        {          
          if (GRAPH_ORDER(pt) == NOT_ORDERED) 
            {
              for (i = 0; i < MAXSUC ; i++)
                {
                  if (GRAPH_PRED(pt)[i])
                    {
                      if ((GRAPH_MARK(GRAPH_PRED(pt)[i]) != num)
                          && (GRAPH_ORDER(GRAPH_PRED(pt)[i]) 
                              == NOT_ORDERED))
                        return FALSE;
                    }
                }
            } else
              return FALSE; /* already ordered */
        }
    }
  
 return TRUE;
}


Number_of_SCC(graph)
     PTR_GRAPH graph;
{
  int max =0;
  
  PTR_GRAPH pt;
  for (pt = graph; pt; pt = CHAIN_LIST(pt))
    {
     if ( GRAPH_MARK(pt) > max)
       max = GRAPH_MARK(pt);     
    }
  return max;
}

/* indique l'ordre pour chacune des composantes
   fortement connexe, pour la distribution des boucles */

Ordered_SCC(graph)
     PTR_GRAPH graph;
{
  int i;
  int num, change, date;
  PTR_GRAPH  pt;
  
  /* remettre l'ordre a jour */
  Reset_order(graph);
  
  num = Number_of_SCC(graph);
  
  change = TRUE;
  date = 1;
  while (change)
    {
      change = FALSE;
      for (i = 1; i <= num ; i++)
        {
          if (Predecessor_Ordered_and_not_Ordered(graph,i))
            {
              for (pt = graph; pt; pt = CHAIN_LIST(pt))
                {
                  if ( GRAPH_MARK(pt) == i)
                    GRAPH_ORDER(pt) = date;                  
                }
              change = TRUE;    
              if (date > MAXSUC)
                abort();              
              date++;              
            }
        }
    }
}

/* Test si le vecteur est positif composante > , >=, = */
Test_Vect_Positif(vect,nb_loop,nb)
     char *vect;
     int nb_loop,nb;
{
  int nb_elem_in_vect;
  int i;
  
  nb_elem_in_vect = 0;
  for (i = 1 ; i < MAX_DEP; i++)
    {
      switch (vect[i])
        {
        case 1 :
        case 2 :
        case 3 :
        case 4 :
        case 5 :
        case 6 :
        case 7 :
          nb_elem_in_vect++;            
          break;
        }
    }
  for (i = nb_elem_in_vect - nb +1;i < nb_elem_in_vect + nb_loop -nb+1; i++)
    {
      switch (vect[i])
        {
          case 1 :
          case 4 :
          case 5 :
            break;
            default : 
              return FALSE;                           
        }
    }

  return TRUE;
}


/* verifie que les boucles sont permutables a partir de boucles est
jusqua la boucle boucle + nb_loop, vers l'interieur (Voir papier de lam
pour la validitee*/
int 
Interchange_Validity (loop,nb_loop)
     PTR_BFND loop;
     int nb_loop;
{
  PTR_BFND temp;
  PTR_GRAPH graph, pt;
  int i ;
  int nb; /* nombre de boucle */
   
  nb= 0;
  for (temp = loop; temp; temp = Get_Next_Nested_Loop(temp)) nb++;

  graph = Dep_Graph_For_Loop (loop);


  /* on parcour l'ensemble des dependance du graphe */
  for (pt = graph; pt; pt = CHAIN_LIST(pt))
    {
      for (i = 0; i < MAXSUC ; i++)
        {
          if (GRAPH_SUC(pt)[i])
            {
              /* on examine la dependence */
              if (!Test_Vect_Positif(GRAPH_VECT(pt)[i],nb_loop,nb))
                {
                  myfree(); /* on desaloue le graphe */
                  return FALSE;
                }
            }
        }
    }
  myfree(); /* on desaloue le graphe */
  return TRUE;
  
}


Graph_Traversal(graphe)
     PTR_GRAPH graphe;
{
  int i;
  
  if (!graphe)
    return FALSE;
  
  if (GRAPH_MARK(graphe) != 0)
    return TRUE;
  
  GRAPH_MARK(graphe) = 1;
  
  for (i = 0; i < MAXSUC ; i++)
    {
      if (GRAPH_SUC(graphe)[i])
       if (Graph_Traversal(GRAPH_SUC(graphe)[i]))
         {
           GRAPH_MARK(graphe) = 0;
           return TRUE;
         }
      
       else
        break;
    }
  GRAPH_MARK(graphe) = 0;
  return FALSE;
}


/* Verifie qu'il n'y a pas de cycle entre les instructions */
Check_Cycle_In_Dep_Graph(graph)
     PTR_GRAPH graph;
{
  PTR_GRAPH  pt;
  for (pt = graph; pt; pt = CHAIN_LIST(pt))
    {
      Reset_Mark(graph);
      if (Graph_Traversal(pt))
        return TRUE;      
    }
  
  return FALSE;  
}


/* indique si une boucle est parallele */
int 
  Is_Loop_Parallel(loop)
PTR_BFND loop;
{
  PTR_BFND temp,last;
  PTR_GRAPH graph, pt;
  int i ;
  int nbu; /* nombre de boucle */
  int level;
  
  nbu =0;
  for (temp = loop; temp; temp = Get_Previous_Nested_Loop(temp)) 
    {      
      nbu++;
      last = temp;      
    }
  graph = Dep_Graph_For_Loop (last);

  /* on veut la composante du vecteur associe a la boucle */
  level = nbu;
  
  /* on parcour l'ensemble des dependance du graphe */
  for (pt = graph; pt; pt = CHAIN_LIST(pt))
    {
      for (i = 0; i < MAXSUC ; i++)
        {
          if (GRAPH_SUC(pt)[i])
            {
              /* on examine la dependence */
              if (GRAPH_VECT(pt)[i][level] != 1)
                {
                  myfree(); /* on desaloue le graphe */
                  return FALSE;
                }
            }
        }
    }
  myfree(); /* on desaloue le graphe */
  return TRUE;
  
}


/* indique si une boucle est parallele */
int 
  Is_Loop_Vector(loop)
PTR_BFND loop;
{
  PTR_BFND temp;
  PTR_GRAPH graph;
  int nb, resul; /* nombre de boucle */
  
  nb =0;
  for (temp = loop; temp; temp = Get_Next_Nested_Loop(temp)) nb++;
  
  if (nb >1)
    return FALSE;
  
  graph = Dep_Graph_For_Loop (loop);
  resul = Check_Cycle_In_Dep_Graph(graph);
  

  myfree(); /* on desaloue le graphe */
  return  !resul;
  
}



Look_Array_Ref_Linear(expr,symb,linear,notlinear)
     PTR_SYMB *symb;
     PTR_LLND expr;
     int *linear;
     int *notlinear;
{
  int coef[MAXTILE];
  PTR_LLND expr_list;
  int lin;
  int cconst = 0;
   
  if (!expr)
    return;

  if (NODE_CODE(expr) == ARRAY_REF)
    {
      expr_list = NODE_OPERAND0(expr);
      lin = TRUE;      
      while(expr_list)
        {
          if(!Build_Linear_Rep(NODE_OPERAND0(expr_list),coef,symb,
                               MAXTILE, &cconst))
            lin = FALSE;          
          expr_list = NODE_OPERAND1(expr_list); 
        }
      if (lin)
        *linear = *linear +1 ;      
        else
          *notlinear = *notlinear +1 ;      
    } else
      {
        Look_Array_Ref_Linear(NODE_OPERAND0(expr),symb,linear,notlinear);
        Look_Array_Ref_Linear(NODE_OPERAND1(expr),symb,linear,notlinear);
      }  
}


Are_All_Index_Function_Linear(loop,linear,notlinear)
     PTR_BFND loop;
     int *linear;
     int *notlinear;
{
  PTR_BFND first,temp,last;
  PTR_SYMB symb[MAXTILE];
  int i;  

  first = loop;
  for (temp = loop; temp; temp = Get_Previous_Nested_Loop(temp)) 
    {      
      first = temp;      
    }
  
  for (i=0; i< MAXTILE; i++) symb[i]= NULL;
  i = 0;
  for (temp = first; temp; temp = Get_Next_Nested_Loop(temp))
    {      
      symb[i]= BIF_SYMB(temp);
      i++;
    }
  
  last = Get_Last_Node_Of_Stmt(first);
  /* on examine toutes les expressions */
  for (temp = first; temp &&(temp != last); temp = BIF_NEXT(temp))
    {
      Look_Array_Ref_Linear(BIF_LL1(temp),symb,linear,notlinear);
      Look_Array_Ref_Linear(BIF_LL2(temp),symb,linear,notlinear);
      Look_Array_Ref_Linear(BIF_LL3(temp),symb,linear,notlinear);
    }  

  if (*notlinear == 0)
    return TRUE;

  return FALSE;
}


/*********************************************************************
 *************************** Take a look to sets *********************
 *********************************************************************/

Compute_Def_Use_Sets()
{
  (void)build_sets(0,CUR_FILE_GLOBAL_BFND(),NULL,NULL,NULL,NULL,1); 
  (void)build_sets(0, CUR_FILE_GLOBAL_BFND(),NULL,NULL,NULL,NULL,2);
}

Get_Def_Of_Var(var,stmt)
     PTR_SYMB var;
     PTR_BFND stmt;
{
 PTR_REFL defs;
  PTR_REFL temp;
  
  if (!stmt)
     return 0;

  if (!BIF_SETS(stmt))
    {
      printf("No sets for Stmt\n");
      return 0;
    }
  
  printf(" --------------------------- At stmt (line %d) \n  %s \n",
         BIF_LINE(stmt),
         funparse_blck(stmt));
  defs =  SETS_INDEF(BIF_SETS(stmt));
  for (temp = defs; temp; temp = SETS_REFL_NEXT(temp))
    {
      
      
      if (var && (SETS_REFL_SYMB(temp) == var))
      printf("definition  to symbol %s at line %d\n", SYMB_IDENT(var),
             BIF_LINE(SETS_REFL_STMT(temp)));
      else
        if (SETS_REFL_SYMB(temp))
          printf("Contains (node KIND %d) %s from stmt %d (line %d)\n", 
                 BIF_CODE(stmt), 
                 SYMB_IDENT(SETS_REFL_SYMB(temp)), 
                 BIF_CODE(SETS_REFL_STMT(temp)), 
                 BIF_LINE(SETS_REFL_STMT(temp)));       
    }
}



PTR_GRAPH 
  Layered_Graph (graphe, level)
     PTR_GRAPH graphe;
     int level;
{
  PTR_GRAPH pt;
  int i,k;
  char *vect;
  
  if (!graphe) 
    return NULL;
  
  for (pt = graphe; pt; pt = CHAIN_LIST (pt))
    {
      for (k = 0; k < MAXSUC; k++)
        {
          if (GRAPH_SUC (pt) [k])
            {
              vect = GRAPH_VECT (pt) [k];
              for (i = 1; i < level; i++)
                if (vect[i] != 1)
                  {
                    GRAPH_DEP (pt) [k] = NULL;
                    GRAPH_SUC (pt) [k] = NULL;
                    GRAPH_VECT (pt) [k] = NULL;
                  }
            }
        }
    }
  return graphe;
}
