/* Copyright (C) 1992, Irisa, Indiana University, 
   subroutines of the Sigma toolbox */
/***************************************************************************
  *                                                                        *
  *   High level of Sigmacs toolbox for                                    *
  *                                                                        *
  *                                                                        *
  **************************************************************************/
#include <stdio.h>
#include <malloc.h>
#include "gvarargs.h"  /* for special arg passing */
#include "macro.h"
#include "dependence.h"
#include "ext_lib.h"
#include "ext_low.h"
#include "ext_mid.h"

#define MAXTILE 10 /* nombre maximum de boucle que l'on peut tiler */
#define MAX_STMT 100 /* nombre d'instruction d'une boucle */
#define NO_STEP 10000 /* utiliser pour le tilling des windows */
#define MAX_LOOP 50

extern PTR_GRAPH Dep_Graph_For_Loop();
extern PTR_GRAPH Layered_Graph ();
extern PTR_GRAPH Strongly_Connected ();

/* MAPLE_FILE_PREFIX holds the prefix of the path to the maple files used in unimodiular */
#ifndef MAPLE_FILE_PREFIX
#define MAPLE_FILE_PREFIX "./"
#endif
static char *maple_file_prefix = MAPLE_FILE_PREFIX ;


/******************************************************************* 
 *                                                                 *
 * test de la toolbox avec les transformations de haut niveau      *
 * attention ne remet pas a jour les dependances de donnee         *
 *                                                                 *
 *******************************************************************/


/* RQUE pour les predicats on disposera de deux predicat par ensemble de
   de tranformation
       - Le premier lier a la semantique 
       - le deuxieme lier a la structure de la boucle, par exemple normalise
*/

/*
 * predicat de tiling (dependence, parfaitement imbrique, ... 
 */

int
  tiling_p(b, permut)
PTR_BFND b;
int *permut;

{
  return 1;
}

/*
  b boucle la plus externe du tile
  nb nombre de boucle
  size[] taille des tile
*/

/* ATTENTION ON SUPPOSE LA BOUCLE NORMALISEE 
 Utiliser le predicat pour savoir si c'est valide */
void 
tiling(b,nb,size)
     PTR_BFND b;
     int nb;
     int size[];
{
  int i;
  PTR_SYMB symb[MAXTILE]; /* symbole du tiling */
  PTR_BFND body, internal_loop, temp, last_loop;
  PTR_BFND loop;
  PTR_LLND borne_inf[MAXTILE];
  PTR_LLND borne_sup[MAXTILE];
  PTR_LLND step[MAXTILE];
  PTR_LLND expr_index[MAXTILE];
  PTR_TYPE type_entier;
  PTR_SYMB  min_fonc;

  for (i = 0 ; i<MAXTILE; i++)
    {
      symb[i] = NULL;
      borne_inf[i] = NULL;
      borne_sup[i] = NULL;
      step[i] = NULL;
      expr_index[i] = NULL;
    }
  
  /* on va chercher un type */
  type_entier = Return_to_me_type(T_INT);
  if (!type_entier)
    abort();
  
  if (nb > Perfectly_Nested (b))
    {
      Message("Loop are not Perfectly nested or uses continue statement",
	      BIF_LINE(b));
      return;
    }

  /* on recupere les index de boucle */
  temp = b;
  for ( i = 0; i < nb; i++)
    {/* on les suppose parfaitement imbriquees */
      symb[i] = BIF_SYMB (temp);
      temp = Get_Next_Nested_Loop(temp);
    }
 

  /* on cree les nouveaux 
     symboles et on les declare  */

  for ( i = 0; i < nb; i++)
    {
      symb[i+nb] = Generate_Variable_Name (BIF_CP(b), type_entier);
      Declare_Scalar(symb[i+nb],BIF_CP(b));
    }

  /* on cherche la boucle la plus interne 
   et on change le step*/
  internal_loop = Get_Internal_Loop(b);

  last_loop = b;
  for ( i = 0; i < nb; i++)
    {
      if (size[i] != NO_STEP)
        Loop_Set_Step(last_loop,bif_build_int(size[i]+1));
      last_loop = Get_Next_Nested_Loop(last_loop); 
    }
  
  /* on enleve le corps de la boucle la plus interne */
   body = Extract_Loop_Body(internal_loop);

  /* on cree le symbole associe a min si n'exite pas */

  min_fonc = (PTR_SYMB) bif_build_decl (FUNCTION_NAME,"min",type_entier);
  /* attention devrait etre declarer comme externe */
  /* creation des expressions*/
  last_loop = b;
  for ( i = 0; i < nb; i++)
    {
      if (size[i] != NO_STEP)
        {
          borne_inf[i+nb] = bif_build_int(0);
          borne_sup[i+nb] = Make_Function_Call(min_fonc, type_entier,2,
                                               bif_build(SUBT_OP,type_entier,
                                                         bif_copy_expr(Loop_Get_Borne_Sup(last_loop)),
                                                         bif_build(VAR_REF,
                                                                   type_entier,
                                                                   symb[i])
                                                         ),
                                               bif_build_int(size[i])    
                                               );
        }
      else
        {          
          borne_inf[i+nb] = bif_build_int(1);
          borne_sup[i+nb] = bif_copy_expr(Loop_Get_Borne_Sup(last_loop));
        }      
      step[i+nb] =  NULL;
      last_loop = Get_Next_Nested_Loop(last_loop); 
    }
  
  /* borne sup des boucles */
  last_loop = b;
  for ( i = 0; i < nb; i++)
    {
      if (size[i] == NO_STEP)
        Loop_Set_Borne_Sup(last_loop,bif_build_int(1));
      last_loop = Get_Next_Nested_Loop(last_loop); 
    }
  /* creation des nouvelles boucles */
  last_loop = internal_loop;
  for ( i = 0; i < nb; i++)
    {
      loop = Make_For_Loop(symb[i+nb],NULL,NULL,NULL, NULL);
/*      printf("%s",funparse_llnd(borne_sup[i+nb]));*/
      Loop_Set_Step(loop, step[i+nb]);
      Loop_Set_Borne_Inf(loop, borne_inf[i+nb]);
      Loop_Set_Borne_Sup(loop, borne_sup[i+nb]);
      Insert_List_of_BifNode(loop,last_loop,last_loop);      
      last_loop = loop; 
    }

  /* modification du corps de la boucle 
     on change les references au variables */

  for ( i = 0; i < nb; i++)
    {
      if (size[i] != NO_STEP)
        {
          Replace_Var_In_Section(body,NULL,symb[i], 
                             bif_build(ADD_OP,type_entier,
                                       bif_build(VAR_REF,type_entier,
                                                 symb[i]),
                                       bif_build(VAR_REF,type_entier,
                                                 symb[i+nb]))
                             );
        } else
          {
            Replace_Var_In_Section(body,NULL,symb[i],
                                   bif_build(VAR_REF,type_entier,symb[i+nb])
                                   );
          }
    }

  /* on replace le corps de la boucle */
  Insert_List_of_BifNode(body,last_loop,last_loop); 
}

/* ne tile qu'une seule boucle */

void 
strip_mining(b,size)
     PTR_BFND b;
     int size;
{
  int i;
  PTR_SYMB symb[MAXTILE]; /* symbole du tiling */
  PTR_BFND body, internal_loop, last_loop;
  PTR_BFND loop;
  PTR_LLND borne_inf[MAXTILE];
  PTR_LLND borne_sup[MAXTILE];
  PTR_LLND step[MAXTILE];
  PTR_LLND expr_index[MAXTILE];
  PTR_TYPE type_entier;
  PTR_SYMB  min_fonc;

; /* on ne tile qu'une seule boucle */
  
  if (size == NO_STEP)
    return;

  for (i = 0 ; i<MAXTILE; i++)
    {
      symb[i] = NULL;
      borne_inf[i] = NULL;
      borne_sup[i] = NULL;
      step[i] = NULL;
      expr_index[i] = NULL;
    }
  
  /* on va chercher un type */
  type_entier = Return_to_me_type(T_INT);
  if (!type_entier)
    abort();
  
  /* on recupere les index de boucle */
  symb[0] = BIF_SYMB (b);
  /* on cree les nouveaux symboles et on les declare  */
  i =0;
  symb[1] = Generate_Variable_Name (BIF_CP(b), type_entier);
  Declare_Scalar(symb[1],BIF_CP(b));
  internal_loop = b;
  last_loop = b;
  i = 0;
  Loop_Set_Step(b,bif_build_int(size+1));
  /* on enleve le corps de la boucle la plus interne */
  body = Extract_Loop_Body(internal_loop);
  /* on cree le symbole associe a min si n'exite pas */
  min_fonc = (PTR_SYMB) bif_build_decl (FUNCTION_NAME,"min",type_entier);
  /* attention devrait etre declarer comme externe */
  /* creation des expressions*/
  last_loop = b;  i = 0;
  borne_inf[1] = bif_build_int(0);
  borne_sup[1] = Make_Function_Call(min_fonc, type_entier,2,
				       bif_build(SUBT_OP,type_entier,
						 bif_copy_expr(Loop_Get_Borne_Sup(last_loop)),
						 bif_build(VAR_REF,
							   type_entier,
							   symb[0])
						 ),
				       bif_build_int(size)    
				       );
  step[1] =  NULL;
  i = 0;
  /* creation des nouvelles boucles */
  last_loop = internal_loop;
  loop = Make_For_Loop(symb[1],NULL,NULL,NULL, NULL);
  Loop_Set_Step(loop, step[1]);
  Loop_Set_Borne_Inf(loop, borne_inf[1]);
  Loop_Set_Borne_Sup(loop, borne_sup[1]);
  Insert_List_of_BifNode(loop,last_loop, last_loop); 
  last_loop = loop; 
  /* modification du corps de la boucle 
     on change les references au variables */
  Replace_Var_In_Section(body,NULL,symb[0], 
			 bif_build(ADD_OP,type_entier,
				   bif_build(VAR_REF,type_entier,
					     symb[0]),
				   bif_build(VAR_REF,type_entier,
					     symb[1]))
			 );
  /* on replace le corps de la boucle */
  
  Insert_List_of_BifNode(body,last_loop, last_loop); 
}



/* Loop Distribution 
   Prends une boucle et distribue les instruction de cette boucle
   Distribue toutes les instructions si specifier autrement 
   (On donne alors la liste des noeuds a regrouper A voir)
   Pour l'instant all_state est suppose a vrai
   Il faudra aussi faire le predicat associe
*/

PTR_BLOB
  Distribute_Loop(b,all_stmt)
     PTR_BFND b;
     int all_stmt;
{
  PTR_BFND initial_loop, where, current;
  PTR_BFND loop, last_loop,last;
  PTR_BLOB temp, first, blob;
  
  /* on extrait la boucle */
  if (!b)
    return NULL;
  
  last = Get_Last_Node_Of_Stmt(b); /* the CONTROL_END*/
  where = Get_Node_Before(b);
  
  if (!where)
    {
      Message(" Pb in Distribute_Loop",0);
      return NULL;
    }
  
  initial_loop = Extract_Bif_Section (b,last);
  
  /* on cree une boucle pour chaque intruction on extrait chaque 
     instruction que l'on ajoute */
  first = NULL;
  last_loop = where;
  for (temp = BIF_BLOB1(initial_loop); temp && BLOB_NEXT(temp) ; 
       temp = BLOB_NEXT(temp))
    {
      current = BLOB_VALUE(temp);
      Extract_Bif_Section (current,current);
      loop = Make_For_Loop(BIF_SYMB(initial_loop),
                           bif_copy_expr(Loop_Get_Borne_Inf(initial_loop)),
                           bif_copy_expr(Loop_Get_Borne_Sup(initial_loop)),
                           bif_copy_expr(Loop_Get_Step(initial_loop)), 
                           current);
      Insert_List_of_BifNode (loop, last_loop, NULL);
      last_loop = Get_Last_Node_Of_Stmt(loop);
      if (! first)
        {
          first = (PTR_BLOB) bif_make_node (BLOB_KIND);
          BLOB_VALUE (first) = loop;
          blob = first;
          BLOB_NEXT (blob) = NULL;
       }
      else
        {
          BLOB_NEXT (blob) = (PTR_BLOB) bif_make_node (BLOB_KIND);
          blob = BLOB_NEXT (blob);
          BLOB_VALUE (blob) = loop;
          BLOB_NEXT (blob) = NULL;
        }
    }

  return first;
}



PTR_BLOB
  Distribute_Loop_SCC(b)
     PTR_BFND b;
{
  PTR_BFND initial_loop, body, where, current;
  PTR_BFND last_loop,last, internal_loop;
  PTR_GRAPH graph, temp;
  int num_SCC, i;
  PTR_BFND loop[MAX_STMT];
  PTR_BFND last_l[MAX_STMT];
  PTR_BLOB blob, first;
    
  /* on extrait la boucle */
  if (!b)
    return;
  /* allocate space */

  /* the dependency graph first*/
  graph = Dep_Graph_For_Loop(b);
  Strongly_Connected(graph);
  Ordered_SCC(graph);
  Print_Mark(graph);
  num_SCC = Number_of_SCC(graph);
  if (num_SCC == 1)
    return NULL;

  internal_loop = Get_Internal_Loop(b);
  body = Extract_Loop_Body(internal_loop);
  last = Get_Last_Node_Of_Stmt(b); /* the CONTROL_END*/
  where = Get_Node_Before(b);
  if (!body || !where)
    {
      Message(" Pb in Distribute_Loop_SCC",0);
      return NULL;
    }

  initial_loop = Extract_Bif_Section (b,last);
  
  /* on cree une boucle pour chaque SCC 
     instruction que l'on ajoute */

  last_loop = where;
  first = NULL;
  for (i= 1 ; i <= num_SCC; i++)
    { /* on cree une boucle pour chaque composantes */
      loop[i] =  Duplicate_stmts(initial_loop);
      Insert_List_of_BifNode (loop[i], last_loop, NULL);
      last_loop = Get_Last_Node_Of_Stmt(loop[i]);
      loop[i]  = Get_Internal_Loop(loop[i]);    
      last_l[i] = loop[i];
      if (! Check_Cycle_In_Dep_Graph(Dep_Graph_For_Loop (loop [i])))
        if (! first)
        {
          first = (PTR_BLOB) bif_make_node (BLOB_KIND);
          BLOB_VALUE (first) = loop[i];
          BLOB_NEXT (first) = NULL;
          blob = first;
         }
      else
        {
          BLOB_NEXT (blob) = (PTR_BLOB) bif_make_node (BLOB_KIND);
          blob = BLOB_NEXT (blob);
          BLOB_VALUE (blob) = loop[i];
          BLOB_NEXT (blob) = NULL;
        }
    }
  
  for (temp = graph; temp ;  temp = CHAIN_LIST(temp))
    {
      current = GRAPH_BIF(temp);
      Extract_Bif_Section (current,current);
      Insert_List_of_BifNode (current, last_l[GRAPH_ORDER(temp)], NULL); 
      last_l[GRAPH_ORDER(temp)] = current;      
    }
  
  myfree();
  return first;
  
}


Loop_Fusion(loop1,loop2)
     PTR_BFND loop1,loop2;

{
  PTR_BFND body2,before_cend;
  
  if (!loop1 || !loop2)
    return;
  
  if (BIF_CODE(loop1) != FOR_NODE)
    {
      Message("Pb in Loop_Fusion",0);
      return;
    }
  
  if (BIF_CODE(loop2) != FOR_NODE)
    {
      Message("Pb in Loop_Fusion",0);
      return;
    }

  /* on cherche le cend de la boucle interne */
  before_cend = Get_Node_Before(Get_Last_Node_Of_Stmt(loop2));
  /* on enleve le corps de la boucle la plus interne */

  if (before_cend != loop2)
    body2 =  Extract_Bif_Section (BIF_NEXT(loop2), before_cend);
  else
    body2 =  NULL;
  
  /* on extrait le boucle 2 */
  Extract_Bif_Section (loop2, Get_Last_Node_Of_Stmt(loop2));

  /* on remplace la variable d'index par celle de la boucle 1 dans 
   la boucle 2 */
  
  Replace_Var_In_Section(body2,NULL, BIF_SYMB(loop2), 
                         bif_build(VAR_REF,SYMB_TYPE(BIF_SYMB(loop1)),
                                   BIF_SYMB(loop1)));
  

  /* on insere le corps de la boucle 2 dans la boucle 1 */
  before_cend = Get_Node_Before(Get_Last_Node_Of_Stmt(loop1));
  Insert_List_of_BifNode (body2, before_cend, NULL);
}



/* 
 * Procedure de unrolling de boucle (interne ou externe )
 * on suppose les boucles parfaitement imbriquee
 * et sans increment (a voir)
 */


Unroll_Loop(loop_to_unroll, degree)
     PTR_BFND loop_to_unroll;
     int degree;
{
  
  PTR_BFND internal_loop, body, new_body, last, compense;
  PTR_BFND loop;
  PTR_LLND expr, last_step,last_bs;
  PTR_SYMB  symb;
  int i;
  
  
  if (!loop_to_unroll || (degree < 1))
    return 0;

  last_step = Loop_Get_Step(loop_to_unroll);
  last_bs = Loop_Get_Borne_Sup(loop_to_unroll);
    
  compense = Duplicate_stmts(body = Extract_Loop_Body(loop_to_unroll));
  /* on remet le corps de la boucle */
  Insert_List_of_BifNode (body, loop_to_unroll, NULL); 
  /* on cherche la boucle interne */
  symb = BIF_SYMB(loop_to_unroll);
  internal_loop = Get_Internal_Loop(loop_to_unroll);
  body = Extract_Loop_Body(internal_loop);
  
  for (i =degree -1  ; i >= 1 ; i--)
    {
      expr = bif_build(ADD_OP, SYMB_TYPE(symb), 
                       bif_build(VAR_REF,SYMB_TYPE(symb),symb),
                       bif_build_int(i));
      new_body = Duplicate_stmts(body);
      Replace_Var_In_Section(new_body,NULL, symb, expr);
      Insert_List_of_BifNode (new_body, internal_loop,internal_loop);
      

    }
  /* on fait la derniere iteration */
  Insert_List_of_BifNode (body, internal_loop, NULL);
    
  /* on met maintenant a jour la boucle */
  if (last_step )
    {
      Loop_Set_Step(loop_to_unroll, bif_build(MULT_OP, SYMB_TYPE(symb),
                                            bif_copy_expr(last_step),
                                            bif_build_int(degree))); 
    }
  else
    Loop_Set_Step(loop_to_unroll,bif_build_int(degree));

  Loop_Set_Borne_Sup(loop_to_unroll,
                     bif_build(SUBT_OP, SYMB_TYPE(symb),
                               bif_copy_expr(last_bs),
                               bif_build_int(degree)));

  /* on ajoute maintenant le code de compensation */
  last = Get_Last_Node_Of_Stmt(loop_to_unroll);
  /* on cree la boucle de compensation */
  loop = Make_For_Loop(symb,
                       bif_build(ADD_OP, SYMB_TYPE(symb), 
                                 bif_build(VAR_REF,SYMB_TYPE(symb),symb),
                                 bif_build_int(1)),
                       bif_copy_expr(last_bs),
                       last_step, 
                       compense);
  Insert_List_of_BifNode (loop, last, NULL);
  return 1;
}


/* 
 * Procedure qui interchange des boucles 
 * b la boucle de depart
 * permut le tableau de permutation (numerotation commence a 0)
 * n le nombre de boucle consideree a partir de b compris
 * on suppose les boucles imbriquees
 * version trivial de l'interchanging 
 */

Interchange_Loops(b,permut, n)
     PTR_BFND b;
     int * permut;
     int n;
{
  PTR_SYMB tab_symb[MAXTILE];  
  PTR_LLND tab_bf[MAXTILE];
  PTR_LLND tab_bi[MAXTILE];
  PTR_LLND tab_step[MAXTILE];
  PTR_BFND temp;
  int i;
  

  if (!b)
     return 0;
  
  if (n > Perfectly_Nested (b))
    {
      Message("Loop are not Perfectly nested or uses continue statement",
	      BIF_LINE(b));
      return;
    }

  if (BIF_CODE(b) != FOR_NODE)
    {
      Message("Not a for loop in  Interchange_Loops",0);
      return 0;
    }

  temp = b;
  for (i = 0; (i < n) && temp; i++)
    {
      tab_symb[i] = BIF_SYMB(temp);
      tab_bf[i] = Loop_Get_Borne_Sup(temp);
      tab_bi[i] = Loop_Get_Borne_Inf(temp);
      tab_step[i] = Loop_Get_Step(temp);
      temp = Get_Next_Nested_Loop(temp);        
    }

  /* on permute */
  temp = b;
  for (i = 0; (i < n) && temp ; i++)
    {
      BIF_SYMB(temp) = tab_symb[permut[i]];      
      Loop_Set_Borne_Sup(temp, tab_bf[permut[i]]);
      Loop_Set_Borne_Inf(temp, tab_bi[permut[i]]);
      Loop_Set_Step(temp, tab_step[permut[i]]);
      temp = Get_Next_Nested_Loop(temp); 
    }

  return 1;
}


/***************************************************************************
 *                                                                         *
 *                                                                         *
 * Version de transformation pour une matrice unimodulaire                 *
 * on suppose cette transformation comme correcte (appliquer le            *
 * predicat d'abord                                                        *
 * b la boucle externe                                                     *
 * matrice unimodular  (a coefficient entier)                              *
 * n taille de la matrice nxn (on suppose n =2 ou 3 pour le moment)        *
 * La boucle est supposee normalisee                                       *
 *                                                                         *
 ***************************************************************************/

Compute_With_Maple(b,matrice,n,inf,sup,matinv)
     PTR_BFND b;
     int *matrice;
     int n;
     PTR_LLND sup[],inf[];
     int *matinv;
{
  PTR_BFND loop;
  char s[4096];
  char *pt; 
  char *string;
  int i;
  int count;
  char sst[256], name[256];
  char maple_cmd[256];

  extern char *Goto_Next_Char();
  extern char *Goto_Next_Param();
  extern PTR_LLND Convert_expr_in_llnd();
  
  bzero(s,4096);
  bzero(sst,256);
  bzero(name,256);
  bzero(maple_cmd,256);
  /* extract parameters */

  /* send it to maple */
  Start_Maple();
  
  strcpy(maple_cmd, "read `");
  strcat(maple_cmd, maple_file_prefix);
  strcat(maple_cmd, "fourier`;\n");
  Write_to_maple(maple_cmd);

  bzero(maple_cmd, 256);
  strcpy(maple_cmd, "read `");
  strcat(maple_cmd, maple_file_prefix);
  strcat(maple_cmd, "unimodular`;\n");
  Write_to_maple(maple_cmd);

  loop = b;
  Write_to_maple("loop := array ([");  
  for (i=0; i < n; i++)
    {
      Write_to_maple("[");
/*      Write_to_maple("`");*/
      Write_to_maple(SYMB_IDENT(BIF_SYMB(loop)));
/*      Write_to_maple("`");*/
      Write_to_maple(", ");
      string = funparse_llnd(Loop_Get_Borne_Inf(loop));
      Write_to_maple(string);
      printf("%s",string);
      Write_to_maple(", ");
      string = funparse_llnd(Loop_Get_Borne_Sup(loop));
      Write_to_maple(string);
      printf("%s",string);
      Write_to_maple("]");
      loop= Get_Next_Nested_Loop(loop); 
      if (loop)
        Write_to_maple(",");
    }
  Write_to_maple("]);\n");
  Write_Matrix_to_Maple("T",matrice,n);
  Write_to_maple("compute_bound(loop,T);\n");

  Read_from_maple(s,4096);
  printf("%s",s);
  /* on lit la sortie */
  count = 0;
  pt = s;  
  sscanf(pt,"%s",sst);
  while (strcmp(sst,"begin")!= 0)
    {
      pt = Goto_Next_Char(pt);
      if (strcmp(sst,"end") == 0)
        {
          printf("BEGIN not found %s\n", sst);
          Message("Unimodular failed", 0);
          Kill_Maple_Process();
          exit(0);
        }
      if (strlen(sst) == 0)
        {
          printf("BEGIN not found %s\n", sst);
          Message("Unimodular failed", 0);
          Kill_Maple_Process();
          exit(0);
        }
      if (count > 1000)
        {
          printf("BEGIN not found (count...) %s\n", sst);
          Message("Unimodular failed", 0);
          Kill_Maple_Process();
          exit(0);
        }
      count++;     
      sst[0] = '\0';          
      sscanf(pt,"%s",sst);
    }
  /* on lit begin */
  pt = Goto_Next_Char(pt);
  for (i=0; i < n; i++)
    {      
      if (i == 63)
        abort();
      /* on lit niveau */
      pt = Goto_Next_Char(pt);
      /* on lit la valeur de niveau */
      sscanf(pt,"%s",sst);
      pt = Goto_Next_Char(pt);
      /* on lit inf */
      pt = Goto_Next_Char(pt);
      /* on lit trouve l'expression */
      inf[i] = Convert_expr_in_llnd(pt,b);
      printf("%s",funparse_llnd(inf[i]));
      pt =  Goto_Next_Param(pt);
      /* on lit sup */
      pt = Goto_Next_Char(pt);
      /* on lit trouve l'expression */
      sup[i] = Convert_expr_in_llnd(pt,b);
      printf("%s",funparse_llnd(sup[i]));
      pt = Goto_Next_Param(pt);
    } 
  /* skip to get the matrice */
  pt = Goto_Next_Char(pt);
  pt = Goto_Next_Char(pt);
  for (i=0; i < n*n; i++)
    {
      sscanf(pt,"%d",&matinv[i]);
      pt = Goto_Next_Char(pt);
    }
  

  Kill_Maple_Process();
}



/*********************** Fin de la parti maple ****************************/

Unimodular(b,matrice,size)
     PTR_BFND b;
     int *matrice;
     int size;     
{
  /* Aller voir algorithme dans IEEE TRANSACTION ON PARALLEL AND
     DISTRIBUTED SYSTEM VOL 2 NO 4 OCTOBER 91
     PAPER WOLF AND LAM
     */
  int i,j;
  PTR_SYMB symb[MAXTILE],oldsymb[MAXTILE];
  PTR_BFND temp, il, body,main_body; /* il loop interne */
  int M[MAXTILE*MAXTILE];
  int Minv[MAXTILE][MAXTILE]; /* inverse la matrice M */
  int ID[MAXTILE][MAXTILE];
  PTR_LLND expr;
  PTR_TYPE type_entier;
  int n;
  PTR_LLND sup[64],inf[64];
  if (!b)
    return 0;
    
  type_entier = Return_to_me_type(T_INT);
  
  n= 0;
  temp = b;
  for (temp = b; temp; temp = Get_Next_Nested_Loop(temp)) n++;
  
  if (n > Perfectly_Nested (b))
    {
      Message("Loop are not Perfectly nested or uses continue statement",
	      BIF_LINE(b));
      return 0;
    }

  if ((n!=2) &&(n!= 3))
    return 0;

  Compute_With_Maple(b,matrice,n,inf,sup,M);

  for (i=0; i<n;i++)
    for(j=0;j<n;j++)
      {
        if (i==j)
          ID[i][j] = 1;
        else 
          ID[i][j] = 0;
        Minv[i][j] = M[i*n+j];
      }
  
  
  /* Pour tout les boucle on remet a jour borne sup et inf */
  temp = b;
  for ( i = 0; i < n; i++)
    {/* on les suppose parfaitement imbriquees */
      if (sup[i])
        Loop_Set_Borne_Sup(temp,sup[i]);
      if(inf[i])
        Loop_Set_Borne_Inf(temp,inf[i]);
      temp = Get_Next_Nested_Loop(temp);
    }  
  
  /* on extrait le corps de la boucle interne */
 
  /* on cree les nouveau symbole lies au boucle 
     et on les remplaces */
  temp = b;
  for ( i = 0; i < n ; i++)
    {
      symb[i] = Generate_Variable_Name (BIF_CP(b), type_entier);      
      Declare_Scalar(symb[i],BIF_CP(b));
      oldsymb[i] = BIF_SYMB(temp);
      BIF_SYMB(temp) = symb[i];
      temp = Get_Next_Nested_Loop(temp); 
    }
  
  il = Get_Internal_Loop(b);
  body = Extract_Loop_Body(il);
  /* Remplacement des indices dans le corps de la boucles */
  for (i = 0; i < n; i++)
    {
      expr = Make_Linear_Expression_From_Int (n,0,symb,Minv[i]);
      Replace_Var_In_Section(body,NULL, oldsymb[i], expr);
    }

  main_body = Extract_Loop_Body(b);
  for (i = 0; i < n; i++)
    {
      expr = Make_Linear_Expression_From_Int (n,0,symb,ID[i]);
      Replace_Var_In_Section(main_body, NULL,oldsymb[i], expr);
    }

  Insert_List_of_BifNode (main_body, b, NULL);
  Insert_List_of_BifNode (body, il, NULL);
  

  return 1;
  
}


              
/* loop est supposee etre la boucle la plus interne a cause de l'interface
   , on ne regarde 
 * donc pas si des boucles se trouvent dans le corps, les indices du
 * tableau sont ceux de loop et des autres boucles qui se trouvent au dessus
 */

 Expand_Scalar (loop, symb, update)
     PTR_BFND loop;     
     PTR_SYMB symb; 
     int update;
{
  PTR_BFND temp, body, decl,instr, first_loop,bif_use;
  PTR_LLND ll, tab, lhs, tab_use, tab_init, condition;
  PTR_LLND indice_tab[MAX_LOOP];
  PTR_SYMB array, *var_index;
  int i, j, nb_loop, *sup_dim_val, *res, *inf_dim_val;
  int inf_default_dim, sup_default_dim, def, use;
  /* some bugs to check >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
  if (!loop)
    return;
 
  for (i = 0; i < MAX_LOOP; i++)
    indice_tab [i] = NULL;
  
  body = Extract_Loop_Body (loop);
  if (body)
      {
        ll = BIF_LL1 (body);
        if (ll && NODE_SYMB(ll) == symb)
          {
            ll =  BIF_LL2 (body);
            if (Is_Var_In_Expression (ll, symb))
              {
                Message ("Warning variable used as recursive in the body", 
                         BIF_LINE (body));
                Insert_List_of_BifNode (body ,loop, NULL);
                return;
              }
          }
      }

  j = 1;
  sup_default_dim = 100;
  inf_default_dim = 1;
  temp = loop;
  while (temp && (BIF_CODE (BIF_CP (temp)) == FOR_NODE))
    {
      j++;
      temp = BIF_CP (temp);
    }
  first_loop = temp;
  nb_loop = j;
  sup_dim_val = (int *) xmalloc ((j+1) * sizeof(int));
  inf_dim_val = (int *) xmalloc ((j+1) * sizeof(int));
  var_index = (PTR_SYMB *) xmalloc ((j+1) * sizeof (PTR_SYMB));
  bzero (sup_dim_val, (j+1) * sizeof (int));
  bzero (var_index, (j+1) * sizeof (PTR_SYMB));
  bzero (inf_dim_val, (j+1) * sizeof (int));

  temp = body;
  bif_use = NULL;
  def = use = FALSE;
  while (temp && !def && !use)
    {
      if (BIF_CODE (temp) == ASSIGN_STAT)
        {
          ll = BIF_LL1 (temp);
          if (ll && NODE_SYMB (ll) == symb)
            def = TRUE;
          ll = BIF_LL2(temp);
          if (Is_Var_In_Expression (ll, symb))
            {
              bif_use = temp;
              use = TRUE;
            }
        }
      temp = BIF_NEXT (temp);
    }

  temp = loop;
  for (i = 1; i <= nb_loop; i++)
    {
      res =  Evaluate_Expression (Loop_Get_Borne_Sup (temp));
      if (res [0] == -1) 
        sup_dim_val [nb_loop-i+1] = sup_default_dim;
      else
        sup_dim_val [nb_loop-i+1] = res [1];
      res =  Evaluate_Expression (Loop_Get_Borne_Inf (temp));
      if (res [0] == -1)
        inf_dim_val [nb_loop-i+1] = inf_default_dim;
      else
        inf_dim_val [nb_loop-i+1] =  res [1];
      var_index [nb_loop-i+1] = Loop_Get_Index (temp);
      temp = BIF_CP (temp);
    }
  temp = loop;
  while ((BIF_CODE (temp) != PROC_HEDR) && (BIF_CODE (temp) != FUNC_HEDR)
          && (BIF_CODE (temp) != PROG_HEDR) && (BIF_CODE (temp) != GLOBAL))
    {
      temp = BIF_CP (temp);
    }
  
  array = Generate_Variable_Name (temp, Return_to_me_type (T_ARRAY));
  decl = Declare_Array (temp,SYMB_IDENT (array),SYMB_TYPE (symb),
                        nb_loop,sup_dim_val);
  Insert_List_of_BifNode (decl, temp, NULL);
  
  if (use)
    {
      for (i = 1; i <= nb_loop; i++)
        indice_tab [i] = bif_build (VAR_REF,SYMB_TYPE (var_index [i]),
                                    var_index [i]);
      indice_tab [nb_loop] = bif_build (SUBT_OP, Return_to_me_type (T_INT),
                                        indice_tab [nb_loop],
                                        bif_build_int (1));
      tab_use = Make_Array_Ref_With_Tab (array, nb_loop, indice_tab);
      
      for (i = 1; i <= nb_loop; i++)
        indice_tab [i] = bif_build_int (inf_dim_val [i]);
      indice_tab [nb_loop] = bif_build (SUBT_OP, Return_to_me_type (T_INT),
                                        indice_tab [nb_loop],
                                        bif_build_int (1));
      res =  Evaluate_Expression (indice_tab [nb_loop]);
      if (res [0] != -1) 
        indice_tab [nb_loop] = bif_build_int (res [1]);
      tab_init = Make_Array_Ref_With_Tab (array, nb_loop, indice_tab);
    }
   
  for (i = 1; i <= nb_loop; i++)
    indice_tab [i] = bif_build (VAR_REF,SYMB_TYPE (var_index [i]),
                                var_index [i]);
  tab = Make_Array_Ref_With_Tab (array, nb_loop, indice_tab);

  if (use)
    {
      lhs =  bif_build (VAR_REF, SYMB_TYPE (symb),symb);
      instr = Make_Assign (tab_init, lhs);
      Insert_Bifnode_Before (first_loop, instr, NULL);
      Replace_Var_In_Section (body, bif_use, symb,tab_use);
      if (def)
        BIF_LL1 (bif_use) = tab;
      if (BIF_NEXT (bif_use))
        Replace_Var_In_Section (BIF_NEXT(bif_use), NULL,symb,tab);
    }
  else  
    Replace_Var_In_Section (body, NULL,symb, tab);

  Insert_List_of_BifNode (body ,loop, NULL);
  if (update == 1)
    {
      lhs = bif_build (VAR_REF, SYMB_TYPE (symb),symb);
      instr = Make_Assign (lhs, tab);
      temp = Get_Last_Node_Of_Stmt (first_loop);
      Insert_List_of_BifNode (instr, temp, NULL);
    }
  if (use)
    {
      temp = first_loop;
      j = 1;
      while (BIF_CODE (BIF_NEXT(temp)) == FOR_NODE)
        {
          bif_use = Get_Last_Node_Of_Stmt (temp);
          condition = bif_build (LT_OP, Return_to_me_type (T_BOOL),
                                 bif_build (VAR_REF,
                                            SYMB_TYPE (var_index [j]),
                                            var_index [j]),
                                 Loop_Get_Borne_Sup (temp));
          for (i = 1; i <= j; i++)
            indice_tab [i] = bif_build (VAR_REF,SYMB_TYPE (var_index [i]),
                                        var_index [i]);
          for (i = j+1 ; i <= nb_loop; i++)
            indice_tab [i] = bif_build_int (inf_dim_val [i]);
          indice_tab [j] = bif_build (ADD_OP, Return_to_me_type (T_INT),
                                      indice_tab [j],
                                       bif_build_int (1));
          res =  Evaluate_Expression (indice_tab [j]);
          if (res [0] != -1) 
            indice_tab [j] = bif_build_int (res [1]);
          indice_tab [nb_loop] = bif_build (SUBT_OP, Return_to_me_type (T_INT),
                                            indice_tab [nb_loop],
                                            bif_build_int (1));
          res =  Evaluate_Expression (indice_tab [nb_loop]);
          if (res [0] != -1) 
            indice_tab [nb_loop] = bif_build_int (res [1]);
          lhs = Make_Array_Ref_With_Tab (array, nb_loop, indice_tab);
          indice_tab [j] = bif_build (VAR_REF,SYMB_TYPE (var_index [j]),
                                      var_index [j]);
          for (i = j+1; i <= nb_loop; i++)
            indice_tab [i] = bif_build_int (sup_dim_val [i]);
          ll = Make_Array_Ref_With_Tab (array, nb_loop, indice_tab);
          instr = Make_Assign (lhs, ll);
          decl = Make_If_Then_Else (instr, NULL, condition);
          Insert_Bifnode_Before (bif_use, decl, NULL);
          temp = BIF_NEXT (temp);
          j++;
        }
    }
  free (inf_dim_val);        
  free (sup_dim_val);
  free (var_index);
  free (res);
}


PTR_BFND
  Scalar_Forward_Substitution (loop, symb)
     PTR_BFND loop;     
     PTR_SYMB symb; 
{     
     PTR_BFND temp, first, bif;
     PTR_LLND ll, ll1;
     int cpt = 0;
     
     first = Extract_Loop_Body (Get_Internal_Loop (loop));
     for (temp = first;temp;temp = BIF_NEXT (temp))
       {
         if (BIF_CODE (temp) == ASSIGN_STAT)
           {
             ll = BIF_LL1 (temp);
             if ((NODE_CODE (ll) == VAR_REF) && (NODE_SYMB (ll) == symb))
               {
                 cpt++;
                 bif = temp;
                 ll1 = BIF_LL2 (temp);
               }
           }
       }
     if (cpt == 1)
       {
         for (temp = BIF_NEXT (bif);temp;temp = BIF_NEXT (temp))
           {
             ll = BIF_LL1 (temp);
             Replace_Var_In_Expression (ll ,symb, ll1);
             ll = BIF_LL2 (temp);
             Replace_Var_In_Expression (ll ,symb, ll1);
           }
         Insert_List_of_BifNode (first, loop, loop);
         /* sert a replacer la boucle apres l'avoir extraite*/
         temp = Delete_Bif_Node (bif);
       }
     return loop;
   }


/* 
 *
 *  loop normalize transformation  
 *
 */


int 
 Normalized (loop)
     PTR_BFND loop;
{
     int *res;
     
     if (! loop) 
          return FALSE;
     if (BIF_CODE (loop) != FOR_NODE) 
          return FALSE;
     if (BIF_LL2 (loop) != NULL) 
          return FALSE;
     res = Evaluate_Expression (NODE_OPERAND0(BIF_LL1(loop)));
     if ((res [0] == -1) || (res [1] != 1)) 
          return FALSE;
     return TRUE;
}



Normalize_Loop (loop)
     PTR_BFND loop;
{
  PTR_BFND body, instr, bif;
  PTR_LLND borne_inf, borne_sup, step, new_borne_inf, new_borne_sup;
  PTR_LLND var, expr, temp, temp1, func, rhs, lhs;
  PTR_SYMB new_varindex, varindex, symb;
  int max, *res;
  
  if (Normalized (loop)) return;
  borne_sup = Loop_Get_Borne_Sup (loop);
  new_borne_inf = bif_build_int (1);
  borne_inf = Loop_Set_Borne_Inf (loop,new_borne_inf);
  step = Loop_Set_Step (loop,NULL);

  if (step != NULL)
    {
      new_borne_sup = bif_build (DIV_OP, Return_to_me_type (T_INT),NULL,step);
      temp = bif_build (ADD_OP, Return_to_me_type (T_INT),NULL,step);
      NODE_OPERAND0 (new_borne_sup) = temp;
      temp1 = bif_build (SUBT_OP, Return_to_me_type (T_INT),
                         borne_sup,borne_inf);
      NODE_OPERAND0 (temp) = temp1;
    }
  else
    {
      new_borne_sup = bif_build (ADD_OP, Return_to_me_type (T_INT),NULL,
                                 bif_build_int (1));
      temp1 = bif_build (SUBT_OP, Return_to_me_type (T_INT),
                         borne_sup,borne_inf);
      NODE_OPERAND0 (new_borne_sup) = temp1;
    }
  
  res = Evaluate_Expression (new_borne_sup);
  if (res [0] != -1) 
    new_borne_sup = bif_build_int (res [1]);
  borne_sup = Loop_Set_Borne_Sup (loop,new_borne_sup);

  new_varindex = Generate_Variable_Name (BIF_CP(loop), Return_to_me_type (T_INT)); 

  Declare_Scalar (new_varindex, BIF_CP(loop));
  varindex = Loop_Set_Index (loop,new_varindex);
  var = bif_build (VAR_REF, Return_to_me_type (T_INT),new_varindex);
  expr = bif_build (ADD_OP, Return_to_me_type (T_INT),borne_inf,NULL);
  if (step != NULL)
    {
      temp = bif_build (MULT_OP, Return_to_me_type (T_INT),NULL,step);
      NODE_OPERAND1 (expr) = temp;
      temp1 = bif_build (SUBT_OP, Return_to_me_type (T_INT),var,
                         bif_build_int (1));
      NODE_OPERAND0 (temp) = temp1;
    }
  else
    {
      temp1 = bif_build (SUBT_OP, Return_to_me_type (T_INT),var,
                         bif_build_int (1));
      NODE_OPERAND1 (expr) = temp1;
    }
  
  body = Extract_Loop_Body (loop);
  Replace_Var_In_Section (body, NULL,varindex, expr);
  Insert_List_of_BifNode (body, loop, loop);
  
  if (res [0] != -1)
    {
      if (res [1] < 0)
        {
          max = 0;
          res = Evaluate_Expression (borne_inf);
          if (res [0] != -1)
            rhs = bif_build_int (res[1]);
          else
            rhs = borne_inf;
        }
      else
        {
          max = res [1];
          temp = bif_build (ADD_OP, Return_to_me_type (T_INT),borne_inf,NULL);
          if (step != NULL)
            temp1 = bif_build (MULT_OP, Return_to_me_type (T_INT),
                               bif_build_int (max),step);
          else
            temp1 = bif_build_int (max);
          NODE_OPERAND1 (temp) = temp1;
          res = Evaluate_Expression (temp);
          if (res [0] != -1)
             rhs = bif_build_int (res[1]);
          else
            rhs = temp;
        }
    }
  else  /* res [0] == -1 */
    {
      symb = get_function_symb ("max");
      if (symb == NULL)
        {
          symb = (PTR_SYMB) bif_build_decl (FUNCTION_NAME,"max",
                                            Return_to_me_type (T_INT));
        }
      func = Make_Function_Call (symb,Return_to_me_type (T_INT),2,
                                 new_borne_sup,bif_build_int (0));
      rhs = bif_build (ADD_OP,Return_to_me_type (T_INT),borne_inf,NULL);
      if (step != NULL)
        {
          temp = bif_build (MULT_OP, Return_to_me_type (T_INT),func,step);
          NODE_OPERAND1 (rhs) = temp;
        }
      else
        NODE_OPERAND1 (rhs) = func;
    }
  lhs = bif_build (VAR_REF,Return_to_me_type (T_INT));
  NODE_SYMB (lhs) = varindex;
  instr = Make_Assign (lhs,rhs);
  bif = Get_Last_Node_Of_Stmt (loop);
  Insert_List_of_BifNode (instr, bif, NULL);
  free (res);
}


 
/* should be use with the predicate Is_Loop_Vector (bif)*/

int 
  Vectorize (bifin)
     PTR_BFND bifin;
{
  PTR_LLND borne_inf, borne_sup;
  PTR_SYMB varindex;
  PTR_BFND temp, temp1;
  PTR_BFND bif,last;

  bif = bifin;
  if (!bif)
    return FALSE;
  
  if (BIF_CODE (bif) != FOR_NODE)
    return FALSE;

  varindex = NULL;
  borne_inf = borne_sup = NULL;
  temp = temp1 = NULL;
  
  borne_inf = Loop_Get_Borne_Inf (bif);
  borne_sup = Loop_Get_Borne_Sup (bif);
  varindex = Loop_Get_Index (bif);
  temp = Extract_Loop_Body (bif);
  if (temp)
    {      
      temp1 = Get_Last_Node_Of_Stmt (bif);      
      Replace_Index_In_Section (temp, varindex, borne_inf, borne_sup);
      temp1 = Delete_Bif_Node (temp1);
      bif = Delete_Bif_Node (bif);
      Insert_List_of_BifNode (temp,bif,NULL);
/*      if (BIF_CP(temp))
        printf("%s",funparse_blck(BIF_CP(temp)));
      else
         printf("%s",funparse_blck(temp));*/
    } else
      return FALSE;
      
  return TRUE;
}

int 
  Flow_Cycle (graphe)
     PTR_GRAPH graphe;
{
  PTR_GRAPH pt;
  int i;
  PTR_DEP dep;
  
  if (!graphe)
    return FALSE;
  
  if (GRAPH_MARK (graphe) != 0) 
    return TRUE;

  GRAPH_MARK (graphe) = 1;
  
  for (i = 0; i < MAXSUC; i++)
    {
      if (GRAPH_SUC (graphe) [i])
        {
          pt = GRAPH_SUC (graphe) [i];
          dep = GRAPH_DEP (graphe) [i];
          if (DEP_TYPE (dep) == 0)
            if (Flow_Cycle (pt))
              {
                GRAPH_MARK (graphe) = 0;
                return TRUE;
              }
        }
      else
        break;
    }
  
  GRAPH_MARK (graphe) = 0;          
  return FALSE;
} 

int 
  Vectorizable_Loop (loop,level)
     PTR_BFND loop;
     int level;
{
  PTR_GRAPH graphe,pt;
    
  graphe = Dep_Graph_For_Loop(loop);
  graphe = Layered_Graph (graphe, level);
  for (pt = graphe; pt; pt = CHAIN_LIST (pt))
    {
      Reset_Mark (graphe);
      if (Flow_Cycle (graphe))
          return FALSE;
    }
  return TRUE;
}


int 
  *Vectorizable_Nest (loop)
     PTR_BFND loop;
{
  int nb_loop, i, k;
  int *list;

  nb_loop = Perfectly_Nested (loop);
  list = (int *) xmalloc ((nb_loop+1) * sizeof(int));
  bzero (list, (nb_loop+1) * sizeof (int));
  k = 0;
  for (i = 1; i <= nb_loop+1; i++)
    {
      list [i] = Vectorizable_Loop (loop, i);
      if (list [i])
        k++;
     }
  list [0] = k;
  return list;
}

int
  Vectorize_Nest (loop)
     PTR_BFND loop;
{
  int nb_loop, i, k, limit;
  PTR_BFND b;
  PTR_GRAPH graphe, pt;
  char *vect;
  int *list,*permut;
    
  if (!loop)
    return 0;
  
  if (BIF_CODE (loop) != FOR_NODE) 
    return 0;

  nb_loop = Perfectly_Nested (loop);

    
 /* the dependency graph first*/
  graphe = Dep_Graph_For_Loop(loop);
  
  Print_Dep_For_Loop (stdout, loop);


  list = Vectorizable_Nest (loop);

  for (i = 1; i <= nb_loop;i++)
    if (list [i])
       printf ("Vectorizable loop %d\n",i);
      else
        printf ("Non vectorizable loop %d\n",i);
  printf ( "%d loops vectorizables\n",list[0]);

  permut = (int *) xmalloc ((nb_loop) * sizeof(int));
  bzero (permut, (nb_loop) * sizeof (int));
  
  for (i = nb_loop; i >= 1; i--)
    {
      if (!list [i])
        for (k = 1; k <= nb_loop; k++)
          if (list [k])
            {
              permut [k-1] = i-1;
              permut [i-1] = k-1;
            }
      else 
        permut [i-1] = i-1;
    }
  
  Interchange_Loops (loop,permut,nb_loop);
 
  for (pt = graphe; pt; pt = CHAIN_LIST (pt))
    for (k = 0; k < MAXSUC; k++)
      if (GRAPH_SUC (pt) [k])
        Print_Dep (stdout,GRAPH_DEP (pt) [k]);
  
  for (i = nb_loop; i<= list[0]-1; i--)
    {
      graphe = Layered_Graph (graphe,i);
/*      if (! Check_Cycle_In_Dep_Graph (graphe))
        {*/
          
      
    }
  
    
   /* Strongly_Connected(graphe);
  Ordered_SCC(graphe);
  Print_Mark(graphe);
  num_SCC = Number_of_SCC(graphe);
  if (num_SCC == 1)
    return NULL;*/

  
/*
  nb_loop = Perfectly_Nested (loop);
  for (i = nb_loop; i > 0; i--)
    {
      b = loop;
      for (k = i; k > 1 ; k--)
          b =  Get_Next_Nested_Loop(b);
      if (Verif_Assign (b))
        Vectorize (b);
    }*/
return 1;  
}

     
       

/*
 * 
 * Donne les proprietes d'un nid de boucle 
 * et indique les transformations valides.
 *
 *
 */

Print_Property_For_Loop(f,b, level)
     FILE *f;
     PTR_BFND b;
     int level;
{
  PTR_BFND loop,temp;
  int Ok;
  int nb;
  int linear = 0;
  int notlinear = 0;
  int inter;
  int cl;
  int converti;
  extern char * funparse_blck();

  if (!b)
    return;

  if (BIF_CODE(b) != FOR_NODE)
    return;
  
  if (level == 1)
    {      
      fprintf(f,"\n\n*********************************************************\n");
      fprintf(f,"Property of the loop nest line %d \n", BIF_LINE(b));
      fprintf(f,"-------------------------------------\n");
      
      fprintf(f,"%s",funparse_blck(b));

      converti = Loop_Conversion (b);
      if (converti)
        {
          fprintf(f,"-------------------------------------\n");
          fprintf(f,"Converted loop if not enddo ---------\n");
          fprintf(f,"-------------------------------------\n");
          fprintf(f,"%s",funparse_blck(b));
        }
      

      fprintf(f,"\n");
      fprintf(f,"Loops perfectly Nested on %d level(s)\n", Perfectly_Nested (b));
      nb= 0;
      for (temp = b; temp; temp = Get_Next_Nested_Loop(temp)) nb++;
      if (inter = Interchange_Validity (b,nb))
        fprintf(f,"Loop Is Fully Interchangeable\n");
      else 
        fprintf(f,"Loop Is not Fully Interchangeable\n");

      
      if (!No_Dependent_Index (b))
        {
         fprintf(f,"-----> Loop Nest Has Dependent Indexes\n");           
        }
      
      if (!No_Basic_Induction (b))
        {
          fprintf(f,"Loop Nest Has Basic_Induction Variables (right)\n"); 
        }
      
      
        if (!No_Def_Of_Induction (b))
          {
            fprintf(f,"Loop Nest Has Dependant Induction Variables\n"); 
          }
      
      

      Are_All_Index_Function_Linear(b,&linear,&notlinear);
      fprintf(f,"Nombre de references tableaux lineaire : %d\n",linear);
      fprintf(f,"------------------------- non lineaire : %d\n",notlinear);

      Ok = TRUE;
      cl = 0;      
      for (temp = b; temp && Ok; temp = Get_Next_Nested_Loop(temp))
        {
          cl++;          
          if (Verif_Assign(temp) || Verif_Assign_If(temp))
            Ok = FALSE;          
        }

      if ((notlinear == 0) && (inter) && Ok && (Perfectly_Nested (b) == cl))
        {
          fprintf(f,"Locality Estimation  : Sorry skipped\n");
          fprintf(f,"-------------------\n");
          /* Print_Window_Report(f,b);*/
        }
      
      else
        {
          fprintf(f,"Locality Estimation  : Aborted (Loop properties)\n");
          fprintf(f,"-------------------\n");
        }
            
      
      fprintf(f,"\n\n\n");
    }
  
  fprintf(f,"Property of the loop line %d \n", BIF_LINE(b));
  fprintf(f,"-------------------------------\n");
  if (Normalized(b))
    fprintf(f,"Loop Normalized\n");
  if (!Is_Good_Loop (b))
    fprintf(f,"Warning ENDDO is preferred\n");
  if (Is_Loop_Parallel(b))
    fprintf(f,"Parallel Loop\n");
  if (Is_Loop_Vector(b))
    {
      fprintf(f,"Vector Loop\n");
    }
  
  
  fprintf(f,"\n\n\n");

  Print_Property_For_Loop(f,Get_Next_Nested_Loop(b), level+1);  

  if (level == 1)
    { 
      loop = Get_Internal_Loop(b);
      if (Verif_Assign(loop) == FALSE) 
        fprintf(f,"Warning Not Only Assign Statement\n");
      if ( Verif_Assign_If(loop) == FALSE) 
        fprintf(f,"Warning Not Only Assign and If Statement\n");
      Print_Dep_For_Loop(f,b);
    }
}
