/* Copyright (C) 1992, Irisa, Indiana University, 
   subroutines of the Sigma toolbox */
 /**************************************************************************
  *                                                                        *
  * interface entre le systeme et maple                                    *
  *                                                                        *
  *                                                                        *
  **************************************************************************/


/***************************** Set the machine **********************/
/* define SUN 1 for a sun that wants vfork.h 
/********************************************************************/

#include <stdio.h>
#include <malloc.h>
#include "gvarargs.h"  /* for special arg passing */
#include "macro.h"
#include "dependence.h"
#include <sys/ioctl.h>
#include <fcntl.h>
#ifdef SUN
#include <vfork.h>
#endif
#include <signal.h>

#define MAPLE_ACTIF 1
#define MAPLE_DOWN  0
#define MAPLE_ABORT -1

/* the following two defines are installation dependent */
#ifndef MAPLE_BIN
#define MAPLE_BIN "/usr/site/share/maple/v4.3/sun4/maple"
#endif

#ifndef MAPLE_LIB
#define MAPLE_LIB "/usr/site/share/maple/v4.3/lib4.3"
#endif

#define MAPLE_FLUSH "lprint(`!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!`);"
#define MAPLE_FLUSH_MARKER "lprint(`$$$`);"

static int Maple_Process_Status = MAPLE_DOWN;
static int Maple_Process_id = -1;
int pipe_maple_read[2];
int pipe_maple_write[2];


/*
 *
 * Create the maple process (if fork() make the process being maple)
 *
 */

Create_Maple_Process()
{
  char *maple_args[5];

  maple_args[0] = (char *) xmalloc(256);
  maple_args[1] = (char *) xmalloc(256);
  maple_args[2] = (char *) xmalloc(256);
  maple_args[3] = (char *) xmalloc(256);
  strcpy(maple_args[0],MAPLE_BIN);
  strcpy(maple_args[1],"-b");
  strcpy(maple_args[2],MAPLE_LIB);
  strcpy(maple_args[3],"-q");
  maple_args[4] = NULL;
  /* -b MAPLE_LIB */
  execv(MAPLE_BIN,maple_args);
}

int 
Check_Maple_Status()
{
  

}

Kill_Maple_Process()
{
  if (kill(Maple_Process_id, SIGKILL) == -1)
    {         
      Message("kill aborting \n",0);
      exit(0);
    }
  Maple_Process_Status = MAPLE_DOWN;
  close(pipe_maple_read[0]);
  close(pipe_maple_read[1]);
  close(pipe_maple_write[0]);
  close(pipe_maple_write[1]);
  
}


/*
 *
 * Start the maple process if not active 
 * report success or failure 
 *
 */
int 
  Start_Maple()
{
  
  if (Maple_Process_Status == MAPLE_DOWN)
   {
     if ((pipe(pipe_maple_read) == -1) || (pipe(pipe_maple_write) == -1))
       {         
         Message("Pipe aborting \n",0);
         exit(0);
       }
     
     if ((Maple_Process_id=fork()) != 0)
       {/* processus parent */
         /* on cherche maintenant a modifer les entrees sortie de maple */
         if (Maple_Process_id == -1)
           {             
             Maple_Process_Status = MAPLE_ABORT;
             Message("Cannot Start Maple (abort()) \n",0);
             exit(1);             
           }         
         else
           {
             Maple_Process_Status = MAPLE_ACTIF;
           }    
       }     
     else
       { 
         /* redirection en lecture */
         /* ici c'est le fils qui commence par fermer le cote lecture
            du pipe en laissant le cote ecriture ouvert.Cette
            fermeture est evidemment non obligatoire.. */

         close(pipe_maple_read[0]);

         /* puis il ferme stdout */
         close(1);

         /* puis dup qui prend le premier descripteur de fichier libre
            (1==stdout qu'on vient de fermer)
            l'associe avec pipdes[1] */
         dup(pipe_maple_read[1]);

         /* puis il ferme le pipe (encore une fois facultatif)*/
         close(pipe_maple_read[1]);

         /* redirection pour l'envoi a maple */
         close(pipe_maple_write[1]);
         close(0);
         dup(pipe_maple_write[0]);
         close(pipe_maple_write[0]);
         Create_Maple_Process();
       }
   } 
  else
    /* report status */
    return Maple_Process_Status; 
}


Write_to_maple(s)
     char *s;
{
  if (Maple_Process_Status == MAPLE_ACTIF)
    {
      write(pipe_maple_write[1],s,strlen(s));
      return TRUE;      
    } else
      {
        return FALSE;        
      }
}

Maple_Flush_Command()
{
  int i;
  /* UGLY TRICK... maple command to flush the output */
  Write_to_maple(MAPLE_FLUSH_MARKER);
  for (i=1; i< 100; i++)
    Write_to_maple(MAPLE_FLUSH);
}

Read_from_maple(s,size)
     char *s;
     int size;
{
  int i;
  int count;
    
  if (Maple_Process_Status == MAPLE_ACTIF)
    {
      Maple_Flush_Command();
      count = 0;
      for (i = 0; i < size; i++)
        {
          read(pipe_maple_read[0],&s[i],1);
          if (s[i] == '$')
            count++;
          else
            count =0;
          if (count == 3)
            break;
          if (s[i] == '!')
            i--;     
          if (s[i] == '\n')
            s[i] = ' ';
        }
      s[i] = '\0';      
      return TRUE;      
    } else
      {
        return FALSE;     
      }
}


Write_Matrix_to_Maple(name,matrice,n)
     char *name;
     int *matrice;
     int n;
{
  int i,j;
  char string[256];
  
  Write_to_maple(name);
  Write_to_maple(" := array([");
  
  for (i=0; i <n; i++)
    {   
      Write_to_maple("[");
      for (j=0;j<n;j++)
        {
          sprintf(string,"%d",matrice[j+n*i]);
          Write_to_maple(string);          
          if (j< n-1)
            Write_to_maple(", ");
        }
      Write_to_maple("]");
      if (i< n-1)
        Write_to_maple(", ");
    }
  Write_to_maple("]);\n");

}

/* convertie une chaine sous la forme d'un arbre de syntaxe 
 par exemple (min x y z (+ 2 (- id 3)))*/

static PTR_SYMB min_func = NULL;
static PTR_SYMB max_func = NULL;
static PTR_SYMB ceil_func = NULL;
static PTR_SYMB floor_func = NULL;

char *
Goto_Next_Char(pt)
     char *pt;
{
  int i;
  char c;
  
  
  c = pt[0];
  i = 0;
  /* on trouve d'abord la chaine */
  while (( c == ' ')
         || ( c == '\t')
         || ( c == '\n')
         )
    {
      i++;           
      c  =  pt[i];           
    }
  while (( c != ' ')
         && ( c != '\t')
         && ( c != '\n')
         )
    {
      i++;           
      c  =  pt[i];           
    }
  return &pt[i];
}


char *
Goto_Next_Param(pt)
     char *pt;
{
  int i;
  char c;
  int count = 1;  
  
  while (*pt == ' ') pt++;
  c = pt[0];
  if (c != '(')
    return Goto_Next_Char(pt);
  i = 1;
  c =  pt[1];
  /* on trouve d'abord la chaine */
  while (count != 0)
    {
      if (c == ')')
        count--;
      if (c == '(')
        count++;
      if ((c == '\0') || (c == EOF))
        return NULL;      
      i++;           
      c  =  pt[i];           
    }
  return &pt[i];
}

#define MAXLEN 256

PTR_LLND
Convert_expr_in_llnd(string,scope)
     char *string;    
     PTR_BFND scope;     
{
  PTR_TYPE type_int;
  int i = 0;
  char s1[MAXLEN],s2[MAXLEN];
  char *pt;
  PTR_LLND llnd=0;
  PTR_LLND parameter[10];
  PTR_SYMB symb;
  int nb_param;
  extern PTR_TYPE Return_to_me_type();
  extern PTR_SYMB bif_build_decl();
  extern PTR_LLND bif_build();
  extern PTR_LLND bif_build_int();
  extern PTR_SYMB Get_Symb();
  extern PTR_LLND Make_Function_Call_bis();
  
  bzero(s1,MAXLEN);
  bzero(s2,MAXLEN);

  if (!min_func)
    {      
      type_int = Return_to_me_type(T_INT);
      min_func = (PTR_SYMB) bif_build_decl (FUNCTION_NAME,"min",type_int);
      max_func = (PTR_SYMB) bif_build_decl (FUNCTION_NAME,"max",type_int);
      ceil_func = (PTR_SYMB) bif_build_decl (FUNCTION_NAME,"ceil",type_int);
      floor_func = (PTR_SYMB) bif_build_decl (FUNCTION_NAME,"floor",type_int);
    }
  
  pt = string;
  while (*pt == ' ') pt++; /* skip the blank */
  switch (pt[0])
    {
    case ')':
      return NULL;      
      break;
    case '(':      
      /* we have an operator */
      pt++;      
      while (*pt == ' ') pt++;
      for (i = 0; i < MAXLEN; i++)
        {
          if ((pt[i] == ' ') || (pt[i] == ')'))
            break;          
          s1[i] = pt[i];          
        }
      s1[i] = '\0';      
      break;
    case '0':
    case '1':
    case '2':
    case '3':
    case '4':
    case '5':
    case '6':
    case '7':
    case '8':
    case '9':
    case '-':
    case '+':
      for (i = 0; i < MAXLEN; i++)
        {
          if ((pt[i] == ' ') || (pt[i] == ')'))
            break;          
          s1[i] = pt[i];          
        }
      s1[i] = '\0';
      /* on examine s1 pour savoir si div */
      {
        int div_found = 0;
        for (i = 0; i < MAXLEN; i++) s2[i] = '\0';
        for (i = 0; i < MAXLEN; i++)
          {
            s2[i] = s1[i];
            if (s1[i] == '/')
              {
                div_found = 1;
                break;          
              }
          }
        if (div_found == 1)
          { /* generate a div_op */
            
            return bif_build(DIV_OP,NULL,bif_build_int(atoi(s2)),
                             bif_build_int(atoi(&(s1[i+1])))); 
          }        
      }     
      return bif_build_int(atoi(s1));
      break;
    default:
      /* an identifier */
      for (i = 0; i < MAXLEN; i++)
        {
          if ((pt[i] == ' ') || (pt[i] == ')'))
            break;          
          s1[i] = pt[i];          
        }
      s1[i] = '\0';
      symb = Get_Symb(s1, scope);
      if (!symb)
        symb = bif_build_decl (VARIABLE_NAME,s1,Return_to_me_type(T_INT));
      return bif_build(VAR_REF,Return_to_me_type(T_INT),symb);
    }

  if (0 == strcmp("min", s1))
    {
      nb_param = 1;
      /* on lit le parametre 1*/
      pt = Goto_Next_Char(pt);    
      parameter[0] = llnd = Convert_expr_in_llnd(pt,scope);
      /* on traite les parametre */
      while ((llnd != NULL) && (nb_param <10))
        {
          pt = Goto_Next_Param(pt);
          parameter[nb_param] =  llnd = Convert_expr_in_llnd(pt,scope);
          nb_param ++;          
        }
      if (nb_param >= 10)
        abort();  
      return Make_Function_Call_bis(min_func,NULL,nb_param-1,parameter);      
    }
  
  
  if (0 == strcmp("max", s1))
    {
      nb_param = 1;
      /* on lit le parametre 1*/
      pt = Goto_Next_Char(pt);    
      parameter[0] = llnd = Convert_expr_in_llnd(pt,scope);
      /* on traite les parametre */
      while ((llnd != NULL) && (nb_param <10))
        {
          pt = Goto_Next_Param(pt);
          parameter[nb_param] =  llnd = Convert_expr_in_llnd(pt,scope);
          nb_param ++;          
        }
      if (nb_param >= 10)
        abort();  
      return Make_Function_Call_bis(max_func,NULL,nb_param-1,parameter);      
    }
  
  
  if (0 == strcmp("ceil", s1))
    {
      nb_param = 1;
      /* on lit le parametre 1*/
      pt = Goto_Next_Char(pt);    
      parameter[0] = llnd = Convert_expr_in_llnd(pt,scope);
      return Make_Function_Call_bis(ceil_func,NULL,1,parameter);      
    }
  
  
  if (0 == strcmp("floor", s1))
    {
      nb_param = 1;
      /* on lit le parametre 1*/
      pt = Goto_Next_Char(pt);    
      parameter[0] = llnd = Convert_expr_in_llnd(pt,scope);
      return Make_Function_Call_bis(floor_func,NULL,1,parameter);      
    }
  
  if (0 == strcmp("+", s1))
    {

      /* on lit le parametre 1*/
      pt = Goto_Next_Char(pt);    
      parameter[0] = llnd = Convert_expr_in_llnd(pt,scope);
      /* on traite les parametre */
      pt = Goto_Next_Param(pt);
      parameter[1] =  llnd = Convert_expr_in_llnd(pt,scope);
      if (!parameter[1]) return parameter[0];
      return bif_build(ADD_OP,NULL,parameter[0],parameter[1]);      
    }

  
  if (0 == strcmp("*", s1))
    {

      /* on lit le parametre 1*/
      pt = Goto_Next_Char(pt);    
      parameter[0] = llnd = Convert_expr_in_llnd(pt,scope);
      /* on traite les parametre */
      pt = Goto_Next_Param(pt);
      parameter[1] =  llnd = Convert_expr_in_llnd(pt,scope);
      if (!parameter[1]) return parameter[0];
      return bif_build(MULT_OP,NULL,parameter[0],parameter[1]);      
    }

  
  if (0 == strcmp("-", s1))
    {

      /* on lit le parametre 1*/
      pt = Goto_Next_Char(pt);    
      parameter[0] = llnd = Convert_expr_in_llnd(pt,scope);
      /* on traite les parametre */
      pt = Goto_Next_Param(pt);
      parameter[1] =  llnd = Convert_expr_in_llnd(pt,scope);
      return bif_build(SUBT_OP,NULL,parameter[0],parameter[1]);      
    }

  
  if (0 == strcmp("/", s1))
    {

      /* on lit le parametre 1*/
      pt = Goto_Next_Char(pt);    
      parameter[0] = llnd = Convert_expr_in_llnd(pt,scope);
      /* on traite les parametre */
      pt = Goto_Next_Param(pt);
      parameter[1] =  llnd = Convert_expr_in_llnd(pt,scope);
      if (!parameter[1]) return parameter[0];
      return bif_build(DIV_OP,NULL,parameter[0],parameter[1]);      
    }

  return NULL;  
}
