/* Employed by Jenq-Kuen Lee for VPC++ Nov 19 ,1987 */
/* Original Filename : hash.c                       */
/* New filename      : vhash.c                      */
/* $Header: /Users/nsundare/faust/vpc/RCS/vhash.c,v 2.0 92/04/02 23:09:10 nsundare Exp $ */

/*
 * hash.c -- hash table routines
 */

#include <stdio.h>
#include <strings.h>

#include "vpc.h"
#include "db.h"
#include "defines.h"
#include "vextern.h"

PTR_BFND cur_scope(),pre_scope(),real_scope() ;
PTR_SYMB make_symb();
PTR_BLOB follow_blob();
PTR_SYMB re_install_entry();
PTR_HASH hash_table[hashMax];


/*
 * init_hash -- initialize the hash table
 */
void
init_hash()
{
	register int i;

	for (i = 0; i < hashMax; i++)
		hash_table[i] = HSNULL;
}


/*
 * Hash(string) -- compute hash value of string.
 */
int 
hash(string)
	register char *string;
{
	register int i;

	for (i = 0; *string;)
		i += *string++;
	return (i % hashMax);
}


/*
 * look_up(string) -- lookup string in the hash table and
 *		      install it if not there
 */
PTR_HASH 
look_up(string)     
	register char *string;
{
	int     index;
	register PTR_HASH entry;

	index = hash(string);
	for (entry = hash_table[index]; entry; entry = entry->next_entry)
		if (!strcmp(string, entry->ident))
			break;

	if (!entry) {	 /* since hash entry was not found make one */
		entry = ALLOC(hash_entry);
		entry->ident = copys(string);
		entry->next_entry = hash_table[index];
		hash_table[index] = entry;
	}
	return (entry);
}


/*
 * look_up_type(string,type_status) 
 *     <1> lookup string in the hash table and install it if not there
 *         exactly same behavior like look_up
 *     <2> type_status called by out
 *         ON  : If the string has been a typename already.
 *         OFF : If not foumd or Has been found but not typename
 *
 */
PTR_HASH 
look_up_type(string,type_status)     
	register char *string;
        int *type_status ;
{
	int     index;
	register PTR_HASH entry;
        PTR_SYMB symb_ptr;


        *type_status = OFF;
	index = hash(string);
	for (entry = hash_table[index]; entry; entry = entry->next_entry)
		if (!strcmp(string, entry->ident))
			break;

	if (!entry) {	 /* since hash entry was not found make one */
		entry = ALLOC(hash_entry);
		entry->ident = copys(string);
                entry->id_attr = SMNULL ;
		entry->next_entry = hash_table[index];
		hash_table[index] = entry;
	}
        else {
	  for (symb_ptr =entry->id_attr ;(symb_ptr && (! (*type_status))); symb_ptr = symb_ptr->outer)
	   { 	  if ((symb_ptr->variant==TYPE_NAME)||
		     (symb_ptr->variant == VARIABLE_NAME)||
		     (symb_ptr->variant == UNION_NAME)||
		     (symb_ptr->variant == CLASS_NAME)||
		     (symb_ptr->variant == STRUCT_NAME)||
                     (symb_ptr->variant == ENUM_NAME))
		       {  PTR_BFND flow ;
                          PTR_BFND type_scope ;                    
                          
                          type_scope = real_scope(symb_ptr->scope) ;
                          for ( flow = cur_scope(); flow ; flow = pre_scope(flow) )
                           {   if (flow == type_scope) 
                                {   
				  if (symb_ptr->variant == TYPE_NAME)
				    *type_status = ON ;
				  return(entry);
                                }
                           }
                       }
            }     
	}

	return (entry);
      }






/* Given a TYPENAME hash_entry, to find the exact symbol for this type
 * flag : OFF  stands for find a id don't care any kind of id
 *
 */
PTR_SYMB
find_type_symbol(entry,flag)
PTR_HASH entry ;
int  flag ;
{  PTR_SYMB symb_ptr ,old_symb ;
   int type_status ;


  type_status= OFF ;
  for (old_symb=symb_ptr =entry->id_attr ;(symb_ptr && (! type_status)); symb_ptr = symb_ptr->outer)
	   { 	  if (is_same_variant_group(symb_ptr->variant,flag) || (flag==OFF))
		       {  PTR_BFND flow ;
                          PTR_BFND type_scope ;                    
                          
                          type_scope = real_scope(symb_ptr->scope) ;
                          for ( flow = cur_scope(); flow ; flow = pre_scope(flow) )
                              if (flow == type_scope) 
                                {   type_status = ON ;
                                    break ;
                                 }
                         }   
                  old_symb = symb_ptr ;
            }
     if (type_status == ON)  return(old_symb);
     else {
         /* default integer function */
         return(SMNULL);
       }

 }




/* Very Imortant basic routine
 * install_entry takes
 *   <1>  a pointer to a hash entry and makes another symbol table entry for 
 *        the same id
 *   <2>  var_decl : variant for new symbol
 *                   if (variant == FUNCTION_NAME) and (sw== ONE)
 *                      1. if it is a new symbol, then 
 *                           FUNCTION_NAME installed, scope = Global_Bfnd,
 *                           seen = NO ;
 *                      2. if has same name symbol,
 *                         a. if not visible don't care same as 1.
 *                         b. report error except for the same name id
 *                            is a function and its scope is global_bfnd
 *                         c. if so, check seen or not seen, anyway nothing
 *                            can do at this time.
 *                  if (variant == FUNCTION_NAME) and (sw >= TWO)
 *                      1. check as normal field_name is ok.
 *                      2. MEMBER_FUNC installed, scope is same as normal field
 *                      3. Seen = No
 *                  if normal data then
 *                      1. only check same name id if appear in the same scope
 *                      2. install entry
 *   <3>  first_name : if the new entry is only a field of a T_ENUM, then
 *        first_name represent the first member in the same structure
 *        it will be NULL if not a member.
 *   <4>  sw         : ONE if  the entry is not a member
 *                     TWO if  the entry is the first member of the structure
 *                    THREE if the entry is a member but not first one
 */
PTR_SYMB 
install_entry(entry,var_decl,err_sw)
	PTR_HASH entry;
        int      var_decl ;
        int      err_sw;

{
	register PTR_SYMB symb_ptr;
	PTR_BFND decl_block();
	PTR_SYMB temp_ptr,old_symb;
        PTR_BFND flow,type_scope ;
        int type_status;
      

      type_status= OFF ;
      for (old_symb=symb_ptr =entry->id_attr ;(symb_ptr && (! type_status)); )
	   { 
                type_scope = symb_ptr->scope ;
                if (type_scope == decl_block() )
		  {
                     if (err_sw == YES) {
		       err("Redeclaration of Identifier");
		       type_status = ON ;
		       return(SMNULL);
		     }
		     else
		       { type_status = ON ;
			 return(symb_ptr);
		       }
                  }
                else      
                  {
                    symb_ptr =symb_ptr->outer ;
                    old_symb = symb_ptr ;
                  }
           }
       
	symb_ptr = make_symb(fi,var_decl, entry->ident);
        symb_ptr->scope = decl_block();
	symb_ptr->outer = entry->id_attr;
	symb_ptr->parent = entry;
	symb_ptr->attr = NULL;
	entry->id_attr = symb_ptr;
        symb_ptr->dovar = (int ) NULL ;
	symb_ptr->id_list = SMNULL;
	return (symb_ptr);
      }



/* install_entry_2 is a division from install_entry
 * <1> No check for scope is done here. Absolutely install one entry
 * <2> For parameter,  empty string name, and label name
 * <3> If scope_sw == YES we use normal way <pred_bfnd> for variable scope
 * <4> If scope_sw == NO  we use current_function bif_node as scope
 * <5> if scope_sw == OTHER the global_bfnd as scope ;
 * <6> if scope_sw == UP_TO_FUNC_HEDR, we use the near predecessor func_hedr
 * <5> if scope_sw == UP_TO_NODECL, we use the first non-decl predecessor
 */
PTR_SYMB 
install_entry_2(entry,var_decl,scope_sw)
PTR_HASH entry;
int      var_decl ;
int      scope_sw ;
{
	register PTR_SYMB symb_ptr;
	PTR_BFND decl_block(),non_decl_scope(),function_block();
	PTR_SYMB temp_ptr,old_symb;
        PTR_BLOB blob ;

      for (symb_ptr =entry->id_attr ;(symb_ptr) ; )
	   { 
                symb_ptr =symb_ptr->outer ;
           }
       
	symb_ptr = make_symb(fi,var_decl, entry->ident);
	switch (scope_sw) {
	case YES:
              symb_ptr->scope = decl_block() ;
	      break;
	case NO:
              blob = follow_blob(pred_bfnd);
              if (!blob) err("system crash at install_entry_2 ");
              symb_ptr->scope = blob->ref ;
	      break;
	case OTHER:
              symb_ptr->scope =  global_bfnd;
	      break;
	case UP_TO_FUNC_HEDR:
              symb_ptr->scope = function_block(pred_bfnd);
	      break;
	case UP_TO_NODECL:
	      symb_ptr->scope = non_decl_scope(pred_bfnd);
	      break;
        }
	symb_ptr->outer = entry->id_attr;
	symb_ptr->parent = entry;
	symb_ptr->attr = NULL;
	entry->id_attr = symb_ptr;
        symb_ptr->dovar = (int ) NULL ;
	symb_ptr->id_list = SMNULL;
	return (symb_ptr);
      }

/* Needs change */
PTR_SYMB 
install_procedure_entry(entry)
	PTR_HASH entry;
{
	register PTR_SYMB symb_ptr;
	PTR_BFND cur_scope();

	for (symb_ptr = entry->id_attr; symb_ptr; symb_ptr = symb_ptr->outer)
	  ;
	if (!symb_ptr || symb_ptr->variant != PROCEDURE_NAME)
	  symb_ptr = make_symb(fi,PROCEDURE_NAME, entry->ident);

	symb_ptr->scope = cur_scope();
	symb_ptr->outer = entry->id_attr;
	symb_ptr->parent = entry;
	entry->id_attr = symb_ptr;

	return (symb_ptr);
}

PTR_BFND
decl_block()
{
   return(pred_bfnd);
 }


/* Mofified */
PTR_BFND
cur_scope()
{
	register PTR_BFND p;
        p = real_scope(pred_bfnd);
        return(p);
}


/* Mofified */
PTR_BFND
real_scope(in_bfnd)
PTR_BFND in_bfnd;

{
	register PTR_BFND p;

/*	for (p = in_bfnd;(is_var_decl(p->variant) == YES ) ; p = p->control_parent) ;
	return (p);
*/
/* Modified Apr 16,1988 */
        return(in_bfnd);
}


/* Mofified */
PTR_BFND
pre_scope(fix_bfnd)
        PTR_BFND fix_bfnd;     
{
	register PTR_BFND p;
        if (fix_bfnd->variant == GLOBAL) return(BFNULL);
/*	for (p = fix_bfnd->control_parent ; (is_var_decl(p->variant) == YES) ; p = p->control_parent);
	return (p);
 */
        p = fix_bfnd->control_parent ;
        return(p);
}


int
is_var_decl(variant_value)
int variant_value;
{
   switch (variant_value) {
   case UNION_DECL :
   case VAR_DECL :
   case CLASS_DECL :
   case STRUCT_DECL :
   case ENUM_DECL :
                     return(YES);
   default :         return(NO);

   }

 }



/*****************************************************************************/
/* find_file_entry: find the file_entry at include file_table                */
/*                  A new one will be allocated if it doesnot exist          */
/*****************************************************************************/
PTR_FNAME
find_file_entry( name)
	char *name ;
{
  register PTR_FNAME file_entry ;
  PTR_FNAME make_file();

  for( file_entry = head_file; file_entry ; file_entry = file_entry->next)
      if (!strcmp(name,file_entry->name)) break;
  if (!file_entry) file_entry = make_file(name);
  return(file_entry);
}



/* Mofified */
PTR_BFND
non_decl_scope(in_bfnd)
PTR_BFND in_bfnd;

{
	register PTR_BFND p;

	for (p = in_bfnd;(is_var_decl(p->variant) == YES ) ; p = p->control_parent) ;
	return (p);
}



/* Mofified */
PTR_BFND
function_block(in_bfnd)
PTR_BFND in_bfnd;

{
	register PTR_BFND p;

	for (p = in_bfnd;((p->variant != FUNC_HEDR)&&(p->variant != GLOBAL)) ; p = p->control_parent) ;
	return (p);

}

/**************************************************************************/
/* check scope for struct tag .....                                       */
/**************************************************************************/
same_scope(old_scope,parent_node)
PTR_BFND old_scope,parent_node;
{
  register PTR_BFND p ;
  int sw;
  p = non_decl_scope(parent_node);
  sw = (p == old_scope) ? 1 : 0 ;
  return(sw);
}


/*****************************************************************************/
/* Remove_entry: Remove a installed symbol entry based on that symbol        */
/*               after the remove will be re-installed again.                */
/*               It is limited used for change the scope of a symbol.        */
/*               Remove it and then re-insert it                             */
/*****************************************************************************/
PTR_SYMB 
remove_entry(symb)
PTR_SYMB symb;
{
  PTR_SYMB old_symb, first;

  old_symb = (PTR_SYMB) NULL;
  for (first =symb->parent->id_attr ;first ;first = first->outer )
    { 
      if (first == symb) break;
      old_symb = first;
    }
  if (!old_symb) {
   /* symb is supposed to be the first one */
    symb->parent->id_attr = first->outer;
  }
  else {
    old_symb->outer =first->outer;
  }
}

change_scope(symb, new_scope)
PTR_SYMB symb;
PTR_BFND new_scope;
{

  remove_entry(symb);
  re_install_entry(symb,new_scope);

}

/****************************************************************************/
/* re_install_entry with symbol instead of hash_entry                       */
/* It is just effective for installing the scope of GLOBAL                  */
/* The symbol is installed as to be found as last one, since it is GLOBAL   */
/****************************************************************************/
PTR_SYMB 
re_install_entry(symb,new_scope)
PTR_SYMB symb;
PTR_BFND new_scope;
{
	register PTR_SYMB symb_ptr;
	PTR_BFND decl_block();
	PTR_SYMB temp_ptr,old_symb;
        PTR_BFND flow,type_scope ;
        int type_status;

      type_status= OFF ;
      for (old_symb=symb_ptr =symb->parent->id_attr ;(symb_ptr && (! type_status)); )
	   { 
                type_scope = symb_ptr->scope ;
                if (type_scope == new_scope)
		  {
                     err("Redeclaration of Identifier");
                     type_status = ON ;
                     return(SMNULL);
                  }
                else      
                  {
                    old_symb = symb_ptr ;
                    symb_ptr =symb_ptr->outer ;
                  }
           }
       
        symb->scope = new_scope;
	symb->outer = (PTR_SYMB) NULL;
	if (old_symb) old_symb->outer = symb;
	else          symb->parent->id_attr = symb;
   }


